function shapelets_polar_matrix, n_max,              $
                                 CALCULATE=calculate,$
                                 C2P=c2p,            $
                                 P2C=p2c,            $
                                 INVERSE=inverse,    $
                                 N1=n1,              $
                                 N2=n2

;$Id: shapelets_c2p.pro, v2$
;
; Copyright  2005 Richard Massey and Alexandre Refregier.
;
; This file is a part of the Shapelets analysis code.
; www.astro.caltech.edu/~rjm/shapelets/
;
; The Shapelets code is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public Licence as published
; by the Free Software Foundation; either version 2 of the Licence, or
; (at your option) any later version.
;
; The Shapelets code is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
; GNU General Public Licence for more details.
;
; You should have received a copy of the GNU General Public Licence
; along with the Shapelets code; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
;
;+
; NAME:
;      SHAPELETS_POLAR_MATRIX
;
; CATEGORY:
;      Shapelets.
;
; PURPOSE:
;      Calculates a matrix to convert Cartesian shapelet coefficients to their
;      equivalent polar shapelet representation.
;
; INPUTS:
;       N_MAX     - The maximum order used in the shapelet expansion.
;
; OPTIONAL INPUTS:
;       N1 and N2 - The shapelet coefficients are rearranged and stored in a
;                   1D array, even if the basis functions are 2D. This 
;                   simplifies many tasks, including the conversion from
;                   Cartesian to polar shapelets, which then has a 2D matrix.
;                   These vectors specify the order in which the shapelet 
;                   coefficients are stored. For polar shapelets, we assume the
;                   convention nl=n2 and nr=n1. If this is not set, the default
;                   order is assumed. Note that this is much faster, because 
;                   the matrix can then be read from a pre-calculated look-up
;                   table.
;
; KEYWORD PARAMETERS:
;       CALCULATE - By default, the matrix is just read in from a 
;                   pre-calculated look-up table. Setting this option forces 
;                   the routine to calculate the matrix anew.
;       C2P       - Calculate the matrix to convert from Cartesian to polar
;                   shapelet representation. This is the default option.
;       P2C       - Calculate the matrix to convert from polar to Cartesian
;                   shapelet representation.
;       INVERT    - A synonym for P2C (works interchangeably).
;
; OUTPUTS:
;       MATRIX    - The requested conversion matrix. It is independent of beta  
;                   etc, so once it has been calculated, simply use e.g.
;                   polar_coeffs=matrix#cartesian_coeffs
;
; NOTES:
;      A vector of polar shapelet coefficients may be indexed using:
;        shapelets_make_nvec,n_max,n_r,n_l,n_coeffs
;        n=n_r+n_l
;        m=n_r-n_l
;      or
;        shapelets_make_nvec,n_max,n,m,n_coeffs,/POLAR
;
; MODIFICATION HISTORY:
;      Mar 06 - Values in look-up table hard-wired to n_max=30 by Joel Berge.
;      Jul 05 - Noticed should be using la_invert to invert complex matrices.
;      Apr 05 - Faster calculation & look-up tables (using invert) added by RM.
;      Feb 02 - Written by Richard Massey.
;-

COMPILE_OPT idl2

; Work out what n_max was and set up index variables
if keyword_set(n1) and keyword_set(n2) then begin
  n_max=max(n1+n2)
  n_coeffs=n_elements(n1)<n_elements(n2)
  calculate=1B
endif else begin
  shapelets_make_nvec, n_max, n1, n2, n_coeffs   
endelse

; Decide whether we want the forward or the reverse transformation
if keyword_set(c2p) or (not keyword_set(p2c) and not keyword_set(inverse)) then begin
  ; Create matrix to convert Cartesian shapelets into polar shapelets

  ; Initialise empty array
  matrix=complexarr(n_coeffs,n_coeffs)
  
  if n_max gt 30 or keyword_set(calculate) then begin
    ; Old code to do calculate the matrix. It explicitly follows the notation in 
    ; published papers, and is very clear. However, its nested for loops are slow.
    ;for i=0,n_coeffs-1 do begin    ; Cartesian
    ;  for j=0,n_coeffs-1 do begin  ; Polar
    ;
    ;    ; Get into form of eqn(69) in Paper I
    ;    n_r=n1[j]
    ;    n_l=n2[j]
    ;    n_1=n1[i]
    ;    n_2=n2[i]
    ;
    ;    if n_1+n_2 eq n_r+n_l then begin
    ;
    ;      factor=complex(0.,0.)
    ;      good=0
    ;      for n_rd=0,n_r do begin
    ;        for n_ld=0,n_l do begin
    ;          if n_rd+n_ld eq n_1 then begin
    ;            good=good+1
    ;            factor = factor                $
    ;              + complex(0.,1.)^(n_ld-n_rd) $
    ;              * comb(n_r,n_rd)             $
    ;              * comb(n_l,n_ld)
    ;          endif
    ;        endfor
    ;      endfor
    ;
    ;      matrix[j,i] =               $
    ;          2.^(-0.5*(n_r+n_l))     $
    ;        * complex(0,1.)^(n_r-n_l)  $
    ;        * sqrt( factorial(n_1)    $
    ;             *  factorial(n_2)    $
    ;             /  factorial(n_r)    $
    ;             /  factorial(n_l)  ) $
    ;        * factor
    ;
    ;    endif
    ;
    ;  endfor
    ;endfor  
    
    ; Faster new alternative
    for i=0,n_coeffs-1 do begin    ; Cartesian shapelet coefficients
      ; Only states with the same n get mixed
      same_n=where(n1+n2 eq n1[i]+n2[i],n_same_n)
      for j=0,n_same_n-1 do begin  ; Polar shapelets coefficients
        ; Get into form of eqn(69) in Paper I
        n_r=n1[same_n[j]]
        n_l=n2[same_n[j]]
        n_1=n1[i]
        n_2=n2[i]
        n_ld=replicate(1,n_r+1)#indgen(n_l+1)
        n_rd=indgen(n_r+1)#replicate(1,n_l+1)
        nonzero=where(n_rd+n_ld eq n_1,n_nonzero)
        if n_nonzero gt 0 then $
        matrix[same_n[j],i] = 2.^(-0.5*(n_r+n_l))         $
                                * complex(0,1)^(n_r-n_l)  $
                                * sqrt( factorial(n_1)    $
                                     *  factorial(n_2)    $
                                     /  factorial(n_r)    $
                                     /  factorial(n_l)  ) $
                                * total(complex(0.,1.)^(n_ld[nonzero]-n_rd[nonzero]) $
                                     * comb(n_r,n_rd[nonzero]) * comb(n_l,n_ld[nonzero]) )
      endfor
    endfor
  
  endif else begin
    ; Use look-up table
    switch n_max of

      30: begin
            matrix[465:495,465:495]=[$
            [complex(-3.05176e-05,-0.00000),complex(0.000167152,0.00000),complex(-0.000636494,-0.00000),complex(0.00194452,0.00000),complex(-0.00505202,-0.00000),complex(0.0115204,0.00000),complex(-0.0235159,-0.00000),complex(0.0435429,0.00000),complex(-0.0738306,-0.00000),complex(0.115432,0.00000),complex(-0.167277,-0.00000),complex(0.225556,0.00000),complex(-0.283819,-0.00000),complex(0.333969,0.00000),complex(-0.368015,-0.00000),complex(0.380085,0.00000),complex(-0.368015,0.00000),complex(0.333969,0.00000),complex(-0.283819,0.00000),complex(0.225556,0.00000),complex(-0.167277,0.00000),complex(0.115432,0.00000),complex(-0.0738306,0.00000),complex(0.0435429,0.00000),complex(-0.0235159,0.00000),complex(0.0115204,0.00000),complex(-0.00505202,0.00000),complex(0.00194452,0.00000),complex(-0.000636494,0.00000),complex(0.000167152,0.00000),complex(-3.05176e-05,0.00000)],$
            [complex(0.00000,-0.000167152),complex(0.00000,0.000854492),complex(0.00000,-0.00302139),complex(0.00000,0.00852047),complex(0.00000,-0.0202921),complex(0.00000,0.0420665),complex(0.00000,-0.0772810),complex(0.00000,0.127197),complex(0.00000,-0.188714),complex(0.00000,0.252899),complex(0.00000,-0.305405),complex(0.00000,0.329446),complex(0.00000,-0.310908),complex(0.00000,0.243896),complex(0.00000,-0.134380),0,complex(0.00000,0.134380),complex(0.00000,-0.243896),complex(0.00000,0.310908),complex(0.00000,-0.329446),complex(0.00000,0.305405),complex(0.00000,-0.252899),complex(0.00000,0.188714),complex(0.00000,-0.127197),complex(0.00000,0.0772810),complex(0.00000,-0.0420665),complex(0.00000,0.0202921),complex(0.00000,-0.00852047),complex(0.00000,0.00302139),complex(0.00000,-0.000854492),complex(0.00000,0.000167152)],$
            [complex(0.000636494,0.00000),complex(-0.00302139,0.00000),complex(0.00985718,0.00000),complex(-0.0254525,0.00000),complex(0.0549853,0.00000),complex(-0.102187,0.00000),complex(0.165742,0.00000),complex(-0.235913,0.00000),complex(0.293812,0.00000),complex(-0.315469,0.00000),complex(0.280711,0.00000),complex(-0.183848,0.00000),complex(0.0408242,0.00000),complex(0.112088,0.00000),complex(-0.229385,-0.00000),complex(0.273355,0.00000),complex(-0.229385,0.00000),complex(0.112088,0.00000),complex(0.0408242,-0.00000),complex(-0.183848,0.00000),complex(0.280711,-0.00000),complex(-0.315469,0.00000),complex(0.293812,-0.00000),complex(-0.235913,0.00000),complex(0.165742,-0.00000),complex(-0.102187,0.00000),complex(0.0549853,-0.00000),complex(-0.0254525,0.00000),complex(0.00985718,-0.00000),complex(-0.00302139,0.00000),complex(0.000636494,-0.00000)],$
            [complex(-0.00000,0.00194452),complex(0.00000,-0.00852047),complex(-0.00000,0.0254525),complex(0.00000,-0.0595703),complex(-0.00000,0.115125),complex(0.00000,-0.188034),complex(-0.00000,0.261295),complex(0.00000,-0.306149),complex(-0.00000,0.291994),complex(0.00000,-0.202900),complex(-0.00000,0.0525053),complex(0.00000,0.113277),complex(0.00000,-0.231623),complex(0.00000,0.251585),complex(0.00000,-0.161719),0,complex(0.00000,0.161719),complex(0.00000,-0.251585),complex(0.00000,0.231623),complex(0.00000,-0.113277),complex(-0.00000,-0.0525053),complex(0.00000,0.202900),complex(-0.00000,-0.291994),complex(0.00000,0.306149),complex(-0.00000,-0.261295),complex(0.00000,0.188034),complex(-0.00000,-0.115125),complex(0.00000,0.0595703),complex(-0.00000,-0.0254525),complex(0.00000,0.00852047),complex(-0.00000,-0.00194452)],$
            [complex(-0.00505202,-0.00000),complex(0.0202921,0.00000),complex(-0.0549853,-0.00000),complex(0.115125,0.00000),complex(-0.195221,-0.00000),complex(0.271752,0.00000),complex(-0.306405,-0.00000),complex(0.263292,0.00000),complex(-0.134242,-0.00000),complex(-0.0439291,0.00000),complex(0.197041,0.00000),complex(-0.249340,0.00000),complex(0.169731,0.00000),complex(0.00201739,0.00000),complex(-0.171176,-0.00000),complex(0.241076,0.00000),complex(-0.171176,0.00000),complex(0.00201739,0.00000),complex(0.169731,-0.00000),complex(-0.249340,0.00000),complex(0.197041,-0.00000),complex(-0.0439291,0.00000),complex(-0.134242,0.00000),complex(0.263292,0.00000),complex(-0.306405,0.00000),complex(0.271752,0.00000),complex(-0.195221,0.00000),complex(0.115125,0.00000),complex(-0.0549853,0.00000),complex(0.0202921,0.00000),complex(-0.00505202,0.00000)],$
            [complex(0.00000,-0.0115204),complex(0.00000,0.0420665),complex(0.00000,-0.102187),complex(0.00000,0.188034),complex(0.00000,-0.271752),complex(0.00000,0.305298),complex(0.00000,-0.245562),complex(0.00000,0.0904310),complex(-0.00000,0.101309),complex(0.00000,-0.231170),complex(-0.00000,0.220673),complex(0.00000,-0.0717001),complex(0.00000,-0.121798),complex(0.00000,0.230018),complex(0.00000,-0.177428),0,complex(0.00000,0.177428),complex(0.00000,-0.230018),complex(0.00000,0.121798),complex(0.00000,0.0717001),complex(-0.00000,-0.220673),complex(0.00000,0.231170),complex(-0.00000,-0.101309),complex(0.00000,-0.0904310),complex(0.00000,0.245562),complex(0.00000,-0.305298),complex(0.00000,0.271752),complex(0.00000,-0.188034),complex(0.00000,0.102187),complex(0.00000,-0.0420665),complex(0.00000,0.0115204)],$
            [complex(0.0235159,0.00000),complex(-0.0772810,0.00000),complex(0.165742,0.00000),complex(-0.261295,0.00000),complex(0.306405,0.00000),complex(-0.245562,0.00000),complex(0.0756531,0.00000),complex(0.126972,0.00000),complex(-0.240779,-0.00000),complex(0.185604,0.00000),complex(0.00325624,0.00000),complex(-0.185288,0.00000),complex(0.217680,0.00000),complex(-0.0732456,0.00000),complex(-0.130382,-0.00000),complex(0.224430,0.00000),complex(-0.130382,0.00000),complex(-0.0732456,0.00000),complex(0.217680,-0.00000),complex(-0.185288,0.00000),complex(0.00325624,-0.00000),complex(0.185604,0.00000),complex(-0.240779,0.00000),complex(0.126972,0.00000),complex(0.0756531,-0.00000),complex(-0.245562,0.00000),complex(0.306405,-0.00000),complex(-0.261295,0.00000),complex(0.165742,-0.00000),complex(-0.0772810,0.00000),complex(0.0235159,-0.00000)],$
            [complex(-0.00000,0.0435429),complex(0.00000,-0.127197),complex(-0.00000,0.235913),complex(0.00000,-0.306149),complex(-0.00000,0.263292),complex(0.00000,-0.0904310),complex(0.00000,-0.126972),complex(0.00000,0.242188),complex(0.00000,-0.164342),complex(0.00000,-0.0465995),complex(-0.00000,0.211029),complex(0.00000,-0.182113),complex(0.00000,-0.0143221),complex(0.00000,0.194743),complex(0.00000,-0.187772),0,complex(0.00000,0.187772),complex(0.00000,-0.194743),complex(0.00000,0.0143221),complex(0.00000,0.182113),complex(-0.00000,-0.211029),complex(0.00000,0.0465995),complex(0.00000,0.164342),complex(0.00000,-0.242188),complex(0.00000,0.126972),complex(0.00000,0.0904310),complex(-0.00000,-0.263292),complex(0.00000,0.306149),complex(-0.00000,-0.235913),complex(0.00000,0.127197),complex(-0.00000,-0.0435429)],$
            [complex(-0.0738306,-0.00000),complex(0.188714,0.00000),complex(-0.293812,-0.00000),complex(0.291994,0.00000),complex(-0.134242,-0.00000),complex(-0.101309,0.00000),complex(0.240779,0.00000),complex(-0.164342,0.00000),complex(-0.0604553,-0.00000),complex(0.218575,0.00000),complex(-0.152461,-0.00000),complex(-0.0696449,0.00000),complex(0.214335,0.00000),complex(-0.127415,0.00000),complex(-0.0968990,-0.00000),complex(0.214450,0.00000),complex(-0.0968990,0.00000),complex(-0.127415,0.00000),complex(0.214335,-0.00000),complex(-0.0696449,0.00000),complex(-0.152461,0.00000),complex(0.218575,0.00000),complex(-0.0604553,0.00000),complex(-0.164342,0.00000),complex(0.240779,-0.00000),complex(-0.101309,0.00000),complex(-0.134242,0.00000),complex(0.291994,0.00000),complex(-0.293812,0.00000),complex(0.188714,0.00000),complex(-0.0738306,0.00000)],$
            [complex(0.00000,-0.115432),complex(0.00000,0.252899),complex(0.00000,-0.315469),complex(0.00000,0.202900),complex(-0.00000,0.0439291),complex(0.00000,-0.231170),complex(-0.00000,0.185604),complex(0.00000,0.0465995),complex(0.00000,-0.218575),complex(0.00000,0.141480),complex(-0.00000,0.0950820),complex(0.00000,-0.215152),complex(-0.00000,0.0775863),complex(0.00000,0.151512),complex(0.00000,-0.194784),0,complex(0.00000,0.194784),complex(0.00000,-0.151512),complex(-0.00000,-0.0775863),complex(0.00000,0.215152),complex(-0.00000,-0.0950820),complex(0.00000,-0.141480),complex(0.00000,0.218575),complex(0.00000,-0.0465995),complex(-0.00000,-0.185604),complex(0.00000,0.231170),complex(-0.00000,-0.0439291),complex(0.00000,-0.202900),complex(0.00000,0.315469),complex(0.00000,-0.252899),complex(0.00000,0.115432)],$
            [complex(0.167277,0.00000),complex(-0.305405,0.00000),complex(0.280711,0.00000),complex(-0.0525053,0.00000),complex(-0.197041,-0.00000),complex(0.220673,0.00000),complex(0.00325624,0.00000),complex(-0.211029,0.00000),complex(0.152461,0.00000),complex(0.0950820,0.00000),complex(-0.213654,-0.00000),complex(0.0511493,0.00000),complex(0.175997,0.00000),complex(-0.165542,0.00000),complex(-0.0672069,-0.00000),complex(0.208233,0.00000),complex(-0.0672069,0.00000),complex(-0.165542,0.00000),complex(0.175997,-0.00000),complex(0.0511493,0.00000),complex(-0.213654,0.00000),complex(0.0950820,0.00000),complex(0.152461,-0.00000),complex(-0.211029,0.00000),complex(0.00325624,-0.00000),complex(0.220673,0.00000),complex(-0.197041,0.00000),complex(-0.0525053,0.00000),complex(0.280711,-0.00000),complex(-0.305405,0.00000),complex(0.167277,-0.00000)],$
            [complex(-0.00000,0.225556),complex(0.00000,-0.329446),complex(-0.00000,0.183848),complex(0.00000,0.113277),complex(0.00000,-0.249340),complex(0.00000,0.0717001),complex(-0.00000,0.185288),complex(0.00000,-0.182113),complex(0.00000,-0.0696449),complex(0.00000,0.215152),complex(0.00000,-0.0511493),complex(0.00000,-0.182617),complex(-0.00000,0.146997),complex(0.00000,0.103385),complex(0.00000,-0.199368),0,complex(0.00000,0.199368),complex(0.00000,-0.103385),complex(-0.00000,-0.146997),complex(0.00000,0.182617),complex(0.00000,0.0511493),complex(0.00000,-0.215152),complex(0.00000,0.0696449),complex(0.00000,0.182113),complex(-0.00000,-0.185288),complex(0.00000,-0.0717001),complex(0.00000,0.249340),complex(0.00000,-0.113277),complex(-0.00000,-0.183848),complex(0.00000,0.329446),complex(-0.00000,-0.225556)],$
            [complex(-0.283819,-0.00000),complex(0.310908,0.00000),complex(-0.0408242,-0.00000),complex(-0.231623,0.00000),complex(0.169731,0.00000),complex(0.121798,0.00000),complex(-0.217680,-0.00000),complex(-0.0143221,0.00000),complex(0.214335,0.00000),complex(-0.0775863,0.00000),complex(-0.175997,-0.00000),complex(0.146997,0.00000),complex(0.114471,0.00000),complex(-0.190000,0.00000),complex(-0.0396104,-0.00000),complex(0.204547,0.00000),complex(-0.0396104,0.00000),complex(-0.190000,0.00000),complex(0.114471,-0.00000),complex(0.146997,0.00000),complex(-0.175997,0.00000),complex(-0.0775863,0.00000),complex(0.214335,-0.00000),complex(-0.0143221,0.00000),complex(-0.217680,0.00000),complex(0.121798,0.00000),complex(0.169731,-0.00000),complex(-0.231623,0.00000),complex(-0.0408242,0.00000),complex(0.310908,0.00000),complex(-0.283819,0.00000)],$
            [complex(0.00000,-0.333969),complex(0.00000,0.243896),complex(-0.00000,0.112088),complex(0.00000,-0.251585),complex(0.00000,-0.00201739),complex(0.00000,0.230018),complex(0.00000,-0.0732456),complex(0.00000,-0.194743),complex(-0.00000,0.127415),complex(0.00000,0.151512),complex(0.00000,-0.165542),complex(0.00000,-0.103385),complex(-0.00000,0.190000),complex(0.00000,0.0523682),complex(0.00000,-0.201974),0,complex(0.00000,0.201974),complex(0.00000,-0.0523682),complex(-0.00000,-0.190000),complex(0.00000,0.103385),complex(0.00000,0.165542),complex(0.00000,-0.151512),complex(-0.00000,-0.127415),complex(0.00000,0.194743),complex(0.00000,0.0732456),complex(0.00000,-0.230018),complex(0.00000,0.00201739),complex(0.00000,0.251585),complex(-0.00000,-0.112088),complex(0.00000,-0.243896),complex(0.00000,0.333969)],$
            [complex(0.368015,0.00000),complex(-0.134380,0.00000),complex(-0.229385,-0.00000),complex(0.161719,0.00000),complex(0.171176,0.00000),complex(-0.177428,0.00000),complex(-0.130382,-0.00000),complex(0.187772,0.00000),complex(0.0968990,0.00000),complex(-0.194784,0.00000),complex(-0.0672069,-0.00000),complex(0.199368,0.00000),complex(0.0396104,0.00000),complex(-0.201974,0.00000),complex(-0.0130920,-0.00000),complex(0.202821,0.00000),complex(-0.0130920,0.00000),complex(-0.201974,0.00000),complex(0.0396104,-0.00000),complex(0.199368,0.00000),complex(-0.0672069,0.00000),complex(-0.194784,0.00000),complex(0.0968990,-0.00000),complex(0.187772,0.00000),complex(-0.130382,0.00000),complex(-0.177428,0.00000),complex(0.171176,-0.00000),complex(0.161719,0.00000),complex(-0.229385,0.00000),complex(-0.134380,0.00000),complex(0.368015,-0.00000)],$
            [complex(-0.00000,0.380085),0,complex(0.00000,-0.273355),complex(0.00000,-1.16325e-15),complex(-0.00000,0.241076),complex(0.00000,-1.72293e-15),complex(0.00000,-0.224430),complex(0.00000,-1.89256e-14),complex(-0.00000,0.214450),complex(0.00000,-8.42939e-14),complex(0.00000,-0.208233),0,complex(-0.00000,0.204547),0,complex(0.00000,-0.202821),0,complex(0.00000,0.202821),0,complex(-0.00000,-0.204547),0,complex(0.00000,0.208233),complex(0.00000,5.38132e-14),complex(-0.00000,-0.214450),complex(0.00000,2.03501e-14),complex(0.00000,0.224430),complex(0.00000,-1.72293e-15),complex(-0.00000,-0.241076),complex(0.00000,1.16325e-15),complex(0.00000,0.273355),0,complex(-0.00000,-0.380085)],$
            [complex(-0.368015,-0.00000),complex(-0.134380,0.00000),complex(0.229385,0.00000),complex(0.161719,0.00000),complex(-0.171176,-0.00000),complex(-0.177428,0.00000),complex(0.130382,0.00000),complex(0.187772,0.00000),complex(-0.0968990,-0.00000),complex(-0.194784,0.00000),complex(0.0672069,0.00000),complex(0.199368,0.00000),complex(-0.0396104,-0.00000),complex(-0.201974,0.00000),complex(0.0130920,0.00000),complex(0.202821,0.00000),complex(0.0130920,-0.00000),complex(-0.201974,0.00000),complex(-0.0396104,0.00000),complex(0.199368,0.00000),complex(0.0672069,-0.00000),complex(-0.194784,0.00000),complex(-0.0968990,0.00000),complex(0.187772,0.00000),complex(0.130382,-0.00000),complex(-0.177428,0.00000),complex(-0.171176,0.00000),complex(0.161719,0.00000),complex(0.229385,-0.00000),complex(-0.134380,0.00000),complex(-0.368015,0.00000)],$
            [complex(0.00000,-0.333969),complex(0.00000,-0.243896),complex(-0.00000,0.112088),complex(0.00000,0.251585),complex(0.00000,-0.00201739),complex(0.00000,-0.230018),complex(0.00000,-0.0732456),complex(0.00000,0.194743),complex(-0.00000,0.127415),complex(0.00000,-0.151512),complex(0.00000,-0.165542),complex(0.00000,0.103385),complex(-0.00000,0.190000),complex(0.00000,-0.0523682),complex(0.00000,-0.201974),0,complex(0.00000,0.201974),complex(0.00000,0.0523682),complex(-0.00000,-0.190000),complex(0.00000,-0.103385),complex(0.00000,0.165542),complex(0.00000,0.151512),complex(-0.00000,-0.127415),complex(0.00000,-0.194743),complex(0.00000,0.0732456),complex(0.00000,0.230018),complex(0.00000,0.00201739),complex(0.00000,-0.251585),complex(-0.00000,-0.112088),complex(0.00000,0.243896),complex(0.00000,0.333969)],$
            [complex(0.283819,0.00000),complex(0.310908,0.00000),complex(0.0408242,0.00000),complex(-0.231623,0.00000),complex(-0.169731,-0.00000),complex(0.121798,0.00000),complex(0.217680,0.00000),complex(-0.0143221,0.00000),complex(-0.214335,-0.00000),complex(-0.0775863,0.00000),complex(0.175997,0.00000),complex(0.146997,0.00000),complex(-0.114471,-0.00000),complex(-0.190000,0.00000),complex(0.0396104,0.00000),complex(0.204547,0.00000),complex(0.0396104,-0.00000),complex(-0.190000,0.00000),complex(-0.114471,0.00000),complex(0.146997,0.00000),complex(0.175997,-0.00000),complex(-0.0775863,0.00000),complex(-0.214335,0.00000),complex(-0.0143221,0.00000),complex(0.217680,-0.00000),complex(0.121798,0.00000),complex(-0.169731,0.00000),complex(-0.231623,0.00000),complex(0.0408242,-0.00000),complex(0.310908,0.00000),complex(0.283819,-0.00000)],$
            [complex(-0.00000,0.225556),complex(0.00000,0.329446),complex(-0.00000,0.183848),complex(0.00000,-0.113277),complex(0.00000,-0.249340),complex(0.00000,-0.0717001),complex(-0.00000,0.185288),complex(0.00000,0.182113),complex(0.00000,-0.0696449),complex(0.00000,-0.215152),complex(0.00000,-0.0511493),complex(0.00000,0.182617),complex(-0.00000,0.146997),complex(0.00000,-0.103385),complex(0.00000,-0.199368),0,complex(0.00000,0.199368),complex(0.00000,0.103385),complex(-0.00000,-0.146997),complex(0.00000,-0.182617),complex(0.00000,0.0511493),complex(0.00000,0.215152),complex(0.00000,0.0696449),complex(0.00000,-0.182113),complex(-0.00000,-0.185288),complex(0.00000,0.0717001),complex(0.00000,0.249340),complex(0.00000,0.113277),complex(-0.00000,-0.183848),complex(0.00000,-0.329446),complex(-0.00000,-0.225556)],$
            [complex(-0.167277,-0.00000),complex(-0.305405,0.00000),complex(-0.280711,-0.00000),complex(-0.0525053,0.00000),complex(0.197041,0.00000),complex(0.220673,0.00000),complex(-0.00325624,-0.00000),complex(-0.211029,0.00000),complex(-0.152461,-0.00000),complex(0.0950820,0.00000),complex(0.213654,0.00000),complex(0.0511493,0.00000),complex(-0.175997,-0.00000),complex(-0.165542,0.00000),complex(0.0672069,0.00000),complex(0.208233,0.00000),complex(0.0672069,-0.00000),complex(-0.165542,0.00000),complex(-0.175997,0.00000),complex(0.0511493,0.00000),complex(0.213654,-0.00000),complex(0.0950820,0.00000),complex(-0.152461,0.00000),complex(-0.211029,0.00000),complex(-0.00325624,0.00000),complex(0.220673,0.00000),complex(0.197041,-0.00000),complex(-0.0525053,0.00000),complex(-0.280711,0.00000),complex(-0.305405,0.00000),complex(-0.167277,0.00000)],$
            [complex(0.00000,-0.115432),complex(0.00000,-0.252899),complex(0.00000,-0.315469),complex(0.00000,-0.202900),complex(-0.00000,0.0439291),complex(0.00000,0.231170),complex(-0.00000,0.185604),complex(0.00000,-0.0465995),complex(0.00000,-0.218575),complex(0.00000,-0.141480),complex(-0.00000,0.0950820),complex(0.00000,0.215152),complex(-0.00000,0.0775863),complex(0.00000,-0.151512),complex(0.00000,-0.194784),0,complex(0.00000,0.194784),complex(0.00000,0.151512),complex(-0.00000,-0.0775863),complex(0.00000,-0.215152),complex(-0.00000,-0.0950820),complex(0.00000,0.141480),complex(0.00000,0.218575),complex(0.00000,0.0465995),complex(-0.00000,-0.185604),complex(0.00000,-0.231170),complex(-0.00000,-0.0439291),complex(0.00000,0.202900),complex(0.00000,0.315469),complex(0.00000,0.252899),complex(0.00000,0.115432)],$
            [complex(0.0738306,0.00000),complex(0.188714,0.00000),complex(0.293812,0.00000),complex(0.291994,0.00000),complex(0.134242,0.00000),complex(-0.101309,0.00000),complex(-0.240779,-0.00000),complex(-0.164342,0.00000),complex(0.0604553,0.00000),complex(0.218575,0.00000),complex(0.152461,0.00000),complex(-0.0696449,0.00000),complex(-0.214335,-0.00000),complex(-0.127415,0.00000),complex(0.0968990,0.00000),complex(0.214450,0.00000),complex(0.0968990,-0.00000),complex(-0.127415,0.00000),complex(-0.214335,0.00000),complex(-0.0696449,0.00000),complex(0.152461,-0.00000),complex(0.218575,0.00000),complex(0.0604553,-0.00000),complex(-0.164342,0.00000),complex(-0.240779,0.00000),complex(-0.101309,0.00000),complex(0.134242,-0.00000),complex(0.291994,0.00000),complex(0.293812,-0.00000),complex(0.188714,0.00000),complex(0.0738306,-0.00000)],$
            [complex(-0.00000,0.0435429),complex(0.00000,0.127197),complex(-0.00000,0.235913),complex(0.00000,0.306149),complex(-0.00000,0.263292),complex(0.00000,0.0904310),complex(0.00000,-0.126972),complex(0.00000,-0.242188),complex(0.00000,-0.164342),complex(0.00000,0.0465995),complex(-0.00000,0.211029),complex(0.00000,0.182113),complex(0.00000,-0.0143221),complex(0.00000,-0.194743),complex(0.00000,-0.187772),0,complex(0.00000,0.187772),complex(0.00000,0.194743),complex(0.00000,0.0143221),complex(0.00000,-0.182113),complex(-0.00000,-0.211029),complex(0.00000,-0.0465995),complex(0.00000,0.164342),complex(0.00000,0.242188),complex(0.00000,0.126972),complex(0.00000,-0.0904310),complex(-0.00000,-0.263292),complex(0.00000,-0.306149),complex(-0.00000,-0.235913),complex(0.00000,-0.127197),complex(-0.00000,-0.0435429)],$
            [complex(-0.0235159,-0.00000),complex(-0.0772810,0.00000),complex(-0.165742,-0.00000),complex(-0.261295,0.00000),complex(-0.306405,-0.00000),complex(-0.245562,0.00000),complex(-0.0756531,-0.00000),complex(0.126972,0.00000),complex(0.240779,0.00000),complex(0.185604,0.00000),complex(-0.00325624,-0.00000),complex(-0.185288,0.00000),complex(-0.217680,-0.00000),complex(-0.0732456,0.00000),complex(0.130382,0.00000),complex(0.224430,0.00000),complex(0.130382,-0.00000),complex(-0.0732456,0.00000),complex(-0.217680,0.00000),complex(-0.185288,0.00000),complex(-0.00325624,0.00000),complex(0.185604,0.00000),complex(0.240779,-0.00000),complex(0.126972,0.00000),complex(-0.0756531,0.00000),complex(-0.245562,0.00000),complex(-0.306405,0.00000),complex(-0.261295,0.00000),complex(-0.165742,0.00000),complex(-0.0772810,0.00000),complex(-0.0235159,0.00000)],$
            [complex(0.00000,-0.0115204),complex(0.00000,-0.0420665),complex(0.00000,-0.102187),complex(0.00000,-0.188034),complex(0.00000,-0.271752),complex(0.00000,-0.305298),complex(0.00000,-0.245562),complex(0.00000,-0.0904310),complex(-0.00000,0.101309),complex(0.00000,0.231170),complex(-0.00000,0.220673),complex(0.00000,0.0717001),complex(0.00000,-0.121798),complex(0.00000,-0.230018),complex(0.00000,-0.177428),0,complex(0.00000,0.177428),complex(0.00000,0.230018),complex(0.00000,0.121798),complex(0.00000,-0.0717001),complex(-0.00000,-0.220673),complex(0.00000,-0.231170),complex(-0.00000,-0.101309),complex(0.00000,0.0904310),complex(0.00000,0.245562),complex(0.00000,0.305298),complex(0.00000,0.271752),complex(0.00000,0.188034),complex(0.00000,0.102187),complex(0.00000,0.0420665),complex(0.00000,0.0115204)],$
            [complex(0.00505202,0.00000),complex(0.0202921,0.00000),complex(0.0549853,0.00000),complex(0.115125,0.00000),complex(0.195221,0.00000),complex(0.271752,0.00000),complex(0.306405,0.00000),complex(0.263292,0.00000),complex(0.134242,0.00000),complex(-0.0439291,0.00000),complex(-0.197041,-0.00000),complex(-0.249340,0.00000),complex(-0.169731,-0.00000),complex(0.00201739,0.00000),complex(0.171176,0.00000),complex(0.241076,0.00000),complex(0.171176,-0.00000),complex(0.00201739,0.00000),complex(-0.169731,0.00000),complex(-0.249340,0.00000),complex(-0.197041,0.00000),complex(-0.0439291,0.00000),complex(0.134242,-0.00000),complex(0.263292,0.00000),complex(0.306405,-0.00000),complex(0.271752,0.00000),complex(0.195221,-0.00000),complex(0.115125,0.00000),complex(0.0549853,-0.00000),complex(0.0202921,0.00000),complex(0.00505202,-0.00000)],$
            [complex(-0.00000,0.00194452),complex(0.00000,0.00852047),complex(-0.00000,0.0254525),complex(0.00000,0.0595703),complex(-0.00000,0.115125),complex(0.00000,0.188034),complex(-0.00000,0.261295),complex(0.00000,0.306149),complex(-0.00000,0.291994),complex(0.00000,0.202900),complex(-0.00000,0.0525053),complex(0.00000,-0.113277),complex(0.00000,-0.231623),complex(0.00000,-0.251585),complex(0.00000,-0.161719),0,complex(0.00000,0.161719),complex(0.00000,0.251585),complex(0.00000,0.231623),complex(0.00000,0.113277),complex(-0.00000,-0.0525053),complex(0.00000,-0.202900),complex(-0.00000,-0.291994),complex(0.00000,-0.306149),complex(-0.00000,-0.261295),complex(0.00000,-0.188034),complex(-0.00000,-0.115125),complex(0.00000,-0.0595703),complex(-0.00000,-0.0254525),complex(0.00000,-0.00852047),complex(-0.00000,-0.00194452)],$
            [complex(-0.000636494,-0.00000),complex(-0.00302139,0.00000),complex(-0.00985718,-0.00000),complex(-0.0254525,0.00000),complex(-0.0549853,-0.00000),complex(-0.102187,0.00000),complex(-0.165742,-0.00000),complex(-0.235913,0.00000),complex(-0.293812,-0.00000),complex(-0.315469,0.00000),complex(-0.280711,-0.00000),complex(-0.183848,0.00000),complex(-0.0408242,-0.00000),complex(0.112088,0.00000),complex(0.229385,0.00000),complex(0.273355,0.00000),complex(0.229385,-0.00000),complex(0.112088,0.00000),complex(-0.0408242,0.00000),complex(-0.183848,0.00000),complex(-0.280711,0.00000),complex(-0.315469,0.00000),complex(-0.293812,0.00000),complex(-0.235913,0.00000),complex(-0.165742,0.00000),complex(-0.102187,0.00000),complex(-0.0549853,0.00000),complex(-0.0254525,0.00000),complex(-0.00985718,0.00000),complex(-0.00302139,0.00000),complex(-0.000636494,0.00000)],$
            [complex(0.00000,-0.000167152),complex(0.00000,-0.000854492),complex(0.00000,-0.00302139),complex(0.00000,-0.00852047),complex(0.00000,-0.0202921),complex(0.00000,-0.0420665),complex(0.00000,-0.0772810),complex(0.00000,-0.127197),complex(0.00000,-0.188714),complex(0.00000,-0.252899),complex(0.00000,-0.305405),complex(0.00000,-0.329446),complex(0.00000,-0.310908),complex(0.00000,-0.243896),complex(0.00000,-0.134380),0,complex(0.00000,0.134380),complex(0.00000,0.243896),complex(0.00000,0.310908),complex(0.00000,0.329446),complex(0.00000,0.305405),complex(0.00000,0.252899),complex(0.00000,0.188714),complex(0.00000,0.127197),complex(0.00000,0.0772810),complex(0.00000,0.0420665),complex(0.00000,0.0202921),complex(0.00000,0.00852047),complex(0.00000,0.00302139),complex(0.00000,0.000854492),complex(0.00000,0.000167152)],$
            [complex(3.05176e-05,0.00000),complex(0.000167152,0.00000),complex(0.000636494,0.00000),complex(0.00194452,0.00000),complex(0.00505202,0.00000),complex(0.0115204,0.00000),complex(0.0235159,0.00000),complex(0.0435429,0.00000),complex(0.0738306,0.00000),complex(0.115432,0.00000),complex(0.167277,0.00000),complex(0.225556,0.00000),complex(0.283819,0.00000),complex(0.333969,0.00000),complex(0.368015,0.00000),complex(0.380085,0.00000),complex(0.368015,-0.00000),complex(0.333969,0.00000),complex(0.283819,-0.00000),complex(0.225556,0.00000),complex(0.167277,-0.00000),complex(0.115432,0.00000),complex(0.0738306,-0.00000),complex(0.0435429,0.00000),complex(0.0235159,-0.00000),complex(0.0115204,0.00000),complex(0.00505202,-0.00000),complex(0.00194452,0.00000),complex(0.000636494,-0.00000),complex(0.000167152,0.00000),complex(3.05176e-05,-0.00000)]$
            ]
          end

      29: begin
            matrix[435:464,435:464]=[$
            [complex(0.00000,-4.31584e-05),complex(-0.00000,0.000232415),complex(0.00000,-0.000869617),complex(-0.00000,0.00260885),complex(0.00000,-0.00665129),complex(-0.00000,0.0148727),complex(0.00000,-0.0297455),complex(-0.00000,0.0539183),complex(0.00000,-0.0894133),complex(-0.00000,0.136581),complex(0.00000,-0.193155),complex(-0.00000,0.253855),complex(0.00000,-0.310908),complex(-0.00000,0.355537),complex(0.00000,-0.380085),complex(0.00000,0.380085),complex(0.00000,-0.355537),complex(0.00000,0.310908),complex(0.00000,-0.253855),complex(0.00000,0.193155),complex(0.00000,-0.136581),complex(0.00000,0.0894133),complex(0.00000,-0.0539183),complex(0.00000,0.0297455),complex(0.00000,-0.0148727),complex(0.00000,0.00665129),complex(0.00000,-0.00260885),complex(0.00000,0.000869617),complex(0.00000,-0.000232415),complex(0.00000,4.31584e-05)],$
            [complex(0.000232415,0.00000),complex(-0.00116528,0.00000),complex(0.00403710,0.00000),complex(-0.0111424,0.00000),complex(0.0259374,0.00000),complex(-0.0524742,0.00000),complex(0.0939011,0.00000),complex(-0.150186,0.00000),complex(0.215847,0.00000),complex(-0.278987,0.00000),complex(0.322812,0.00000),complex(-0.329978,0.00000),complex(0.288671,0.00000),complex(-0.198065,0.00000),complex(0.0705800,0.00000),complex(0.0705800,0.00000),complex(-0.198065,0.00000),complex(0.288671,0.00000),complex(-0.329978,0.00000),complex(0.322812,0.00000),complex(-0.278987,0.00000),complex(0.215847,0.00000),complex(-0.150186,0.00000),complex(0.0939011,0.00000),complex(-0.0524742,0.00000),complex(0.0259374,0.00000),complex(-0.0111424,0.00000),complex(0.00403710,0.00000),complex(-0.00116528,0.00000),complex(0.000232415,0.00000)],$
            [complex(0.00000,0.000869617),complex(0.00000,-0.00403710),complex(0.00000,0.0128612),complex(0.00000,-0.0323688),complex(0.00000,0.0680002),complex(0.00000,-0.122528),complex(0.00000,0.191912),complex(0.00000,-0.262240),complex(0.00000,0.310626),complex(0.00000,-0.311807),complex(0.00000,0.249239),complex(0.00000,-0.125986),complex(0.00000,-0.0308602),complex(-0.00000,0.176450),complex(0.00000,-0.264086),complex(0.00000,0.264086),complex(0.00000,-0.176450),complex(0.00000,0.0308602),complex(0.00000,0.125986),complex(-0.00000,-0.249239),complex(0.00000,0.311807),complex(-0.00000,-0.310626),complex(0.00000,0.262240),complex(-0.00000,-0.191912),complex(0.00000,0.122528),complex(-0.00000,-0.0680002),complex(0.00000,0.0323688),complex(-0.00000,-0.0128612),complex(0.00000,0.00403710),complex(-0.00000,-0.000869617)],$
            [complex(-0.00260885,-0.00000),complex(0.0111424,0.00000),complex(-0.0323688,-0.00000),complex(0.0734555,0.00000),complex(-0.137101,-0.00000),complex(0.215040,0.00000),complex(-0.284423,-0.00000),complex(0.312191,0.00000),complex(-0.269209,-0.00000),complex(0.149125,0.00000),complex(0.0191722,0.00000),complex(-0.176381,0.00000),complex(0.257168,0.00000),complex(-0.223503,0.00000),complex(0.0880287,0.00000),complex(0.0880287,0.00000),complex(-0.223503,0.00000),complex(0.257168,0.00000),complex(-0.176381,0.00000),complex(0.0191722,0.00000),complex(0.149125,-0.00000),complex(-0.269209,0.00000),complex(0.312191,-0.00000),complex(-0.284423,0.00000),complex(0.215040,-0.00000),complex(-0.137101,0.00000),complex(0.0734555,-0.00000),complex(-0.0323688,0.00000),complex(0.0111424,-0.00000),complex(-0.00260885,0.00000)],$
            [complex(0.00000,-0.00665129),complex(-0.00000,0.0259374),complex(0.00000,-0.0680002),complex(-0.00000,0.137101),complex(0.00000,-0.222309),complex(-0.00000,0.292507),complex(0.00000,-0.304763),complex(-0.00000,0.227759),complex(0.00000,-0.0690412),complex(0.00000,-0.114325),complex(0.00000,0.236879),complex(0.00000,-0.232255),complex(0.00000,0.0988523),complex(-0.00000,0.0899722),complex(0.00000,-0.224430),complex(0.00000,0.224430),complex(0.00000,-0.0899722),complex(-0.00000,-0.0988523),complex(0.00000,0.232255),complex(-0.00000,-0.236879),complex(0.00000,0.114325),complex(0.00000,0.0690412),complex(0.00000,-0.227759),complex(0.00000,0.304763),complex(0.00000,-0.292507),complex(0.00000,0.222309),complex(0.00000,-0.137101),complex(0.00000,0.0680002),complex(0.00000,-0.0259374),complex(0.00000,0.00665129)],$
            [complex(0.0148727,0.00000),complex(-0.0524742,0.00000),complex(0.122528,0.00000),complex(-0.215040,0.00000),complex(0.292507,0.00000),complex(-0.300943,0.00000),complex(0.203966,0.00000),complex(-0.0208095,0.00000),complex(-0.165278,-0.00000),complex(0.248503,0.00000),complex(-0.173196,-0.00000),complex(-0.0154696,0.00000),complex(0.190365,0.00000),complex(-0.228008,0.00000),complex(0.100368,0.00000),complex(0.100368,0.00000),complex(-0.228008,0.00000),complex(0.190365,0.00000),complex(-0.0154696,0.00000),complex(-0.173196,0.00000),complex(0.248503,-0.00000),complex(-0.165278,0.00000),complex(-0.0208095,0.00000),complex(0.203966,0.00000),complex(-0.300943,0.00000),complex(0.292507,0.00000),complex(-0.215040,0.00000),complex(0.122528,0.00000),complex(-0.0524742,0.00000),complex(0.0148727,0.00000)],$
            [complex(0.00000,0.0297455),complex(0.00000,-0.0939011),complex(0.00000,0.191912),complex(0.00000,-0.284423),complex(0.00000,0.304763),complex(0.00000,-0.203966),complex(0.00000,0.00500637),complex(-0.00000,0.186190),complex(0.00000,-0.243377),complex(-0.00000,0.121279),complex(0.00000,0.0908018),complex(0.00000,-0.225414),complex(0.00000,0.171419),complex(-0.00000,0.0268245),complex(0.00000,-0.200736),complex(0.00000,0.200736),complex(0.00000,-0.0268245),complex(-0.00000,-0.171419),complex(0.00000,0.225414),complex(-0.00000,-0.0908018),complex(0.00000,-0.121279),complex(0.00000,0.243377),complex(0.00000,-0.186190),complex(-0.00000,-0.00500637),complex(0.00000,0.203966),complex(-0.00000,-0.304763),complex(0.00000,0.284423),complex(-0.00000,-0.191912),complex(0.00000,0.0939011),complex(-0.00000,-0.0297455)],$
            [complex(-0.0539183,-0.00000),complex(0.150186,0.00000),complex(-0.262240,-0.00000),complex(0.312191,0.00000),complex(-0.227759,-0.00000),complex(0.0208095,0.00000),complex(0.186190,0.00000),complex(-0.239788,0.00000),complex(0.0930411,0.00000),complex(0.129878,0.00000),complex(-0.228203,-0.00000),complex(0.109726,0.00000),complex(0.112486,0.00000),complex(-0.221977,0.00000),complex(0.110742,0.00000),complex(0.110742,0.00000),complex(-0.221977,0.00000),complex(0.112486,0.00000),complex(0.109726,-0.00000),complex(-0.228203,0.00000),complex(0.129878,-0.00000),complex(0.0930411,0.00000),complex(-0.239788,0.00000),complex(0.186190,0.00000),complex(0.0208095,-0.00000),complex(-0.227759,0.00000),complex(0.312191,-0.00000),complex(-0.262240,0.00000),complex(0.150186,-0.00000),complex(-0.0539183,0.00000)],$
            [complex(0.00000,-0.0894133),complex(-0.00000,0.215847),complex(0.00000,-0.310626),complex(-0.00000,0.269209),complex(0.00000,-0.0690412),complex(0.00000,-0.165278),complex(0.00000,0.243377),complex(0.00000,-0.0930411),complex(0.00000,-0.141603),complex(-0.00000,0.223685),complex(0.00000,-0.0679667),complex(0.00000,-0.157698),complex(0.00000,0.206347),complex(0.00000,-0.0245405),complex(0.00000,-0.183644),complex(0.00000,0.183644),complex(0.00000,0.0245405),complex(-0.00000,-0.206347),complex(0.00000,0.157698),complex(0.00000,0.0679667),complex(0.00000,-0.223685),complex(0.00000,0.141603),complex(0.00000,0.0930411),complex(-0.00000,-0.243377),complex(0.00000,0.165278),complex(0.00000,0.0690412),complex(0.00000,-0.269209),complex(0.00000,0.310626),complex(0.00000,-0.215847),complex(0.00000,0.0894133)],$
            [complex(0.136581,0.00000),complex(-0.278987,0.00000),complex(0.311807,0.00000),complex(-0.149125,0.00000),complex(-0.114325,-0.00000),complex(0.248503,0.00000),complex(-0.121279,-0.00000),complex(-0.129878,0.00000),complex(0.223685,0.00000),complex(-0.0536459,0.00000),complex(-0.175720,-0.00000),complex(0.186181,0.00000),complex(0.0335012,0.00000),complex(-0.208852,0.00000),complex(0.120223,0.00000),complex(0.120223,0.00000),complex(-0.208852,0.00000),complex(0.0335012,0.00000),complex(0.186181,-0.00000),complex(-0.175720,0.00000),complex(-0.0536459,0.00000),complex(0.223685,0.00000),complex(-0.129878,0.00000),complex(-0.121279,0.00000),complex(0.248503,-0.00000),complex(-0.114325,0.00000),complex(-0.149125,0.00000),complex(0.311807,0.00000),complex(-0.278987,0.00000),complex(0.136581,0.00000)],$
            [complex(0.00000,0.193155),complex(0.00000,-0.322812),complex(0.00000,0.249239),complex(-0.00000,0.0191722),complex(0.00000,-0.236879),complex(-0.00000,0.173196),complex(0.00000,0.0908018),complex(0.00000,-0.228203),complex(0.00000,0.0679667),complex(-0.00000,0.175720),complex(0.00000,-0.177899),complex(0.00000,-0.0611455),complex(0.00000,0.212436),complex(0.00000,-0.0681603),complex(0.00000,-0.170022),complex(0.00000,0.170022),complex(0.00000,0.0681603),complex(-0.00000,-0.212436),complex(0.00000,0.0611455),complex(0.00000,0.177899),complex(0.00000,-0.175720),complex(-0.00000,-0.0679667),complex(0.00000,0.228203),complex(-0.00000,-0.0908018),complex(0.00000,-0.173196),complex(0.00000,0.236879),complex(0.00000,-0.0191722),complex(-0.00000,-0.249239),complex(0.00000,0.322812),complex(-0.00000,-0.193155)],$
            [complex(-0.253855,-0.00000),complex(0.329978,0.00000),complex(-0.125986,-0.00000),complex(-0.176381,0.00000),complex(0.232255,0.00000),complex(0.0154696,0.00000),complex(-0.225414,-0.00000),complex(0.109726,0.00000),complex(0.157698,0.00000),complex(-0.186181,0.00000),complex(-0.0611455,-0.00000),complex(0.211735,0.00000),complex(-0.0407007,-0.00000),complex(-0.190161,0.00000),complex(0.129367,0.00000),complex(0.129367,0.00000),complex(-0.190161,0.00000),complex(-0.0407007,0.00000),complex(0.211735,-0.00000),complex(-0.0611455,0.00000),complex(-0.186181,0.00000),complex(0.157698,0.00000),complex(0.109726,-0.00000),complex(-0.225414,0.00000),complex(0.0154696,-0.00000),complex(0.232255,0.00000),complex(-0.176381,0.00000),complex(-0.125986,0.00000),complex(0.329978,-0.00000),complex(-0.253855,0.00000)],$
            [complex(0.00000,-0.310908),complex(-0.00000,0.288671),complex(0.00000,0.0308602),complex(0.00000,-0.257168),complex(0.00000,0.0988523),complex(-0.00000,0.190365),complex(0.00000,-0.171419),complex(0.00000,-0.112486),complex(0.00000,0.206347),complex(-0.00000,0.0335012),complex(0.00000,-0.212436),complex(-0.00000,0.0407007),complex(0.00000,0.195119),complex(0.00000,-0.105863),complex(0.00000,-0.158442),complex(0.00000,0.158442),complex(0.00000,0.105863),complex(-0.00000,-0.195119),complex(0.00000,-0.0407007),complex(0.00000,0.212436),complex(0.00000,-0.0335012),complex(-0.00000,-0.206347),complex(0.00000,0.112486),complex(0.00000,0.171419),complex(0.00000,-0.190365),complex(-0.00000,-0.0988523),complex(0.00000,0.257168),complex(-0.00000,-0.0308602),complex(0.00000,-0.288671),complex(0.00000,0.310908)],$
            [complex(0.355537,0.00000),complex(-0.198065,0.00000),complex(-0.176450,-0.00000),complex(0.223503,0.00000),complex(0.0899722,0.00000),complex(-0.228008,0.00000),complex(-0.0268245,-0.00000),complex(0.221977,0.00000),complex(-0.0245405,-0.00000),complex(-0.208852,0.00000),complex(0.0681603,0.00000),complex(0.190161,0.00000),complex(-0.105863,-0.00000),complex(-0.166635,0.00000),complex(0.138553,0.00000),complex(0.138553,0.00000),complex(-0.166635,0.00000),complex(-0.105863,0.00000),complex(0.190161,-0.00000),complex(0.0681603,0.00000),complex(-0.208852,0.00000),complex(-0.0245405,0.00000),complex(0.221977,-0.00000),complex(-0.0268245,0.00000),complex(-0.228008,0.00000),complex(0.0899722,0.00000),complex(0.223503,-0.00000),complex(-0.176450,0.00000),complex(-0.198065,0.00000),complex(0.355537,0.00000)],$
            [complex(0.00000,0.380085),complex(0.00000,-0.0705800),complex(0.00000,-0.264086),complex(-0.00000,0.0880287),complex(0.00000,0.224430),complex(0.00000,-0.100368),complex(0.00000,-0.200736),complex(-0.00000,0.110742),complex(0.00000,0.183644),complex(0.00000,-0.120223),complex(0.00000,-0.170022),complex(-0.00000,0.129367),complex(0.00000,0.158442),complex(0.00000,-0.138553),complex(0.00000,-0.148119),complex(0.00000,0.148119),complex(0.00000,0.138553),complex(-0.00000,-0.158442),complex(0.00000,-0.129367),complex(0.00000,0.170022),complex(0.00000,0.120223),complex(-0.00000,-0.183644),complex(0.00000,-0.110742),complex(0.00000,0.200736),complex(0.00000,0.100368),complex(-0.00000,-0.224430),complex(0.00000,-0.0880287),complex(0.00000,0.264086),complex(0.00000,0.0705800),complex(-0.00000,-0.380085)],$
            [complex(-0.380085,-0.00000),complex(-0.0705800,0.00000),complex(0.264086,0.00000),complex(0.0880287,0.00000),complex(-0.224430,-0.00000),complex(-0.100368,0.00000),complex(0.200736,0.00000),complex(0.110742,0.00000),complex(-0.183644,-0.00000),complex(-0.120223,0.00000),complex(0.170022,0.00000),complex(0.129367,0.00000),complex(-0.158442,-0.00000),complex(-0.138553,0.00000),complex(0.148119,0.00000),complex(0.148119,0.00000),complex(-0.138553,0.00000),complex(-0.158442,0.00000),complex(0.129367,-0.00000),complex(0.170022,0.00000),complex(-0.120223,0.00000),complex(-0.183644,0.00000),complex(0.110742,-0.00000),complex(0.200736,0.00000),complex(-0.100368,0.00000),complex(-0.224430,0.00000),complex(0.0880287,-0.00000),complex(0.264086,0.00000),complex(-0.0705800,0.00000),complex(-0.380085,0.00000)],$
            [complex(0.00000,-0.355537),complex(0.00000,-0.198065),complex(0.00000,0.176450),complex(-0.00000,0.223503),complex(0.00000,-0.0899722),complex(0.00000,-0.228008),complex(0.00000,0.0268245),complex(-0.00000,0.221977),complex(0.00000,0.0245405),complex(0.00000,-0.208852),complex(0.00000,-0.0681603),complex(-0.00000,0.190161),complex(0.00000,0.105863),complex(0.00000,-0.166635),complex(0.00000,-0.138553),complex(0.00000,0.138553),complex(0.00000,0.166635),complex(-0.00000,-0.105863),complex(0.00000,-0.190161),complex(0.00000,0.0681603),complex(0.00000,0.208852),complex(-0.00000,-0.0245405),complex(0.00000,-0.221977),complex(-0.00000,-0.0268245),complex(0.00000,0.228008),complex(0.00000,0.0899722),complex(0.00000,-0.223503),complex(-0.00000,-0.176450),complex(0.00000,0.198065),complex(0.00000,0.355537)],$
            [complex(0.310908,0.00000),complex(0.288671,0.00000),complex(-0.0308602,-0.00000),complex(-0.257168,0.00000),complex(-0.0988523,-0.00000),complex(0.190365,0.00000),complex(0.171419,0.00000),complex(-0.112486,0.00000),complex(-0.206347,-0.00000),complex(0.0335012,0.00000),complex(0.212436,0.00000),complex(0.0407007,0.00000),complex(-0.195119,-0.00000),complex(-0.105863,0.00000),complex(0.158442,0.00000),complex(0.158442,0.00000),complex(-0.105863,0.00000),complex(-0.195119,0.00000),complex(0.0407007,-0.00000),complex(0.212436,0.00000),complex(0.0335012,-0.00000),complex(-0.206347,0.00000),complex(-0.112486,0.00000),complex(0.171419,0.00000),complex(0.190365,-0.00000),complex(-0.0988523,0.00000),complex(-0.257168,0.00000),complex(-0.0308602,0.00000),complex(0.288671,-0.00000),complex(0.310908,0.00000)],$
            [complex(0.00000,0.253855),complex(-0.00000,0.329978),complex(0.00000,0.125986),complex(0.00000,-0.176381),complex(0.00000,-0.232255),complex(-0.00000,0.0154696),complex(0.00000,0.225414),complex(-0.00000,0.109726),complex(0.00000,-0.157698),complex(0.00000,-0.186181),complex(0.00000,0.0611455),complex(-0.00000,0.211735),complex(0.00000,0.0407007),complex(0.00000,-0.190161),complex(0.00000,-0.129367),complex(0.00000,0.129367),complex(0.00000,0.190161),complex(-0.00000,-0.0407007),complex(0.00000,-0.211735),complex(-0.00000,-0.0611455),complex(0.00000,0.186181),complex(0.00000,0.157698),complex(0.00000,-0.109726),complex(-0.00000,-0.225414),complex(0.00000,-0.0154696),complex(0.00000,0.232255),complex(0.00000,0.176381),complex(-0.00000,-0.125986),complex(0.00000,-0.329978),complex(-0.00000,-0.253855)],$
            [complex(-0.193155,-0.00000),complex(-0.322812,0.00000),complex(-0.249239,-0.00000),complex(0.0191722,0.00000),complex(0.236879,0.00000),complex(0.173196,0.00000),complex(-0.0908018,-0.00000),complex(-0.228203,0.00000),complex(-0.0679667,-0.00000),complex(0.175720,0.00000),complex(0.177899,0.00000),complex(-0.0611455,0.00000),complex(-0.212436,-0.00000),complex(-0.0681603,0.00000),complex(0.170022,0.00000),complex(0.170022,0.00000),complex(-0.0681603,0.00000),complex(-0.212436,0.00000),complex(-0.0611455,0.00000),complex(0.177899,0.00000),complex(0.175720,-0.00000),complex(-0.0679667,0.00000),complex(-0.228203,0.00000),complex(-0.0908018,0.00000),complex(0.173196,-0.00000),complex(0.236879,0.00000),complex(0.0191722,-0.00000),complex(-0.249239,0.00000),complex(-0.322812,0.00000),complex(-0.193155,0.00000)],$
            [complex(0.00000,-0.136581),complex(0.00000,-0.278987),complex(0.00000,-0.311807),complex(0.00000,-0.149125),complex(0.00000,0.114325),complex(-0.00000,0.248503),complex(0.00000,0.121279),complex(0.00000,-0.129878),complex(0.00000,-0.223685),complex(0.00000,-0.0536459),complex(0.00000,0.175720),complex(-0.00000,0.186181),complex(0.00000,-0.0335012),complex(0.00000,-0.208852),complex(0.00000,-0.120223),complex(0.00000,0.120223),complex(0.00000,0.208852),complex(0.00000,0.0335012),complex(0.00000,-0.186181),complex(-0.00000,-0.175720),complex(0.00000,0.0536459),complex(0.00000,0.223685),complex(0.00000,0.129878),complex(-0.00000,-0.121279),complex(0.00000,-0.248503),complex(-0.00000,-0.114325),complex(0.00000,0.149125),complex(0.00000,0.311807),complex(0.00000,0.278987),complex(0.00000,0.136581)],$
            [complex(0.0894133,0.00000),complex(0.215847,0.00000),complex(0.310626,0.00000),complex(0.269209,0.00000),complex(0.0690412,0.00000),complex(-0.165278,0.00000),complex(-0.243377,-0.00000),complex(-0.0930411,0.00000),complex(0.141603,0.00000),complex(0.223685,0.00000),complex(0.0679667,0.00000),complex(-0.157698,0.00000),complex(-0.206347,-0.00000),complex(-0.0245405,0.00000),complex(0.183644,0.00000),complex(0.183644,0.00000),complex(-0.0245405,0.00000),complex(-0.206347,0.00000),complex(-0.157698,0.00000),complex(0.0679667,0.00000),complex(0.223685,-0.00000),complex(0.141603,0.00000),complex(-0.0930411,0.00000),complex(-0.243377,0.00000),complex(-0.165278,0.00000),complex(0.0690412,0.00000),complex(0.269209,-0.00000),complex(0.310626,0.00000),complex(0.215847,-0.00000),complex(0.0894133,0.00000)],$
            [complex(0.00000,0.0539183),complex(-0.00000,0.150186),complex(0.00000,0.262240),complex(-0.00000,0.312191),complex(0.00000,0.227759),complex(-0.00000,0.0208095),complex(0.00000,-0.186190),complex(0.00000,-0.239788),complex(0.00000,-0.0930411),complex(-0.00000,0.129878),complex(0.00000,0.228203),complex(-0.00000,0.109726),complex(0.00000,-0.112486),complex(0.00000,-0.221977),complex(0.00000,-0.110742),complex(0.00000,0.110742),complex(0.00000,0.221977),complex(0.00000,0.112486),complex(0.00000,-0.109726),complex(-0.00000,-0.228203),complex(0.00000,-0.129878),complex(0.00000,0.0930411),complex(0.00000,0.239788),complex(0.00000,0.186190),complex(0.00000,-0.0208095),complex(-0.00000,-0.227759),complex(0.00000,-0.312191),complex(-0.00000,-0.262240),complex(0.00000,-0.150186),complex(-0.00000,-0.0539183)],$
            [complex(-0.0297455,-0.00000),complex(-0.0939011,0.00000),complex(-0.191912,-0.00000),complex(-0.284423,0.00000),complex(-0.304763,-0.00000),complex(-0.203966,0.00000),complex(-0.00500637,-0.00000),complex(0.186190,0.00000),complex(0.243377,0.00000),complex(0.121279,0.00000),complex(-0.0908018,-0.00000),complex(-0.225414,0.00000),complex(-0.171419,-0.00000),complex(0.0268245,0.00000),complex(0.200736,0.00000),complex(0.200736,0.00000),complex(0.0268245,-0.00000),complex(-0.171419,0.00000),complex(-0.225414,0.00000),complex(-0.0908018,0.00000),complex(0.121279,-0.00000),complex(0.243377,0.00000),complex(0.186190,-0.00000),complex(-0.00500637,0.00000),complex(-0.203966,0.00000),complex(-0.304763,0.00000),complex(-0.284423,0.00000),complex(-0.191912,0.00000),complex(-0.0939011,0.00000),complex(-0.0297455,0.00000)],$
            [complex(0.00000,-0.0148727),complex(0.00000,-0.0524742),complex(0.00000,-0.122528),complex(0.00000,-0.215040),complex(0.00000,-0.292507),complex(0.00000,-0.300943),complex(0.00000,-0.203966),complex(0.00000,-0.0208095),complex(0.00000,0.165278),complex(-0.00000,0.248503),complex(0.00000,0.173196),complex(0.00000,-0.0154696),complex(0.00000,-0.190365),complex(0.00000,-0.228008),complex(0.00000,-0.100368),complex(0.00000,0.100368),complex(0.00000,0.228008),complex(0.00000,0.190365),complex(0.00000,0.0154696),complex(-0.00000,-0.173196),complex(0.00000,-0.248503),complex(-0.00000,-0.165278),complex(0.00000,0.0208095),complex(0.00000,0.203966),complex(0.00000,0.300943),complex(0.00000,0.292507),complex(0.00000,0.215040),complex(0.00000,0.122528),complex(0.00000,0.0524742),complex(0.00000,0.0148727)],$
            [complex(0.00665129,0.00000),complex(0.0259374,0.00000),complex(0.0680002,0.00000),complex(0.137101,0.00000),complex(0.222309,0.00000),complex(0.292507,0.00000),complex(0.304763,0.00000),complex(0.227759,0.00000),complex(0.0690412,0.00000),complex(-0.114325,0.00000),complex(-0.236879,-0.00000),complex(-0.232255,0.00000),complex(-0.0988523,-0.00000),complex(0.0899722,0.00000),complex(0.224430,0.00000),complex(0.224430,0.00000),complex(0.0899722,-0.00000),complex(-0.0988523,0.00000),complex(-0.232255,0.00000),complex(-0.236879,0.00000),complex(-0.114325,0.00000),complex(0.0690412,0.00000),complex(0.227759,-0.00000),complex(0.304763,0.00000),complex(0.292507,-0.00000),complex(0.222309,0.00000),complex(0.137101,-0.00000),complex(0.0680002,0.00000),complex(0.0259374,-0.00000),complex(0.00665129,0.00000)],$
            [complex(0.00000,0.00260885),complex(-0.00000,0.0111424),complex(0.00000,0.0323688),complex(-0.00000,0.0734555),complex(0.00000,0.137101),complex(-0.00000,0.215040),complex(0.00000,0.284423),complex(-0.00000,0.312191),complex(0.00000,0.269209),complex(-0.00000,0.149125),complex(0.00000,-0.0191722),complex(0.00000,-0.176381),complex(0.00000,-0.257168),complex(0.00000,-0.223503),complex(0.00000,-0.0880287),complex(0.00000,0.0880287),complex(0.00000,0.223503),complex(0.00000,0.257168),complex(0.00000,0.176381),complex(0.00000,0.0191722),complex(0.00000,-0.149125),complex(-0.00000,-0.269209),complex(0.00000,-0.312191),complex(-0.00000,-0.284423),complex(0.00000,-0.215040),complex(-0.00000,-0.137101),complex(0.00000,-0.0734555),complex(-0.00000,-0.0323688),complex(0.00000,-0.0111424),complex(-0.00000,-0.00260885)],$
            [complex(-0.000869617,-0.00000),complex(-0.00403710,0.00000),complex(-0.0128612,-0.00000),complex(-0.0323688,0.00000),complex(-0.0680002,-0.00000),complex(-0.122528,0.00000),complex(-0.191912,-0.00000),complex(-0.262240,0.00000),complex(-0.310626,-0.00000),complex(-0.311807,0.00000),complex(-0.249239,-0.00000),complex(-0.125986,0.00000),complex(0.0308602,0.00000),complex(0.176450,0.00000),complex(0.264086,0.00000),complex(0.264086,0.00000),complex(0.176450,-0.00000),complex(0.0308602,0.00000),complex(-0.125986,0.00000),complex(-0.249239,0.00000),complex(-0.311807,0.00000),complex(-0.310626,0.00000),complex(-0.262240,0.00000),complex(-0.191912,0.00000),complex(-0.122528,0.00000),complex(-0.0680002,0.00000),complex(-0.0323688,0.00000),complex(-0.0128612,0.00000),complex(-0.00403710,0.00000),complex(-0.000869617,0.00000)],$
            [complex(0.00000,-0.000232415),complex(0.00000,-0.00116528),complex(0.00000,-0.00403710),complex(0.00000,-0.0111424),complex(0.00000,-0.0259374),complex(0.00000,-0.0524742),complex(0.00000,-0.0939011),complex(0.00000,-0.150186),complex(0.00000,-0.215847),complex(0.00000,-0.278987),complex(0.00000,-0.322812),complex(0.00000,-0.329978),complex(0.00000,-0.288671),complex(0.00000,-0.198065),complex(0.00000,-0.0705800),complex(0.00000,0.0705800),complex(0.00000,0.198065),complex(0.00000,0.288671),complex(0.00000,0.329978),complex(0.00000,0.322812),complex(0.00000,0.278987),complex(0.00000,0.215847),complex(0.00000,0.150186),complex(0.00000,0.0939011),complex(0.00000,0.0524742),complex(0.00000,0.0259374),complex(0.00000,0.0111424),complex(0.00000,0.00403710),complex(0.00000,0.00116528),complex(0.00000,0.000232415)],$
            [complex(4.31584e-05,0.00000),complex(0.000232415,0.00000),complex(0.000869617,0.00000),complex(0.00260885,0.00000),complex(0.00665129,0.00000),complex(0.0148727,0.00000),complex(0.0297455,0.00000),complex(0.0539183,0.00000),complex(0.0894133,0.00000),complex(0.136581,0.00000),complex(0.193155,0.00000),complex(0.253855,0.00000),complex(0.310908,0.00000),complex(0.355537,0.00000),complex(0.380085,0.00000),complex(0.380085,0.00000),complex(0.355537,-0.00000),complex(0.310908,0.00000),complex(0.253855,-0.00000),complex(0.193155,0.00000),complex(0.136581,-0.00000),complex(0.0894133,0.00000),complex(0.0539183,-0.00000),complex(0.0297455,0.00000),complex(0.0148727,-0.00000),complex(0.00665129,0.00000),complex(0.00260885,-0.00000),complex(0.000869617,0.00000),complex(0.000232415,-0.00000),complex(4.31584e-05,0.00000)]$
            ]
          end


      28: begin
            matrix[406:434,406:434]=[$
            [complex(6.10352e-05,0.00000),complex(-0.000322968,-0.00000),complex(0.00118666,0.00000),complex(-0.00349343,-0.00000),complex(0.00873357,0.00000),complex(-0.0191343,-0.00000),complex(0.0374629,0.00000),complex(-0.0664145,-0.00000),complex(0.107604,0.00000),complex(-0.160406,-0.00000),complex(0.221105,0.00000),complex(-0.282838,-0.00000),complex(0.336645,0.00000),complex(-0.373474,-0.00000),complex(0.386582,0.00000),complex(-0.373474,0.00000),complex(0.336645,0.00000),complex(-0.282838,0.00000),complex(0.221105,0.00000),complex(-0.160406,0.00000),complex(0.107604,0.00000),complex(-0.0664145,0.00000),complex(0.0374629,0.00000),complex(-0.0191343,0.00000),complex(0.00873357,0.00000),complex(-0.00349343,0.00000),complex(0.00118666,0.00000),complex(-0.000322968,0.00000),complex(6.10352e-05,0.00000)],$
            [complex(0.00000,0.000322968),complex(0.00000,-0.00158691),complex(0.00000,0.00538218),complex(0.00000,-0.0145243),complex(0.00000,0.0330098),complex(0.00000,-0.0650888),complex(0.00000,0.113277),complex(0.00000,-0.175716),complex(0.00000,0.244023),complex(0.00000,-0.303140),complex(0.00000,0.334279),complex(0.00000,-0.320709),complex(0.00000,0.254480),complex(0.00000,-0.141160),0,complex(0.00000,0.141160),complex(0.00000,-0.254480),complex(0.00000,0.320709),complex(0.00000,-0.334279),complex(0.00000,0.303140),complex(0.00000,-0.244023),complex(0.00000,0.175716),complex(0.00000,-0.113277),complex(0.00000,0.0650888),complex(0.00000,-0.0330098),complex(0.00000,0.0145243),complex(0.00000,-0.00538218),complex(0.00000,0.00158691),complex(0.00000,-0.000322968)],$
            [complex(-0.00118666,0.00000),complex(0.00538218,0.00000),complex(-0.0167236,0.00000),complex(0.0409676,0.00000),complex(-0.0835524,0.00000),complex(0.145656,0.00000),complex(-0.219664,0.00000),complex(0.286944,0.00000),complex(-0.321004,0.00000),complex(0.297015,0.00000),complex(-0.204703,0.00000),complex(0.0581906,0.00000),complex(0.103891,0.00000),complex(-0.230513,-0.00000),complex(0.278371,0.00000),complex(-0.230513,0.00000),complex(0.103891,0.00000),complex(0.0581906,-0.00000),complex(-0.204703,0.00000),complex(0.297015,-0.00000),complex(-0.321004,0.00000),complex(0.286944,-0.00000),complex(-0.219664,0.00000),complex(0.145656,-0.00000),complex(-0.0835524,0.00000),complex(0.0409676,-0.00000),complex(-0.0167236,0.00000),complex(0.00538218,-0.00000),complex(-0.00118666,0.00000)],$
            [complex(0.00000,-0.00349343),complex(-0.00000,0.0145243),complex(0.00000,-0.0409676),complex(-0.00000,0.0899658),complex(0.00000,-0.161743),complex(-0.00000,0.242704),complex(0.00000,-0.303701),complex(-0.00000,0.308655),complex(0.00000,-0.233119),complex(-0.00000,0.0840758),complex(0.00000,0.0927124),complex(0.00000,-0.227313),complex(0.00000,0.258793),complex(0.00000,-0.169653),0,complex(0.00000,0.169653),complex(0.00000,-0.258793),complex(0.00000,0.227313),complex(0.00000,-0.0927124),complex(-0.00000,-0.0840758),complex(0.00000,0.233119),complex(-0.00000,-0.308655),complex(0.00000,0.303701),complex(-0.00000,-0.242704),complex(0.00000,0.161743),complex(-0.00000,-0.0899658),complex(0.00000,0.0409676),complex(-0.00000,-0.0145243),complex(0.00000,0.00349343)],$
            [complex(0.00873357,0.00000),complex(-0.0330098,-0.00000),complex(0.0835524,0.00000),complex(-0.161743,-0.00000),complex(0.249695,0.00000),complex(-0.308228,-0.00000),complex(0.291920,0.00000),complex(-0.178695,-0.00000),complex(-0.00375998,0.00000),complex(0.178241,0.00000),complex(-0.254959,0.00000),complex(0.187780,0.00000),complex(-0.0117633,0.00000),complex(-0.169653,-0.00000),complex(0.245851,0.00000),complex(-0.169653,0.00000),complex(-0.0117633,0.00000),complex(0.187780,-0.00000),complex(-0.254959,0.00000),complex(0.178241,-0.00000),complex(-0.00375998,0.00000),complex(-0.178695,0.00000),complex(0.291920,0.00000),complex(-0.308228,0.00000),complex(0.249695,0.00000),complex(-0.161743,0.00000),complex(0.0835524,0.00000),complex(-0.0330098,0.00000),complex(0.00873357,0.00000)],$
            [complex(0.00000,0.0191343),complex(0.00000,-0.0650888),complex(0.00000,0.145656),complex(0.00000,-0.242704),complex(0.00000,0.308228),complex(0.00000,-0.284912),complex(0.00000,0.149136),complex(-0.00000,0.0533865),complex(0.00000,-0.216926),complex(-0.00000,0.239461),complex(0.00000,-0.101561),complex(0.00000,-0.104656),complex(0.00000,0.231949),complex(0.00000,-0.185846),0,complex(0.00000,0.185846),complex(0.00000,-0.231949),complex(0.00000,0.104656),complex(0.00000,0.101561),complex(-0.00000,-0.239461),complex(0.00000,0.216926),complex(-0.00000,-0.0533865),complex(0.00000,-0.149136),complex(0.00000,0.284912),complex(0.00000,-0.308228),complex(0.00000,0.242704),complex(0.00000,-0.145656),complex(0.00000,0.0650888),complex(0.00000,-0.0191343)],$
            [complex(-0.0374629,0.00000),complex(0.113277,0.00000),complex(-0.219664,0.00000),complex(0.303701,0.00000),complex(-0.291920,0.00000),complex(0.149136,0.00000),complex(0.0690918,0.00000),complex(-0.230258,-0.00000),complex(0.218086,0.00000),complex(-0.0376325,-0.00000),complex(-0.168587,0.00000),complex(0.228559,0.00000),complex(-0.0899488,0.00000),complex(-0.126562,-0.00000),complex(0.229257,0.00000),complex(-0.126562,0.00000),complex(-0.0899488,0.00000),complex(0.228559,-0.00000),complex(-0.168587,0.00000),complex(-0.0376325,0.00000),complex(0.218086,0.00000),complex(-0.230258,0.00000),complex(0.0690918,0.00000),complex(0.149136,-0.00000),complex(-0.291920,0.00000),complex(0.303701,-0.00000),complex(-0.219664,0.00000),complex(0.113277,-0.00000),complex(-0.0374629,0.00000)],$
            [complex(0.00000,-0.0664145),complex(-0.00000,0.175716),complex(0.00000,-0.286944),complex(-0.00000,0.308655),complex(0.00000,-0.178695),complex(0.00000,-0.0533865),complex(0.00000,0.230258),complex(0.00000,-0.209229),complex(0.00000,0.00553774),complex(-0.00000,0.196355),complex(0.00000,-0.204822),complex(-0.00000,0.0114369),complex(0.00000,0.190577),complex(0.00000,-0.196324),0,complex(0.00000,0.196324),complex(0.00000,-0.190577),complex(-0.00000,-0.0114369),complex(0.00000,0.204822),complex(-0.00000,-0.196355),complex(0.00000,-0.00553774),complex(0.00000,0.209229),complex(0.00000,-0.230258),complex(0.00000,0.0533865),complex(0.00000,0.178695),complex(-0.00000,-0.308655),complex(0.00000,0.286944),complex(-0.00000,-0.175716),complex(0.00000,0.0664145)],$
            [complex(0.107604,0.00000),complex(-0.244023,-0.00000),complex(0.321004,0.00000),complex(-0.233119,-0.00000),complex(-0.00375998,0.00000),complex(0.216926,0.00000),complex(-0.218086,0.00000),complex(0.00553774,0.00000),complex(0.203674,0.00000),complex(-0.187522,-0.00000),complex(-0.0349909,0.00000),complex(0.213535,0.00000),complex(-0.144933,0.00000),complex(-0.0908803,-0.00000),complex(0.219497,0.00000),complex(-0.0908803,0.00000),complex(-0.144933,0.00000),complex(0.213535,-0.00000),complex(-0.0349909,0.00000),complex(-0.187522,0.00000),complex(0.203674,0.00000),complex(0.00553774,-0.00000),complex(-0.218086,0.00000),complex(0.216926,-0.00000),complex(-0.00375998,0.00000),complex(-0.233119,0.00000),complex(0.321004,0.00000),complex(-0.244023,0.00000),complex(0.107604,0.00000)],$
            [complex(0.00000,0.160406),complex(0.00000,-0.303140),complex(0.00000,0.297015),complex(0.00000,-0.0840758),complex(0.00000,-0.178241),complex(-0.00000,0.239461),complex(0.00000,-0.0376325),complex(0.00000,-0.196355),complex(0.00000,0.187522),complex(-0.00000,0.0499268),complex(0.00000,-0.218741),complex(-0.00000,0.106545),complex(0.00000,0.140904),complex(0.00000,-0.203215),0,complex(0.00000,0.203215),complex(0.00000,-0.140904),complex(-0.00000,-0.106545),complex(0.00000,0.218741),complex(-0.00000,-0.0499268),complex(0.00000,-0.187522),complex(0.00000,0.196355),complex(0.00000,0.0376325),complex(-0.00000,-0.239461),complex(0.00000,0.178241),complex(0.00000,0.0840758),complex(0.00000,-0.297015),complex(0.00000,0.303140),complex(0.00000,-0.160406)],$
            [complex(-0.221105,0.00000),complex(0.334279,0.00000),complex(-0.204703,0.00000),complex(-0.0927124,-0.00000),complex(0.254959,0.00000),complex(-0.101561,-0.00000),complex(-0.168587,0.00000),complex(0.204822,0.00000),complex(0.0349909,0.00000),complex(-0.218741,-0.00000),complex(0.0928955,0.00000),complex(0.161462,0.00000),complex(-0.181956,0.00000),complex(-0.0589710,-0.00000),complex(0.213643,0.00000),complex(-0.0589710,0.00000),complex(-0.181956,0.00000),complex(0.161462,-0.00000),complex(0.0928955,0.00000),complex(-0.218741,0.00000),complex(0.0349909,0.00000),complex(0.204822,-0.00000),complex(-0.168587,0.00000),complex(-0.101561,0.00000),complex(0.254959,0.00000),complex(-0.0927124,0.00000),complex(-0.204703,0.00000),complex(0.334279,-0.00000),complex(-0.221105,0.00000)],$
            [complex(0.00000,-0.282838),complex(-0.00000,0.320709),complex(0.00000,-0.0581906),complex(0.00000,-0.227313),complex(0.00000,0.187780),complex(-0.00000,0.104656),complex(0.00000,-0.228559),complex(-0.00000,0.0114369),complex(0.00000,0.213535),complex(0.00000,-0.106545),complex(0.00000,-0.161462),complex(-0.00000,0.173218),complex(0.00000,0.0863039),complex(0.00000,-0.207449),0,complex(0.00000,0.207449),complex(0.00000,-0.0863039),complex(-0.00000,-0.173218),complex(0.00000,0.161462),complex(0.00000,0.106545),complex(0.00000,-0.213535),complex(-0.00000,-0.0114369),complex(0.00000,0.228559),complex(-0.00000,-0.104656),complex(0.00000,-0.187780),complex(0.00000,0.227313),complex(0.00000,0.0581906),complex(-0.00000,-0.320709),complex(0.00000,0.282838)],$
            [complex(0.336645,0.00000),complex(-0.254480,-0.00000),complex(-0.103891,0.00000),complex(0.258793,0.00000),complex(-0.0117633,0.00000),complex(-0.231949,-0.00000),complex(0.0899488,0.00000),complex(0.190577,0.00000),complex(-0.144933,0.00000),complex(-0.140904,-0.00000),complex(0.181956,0.00000),complex(0.0863039,0.00000),complex(-0.203430,0.00000),complex(-0.0290486,-0.00000),complex(0.210477,0.00000),complex(-0.0290486,0.00000),complex(-0.203430,0.00000),complex(0.0863039,-0.00000),complex(0.181956,0.00000),complex(-0.140904,0.00000),complex(-0.144933,0.00000),complex(0.190577,-0.00000),complex(0.0899488,0.00000),complex(-0.231949,0.00000),complex(-0.0117633,0.00000),complex(0.258793,-0.00000),complex(-0.103891,0.00000),complex(-0.254480,0.00000),complex(0.336645,0.00000)],$
            [complex(0.00000,0.373474),complex(0.00000,-0.141160),complex(0.00000,-0.230513),complex(-0.00000,0.169653),complex(0.00000,0.169653),complex(0.00000,-0.185846),complex(0.00000,-0.126562),complex(-0.00000,0.196324),complex(0.00000,0.0908803),complex(0.00000,-0.203215),complex(0.00000,-0.0589710),complex(-0.00000,0.207449),complex(0.00000,0.0290486),complex(0.00000,-0.209473),0,complex(0.00000,0.209473),complex(0.00000,-0.0290486),complex(-0.00000,-0.207449),complex(0.00000,0.0589710),complex(0.00000,0.203215),complex(0.00000,-0.0908803),complex(-0.00000,-0.196324),complex(0.00000,0.126562),complex(0.00000,0.185846),complex(0.00000,-0.169653),complex(-0.00000,-0.169653),complex(0.00000,0.230513),complex(0.00000,0.141160),complex(0.00000,-0.373474)],$
            [complex(-0.386582,0.00000),complex(-1.89958e-16,-0.00000),complex(0.278371,0.00000),0,complex(-0.245851,0.00000),complex(-5.97874e-15,-0.00000),complex(0.229257,0.00000),complex(-2.34986e-14,-0.00000),complex(-0.219497,0.00000),0,complex(0.213643,0.00000),0,complex(-0.210477,0.00000),0,complex(0.209473,0.00000),0,complex(-0.210477,0.00000),0,complex(0.213643,0.00000),0,complex(-0.219497,0.00000),complex(-1.55640e-14,0.00000),complex(0.229257,0.00000),complex(-2.46183e-15,0.00000),complex(-0.245851,0.00000),0,complex(0.278371,0.00000),complex(-1.89958e-16,0.00000),complex(-0.386582,0.00000)],$
            [complex(0.00000,-0.373474),complex(0.00000,-0.141160),complex(0.00000,0.230513),complex(-0.00000,0.169653),complex(0.00000,-0.169653),complex(0.00000,-0.185846),complex(0.00000,0.126562),complex(-0.00000,0.196324),complex(0.00000,-0.0908803),complex(0.00000,-0.203215),complex(0.00000,0.0589710),complex(-0.00000,0.207449),complex(0.00000,-0.0290486),complex(0.00000,-0.209473),0,complex(0.00000,0.209473),complex(0.00000,0.0290486),complex(-0.00000,-0.207449),complex(0.00000,-0.0589710),complex(0.00000,0.203215),complex(0.00000,0.0908803),complex(-0.00000,-0.196324),complex(0.00000,-0.126562),complex(0.00000,0.185846),complex(0.00000,0.169653),complex(-0.00000,-0.169653),complex(0.00000,-0.230513),complex(0.00000,0.141160),complex(0.00000,0.373474)],$
            [complex(0.336645,0.00000),complex(0.254480,0.00000),complex(-0.103891,0.00000),complex(-0.258793,-0.00000),complex(-0.0117633,0.00000),complex(0.231949,0.00000),complex(0.0899488,0.00000),complex(-0.190577,-0.00000),complex(-0.144933,0.00000),complex(0.140904,0.00000),complex(0.181956,0.00000),complex(-0.0863039,-0.00000),complex(-0.203430,0.00000),complex(0.0290486,0.00000),complex(0.210477,0.00000),complex(0.0290486,-0.00000),complex(-0.203430,0.00000),complex(-0.0863039,0.00000),complex(0.181956,0.00000),complex(0.140904,-0.00000),complex(-0.144933,0.00000),complex(-0.190577,0.00000),complex(0.0899488,0.00000),complex(0.231949,-0.00000),complex(-0.0117633,0.00000),complex(-0.258793,0.00000),complex(-0.103891,0.00000),complex(0.254480,-0.00000),complex(0.336645,0.00000)],$
            [complex(0.00000,0.282838),complex(-0.00000,0.320709),complex(0.00000,0.0581906),complex(0.00000,-0.227313),complex(0.00000,-0.187780),complex(-0.00000,0.104656),complex(0.00000,0.228559),complex(-0.00000,0.0114369),complex(0.00000,-0.213535),complex(0.00000,-0.106545),complex(0.00000,0.161462),complex(-0.00000,0.173218),complex(0.00000,-0.0863039),complex(0.00000,-0.207449),0,complex(0.00000,0.207449),complex(0.00000,0.0863039),complex(-0.00000,-0.173218),complex(0.00000,-0.161462),complex(0.00000,0.106545),complex(0.00000,0.213535),complex(-0.00000,-0.0114369),complex(0.00000,-0.228559),complex(-0.00000,-0.104656),complex(0.00000,0.187780),complex(0.00000,0.227313),complex(0.00000,-0.0581906),complex(-0.00000,-0.320709),complex(0.00000,-0.282838)],$
            [complex(-0.221105,0.00000),complex(-0.334279,-0.00000),complex(-0.204703,0.00000),complex(0.0927124,0.00000),complex(0.254959,0.00000),complex(0.101561,0.00000),complex(-0.168587,0.00000),complex(-0.204822,-0.00000),complex(0.0349909,0.00000),complex(0.218741,0.00000),complex(0.0928955,0.00000),complex(-0.161462,-0.00000),complex(-0.181956,0.00000),complex(0.0589710,0.00000),complex(0.213643,0.00000),complex(0.0589710,-0.00000),complex(-0.181956,0.00000),complex(-0.161462,0.00000),complex(0.0928955,0.00000),complex(0.218741,-0.00000),complex(0.0349909,0.00000),complex(-0.204822,0.00000),complex(-0.168587,0.00000),complex(0.101561,-0.00000),complex(0.254959,0.00000),complex(0.0927124,-0.00000),complex(-0.204703,0.00000),complex(-0.334279,0.00000),complex(-0.221105,0.00000)],$
            [complex(0.00000,-0.160406),complex(0.00000,-0.303140),complex(0.00000,-0.297015),complex(0.00000,-0.0840758),complex(0.00000,0.178241),complex(-0.00000,0.239461),complex(0.00000,0.0376325),complex(0.00000,-0.196355),complex(0.00000,-0.187522),complex(-0.00000,0.0499268),complex(0.00000,0.218741),complex(-0.00000,0.106545),complex(0.00000,-0.140904),complex(0.00000,-0.203215),0,complex(0.00000,0.203215),complex(0.00000,0.140904),complex(-0.00000,-0.106545),complex(0.00000,-0.218741),complex(-0.00000,-0.0499268),complex(0.00000,0.187522),complex(0.00000,0.196355),complex(0.00000,-0.0376325),complex(-0.00000,-0.239461),complex(0.00000,-0.178241),complex(0.00000,0.0840758),complex(0.00000,0.297015),complex(0.00000,0.303140),complex(0.00000,0.160406)],$
            [complex(0.107604,0.00000),complex(0.244023,0.00000),complex(0.321004,0.00000),complex(0.233119,0.00000),complex(-0.00375998,0.00000),complex(-0.216926,-0.00000),complex(-0.218086,0.00000),complex(-0.00553774,-0.00000),complex(0.203674,0.00000),complex(0.187522,0.00000),complex(-0.0349909,0.00000),complex(-0.213535,-0.00000),complex(-0.144933,0.00000),complex(0.0908803,0.00000),complex(0.219497,0.00000),complex(0.0908803,-0.00000),complex(-0.144933,0.00000),complex(-0.213535,0.00000),complex(-0.0349909,0.00000),complex(0.187522,-0.00000),complex(0.203674,0.00000),complex(-0.00553774,0.00000),complex(-0.218086,0.00000),complex(-0.216926,0.00000),complex(-0.00375998,0.00000),complex(0.233119,-0.00000),complex(0.321004,0.00000),complex(0.244023,-0.00000),complex(0.107604,0.00000)],$
            [complex(0.00000,0.0664145),complex(-0.00000,0.175716),complex(0.00000,0.286944),complex(-0.00000,0.308655),complex(0.00000,0.178695),complex(0.00000,-0.0533865),complex(0.00000,-0.230258),complex(0.00000,-0.209229),complex(0.00000,-0.00553774),complex(-0.00000,0.196355),complex(0.00000,0.204822),complex(-0.00000,0.0114369),complex(0.00000,-0.190577),complex(0.00000,-0.196324),0,complex(0.00000,0.196324),complex(0.00000,0.190577),complex(-0.00000,-0.0114369),complex(0.00000,-0.204822),complex(-0.00000,-0.196355),complex(0.00000,0.00553774),complex(0.00000,0.209229),complex(0.00000,0.230258),complex(0.00000,0.0533865),complex(0.00000,-0.178695),complex(-0.00000,-0.308655),complex(0.00000,-0.286944),complex(-0.00000,-0.175716),complex(0.00000,-0.0664145)],$
            [complex(-0.0374629,0.00000),complex(-0.113277,-0.00000),complex(-0.219664,0.00000),complex(-0.303701,-0.00000),complex(-0.291920,0.00000),complex(-0.149136,-0.00000),complex(0.0690918,0.00000),complex(0.230258,0.00000),complex(0.218086,0.00000),complex(0.0376325,0.00000),complex(-0.168587,0.00000),complex(-0.228559,-0.00000),complex(-0.0899488,0.00000),complex(0.126562,0.00000),complex(0.229257,0.00000),complex(0.126562,-0.00000),complex(-0.0899488,0.00000),complex(-0.228559,0.00000),complex(-0.168587,0.00000),complex(0.0376325,-0.00000),complex(0.218086,0.00000),complex(0.230258,-0.00000),complex(0.0690918,0.00000),complex(-0.149136,0.00000),complex(-0.291920,0.00000),complex(-0.303701,0.00000),complex(-0.219664,0.00000),complex(-0.113277,0.00000),complex(-0.0374629,0.00000)],$
            [complex(0.00000,-0.0191343),complex(0.00000,-0.0650888),complex(0.00000,-0.145656),complex(0.00000,-0.242704),complex(0.00000,-0.308228),complex(0.00000,-0.284912),complex(0.00000,-0.149136),complex(-0.00000,0.0533865),complex(0.00000,0.216926),complex(-0.00000,0.239461),complex(0.00000,0.101561),complex(0.00000,-0.104656),complex(0.00000,-0.231949),complex(0.00000,-0.185846),0,complex(0.00000,0.185846),complex(0.00000,0.231949),complex(0.00000,0.104656),complex(0.00000,-0.101561),complex(-0.00000,-0.239461),complex(0.00000,-0.216926),complex(-0.00000,-0.0533865),complex(0.00000,0.149136),complex(0.00000,0.284912),complex(0.00000,0.308228),complex(0.00000,0.242704),complex(0.00000,0.145656),complex(0.00000,0.0650888),complex(0.00000,0.0191343)],$
            [complex(0.00873357,0.00000),complex(0.0330098,0.00000),complex(0.0835524,0.00000),complex(0.161743,0.00000),complex(0.249695,0.00000),complex(0.308228,0.00000),complex(0.291920,0.00000),complex(0.178695,0.00000),complex(-0.00375998,0.00000),complex(-0.178241,-0.00000),complex(-0.254959,0.00000),complex(-0.187780,-0.00000),complex(-0.0117633,0.00000),complex(0.169653,0.00000),complex(0.245851,0.00000),complex(0.169653,-0.00000),complex(-0.0117633,0.00000),complex(-0.187780,0.00000),complex(-0.254959,0.00000),complex(-0.178241,0.00000),complex(-0.00375998,0.00000),complex(0.178695,-0.00000),complex(0.291920,0.00000),complex(0.308228,-0.00000),complex(0.249695,0.00000),complex(0.161743,-0.00000),complex(0.0835524,0.00000),complex(0.0330098,-0.00000),complex(0.00873357,0.00000)],$
            [complex(0.00000,0.00349343),complex(-0.00000,0.0145243),complex(0.00000,0.0409676),complex(-0.00000,0.0899658),complex(0.00000,0.161743),complex(-0.00000,0.242704),complex(0.00000,0.303701),complex(-0.00000,0.308655),complex(0.00000,0.233119),complex(-0.00000,0.0840758),complex(0.00000,-0.0927124),complex(0.00000,-0.227313),complex(0.00000,-0.258793),complex(0.00000,-0.169653),0,complex(0.00000,0.169653),complex(0.00000,0.258793),complex(0.00000,0.227313),complex(0.00000,0.0927124),complex(-0.00000,-0.0840758),complex(0.00000,-0.233119),complex(-0.00000,-0.308655),complex(0.00000,-0.303701),complex(-0.00000,-0.242704),complex(0.00000,-0.161743),complex(-0.00000,-0.0899658),complex(0.00000,-0.0409676),complex(-0.00000,-0.0145243),complex(0.00000,-0.00349343)],$
            [complex(-0.00118666,0.00000),complex(-0.00538218,-0.00000),complex(-0.0167236,0.00000),complex(-0.0409676,-0.00000),complex(-0.0835524,0.00000),complex(-0.145656,-0.00000),complex(-0.219664,0.00000),complex(-0.286944,-0.00000),complex(-0.321004,0.00000),complex(-0.297015,-0.00000),complex(-0.204703,0.00000),complex(-0.0581906,-0.00000),complex(0.103891,0.00000),complex(0.230513,0.00000),complex(0.278371,0.00000),complex(0.230513,-0.00000),complex(0.103891,0.00000),complex(-0.0581906,0.00000),complex(-0.204703,0.00000),complex(-0.297015,0.00000),complex(-0.321004,0.00000),complex(-0.286944,0.00000),complex(-0.219664,0.00000),complex(-0.145656,0.00000),complex(-0.0835524,0.00000),complex(-0.0409676,0.00000),complex(-0.0167236,0.00000),complex(-0.00538218,0.00000),complex(-0.00118666,0.00000)],$
            [complex(0.00000,-0.000322968),complex(0.00000,-0.00158691),complex(0.00000,-0.00538218),complex(0.00000,-0.0145243),complex(0.00000,-0.0330098),complex(0.00000,-0.0650888),complex(0.00000,-0.113277),complex(0.00000,-0.175716),complex(0.00000,-0.244023),complex(0.00000,-0.303140),complex(0.00000,-0.334279),complex(0.00000,-0.320709),complex(0.00000,-0.254480),complex(0.00000,-0.141160),0,complex(0.00000,0.141160),complex(0.00000,0.254480),complex(0.00000,0.320709),complex(0.00000,0.334279),complex(0.00000,0.303140),complex(0.00000,0.244023),complex(0.00000,0.175716),complex(0.00000,0.113277),complex(0.00000,0.0650888),complex(0.00000,0.0330098),complex(0.00000,0.0145243),complex(0.00000,0.00538218),complex(0.00000,0.00158691),complex(0.00000,0.000322968)],$
            [complex(6.10352e-05,0.00000),complex(0.000322968,0.00000),complex(0.00118666,0.00000),complex(0.00349343,0.00000),complex(0.00873357,0.00000),complex(0.0191343,0.00000),complex(0.0374629,0.00000),complex(0.0664145,0.00000),complex(0.107604,0.00000),complex(0.160406,0.00000),complex(0.221105,0.00000),complex(0.282838,0.00000),complex(0.336645,0.00000),complex(0.373474,0.00000),complex(0.386582,0.00000),complex(0.373474,-0.00000),complex(0.336645,0.00000),complex(0.282838,-0.00000),complex(0.221105,0.00000),complex(0.160406,-0.00000),complex(0.107604,0.00000),complex(0.0664145,-0.00000),complex(0.0374629,0.00000),complex(0.0191343,-0.00000),complex(0.00873357,0.00000),complex(0.00349343,-0.00000),complex(0.00118666,0.00000),complex(0.000322968,-0.00000),complex(6.10352e-05,0.00000)]$
            ]
          end


      27: begin
            matrix[378:405,378:405]=[$
            [complex(-0.00000,8.63167e-05),complex(0.00000,-0.000448515),complex(-0.00000,0.00161714),complex(0.00000,-0.00466829),complex(-0.00000,0.0114349),complex(0.00000,-0.0245252),complex(-0.00000,0.0469622),complex(0.00000,-0.0813409),complex(-0.00000,0.128611),complex(0.00000,-0.186868),complex(-0.00000,0.250709),complex(0.00000,-0.311673),complex(-0.00000,0.359889),complex(0.00000,-0.386582),complex(0.00000,0.386582),complex(0.00000,-0.359889),complex(0.00000,0.311673),complex(0.00000,-0.250709),complex(0.00000,0.186868),complex(0.00000,-0.128611),complex(0.00000,0.0813409),complex(0.00000,-0.0469622),complex(0.00000,0.0245252),complex(0.00000,-0.0114349),complex(0.00000,0.00466829),complex(0.00000,-0.00161714),complex(0.00000,0.000448515),complex(0.00000,-8.63167e-05)],$
            [complex(-0.000448515,0.00000),complex(0.00215792,0.00000),complex(-0.00715805,0.00000),complex(0.0188667,0.00000),complex(-0.0418124,0.00000),complex(0.0802379,0.00000),complex(-0.135568,0.00000),complex(0.203503,0.00000),complex(-0.272264,0.00000),complex(0.323664,0.00000),complex(-0.337743,0.00000),complex(0.299907,0.00000),complex(-0.207782,0.00000),complex(0.0743978,0.00000),complex(0.0743978,0.00000),complex(-0.207782,0.00000),complex(0.299907,0.00000),complex(-0.337743,0.00000),complex(0.323664,0.00000),complex(-0.272264,0.00000),complex(0.203503,0.00000),complex(-0.135568,0.00000),complex(0.0802379,0.00000),complex(-0.0418124,0.00000),complex(0.0188667,0.00000),complex(-0.00715805,0.00000),complex(0.00215792,0.00000),complex(-0.000448515,0.00000)],$
            [complex(0.00000,-0.00161714),complex(0.00000,0.00715805),complex(0.00000,-0.0216655),complex(0.00000,0.0515792),complex(0.00000,-0.101929),complex(0.00000,0.171487),complex(0.00000,-0.248159),complex(0.00000,0.308257),complex(0.00000,-0.322644),complex(0.00000,0.269305),complex(0.00000,-0.147201),complex(0.00000,-0.0166359),complex(-0.00000,0.172885),complex(0.00000,-0.268245),complex(0.00000,0.268245),complex(0.00000,-0.172885),complex(0.00000,0.0166359),complex(0.00000,0.147201),complex(-0.00000,-0.269305),complex(0.00000,0.322644),complex(-0.00000,-0.308257),complex(0.00000,0.248159),complex(-0.00000,-0.171487),complex(0.00000,0.101929),complex(-0.00000,-0.0515792),complex(0.00000,0.0216655),complex(-0.00000,-0.00715805),complex(0.00000,0.00161714)],$
            [complex(0.00466829,0.00000),complex(-0.0188667,-0.00000),complex(0.0515792,0.00000),complex(-0.109363,-0.00000),complex(0.188809,0.00000),complex(-0.269815,-0.00000),complex(0.316941,0.00000),complex(-0.293279,-0.00000),complex(0.183108,0.00000),complex(-0.0103656,-0.00000),complex(-0.162247,0.00000),complex(0.259327,0.00000),complex(-0.232902,0.00000),complex(0.0929229,0.00000),complex(0.0929229,0.00000),complex(-0.232902,0.00000),complex(0.259327,0.00000),complex(-0.162247,0.00000),complex(-0.0103656,0.00000),complex(0.183108,-0.00000),complex(-0.293279,0.00000),complex(0.316941,-0.00000),complex(-0.269815,0.00000),complex(0.188809,-0.00000),complex(-0.109363,0.00000),complex(0.0515792,-0.00000),complex(-0.0188667,0.00000),complex(0.00466829,-0.00000)],$
            [complex(-0.00000,0.0114349),complex(0.00000,-0.0418124),complex(-0.00000,0.101929),complex(0.00000,-0.188809),complex(-0.00000,0.276041),complex(0.00000,-0.316570),complex(-0.00000,0.265871),complex(0.00000,-0.116660),complex(0.00000,-0.0796075),complex(0.00000,0.228513),complex(0.00000,-0.246023),complex(0.00000,0.117633),complex(-0.00000,0.0814987),complex(0.00000,-0.227614),complex(0.00000,0.227614),complex(0.00000,-0.0814987),complex(-0.00000,-0.117633),complex(0.00000,0.246023),complex(-0.00000,-0.228513),complex(0.00000,0.0796075),complex(0.00000,0.116660),complex(0.00000,-0.265871),complex(0.00000,0.316570),complex(0.00000,-0.276041),complex(0.00000,0.188809),complex(0.00000,-0.101929),complex(0.00000,0.0418124),complex(0.00000,-0.0114349)],$
            [complex(-0.0245252,0.00000),complex(0.0802379,0.00000),complex(-0.171487,0.00000),complex(0.269815,0.00000),complex(-0.316570,0.00000),complex(0.255325,0.00000),complex(-0.0823114,0.00000),complex(-0.126536,-0.00000),complex(0.248957,0.00000),complex(-0.201251,-0.00000),complex(0.0123532,0.00000),complex(0.182091,0.00000),complex(-0.235594,0.00000),complex(0.106125,0.00000),complex(0.106125,0.00000),complex(-0.235594,0.00000),complex(0.182091,0.00000),complex(0.0123532,-0.00000),complex(-0.201251,0.00000),complex(0.248957,-0.00000),complex(-0.126536,0.00000),complex(-0.0823114,0.00000),complex(0.255325,0.00000),complex(-0.316570,0.00000),complex(0.269815,0.00000),complex(-0.171487,0.00000),complex(0.0802379,0.00000),complex(-0.0245252,0.00000)],$
            [complex(0.00000,-0.0469622),complex(0.00000,0.135568),complex(0.00000,-0.248159),complex(0.00000,0.316941),complex(0.00000,-0.265871),complex(0.00000,0.0823114),complex(-0.00000,0.140696),complex(0.00000,-0.252065),complex(-0.00000,0.164053),complex(0.00000,0.0556412),complex(0.00000,-0.222108),complex(0.00000,0.189042),complex(-0.00000,0.0145525),complex(0.00000,-0.203215),complex(0.00000,0.203215),complex(0.00000,-0.0145525),complex(-0.00000,-0.189042),complex(0.00000,0.222108),complex(-0.00000,-0.0556412),complex(0.00000,-0.164053),complex(0.00000,0.252065),complex(0.00000,-0.140696),complex(-0.00000,-0.0823114),complex(0.00000,0.265871),complex(-0.00000,-0.316941),complex(0.00000,0.248159),complex(-0.00000,-0.135568),complex(0.00000,0.0469622)],$
            [complex(0.0813409,0.00000),complex(-0.203503,-0.00000),complex(0.308257,0.00000),complex(-0.293279,-0.00000),complex(0.116660,0.00000),complex(0.126536,0.00000),complex(-0.252065,0.00000),complex(0.150364,0.00000),complex(0.0870735,0.00000),complex(-0.232010,-0.00000),complex(0.139940,0.00000),complex(0.0945913,0.00000),complex(-0.226851,0.00000),complex(0.117326,0.00000),complex(0.117326,0.00000),complex(-0.226851,0.00000),complex(0.0945913,0.00000),complex(0.139940,-0.00000),complex(-0.232010,0.00000),complex(0.0870735,-0.00000),complex(0.150364,0.00000),complex(-0.252065,0.00000),complex(0.126536,0.00000),complex(0.116660,-0.00000),complex(-0.293279,0.00000),complex(0.308257,-0.00000),complex(-0.203503,0.00000),complex(0.0813409,-0.00000)],$
            [complex(-0.00000,0.128611),complex(0.00000,-0.272264),complex(-0.00000,0.322644),complex(0.00000,-0.183108),complex(0.00000,-0.0796075),complex(0.00000,0.248957),complex(0.00000,-0.164053),complex(0.00000,-0.0870735),complex(-0.00000,0.232969),complex(0.00000,-0.111745),complex(0.00000,-0.135451),complex(0.00000,0.218590),complex(0.00000,-0.0398537),complex(0.00000,-0.185509),complex(0.00000,0.185509),complex(0.00000,0.0398537),complex(-0.00000,-0.218590),complex(0.00000,0.135451),complex(0.00000,0.111745),complex(0.00000,-0.232969),complex(0.00000,0.0870735),complex(0.00000,0.164053),complex(-0.00000,-0.248957),complex(0.00000,0.0796075),complex(0.00000,0.183108),complex(0.00000,-0.322644),complex(0.00000,0.272264),complex(0.00000,-0.128611)],$
            [complex(-0.186868,0.00000),complex(0.323664,0.00000),complex(-0.269305,0.00000),complex(0.0103656,0.00000),complex(0.228513,0.00000),complex(-0.201251,-0.00000),complex(-0.0556412,0.00000),complex(0.232010,0.00000),complex(-0.111745,0.00000),complex(-0.147515,-0.00000),complex(0.207872,0.00000),complex(0.00791812,0.00000),complex(-0.210290,0.00000),complex(0.127676,0.00000),complex(0.127676,0.00000),complex(-0.210290,0.00000),complex(0.00791812,0.00000),complex(0.207872,-0.00000),complex(-0.147515,0.00000),complex(-0.111745,0.00000),complex(0.232010,0.00000),complex(-0.0556412,0.00000),complex(-0.201251,0.00000),complex(0.228513,-0.00000),complex(0.0103656,0.00000),complex(-0.269305,0.00000),complex(0.323664,0.00000),complex(-0.186868,0.00000)],$
            [complex(0.00000,-0.250709),complex(0.00000,0.337743),complex(0.00000,-0.147201),complex(0.00000,-0.162247),complex(-0.00000,0.246023),complex(0.00000,-0.0123532),complex(0.00000,-0.222108),complex(0.00000,0.139940),complex(-0.00000,0.135451),complex(0.00000,-0.207872),complex(0.00000,-0.0235645),complex(0.00000,0.216007),complex(0.00000,-0.0858669),complex(0.00000,-0.171295),complex(0.00000,0.171295),complex(0.00000,0.0858669),complex(-0.00000,-0.216007),complex(0.00000,0.0235645),complex(0.00000,0.207872),complex(0.00000,-0.135451),complex(-0.00000,-0.139940),complex(0.00000,0.222108),complex(0.00000,0.0123532),complex(0.00000,-0.246023),complex(0.00000,0.162247),complex(0.00000,0.147201),complex(-0.00000,-0.337743),complex(0.00000,0.250709)],$
            [complex(0.311673,0.00000),complex(-0.299907,-0.00000),complex(-0.0166359,0.00000),complex(0.259327,0.00000),complex(-0.117633,0.00000),complex(-0.182091,-0.00000),complex(0.189042,0.00000),complex(0.0945913,0.00000),complex(-0.218590,0.00000),complex(-0.00791812,-0.00000),complex(0.216007,0.00000),complex(-0.0712113,-0.00000),complex(-0.187479,0.00000),complex(0.137790,0.00000),complex(0.137790,0.00000),complex(-0.187479,0.00000),complex(-0.0712113,0.00000),complex(0.216007,-0.00000),complex(-0.00791812,0.00000),complex(-0.218590,0.00000),complex(0.0945913,0.00000),complex(0.189042,-0.00000),complex(-0.182091,0.00000),complex(-0.117633,0.00000),complex(0.259327,0.00000),complex(-0.0166359,0.00000),complex(-0.299907,0.00000),complex(0.311673,-0.00000)],$
            [complex(-0.00000,0.359889),complex(0.00000,-0.207782),complex(0.00000,-0.172885),complex(0.00000,0.232902),complex(-0.00000,0.0814987),complex(0.00000,-0.235594),complex(0.00000,-0.0145525),complex(0.00000,0.226851),complex(0.00000,-0.0398537),complex(0.00000,-0.210290),complex(-0.00000,0.0858669),complex(0.00000,0.187479),complex(0.00000,-0.125332),complex(0.00000,-0.159106),complex(0.00000,0.159106),complex(0.00000,0.125332),complex(-0.00000,-0.187479),complex(0.00000,-0.0858669),complex(0.00000,0.210290),complex(0.00000,0.0398537),complex(-0.00000,-0.226851),complex(0.00000,0.0145525),complex(0.00000,0.235594),complex(0.00000,-0.0814987),complex(-0.00000,-0.232902),complex(0.00000,0.172885),complex(0.00000,0.207782),complex(0.00000,-0.359889)],$
            [complex(-0.386582,0.00000),complex(0.0743978,0.00000),complex(0.268245,0.00000),complex(-0.0929229,-0.00000),complex(-0.227614,0.00000),complex(0.106125,0.00000),complex(0.203215,0.00000),complex(-0.117326,-0.00000),complex(-0.185509,0.00000),complex(0.127676,0.00000),complex(0.171295,0.00000),complex(-0.137790,-0.00000),complex(-0.159106,0.00000),complex(0.148119,0.00000),complex(0.148119,0.00000),complex(-0.159106,0.00000),complex(-0.137790,0.00000),complex(0.171295,-0.00000),complex(0.127676,0.00000),complex(-0.185509,0.00000),complex(-0.117326,0.00000),complex(0.203215,-0.00000),complex(0.106125,0.00000),complex(-0.227614,0.00000),complex(-0.0929229,0.00000),complex(0.268245,-0.00000),complex(0.0743978,0.00000),complex(-0.386582,0.00000)],$
            [complex(0.00000,-0.386582),complex(0.00000,-0.0743978),complex(-0.00000,0.268245),complex(0.00000,0.0929229),complex(0.00000,-0.227614),complex(0.00000,-0.106125),complex(-0.00000,0.203215),complex(0.00000,0.117326),complex(0.00000,-0.185509),complex(0.00000,-0.127676),complex(-0.00000,0.171295),complex(0.00000,0.137790),complex(0.00000,-0.159106),complex(0.00000,-0.148119),complex(0.00000,0.148119),complex(0.00000,0.159106),complex(-0.00000,-0.137790),complex(0.00000,-0.171295),complex(0.00000,0.127676),complex(0.00000,0.185509),complex(-0.00000,-0.117326),complex(0.00000,-0.203215),complex(0.00000,0.106125),complex(0.00000,0.227614),complex(-0.00000,-0.0929229),complex(0.00000,-0.268245),complex(0.00000,0.0743978),complex(0.00000,0.386582)],$
            [complex(0.359889,0.00000),complex(0.207782,0.00000),complex(-0.172885,0.00000),complex(-0.232902,-0.00000),complex(0.0814987,0.00000),complex(0.235594,0.00000),complex(-0.0145525,0.00000),complex(-0.226851,-0.00000),complex(-0.0398537,0.00000),complex(0.210290,0.00000),complex(0.0858669,0.00000),complex(-0.187479,-0.00000),complex(-0.125332,0.00000),complex(0.159106,0.00000),complex(0.159106,0.00000),complex(-0.125332,0.00000),complex(-0.187479,0.00000),complex(0.0858669,-0.00000),complex(0.210290,0.00000),complex(-0.0398537,0.00000),complex(-0.226851,0.00000),complex(-0.0145525,0.00000),complex(0.235594,0.00000),complex(0.0814987,-0.00000),complex(-0.232902,0.00000),complex(-0.172885,0.00000),complex(0.207782,0.00000),complex(0.359889,-0.00000)],$
            [complex(-0.00000,0.311673),complex(0.00000,0.299907),complex(0.00000,-0.0166359),complex(0.00000,-0.259327),complex(0.00000,-0.117633),complex(0.00000,0.182091),complex(-0.00000,0.189042),complex(0.00000,-0.0945913),complex(0.00000,-0.218590),complex(0.00000,0.00791812),complex(-0.00000,0.216007),complex(0.00000,0.0712113),complex(0.00000,-0.187479),complex(0.00000,-0.137790),complex(0.00000,0.137790),complex(0.00000,0.187479),complex(-0.00000,-0.0712113),complex(0.00000,-0.216007),complex(-0.00000,-0.00791812),complex(0.00000,0.218590),complex(0.00000,0.0945913),complex(0.00000,-0.189042),complex(-0.00000,-0.182091),complex(0.00000,0.117633),complex(0.00000,0.259327),complex(0.00000,0.0166359),complex(-0.00000,-0.299907),complex(0.00000,-0.311673)],$
            [complex(-0.250709,0.00000),complex(-0.337743,-0.00000),complex(-0.147201,0.00000),complex(0.162247,0.00000),complex(0.246023,0.00000),complex(0.0123532,0.00000),complex(-0.222108,0.00000),complex(-0.139940,-0.00000),complex(0.135451,0.00000),complex(0.207872,0.00000),complex(-0.0235645,0.00000),complex(-0.216007,-0.00000),complex(-0.0858669,0.00000),complex(0.171295,0.00000),complex(0.171295,0.00000),complex(-0.0858669,0.00000),complex(-0.216007,0.00000),complex(-0.0235645,0.00000),complex(0.207872,0.00000),complex(0.135451,-0.00000),complex(-0.139940,0.00000),complex(-0.222108,0.00000),complex(0.0123532,0.00000),complex(0.246023,-0.00000),complex(0.162247,0.00000),complex(-0.147201,0.00000),complex(-0.337743,0.00000),complex(-0.250709,0.00000)],$
            [complex(0.00000,-0.186868),complex(0.00000,-0.323664),complex(0.00000,-0.269305),complex(0.00000,-0.0103656),complex(-0.00000,0.228513),complex(0.00000,0.201251),complex(0.00000,-0.0556412),complex(0.00000,-0.232010),complex(0.00000,-0.111745),complex(0.00000,0.147515),complex(-0.00000,0.207872),complex(0.00000,-0.00791812),complex(0.00000,-0.210290),complex(0.00000,-0.127676),complex(0.00000,0.127676),complex(0.00000,0.210290),complex(0.00000,0.00791812),complex(0.00000,-0.207872),complex(-0.00000,-0.147515),complex(0.00000,0.111745),complex(0.00000,0.232010),complex(0.00000,0.0556412),complex(-0.00000,-0.201251),complex(0.00000,-0.228513),complex(0.00000,0.0103656),complex(0.00000,0.269305),complex(0.00000,0.323664),complex(0.00000,0.186868)],$
            [complex(0.128611,0.00000),complex(0.272264,0.00000),complex(0.322644,0.00000),complex(0.183108,0.00000),complex(-0.0796075,0.00000),complex(-0.248957,-0.00000),complex(-0.164053,0.00000),complex(0.0870735,0.00000),complex(0.232969,0.00000),complex(0.111745,0.00000),complex(-0.135451,0.00000),complex(-0.218590,-0.00000),complex(-0.0398537,0.00000),complex(0.185509,0.00000),complex(0.185509,0.00000),complex(-0.0398537,0.00000),complex(-0.218590,0.00000),complex(-0.135451,0.00000),complex(0.111745,0.00000),complex(0.232969,-0.00000),complex(0.0870735,0.00000),complex(-0.164053,0.00000),complex(-0.248957,0.00000),complex(-0.0796075,0.00000),complex(0.183108,0.00000),complex(0.322644,-0.00000),complex(0.272264,0.00000),complex(0.128611,-0.00000)],$
            [complex(-0.00000,0.0813409),complex(0.00000,0.203503),complex(-0.00000,0.308257),complex(0.00000,0.293279),complex(-0.00000,0.116660),complex(0.00000,-0.126536),complex(0.00000,-0.252065),complex(0.00000,-0.150364),complex(-0.00000,0.0870735),complex(0.00000,0.232010),complex(-0.00000,0.139940),complex(0.00000,-0.0945913),complex(0.00000,-0.226851),complex(0.00000,-0.117326),complex(0.00000,0.117326),complex(0.00000,0.226851),complex(0.00000,0.0945913),complex(0.00000,-0.139940),complex(-0.00000,-0.232010),complex(0.00000,-0.0870735),complex(0.00000,0.150364),complex(0.00000,0.252065),complex(0.00000,0.126536),complex(0.00000,-0.116660),complex(-0.00000,-0.293279),complex(0.00000,-0.308257),complex(-0.00000,-0.203503),complex(0.00000,-0.0813409)],$
            [complex(-0.0469622,0.00000),complex(-0.135568,-0.00000),complex(-0.248159,0.00000),complex(-0.316941,-0.00000),complex(-0.265871,0.00000),complex(-0.0823114,-0.00000),complex(0.140696,0.00000),complex(0.252065,0.00000),complex(0.164053,0.00000),complex(-0.0556412,-0.00000),complex(-0.222108,0.00000),complex(-0.189042,-0.00000),complex(0.0145525,0.00000),complex(0.203215,0.00000),complex(0.203215,0.00000),complex(0.0145525,-0.00000),complex(-0.189042,0.00000),complex(-0.222108,0.00000),complex(-0.0556412,0.00000),complex(0.164053,-0.00000),complex(0.252065,0.00000),complex(0.140696,-0.00000),complex(-0.0823114,0.00000),complex(-0.265871,0.00000),complex(-0.316941,0.00000),complex(-0.248159,0.00000),complex(-0.135568,0.00000),complex(-0.0469622,0.00000)],$
            [complex(0.00000,-0.0245252),complex(0.00000,-0.0802379),complex(0.00000,-0.171487),complex(0.00000,-0.269815),complex(0.00000,-0.316570),complex(0.00000,-0.255325),complex(0.00000,-0.0823114),complex(0.00000,0.126536),complex(-0.00000,0.248957),complex(0.00000,0.201251),complex(-0.00000,0.0123532),complex(0.00000,-0.182091),complex(0.00000,-0.235594),complex(0.00000,-0.106125),complex(0.00000,0.106125),complex(0.00000,0.235594),complex(0.00000,0.182091),complex(0.00000,-0.0123532),complex(-0.00000,-0.201251),complex(0.00000,-0.248957),complex(-0.00000,-0.126536),complex(0.00000,0.0823114),complex(0.00000,0.255325),complex(0.00000,0.316570),complex(0.00000,0.269815),complex(0.00000,0.171487),complex(0.00000,0.0802379),complex(0.00000,0.0245252)],$
            [complex(0.0114349,0.00000),complex(0.0418124,0.00000),complex(0.101929,0.00000),complex(0.188809,0.00000),complex(0.276041,0.00000),complex(0.316570,0.00000),complex(0.265871,0.00000),complex(0.116660,0.00000),complex(-0.0796075,0.00000),complex(-0.228513,-0.00000),complex(-0.246023,0.00000),complex(-0.117633,-0.00000),complex(0.0814987,0.00000),complex(0.227614,0.00000),complex(0.227614,0.00000),complex(0.0814987,-0.00000),complex(-0.117633,0.00000),complex(-0.246023,0.00000),complex(-0.228513,0.00000),complex(-0.0796075,0.00000),complex(0.116660,0.00000),complex(0.265871,-0.00000),complex(0.316570,0.00000),complex(0.276041,-0.00000),complex(0.188809,0.00000),complex(0.101929,-0.00000),complex(0.0418124,0.00000),complex(0.0114349,-0.00000)],$
            [complex(-0.00000,0.00466829),complex(0.00000,0.0188667),complex(-0.00000,0.0515792),complex(0.00000,0.109363),complex(-0.00000,0.188809),complex(0.00000,0.269815),complex(-0.00000,0.316941),complex(0.00000,0.293279),complex(-0.00000,0.183108),complex(0.00000,0.0103656),complex(0.00000,-0.162247),complex(0.00000,-0.259327),complex(0.00000,-0.232902),complex(0.00000,-0.0929229),complex(0.00000,0.0929229),complex(0.00000,0.232902),complex(0.00000,0.259327),complex(0.00000,0.162247),complex(-0.00000,-0.0103656),complex(0.00000,-0.183108),complex(-0.00000,-0.293279),complex(0.00000,-0.316941),complex(-0.00000,-0.269815),complex(0.00000,-0.188809),complex(-0.00000,-0.109363),complex(0.00000,-0.0515792),complex(-0.00000,-0.0188667),complex(0.00000,-0.00466829)],$
            [complex(-0.00161714,0.00000),complex(-0.00715805,-0.00000),complex(-0.0216655,0.00000),complex(-0.0515792,-0.00000),complex(-0.101929,0.00000),complex(-0.171487,-0.00000),complex(-0.248159,0.00000),complex(-0.308257,-0.00000),complex(-0.322644,0.00000),complex(-0.269305,-0.00000),complex(-0.147201,0.00000),complex(0.0166359,0.00000),complex(0.172885,0.00000),complex(0.268245,0.00000),complex(0.268245,0.00000),complex(0.172885,-0.00000),complex(0.0166359,0.00000),complex(-0.147201,0.00000),complex(-0.269305,0.00000),complex(-0.322644,0.00000),complex(-0.308257,0.00000),complex(-0.248159,0.00000),complex(-0.171487,0.00000),complex(-0.101929,0.00000),complex(-0.0515792,0.00000),complex(-0.0216655,0.00000),complex(-0.00715805,0.00000),complex(-0.00161714,0.00000)],$
            [complex(0.00000,-0.000448515),complex(0.00000,-0.00215792),complex(0.00000,-0.00715805),complex(0.00000,-0.0188667),complex(0.00000,-0.0418124),complex(0.00000,-0.0802379),complex(0.00000,-0.135568),complex(0.00000,-0.203503),complex(0.00000,-0.272264),complex(0.00000,-0.323664),complex(0.00000,-0.337743),complex(0.00000,-0.299907),complex(0.00000,-0.207782),complex(0.00000,-0.0743978),complex(0.00000,0.0743978),complex(0.00000,0.207782),complex(0.00000,0.299907),complex(0.00000,0.337743),complex(0.00000,0.323664),complex(0.00000,0.272264),complex(0.00000,0.203503),complex(0.00000,0.135568),complex(0.00000,0.0802379),complex(0.00000,0.0418124),complex(0.00000,0.0188667),complex(0.00000,0.00715805),complex(0.00000,0.00215792),complex(0.00000,0.000448515)],$
            [complex(8.63167e-05,0.00000),complex(0.000448515,0.00000),complex(0.00161714,0.00000),complex(0.00466829,0.00000),complex(0.0114349,0.00000),complex(0.0245252,0.00000),complex(0.0469622,0.00000),complex(0.0813409,0.00000),complex(0.128611,0.00000),complex(0.186868,0.00000),complex(0.250709,0.00000),complex(0.311673,0.00000),complex(0.359889,0.00000),complex(0.386582,0.00000),complex(0.386582,0.00000),complex(0.359889,-0.00000),complex(0.311673,0.00000),complex(0.250709,-0.00000),complex(0.186868,0.00000),complex(0.128611,-0.00000),complex(0.0813409,0.00000),complex(0.0469622,-0.00000),complex(0.0245252,0.00000),complex(0.0114349,-0.00000),complex(0.00466829,0.00000),complex(0.00161714,-0.00000),complex(0.000448515,0.00000),complex(8.63167e-05,-0.00000)]$
            ]
          end

      26: begin
            matrix[351:377,351:377]=[$
            [complex(-0.000122070,-0.00000),complex(0.000622439,0.00000),complex(-0.00220065,-0.00000),complex(0.00622439,0.00000),complex(-0.0149256,-0.00000),complex(0.0313081,0.00000),complex(-0.0585721,-0.00000),complex(0.0990050,0.00000),complex(-0.152577,-0.00000),complex(0.215776,0.00000),complex(-0.281338,-0.00000),complex(0.339306,0.00000),complex(-0.379356,-0.00000),complex(0.393676,0.00000),complex(-0.379356,0.00000),complex(0.339306,0.00000),complex(-0.281338,0.00000),complex(0.215776,0.00000),complex(-0.152577,0.00000),complex(0.0990050,0.00000),complex(-0.0585721,0.00000),complex(0.0313081,0.00000),complex(-0.0149256,0.00000),complex(0.00622439,0.00000),complex(-0.00220065,0.00000),complex(0.000622439,0.00000),complex(-0.000122070,0.00000)],$
            [complex(0.00000,-0.000622439),complex(0.00000,0.00292969),complex(0.00000,-0.00949484),complex(0.00000,0.0244141),complex(0.00000,-0.0526886),complex(0.00000,0.0982404),complex(0.00000,-0.160817),complex(0.00000,0.232998),complex(0.00000,-0.299228),complex(0.00000,0.338538),complex(0.00000,-0.331049),complex(0.00000,0.266174),complex(0.00000,-0.148796),0,complex(0.00000,0.148796),complex(0.00000,-0.266174),complex(0.00000,0.331049),complex(0.00000,-0.338538),complex(0.00000,0.299228),complex(0.00000,-0.232998),complex(0.00000,0.160817),complex(0.00000,-0.0982404),complex(0.00000,0.0526886),complex(0.00000,-0.0244141),complex(0.00000,0.00949484),complex(0.00000,-0.00292969),complex(0.00000,0.000622439)],$
            [complex(0.00220065,0.00000),complex(-0.00949484,0.00000),complex(0.0279541,0.00000),complex(-0.0645649,0.00000),complex(0.123360,0.00000),complex(-0.199716,0.00000),complex(0.276165,0.00000),complex(-0.324017,0.00000),complex(0.313147,0.00000),complex(-0.227413,0.00000),complex(0.0780291,0.00000),complex(0.0941066,0.00000),complex(-0.231472,-0.00000),complex(0.283884,0.00000),complex(-0.231472,0.00000),complex(0.0941066,0.00000),complex(0.0780291,-0.00000),complex(-0.227413,0.00000),complex(0.313147,-0.00000),complex(-0.324017,0.00000),complex(0.276165,-0.00000),complex(-0.199716,0.00000),complex(0.123360,-0.00000),complex(-0.0645649,0.00000),complex(0.0279541,-0.00000),complex(-0.00949484,0.00000),complex(0.00220065,-0.00000)],$
            [complex(-0.00000,0.00622439),complex(0.00000,-0.0244141),complex(-0.00000,0.0645649),complex(0.00000,-0.131836),complex(-0.00000,0.217780),complex(0.00000,-0.294721),complex(-0.00000,0.321634),complex(0.00000,-0.264064),complex(-0.00000,0.119691),complex(0.00000,0.0677076),complex(0.00000,-0.220700),complex(0.00000,0.266174),complex(0.00000,-0.178555),0,complex(0.00000,0.178555),complex(0.00000,-0.266174),complex(0.00000,0.220700),complex(0.00000,-0.0677076),complex(-0.00000,-0.119691),complex(0.00000,0.264064),complex(-0.00000,-0.321634),complex(0.00000,0.294721),complex(-0.00000,-0.217780),complex(0.00000,0.131836),complex(-0.00000,-0.0645649),complex(0.00000,0.0244141),complex(-0.00000,-0.00622439)],$
            [complex(-0.0149256,-0.00000),complex(0.0526886,0.00000),complex(-0.123360,-0.00000),complex(0.217780,0.00000),complex(-0.299560,-0.00000),complex(0.314950,0.00000),complex(-0.225148,-0.00000),complex(0.0437250,0.00000),complex(0.152240,0.00000),complex(-0.257654,0.00000),complex(0.207086,0.00000),complex(-0.0277505,0.00000),complex(-0.167541,-0.00000),complex(0.251139,0.00000),complex(-0.167541,0.00000),complex(-0.0277505,0.00000),complex(0.207086,-0.00000),complex(-0.257654,0.00000),complex(0.152240,-0.00000),complex(0.0437250,0.00000),complex(-0.225148,0.00000),complex(0.314950,0.00000),complex(-0.299560,0.00000),complex(0.217780,0.00000),complex(-0.123360,0.00000),complex(0.0526886,0.00000),complex(-0.0149256,0.00000)],$
            [complex(0.00000,-0.0313081),complex(0.00000,0.0982404),complex(0.00000,-0.199716),complex(0.00000,0.294721),complex(0.00000,-0.314950),complex(0.00000,0.210938),complex(0.00000,-0.00639443),complex(0.00000,-0.191466),complex(-0.00000,0.254616),complex(0.00000,-0.134610),complex(0.00000,-0.0833671),complex(0.00000,0.232840),complex(0.00000,-0.195242),0,complex(0.00000,0.195242),complex(0.00000,-0.232840),complex(0.00000,0.0833671),complex(0.00000,0.134610),complex(-0.00000,-0.254616),complex(0.00000,0.191466),complex(0.00000,0.00639443),complex(0.00000,-0.210938),complex(0.00000,0.314950),complex(0.00000,-0.294721),complex(0.00000,0.199716),complex(0.00000,-0.0982404),complex(0.00000,0.0313081)],$
            [complex(0.0585721,0.00000),complex(-0.160817,0.00000),complex(0.276165,0.00000),complex(-0.321634,0.00000),complex(0.225148,0.00000),complex(-0.00639443,0.00000),complex(-0.202393,-0.00000),complex(0.245540,0.00000),complex(-0.0845842,-0.00000),complex(-0.144803,0.00000),complex(0.238053,0.00000),complex(-0.108901,0.00000),complex(-0.121755,-0.00000),complex(0.234652,0.00000),complex(-0.121755,0.00000),complex(-0.108901,0.00000),complex(0.238053,-0.00000),complex(-0.144803,0.00000),complex(-0.0845842,0.00000),complex(0.245540,0.00000),complex(-0.202393,0.00000),complex(-0.00639443,0.00000),complex(0.225148,-0.00000),complex(-0.321634,0.00000),complex(0.276165,-0.00000),complex(-0.160817,0.00000),complex(0.0585721,-0.00000)],$
            [complex(-0.00000,0.0990050),complex(0.00000,-0.232998),complex(-0.00000,0.324017),complex(0.00000,-0.264064),complex(-0.00000,0.0437250),complex(0.00000,0.191466),complex(0.00000,-0.245540),complex(0.00000,0.0673828),complex(-0.00000,0.170063),complex(0.00000,-0.225607),complex(-0.00000,0.0416258),complex(0.00000,0.184076),complex(0.00000,-0.205803),0,complex(0.00000,0.205803),complex(0.00000,-0.184076),complex(-0.00000,-0.0416258),complex(0.00000,0.225607),complex(-0.00000,-0.170063),complex(0.00000,-0.0673828),complex(0.00000,0.245540),complex(0.00000,-0.191466),complex(-0.00000,-0.0437250),complex(0.00000,0.264064),complex(-0.00000,-0.324017),complex(0.00000,0.232998),complex(-0.00000,-0.0990050)],$
            [complex(-0.152577,-0.00000),complex(0.299228,0.00000),complex(-0.313147,-0.00000),complex(0.119691,0.00000),complex(0.152240,0.00000),complex(-0.254616,0.00000),complex(0.0845842,0.00000),complex(0.170063,0.00000),complex(-0.219116,-0.00000),complex(0.00742324,0.00000),complex(0.208205,0.00000),complex(-0.164236,0.00000),complex(-0.0834643,-0.00000),complex(0.225199,0.00000),complex(-0.0834643,0.00000),complex(-0.164236,0.00000),complex(0.208205,-0.00000),complex(0.00742324,0.00000),complex(-0.219116,0.00000),complex(0.170063,0.00000),complex(0.0845842,-0.00000),complex(-0.254616,0.00000),complex(0.152240,-0.00000),complex(0.119691,0.00000),complex(-0.313147,0.00000),complex(0.299228,0.00000),complex(-0.152577,0.00000)],$
            [complex(0.00000,-0.215776),complex(0.00000,0.338538),complex(0.00000,-0.227413),complex(0.00000,-0.0677076),complex(-0.00000,0.257654),complex(0.00000,-0.134610),complex(0.00000,-0.144803),complex(0.00000,0.225607),complex(0.00000,-0.00742324),complex(0.00000,-0.213867),complex(-0.00000,0.138469),complex(0.00000,0.126690),complex(0.00000,-0.212465),0,complex(0.00000,0.212465),complex(0.00000,-0.126690),complex(-0.00000,-0.138469),complex(0.00000,0.213867),complex(0.00000,0.00742324),complex(0.00000,-0.225607),complex(0.00000,0.144803),complex(0.00000,0.134610),complex(-0.00000,-0.257654),complex(0.00000,0.0677076),complex(0.00000,0.227413),complex(0.00000,-0.338538),complex(0.00000,0.215776)],$
            [complex(0.281338,0.00000),complex(-0.331049,0.00000),complex(0.0780291,0.00000),complex(0.220700,0.00000),complex(-0.207086,-0.00000),complex(-0.0833671,0.00000),complex(0.238053,0.00000),complex(-0.0416258,0.00000),complex(-0.208205,-0.00000),complex(0.138469,0.00000),complex(0.139526,0.00000),complex(-0.199192,0.00000),complex(-0.0488861,-0.00000),complex(0.219836,0.00000),complex(-0.0488861,0.00000),complex(-0.199192,0.00000),complex(0.139526,-0.00000),complex(0.138469,0.00000),complex(-0.208205,0.00000),complex(-0.0416258,0.00000),complex(0.238053,-0.00000),complex(-0.0833671,0.00000),complex(-0.207086,0.00000),complex(0.220700,0.00000),complex(0.0780291,-0.00000),complex(-0.331049,0.00000),complex(0.281338,-0.00000)],$
            [complex(-0.00000,0.339306),complex(0.00000,-0.266174),complex(0.00000,-0.0941066),complex(0.00000,0.266174),complex(0.00000,-0.0277505),complex(0.00000,-0.232840),complex(-0.00000,0.108901),complex(0.00000,0.184076),complex(0.00000,-0.164236),complex(0.00000,-0.126690),complex(-0.00000,0.199192),complex(0.00000,0.0644531),complex(0.00000,-0.216182),0,complex(0.00000,0.216182),complex(0.00000,-0.0644531),complex(-0.00000,-0.199192),complex(0.00000,0.126690),complex(0.00000,0.164236),complex(0.00000,-0.184076),complex(-0.00000,-0.108901),complex(0.00000,0.232840),complex(0.00000,0.0277505),complex(0.00000,-0.266174),complex(0.00000,0.0941066),complex(0.00000,0.266174),complex(-0.00000,-0.339306)],$
            [complex(-0.379356,-0.00000),complex(0.148796,0.00000),complex(0.231472,0.00000),complex(-0.178555,0.00000),complex(-0.167541,-0.00000),complex(0.195242,0.00000),complex(0.121755,0.00000),complex(-0.205803,0.00000),complex(-0.0834643,-0.00000),complex(0.212465,0.00000),complex(0.0488861,0.00000),complex(-0.216182,0.00000),complex(-0.0161133,-0.00000),complex(0.217380,0.00000),complex(-0.0161133,0.00000),complex(-0.216182,0.00000),complex(0.0488861,-0.00000),complex(0.212465,0.00000),complex(-0.0834643,0.00000),complex(-0.205803,0.00000),complex(0.121755,-0.00000),complex(0.195242,0.00000),complex(-0.167541,0.00000),complex(-0.178555,0.00000),complex(0.231472,-0.00000),complex(0.148796,0.00000),complex(-0.379356,0.00000)],$
            [complex(0.00000,-0.393676),0,complex(-0.00000,0.283884),complex(0.00000,4.49374e-16),complex(0.00000,-0.251139),complex(0.00000,4.23808e-15),complex(-0.00000,0.234652),0,complex(0.00000,-0.225199),0,complex(-0.00000,0.219836),0,complex(0.00000,-0.217380),0,complex(0.00000,0.217380),0,complex(-0.00000,-0.219836),0,complex(0.00000,0.225199),0,complex(-0.00000,-0.234652),complex(0.00000,-4.80316e-15),complex(0.00000,0.251139),0,complex(-0.00000,-0.283884),0,complex(0.00000,0.393676)],$
            [complex(0.379356,0.00000),complex(0.148796,0.00000),complex(-0.231472,-0.00000),complex(-0.178555,0.00000),complex(0.167541,0.00000),complex(0.195242,0.00000),complex(-0.121755,-0.00000),complex(-0.205803,0.00000),complex(0.0834643,0.00000),complex(0.212465,0.00000),complex(-0.0488861,-0.00000),complex(-0.216182,0.00000),complex(0.0161133,0.00000),complex(0.217380,0.00000),complex(0.0161133,-0.00000),complex(-0.216182,0.00000),complex(-0.0488861,0.00000),complex(0.212465,0.00000),complex(0.0834643,-0.00000),complex(-0.205803,0.00000),complex(-0.121755,0.00000),complex(0.195242,0.00000),complex(0.167541,-0.00000),complex(-0.178555,0.00000),complex(-0.231472,0.00000),complex(0.148796,0.00000),complex(0.379356,-0.00000)],$
            [complex(-0.00000,0.339306),complex(0.00000,0.266174),complex(0.00000,-0.0941066),complex(0.00000,-0.266174),complex(0.00000,-0.0277505),complex(0.00000,0.232840),complex(-0.00000,0.108901),complex(0.00000,-0.184076),complex(0.00000,-0.164236),complex(0.00000,0.126690),complex(-0.00000,0.199192),complex(0.00000,-0.0644531),complex(0.00000,-0.216182),0,complex(0.00000,0.216182),complex(0.00000,0.0644531),complex(-0.00000,-0.199192),complex(0.00000,-0.126690),complex(0.00000,0.164236),complex(0.00000,0.184076),complex(-0.00000,-0.108901),complex(0.00000,-0.232840),complex(0.00000,0.0277505),complex(0.00000,0.266174),complex(0.00000,0.0941066),complex(0.00000,-0.266174),complex(-0.00000,-0.339306)],$
            [complex(-0.281338,-0.00000),complex(-0.331049,0.00000),complex(-0.0780291,-0.00000),complex(0.220700,0.00000),complex(0.207086,0.00000),complex(-0.0833671,0.00000),complex(-0.238053,-0.00000),complex(-0.0416258,0.00000),complex(0.208205,0.00000),complex(0.138469,0.00000),complex(-0.139526,-0.00000),complex(-0.199192,0.00000),complex(0.0488861,0.00000),complex(0.219836,0.00000),complex(0.0488861,-0.00000),complex(-0.199192,0.00000),complex(-0.139526,0.00000),complex(0.138469,0.00000),complex(0.208205,-0.00000),complex(-0.0416258,0.00000),complex(-0.238053,0.00000),complex(-0.0833671,0.00000),complex(0.207086,-0.00000),complex(0.220700,0.00000),complex(-0.0780291,0.00000),complex(-0.331049,0.00000),complex(-0.281338,0.00000)],$
            [complex(0.00000,-0.215776),complex(0.00000,-0.338538),complex(0.00000,-0.227413),complex(0.00000,0.0677076),complex(-0.00000,0.257654),complex(0.00000,0.134610),complex(0.00000,-0.144803),complex(0.00000,-0.225607),complex(0.00000,-0.00742324),complex(0.00000,0.213867),complex(-0.00000,0.138469),complex(0.00000,-0.126690),complex(0.00000,-0.212465),0,complex(0.00000,0.212465),complex(0.00000,0.126690),complex(-0.00000,-0.138469),complex(0.00000,-0.213867),complex(0.00000,0.00742324),complex(0.00000,0.225607),complex(0.00000,0.144803),complex(0.00000,-0.134610),complex(-0.00000,-0.257654),complex(0.00000,-0.0677076),complex(0.00000,0.227413),complex(0.00000,0.338538),complex(0.00000,0.215776)],$
            [complex(0.152577,0.00000),complex(0.299228,0.00000),complex(0.313147,0.00000),complex(0.119691,0.00000),complex(-0.152240,-0.00000),complex(-0.254616,0.00000),complex(-0.0845842,-0.00000),complex(0.170063,0.00000),complex(0.219116,0.00000),complex(0.00742324,0.00000),complex(-0.208205,-0.00000),complex(-0.164236,0.00000),complex(0.0834643,0.00000),complex(0.225199,0.00000),complex(0.0834643,-0.00000),complex(-0.164236,0.00000),complex(-0.208205,0.00000),complex(0.00742324,0.00000),complex(0.219116,-0.00000),complex(0.170063,0.00000),complex(-0.0845842,0.00000),complex(-0.254616,0.00000),complex(-0.152240,0.00000),complex(0.119691,0.00000),complex(0.313147,-0.00000),complex(0.299228,0.00000),complex(0.152577,-0.00000)],$
            [complex(-0.00000,0.0990050),complex(0.00000,0.232998),complex(-0.00000,0.324017),complex(0.00000,0.264064),complex(-0.00000,0.0437250),complex(0.00000,-0.191466),complex(0.00000,-0.245540),complex(0.00000,-0.0673828),complex(-0.00000,0.170063),complex(0.00000,0.225607),complex(-0.00000,0.0416258),complex(0.00000,-0.184076),complex(0.00000,-0.205803),0,complex(0.00000,0.205803),complex(0.00000,0.184076),complex(-0.00000,-0.0416258),complex(0.00000,-0.225607),complex(-0.00000,-0.170063),complex(0.00000,0.0673828),complex(0.00000,0.245540),complex(0.00000,0.191466),complex(-0.00000,-0.0437250),complex(0.00000,-0.264064),complex(-0.00000,-0.324017),complex(0.00000,-0.232998),complex(-0.00000,-0.0990050)],$
            [complex(-0.0585721,-0.00000),complex(-0.160817,0.00000),complex(-0.276165,-0.00000),complex(-0.321634,0.00000),complex(-0.225148,-0.00000),complex(-0.00639443,0.00000),complex(0.202393,0.00000),complex(0.245540,0.00000),complex(0.0845842,0.00000),complex(-0.144803,0.00000),complex(-0.238053,-0.00000),complex(-0.108901,0.00000),complex(0.121755,0.00000),complex(0.234652,0.00000),complex(0.121755,-0.00000),complex(-0.108901,0.00000),complex(-0.238053,0.00000),complex(-0.144803,0.00000),complex(0.0845842,-0.00000),complex(0.245540,0.00000),complex(0.202393,-0.00000),complex(-0.00639443,0.00000),complex(-0.225148,0.00000),complex(-0.321634,0.00000),complex(-0.276165,0.00000),complex(-0.160817,0.00000),complex(-0.0585721,0.00000)],$
            [complex(0.00000,-0.0313081),complex(0.00000,-0.0982404),complex(0.00000,-0.199716),complex(0.00000,-0.294721),complex(0.00000,-0.314950),complex(0.00000,-0.210938),complex(0.00000,-0.00639443),complex(0.00000,0.191466),complex(-0.00000,0.254616),complex(0.00000,0.134610),complex(0.00000,-0.0833671),complex(0.00000,-0.232840),complex(0.00000,-0.195242),0,complex(0.00000,0.195242),complex(0.00000,0.232840),complex(0.00000,0.0833671),complex(0.00000,-0.134610),complex(-0.00000,-0.254616),complex(0.00000,-0.191466),complex(0.00000,0.00639443),complex(0.00000,0.210938),complex(0.00000,0.314950),complex(0.00000,0.294721),complex(0.00000,0.199716),complex(0.00000,0.0982404),complex(0.00000,0.0313081)],$
            [complex(0.0149256,0.00000),complex(0.0526886,0.00000),complex(0.123360,0.00000),complex(0.217780,0.00000),complex(0.299560,0.00000),complex(0.314950,0.00000),complex(0.225148,0.00000),complex(0.0437250,0.00000),complex(-0.152240,-0.00000),complex(-0.257654,0.00000),complex(-0.207086,-0.00000),complex(-0.0277505,0.00000),complex(0.167541,0.00000),complex(0.251139,0.00000),complex(0.167541,-0.00000),complex(-0.0277505,0.00000),complex(-0.207086,0.00000),complex(-0.257654,0.00000),complex(-0.152240,0.00000),complex(0.0437250,0.00000),complex(0.225148,-0.00000),complex(0.314950,0.00000),complex(0.299560,-0.00000),complex(0.217780,0.00000),complex(0.123360,-0.00000),complex(0.0526886,0.00000),complex(0.0149256,-0.00000)],$
            [complex(-0.00000,0.00622439),complex(0.00000,0.0244141),complex(-0.00000,0.0645649),complex(0.00000,0.131836),complex(-0.00000,0.217780),complex(0.00000,0.294721),complex(-0.00000,0.321634),complex(0.00000,0.264064),complex(-0.00000,0.119691),complex(0.00000,-0.0677076),complex(0.00000,-0.220700),complex(0.00000,-0.266174),complex(0.00000,-0.178555),0,complex(0.00000,0.178555),complex(0.00000,0.266174),complex(0.00000,0.220700),complex(0.00000,0.0677076),complex(-0.00000,-0.119691),complex(0.00000,-0.264064),complex(-0.00000,-0.321634),complex(0.00000,-0.294721),complex(-0.00000,-0.217780),complex(0.00000,-0.131836),complex(-0.00000,-0.0645649),complex(0.00000,-0.0244141),complex(-0.00000,-0.00622439)],$
            [complex(-0.00220065,-0.00000),complex(-0.00949484,0.00000),complex(-0.0279541,-0.00000),complex(-0.0645649,0.00000),complex(-0.123360,-0.00000),complex(-0.199716,0.00000),complex(-0.276165,-0.00000),complex(-0.324017,0.00000),complex(-0.313147,-0.00000),complex(-0.227413,0.00000),complex(-0.0780291,-0.00000),complex(0.0941066,0.00000),complex(0.231472,0.00000),complex(0.283884,0.00000),complex(0.231472,-0.00000),complex(0.0941066,0.00000),complex(-0.0780291,0.00000),complex(-0.227413,0.00000),complex(-0.313147,0.00000),complex(-0.324017,0.00000),complex(-0.276165,0.00000),complex(-0.199716,0.00000),complex(-0.123360,0.00000),complex(-0.0645649,0.00000),complex(-0.0279541,0.00000),complex(-0.00949484,0.00000),complex(-0.00220065,0.00000)],$
            [complex(0.00000,-0.000622439),complex(0.00000,-0.00292969),complex(0.00000,-0.00949484),complex(0.00000,-0.0244141),complex(0.00000,-0.0526886),complex(0.00000,-0.0982404),complex(0.00000,-0.160817),complex(0.00000,-0.232998),complex(0.00000,-0.299228),complex(0.00000,-0.338538),complex(0.00000,-0.331049),complex(0.00000,-0.266174),complex(0.00000,-0.148796),0,complex(0.00000,0.148796),complex(0.00000,0.266174),complex(0.00000,0.331049),complex(0.00000,0.338538),complex(0.00000,0.299228),complex(0.00000,0.232998),complex(0.00000,0.160817),complex(0.00000,0.0982404),complex(0.00000,0.0526886),complex(0.00000,0.0244141),complex(0.00000,0.00949484),complex(0.00000,0.00292969),complex(0.00000,0.000622439)],$
            [complex(0.000122070,0.00000),complex(0.000622439,0.00000),complex(0.00220065,0.00000),complex(0.00622439,0.00000),complex(0.0149256,0.00000),complex(0.0313081,0.00000),complex(0.0585721,0.00000),complex(0.0990050,0.00000),complex(0.152577,0.00000),complex(0.215776,0.00000),complex(0.281338,0.00000),complex(0.339306,0.00000),complex(0.379356,0.00000),complex(0.393676,0.00000),complex(0.379356,-0.00000),complex(0.339306,0.00000),complex(0.281338,-0.00000),complex(0.215776,0.00000),complex(0.152577,-0.00000),complex(0.0990050,0.00000),complex(0.0585721,-0.00000),complex(0.0313081,0.00000),complex(0.0149256,-0.00000),complex(0.00622439,0.00000),complex(0.00220065,-0.00000),complex(0.000622439,0.00000),complex(0.000122070,-0.00000)]$
            ]
          end

      25: begin
            matrix[325:350,325:350]=[$
            [complex(0.00000,-0.000172634),complex(-0.00000,0.000863168),complex(0.00000,-0.00299010),complex(-0.00000,0.00827921),complex(0.00000,-0.0194165),complex(-0.00000,0.0397919),complex(0.00000,-0.0726498),complex(-0.00000,0.119691),complex(0.00000,-0.179537),complex(-0.00000,0.246750),complex(0.00000,-0.312116),complex(-0.00000,0.364473),complex(0.00000,-0.393676),complex(0.00000,0.393676),complex(0.00000,-0.364473),complex(0.00000,0.312116),complex(0.00000,-0.246750),complex(0.00000,0.179537),complex(0.00000,-0.119691),complex(0.00000,0.0726498),complex(0.00000,-0.0397919),complex(0.00000,0.0194165),complex(0.00000,-0.00827921),complex(0.00000,0.00299010),complex(0.00000,-0.000863168),complex(0.00000,0.000172634)],$
            [complex(0.000863168,0.00000),complex(-0.00397057,0.00000),complex(0.0125584,0.00000),complex(-0.0314610,0.00000),complex(0.0660160,0.00000),complex(-0.119376,0.00000),complex(0.188889,0.00000),complex(-0.263321,0.00000),complex(0.323166,0.00000),complex(-0.345450,0.00000),complex(0.312116,0.00000),complex(-0.218684,0.00000),complex(0.0787352,0.00000),complex(0.0787352,0.00000),complex(-0.218684,0.00000),complex(0.312116,0.00000),complex(-0.345450,0.00000),complex(0.323166,0.00000),complex(-0.263321,0.00000),complex(0.188889,0.00000),complex(-0.119376,0.00000),complex(0.0660160,0.00000),complex(-0.0314610,0.00000),complex(0.0125584,0.00000),complex(-0.00397057,0.00000),complex(0.000863168,0.00000)],$
            [complex(0.00000,0.00299010),complex(0.00000,-0.0125584),complex(0.00000,0.0359078),complex(0.00000,-0.0803041),complex(0.00000,0.147973),complex(0.00000,-0.229739),complex(0.00000,0.302000),complex(0.00000,-0.331698),complex(0.00000,0.290236),complex(0.00000,-0.170953),0,complex(-0.00000,0.168343),complex(0.00000,-0.272747),complex(0.00000,0.272747),complex(0.00000,-0.168343),0,complex(0.00000,0.170953),complex(-0.00000,-0.290236),complex(0.00000,0.331698),complex(-0.00000,-0.302000),complex(0.00000,0.229739),complex(-0.00000,-0.147973),complex(0.00000,0.0803041),complex(-0.00000,-0.0359078),complex(0.00000,0.0125584),complex(-0.00000,-0.00299010)],$
            [complex(-0.00827921,-0.00000),complex(0.0314610,0.00000),complex(-0.0803041,-0.00000),complex(0.157442,0.00000),complex(-0.247775,-0.00000),complex(0.315293,0.00000),complex(-0.315089,-0.00000),complex(0.219625,0.00000),complex(-0.0449232,-0.00000),complex(-0.144062,0.00000),complex(0.260323,0.00000),complex(-0.243194,0.00000),complex(0.0985046,0.00000),complex(0.0985046,0.00000),complex(-0.243194,0.00000),complex(0.260323,0.00000),complex(-0.144062,0.00000),complex(-0.0449232,0.00000),complex(0.219625,-0.00000),complex(-0.315089,0.00000),complex(0.315293,-0.00000),complex(-0.247775,0.00000),complex(0.157442,-0.00000),complex(-0.0803041,0.00000),complex(0.0314610,-0.00000),complex(-0.00827921,0.00000)],$
            [complex(0.00000,-0.0194165),complex(-0.00000,0.0660160),complex(0.00000,-0.147973),complex(-0.00000,0.247775),complex(0.00000,-0.317991),complex(-0.00000,0.300724),complex(0.00000,-0.169235),complex(0.00000,-0.0361823),complex(0.00000,0.213901),complex(0.00000,-0.258877),complex(0.00000,0.138753),complex(-0.00000,0.0712924),complex(0.00000,-0.231014),complex(0.00000,0.231014),complex(0.00000,-0.0712924),complex(-0.00000,-0.138753),complex(0.00000,0.258877),complex(-0.00000,-0.213901),complex(0.00000,0.0361823),complex(0.00000,0.169235),complex(0.00000,-0.300724),complex(0.00000,0.317991),complex(0.00000,-0.247775),complex(0.00000,0.147973),complex(0.00000,-0.0660160),complex(0.00000,0.0194165)],$
            [complex(0.0397919,0.00000),complex(-0.119376,0.00000),complex(0.229739,0.00000),complex(-0.315293,0.00000),complex(0.300724,0.00000),complex(-0.151572,0.00000),complex(-0.0737531,-0.00000),complex(0.239902,0.00000),complex(-0.228997,-0.00000),complex(0.0449610,0.00000),complex(0.170615,0.00000),complex(-0.243510,0.00000),complex(0.112723,0.00000),complex(0.112723,0.00000),complex(-0.243510,0.00000),complex(0.170615,0.00000),complex(0.0449610,-0.00000),complex(-0.228997,0.00000),complex(0.239902,-0.00000),complex(-0.0737531,0.00000),complex(-0.151572,0.00000),complex(0.300724,0.00000),complex(-0.315293,0.00000),complex(0.229739,0.00000),complex(-0.119376,0.00000),complex(0.0397919,0.00000)],$
            [complex(0.00000,0.0726498),complex(0.00000,-0.188889),complex(0.00000,0.302000),complex(0.00000,-0.315089),complex(0.00000,0.169235),complex(-0.00000,0.0737531),complex(0.00000,-0.245830),complex(-0.00000,0.207054),complex(0.00000,0.0119454),complex(0.00000,-0.213427),complex(0.00000,0.207666),0,complex(0.00000,-0.205803),complex(0.00000,0.205803),0,complex(-0.00000,-0.207666),complex(0.00000,0.213427),complex(-0.00000,-0.0119454),complex(0.00000,-0.207054),complex(0.00000,0.245830),complex(0.00000,-0.0737531),complex(-0.00000,-0.169235),complex(0.00000,0.315089),complex(-0.00000,-0.302000),complex(0.00000,0.188889),complex(-0.00000,-0.0726498)],$
            [complex(-0.119691,-0.00000),complex(0.263321,0.00000),complex(-0.331698,-0.00000),complex(0.219625,0.00000),complex(0.0361823,0.00000),complex(-0.239902,0.00000),complex(0.207054,0.00000),complex(0.0303835,0.00000),complex(-0.226840,-0.00000),complex(0.172252,0.00000),complex(0.0720277,0.00000),complex(-0.231303,0.00000),complex(0.124918,0.00000),complex(0.124918,0.00000),complex(-0.231303,0.00000),complex(0.0720277,0.00000),complex(0.172252,-0.00000),complex(-0.226840,0.00000),complex(0.0303835,-0.00000),complex(0.207054,0.00000),complex(-0.239902,0.00000),complex(0.0361823,0.00000),complex(0.219625,-0.00000),complex(-0.331698,0.00000),complex(0.263321,-0.00000),complex(-0.119691,0.00000)],$
            [complex(0.00000,-0.179537),complex(-0.00000,0.323166),complex(0.00000,-0.290236),complex(-0.00000,0.0449232),complex(0.00000,0.213901),complex(0.00000,-0.228997),complex(0.00000,-0.0119454),complex(-0.00000,0.226840),complex(0.00000,-0.158650),complex(0.00000,-0.104633),complex(0.00000,0.229588),complex(0.00000,-0.0578258),complex(0.00000,-0.187377),complex(0.00000,0.187377),complex(0.00000,0.0578258),complex(-0.00000,-0.229588),complex(0.00000,0.104633),complex(0.00000,0.158650),complex(0.00000,-0.226840),complex(0.00000,0.0119454),complex(0.00000,0.228997),complex(-0.00000,-0.213901),complex(0.00000,-0.0449232),complex(0.00000,0.290236),complex(0.00000,-0.323166),complex(0.00000,0.179537)],$
            [complex(0.246750,0.00000),complex(-0.345450,0.00000),complex(0.170953,0.00000),complex(0.144062,0.00000),complex(-0.258877,-0.00000),complex(0.0449610,0.00000),complex(0.213427,0.00000),complex(-0.172252,0.00000),complex(-0.104633,-0.00000),complex(0.226323,0.00000),complex(-0.0229284,-0.00000),complex(-0.210372,0.00000),complex(0.136337,0.00000),complex(0.136337,0.00000),complex(-0.210372,0.00000),complex(-0.0229284,0.00000),complex(0.226323,-0.00000),complex(-0.104633,0.00000),complex(-0.172252,0.00000),complex(0.213427,0.00000),complex(0.0449610,-0.00000),complex(-0.258877,0.00000),complex(0.144062,-0.00000),complex(0.170953,0.00000),complex(-0.345450,0.00000),complex(0.246750,0.00000)],$
            [complex(0.00000,0.312116),complex(0.00000,-0.312116),0,complex(-0.00000,0.260323),complex(0.00000,-0.138753),complex(0.00000,-0.170615),complex(0.00000,0.207666),complex(-0.00000,0.0720277),complex(0.00000,-0.229588),complex(-0.00000,0.0229284),complex(0.00000,0.215447),complex(0.00000,-0.106441),complex(0.00000,-0.172454),complex(0.00000,0.172454),complex(0.00000,0.106441),complex(-0.00000,-0.215447),complex(0.00000,-0.0229284),complex(0.00000,0.229588),complex(0.00000,-0.0720277),complex(-0.00000,-0.207666),complex(0.00000,0.170615),complex(0.00000,0.138753),complex(0.00000,-0.260323),0,complex(0.00000,0.312116),complex(-0.00000,-0.312116)],$
            [complex(-0.364473,-0.00000),complex(0.218684,0.00000),complex(0.168343,0.00000),complex(-0.243194,0.00000),complex(-0.0712924,-0.00000),complex(0.243510,0.00000),0,complex(-0.231303,0.00000),complex(0.0578258,0.00000),complex(0.210372,0.00000),complex(-0.106441,-0.00000),complex(-0.182301,0.00000),complex(0.147681,0.00000),complex(0.147681,0.00000),complex(-0.182301,0.00000),complex(-0.106441,0.00000),complex(0.210372,-0.00000),complex(0.0578258,0.00000),complex(-0.231303,0.00000),0,complex(0.243510,-0.00000),complex(-0.0712924,0.00000),complex(-0.243194,0.00000),complex(0.168343,0.00000),complex(0.218684,-0.00000),complex(-0.364473,0.00000)],$
            [complex(0.00000,-0.393676),complex(-0.00000,0.0787352),complex(0.00000,0.272747),complex(0.00000,-0.0985046),complex(0.00000,-0.231014),complex(-0.00000,0.112723),complex(0.00000,0.205803),complex(0.00000,-0.124918),complex(0.00000,-0.187377),complex(-0.00000,0.136337),complex(0.00000,0.172454),complex(0.00000,-0.147681),complex(0.00000,-0.159513),complex(0.00000,0.159513),complex(0.00000,0.147681),complex(-0.00000,-0.172454),complex(0.00000,-0.136337),complex(0.00000,0.187377),complex(0.00000,0.124918),complex(-0.00000,-0.205803),complex(0.00000,-0.112723),complex(0.00000,0.231014),complex(0.00000,0.0985046),complex(-0.00000,-0.272747),complex(0.00000,-0.0787352),complex(0.00000,0.393676)],$
            [complex(0.393676,0.00000),complex(0.0787352,0.00000),complex(-0.272747,-0.00000),complex(-0.0985046,0.00000),complex(0.231014,0.00000),complex(0.112723,0.00000),complex(-0.205803,-0.00000),complex(-0.124918,0.00000),complex(0.187377,0.00000),complex(0.136337,0.00000),complex(-0.172454,-0.00000),complex(-0.147681,0.00000),complex(0.159513,0.00000),complex(0.159513,0.00000),complex(-0.147681,0.00000),complex(-0.172454,0.00000),complex(0.136337,-0.00000),complex(0.187377,0.00000),complex(-0.124918,0.00000),complex(-0.205803,0.00000),complex(0.112723,-0.00000),complex(0.231014,0.00000),complex(-0.0985046,0.00000),complex(-0.272747,0.00000),complex(0.0787352,-0.00000),complex(0.393676,0.00000)],$
            [complex(0.00000,0.364473),complex(-0.00000,0.218684),complex(0.00000,-0.168343),complex(0.00000,-0.243194),complex(0.00000,0.0712924),complex(-0.00000,0.243510),0,complex(0.00000,-0.231303),complex(0.00000,-0.0578258),complex(-0.00000,0.210372),complex(0.00000,0.106441),complex(0.00000,-0.182301),complex(0.00000,-0.147681),complex(0.00000,0.147681),complex(0.00000,0.182301),complex(-0.00000,-0.106441),complex(0.00000,-0.210372),complex(0.00000,0.0578258),complex(0.00000,0.231303),0,complex(0.00000,-0.243510),complex(-0.00000,-0.0712924),complex(0.00000,0.243194),complex(0.00000,0.168343),complex(0.00000,-0.218684),complex(-0.00000,-0.364473)],$
            [complex(-0.312116,-0.00000),complex(-0.312116,0.00000),complex(-4.81332e-16,-0.00000),complex(0.260323,0.00000),complex(0.138753,0.00000),complex(-0.170615,0.00000),complex(-0.207666,-0.00000),complex(0.0720277,0.00000),complex(0.229588,0.00000),complex(0.0229284,0.00000),complex(-0.215447,-0.00000),complex(-0.106441,0.00000),complex(0.172454,0.00000),complex(0.172454,0.00000),complex(-0.106441,0.00000),complex(-0.215447,0.00000),complex(0.0229284,-0.00000),complex(0.229588,0.00000),complex(0.0720277,-0.00000),complex(-0.207666,0.00000),complex(-0.170615,0.00000),complex(0.138753,0.00000),complex(0.260323,-0.00000),complex(-3.85065e-16,0.00000),complex(-0.312116,0.00000),complex(-0.312116,0.00000)],$
            [complex(0.00000,-0.246750),complex(0.00000,-0.345450),complex(0.00000,-0.170953),complex(-0.00000,0.144062),complex(0.00000,0.258877),complex(-0.00000,0.0449610),complex(0.00000,-0.213427),complex(0.00000,-0.172252),complex(0.00000,0.104633),complex(-0.00000,0.226323),complex(0.00000,0.0229284),complex(0.00000,-0.210372),complex(0.00000,-0.136337),complex(0.00000,0.136337),complex(0.00000,0.210372),complex(-0.00000,-0.0229284),complex(0.00000,-0.226323),complex(-0.00000,-0.104633),complex(0.00000,0.172252),complex(0.00000,0.213427),complex(0.00000,-0.0449610),complex(-0.00000,-0.258877),complex(0.00000,-0.144062),complex(0.00000,0.170953),complex(0.00000,0.345450),complex(0.00000,0.246750)],$
            [complex(0.179537,0.00000),complex(0.323166,0.00000),complex(0.290236,0.00000),complex(0.0449232,0.00000),complex(-0.213901,-0.00000),complex(-0.228997,0.00000),complex(0.0119454,0.00000),complex(0.226840,0.00000),complex(0.158650,0.00000),complex(-0.104633,0.00000),complex(-0.229588,-0.00000),complex(-0.0578258,0.00000),complex(0.187377,0.00000),complex(0.187377,0.00000),complex(-0.0578258,0.00000),complex(-0.229588,0.00000),complex(-0.104633,0.00000),complex(0.158650,0.00000),complex(0.226840,-0.00000),complex(0.0119454,0.00000),complex(-0.228997,0.00000),complex(-0.213901,0.00000),complex(0.0449232,-0.00000),complex(0.290236,0.00000),complex(0.323166,-0.00000),complex(0.179537,0.00000)],$
            [complex(0.00000,0.119691),complex(-0.00000,0.263321),complex(0.00000,0.331698),complex(-0.00000,0.219625),complex(0.00000,-0.0361823),complex(0.00000,-0.239902),complex(0.00000,-0.207054),complex(-0.00000,0.0303835),complex(0.00000,0.226840),complex(-0.00000,0.172252),complex(0.00000,-0.0720277),complex(0.00000,-0.231303),complex(0.00000,-0.124918),complex(0.00000,0.124918),complex(0.00000,0.231303),complex(0.00000,0.0720277),complex(0.00000,-0.172252),complex(-0.00000,-0.226840),complex(0.00000,-0.0303835),complex(0.00000,0.207054),complex(0.00000,0.239902),complex(0.00000,0.0361823),complex(0.00000,-0.219625),complex(-0.00000,-0.331698),complex(0.00000,-0.263321),complex(-0.00000,-0.119691)],$
            [complex(-0.0726498,-0.00000),complex(-0.188889,0.00000),complex(-0.302000,-0.00000),complex(-0.315089,0.00000),complex(-0.169235,-0.00000),complex(0.0737531,0.00000),complex(0.245830,0.00000),complex(0.207054,0.00000),complex(-0.0119454,-0.00000),complex(-0.213427,0.00000),complex(-0.207666,-0.00000),0,complex(0.205803,0.00000),complex(0.205803,0.00000),0,complex(-0.207666,0.00000),complex(-0.213427,0.00000),complex(-0.0119454,0.00000),complex(0.207054,-0.00000),complex(0.245830,0.00000),complex(0.0737531,-0.00000),complex(-0.169235,0.00000),complex(-0.315089,0.00000),complex(-0.302000,0.00000),complex(-0.188889,0.00000),complex(-0.0726498,0.00000)],$
            [complex(0.00000,-0.0397919),complex(0.00000,-0.119376),complex(0.00000,-0.229739),complex(0.00000,-0.315293),complex(0.00000,-0.300724),complex(0.00000,-0.151572),complex(0.00000,0.0737531),complex(-0.00000,0.239902),complex(0.00000,0.228997),complex(-0.00000,0.0449610),complex(0.00000,-0.170615),complex(0.00000,-0.243510),complex(0.00000,-0.112723),complex(0.00000,0.112723),complex(0.00000,0.243510),complex(0.00000,0.170615),complex(0.00000,-0.0449610),complex(-0.00000,-0.228997),complex(0.00000,-0.239902),complex(-0.00000,-0.0737531),complex(0.00000,0.151572),complex(0.00000,0.300724),complex(0.00000,0.315293),complex(0.00000,0.229739),complex(0.00000,0.119376),complex(0.00000,0.0397919)],$
            [complex(0.0194165,0.00000),complex(0.0660160,0.00000),complex(0.147973,0.00000),complex(0.247775,0.00000),complex(0.317991,0.00000),complex(0.300724,0.00000),complex(0.169235,0.00000),complex(-0.0361823,0.00000),complex(-0.213901,-0.00000),complex(-0.258877,0.00000),complex(-0.138753,-0.00000),complex(0.0712924,0.00000),complex(0.231014,0.00000),complex(0.231014,0.00000),complex(0.0712924,-0.00000),complex(-0.138753,0.00000),complex(-0.258877,0.00000),complex(-0.213901,0.00000),complex(-0.0361823,0.00000),complex(0.169235,0.00000),complex(0.300724,-0.00000),complex(0.317991,0.00000),complex(0.247775,-0.00000),complex(0.147973,0.00000),complex(0.0660160,-0.00000),complex(0.0194165,0.00000)],$
            [complex(0.00000,0.00827921),complex(-0.00000,0.0314610),complex(0.00000,0.0803041),complex(-0.00000,0.157442),complex(0.00000,0.247775),complex(-0.00000,0.315293),complex(0.00000,0.315089),complex(-0.00000,0.219625),complex(0.00000,0.0449232),complex(0.00000,-0.144062),complex(0.00000,-0.260323),complex(0.00000,-0.243194),complex(0.00000,-0.0985046),complex(0.00000,0.0985046),complex(0.00000,0.243194),complex(0.00000,0.260323),complex(0.00000,0.144062),complex(-0.00000,-0.0449232),complex(0.00000,-0.219625),complex(-0.00000,-0.315089),complex(0.00000,-0.315293),complex(-0.00000,-0.247775),complex(0.00000,-0.157442),complex(-0.00000,-0.0803041),complex(0.00000,-0.0314610),complex(-0.00000,-0.00827921)],$
            [complex(-0.00299010,-0.00000),complex(-0.0125584,0.00000),complex(-0.0359078,-0.00000),complex(-0.0803041,0.00000),complex(-0.147973,-0.00000),complex(-0.229739,0.00000),complex(-0.302000,-0.00000),complex(-0.331698,0.00000),complex(-0.290236,-0.00000),complex(-0.170953,0.00000),0,complex(0.168343,0.00000),complex(0.272747,0.00000),complex(0.272747,0.00000),complex(0.168343,-0.00000),0,complex(-0.170953,0.00000),complex(-0.290236,0.00000),complex(-0.331698,0.00000),complex(-0.302000,0.00000),complex(-0.229739,0.00000),complex(-0.147973,0.00000),complex(-0.0803041,0.00000),complex(-0.0359078,0.00000),complex(-0.0125584,0.00000),complex(-0.00299010,0.00000)],$
            [complex(0.00000,-0.000863168),complex(0.00000,-0.00397057),complex(0.00000,-0.0125584),complex(0.00000,-0.0314610),complex(0.00000,-0.0660160),complex(0.00000,-0.119376),complex(0.00000,-0.188889),complex(0.00000,-0.263321),complex(0.00000,-0.323166),complex(0.00000,-0.345450),complex(0.00000,-0.312116),complex(0.00000,-0.218684),complex(0.00000,-0.0787352),complex(0.00000,0.0787352),complex(0.00000,0.218684),complex(0.00000,0.312116),complex(0.00000,0.345450),complex(0.00000,0.323166),complex(0.00000,0.263321),complex(0.00000,0.188889),complex(0.00000,0.119376),complex(0.00000,0.0660160),complex(0.00000,0.0314610),complex(0.00000,0.0125584),complex(0.00000,0.00397057),complex(0.00000,0.000863168)],$
            [complex(0.000172634,0.00000),complex(0.000863168,0.00000),complex(0.00299010,0.00000),complex(0.00827921,0.00000),complex(0.0194165,0.00000),complex(0.0397919,0.00000),complex(0.0726498,0.00000),complex(0.119691,0.00000),complex(0.179537,0.00000),complex(0.246750,0.00000),complex(0.312116,0.00000),complex(0.364473,0.00000),complex(0.393676,0.00000),complex(0.393676,0.00000),complex(0.364473,-0.00000),complex(0.312116,0.00000),complex(0.246750,-0.00000),complex(0.179537,0.00000),complex(0.119691,-0.00000),complex(0.0726498,0.00000),complex(0.0397919,-0.00000),complex(0.0194165,0.00000),complex(0.00827921,-0.00000),complex(0.00299010,0.00000),complex(0.000863168,-0.00000),complex(0.000172634,0.00000)]$
            ]
          end

       
      24: begin
            matrix[300:324,300:324]=[$
            [complex(0.000244141,0.00000),complex(-0.00119604,-0.00000),complex(0.00405597,0.00000),complex(-0.0109836,-0.00000),complex(0.0251666,0.00000),complex(-0.0503333,-0.00000),complex(0.0895687,0.00000),complex(-0.143629,-0.00000),complex(0.209374,0.00000),complex(-0.279165,-0.00000),complex(0.341906,0.00000),complex(-0.385722,-0.00000),complex(0.401473,0.00000),complex(-0.385722,0.00000),complex(0.341906,0.00000),complex(-0.279165,0.00000),complex(0.209374,0.00000),complex(-0.143629,0.00000),complex(0.0895687,0.00000),complex(-0.0503333,0.00000),complex(0.0251666,0.00000),complex(-0.0109836,0.00000),complex(0.00405597,0.00000),complex(-0.00119604,0.00000),complex(0.000244141,0.00000)],$
            [complex(0.00000,0.00119604),complex(0.00000,-0.00537109),complex(0.00000,0.0165584),complex(0.00000,-0.0403564),complex(0.00000,0.0821939),complex(0.00000,-0.143839),complex(0.00000,0.219398),complex(0.00000,-0.293182),complex(0.00000,0.341906),complex(0.00000,-0.341906),complex(0.00000,0.279165),complex(0.00000,-0.157470),0,complex(0.00000,0.157470),complex(0.00000,-0.279165),complex(0.00000,0.341906),complex(0.00000,-0.341906),complex(0.00000,0.293182),complex(0.00000,-0.219398),complex(0.00000,0.143839),complex(0.00000,-0.0821939),complex(0.00000,0.0403564),complex(0.00000,-0.0165584),complex(0.00000,0.00537109),complex(0.00000,-0.00119604)],$
            [complex(-0.00405597,0.00000),complex(0.0165584,0.00000),complex(-0.0458984,0.00000),complex(0.0991704,0.00000),complex(-0.175723,0.00000),complex(0.260555,0.00000),complex(-0.323484,0.00000),complex(0.328528,0.00000),complex(-0.252057,0.00000),complex(0.100823,0.00000),complex(0.0823214,0.00000),complex(-0.232178,-0.00000),complex(0.289990,0.00000),complex(-0.232178,0.00000),complex(0.0823214,0.00000),complex(0.100823,-0.00000),complex(-0.252057,0.00000),complex(0.328528,-0.00000),complex(-0.323484,0.00000),complex(0.260555,-0.00000),complex(-0.175723,0.00000),complex(0.0991704,-0.00000),complex(-0.0458984,0.00000),complex(0.0165584,-0.00000),complex(-0.00405597,0.00000)],$
            [complex(0.00000,-0.0109836),complex(-0.00000,0.0403564),complex(0.00000,-0.0991704),complex(-0.00000,0.186035),complex(0.00000,-0.277461),complex(-0.00000,0.328925),complex(0.00000,-0.294654),complex(-0.00000,0.159628),complex(0.00000,0.0372312),complex(0.00000,-0.210977),complex(0.00000,0.273593),complex(0.00000,-0.188622),0,complex(0.00000,0.188622),complex(0.00000,-0.273593),complex(0.00000,0.210977),complex(0.00000,-0.0372312),complex(-0.00000,-0.159628),complex(0.00000,0.294654),complex(-0.00000,-0.328925),complex(0.00000,0.277461),complex(-0.00000,-0.186035),complex(0.00000,0.0991704),complex(-0.00000,-0.0403564),complex(0.00000,0.0109836)],$
            [complex(0.0251666,0.00000),complex(-0.0821939,-0.00000),complex(0.175723,0.00000),complex(-0.277461,-0.00000),complex(0.328613,0.00000),complex(-0.271484,-0.00000),complex(0.0990549,0.00000),complex(0.117041,0.00000),complex(-0.255922,0.00000),complex(0.227487,0.00000),complex(-0.0464355,0.00000),complex(-0.164643,-0.00000),complex(0.257048,0.00000),complex(-0.164643,0.00000),complex(-0.0464355,0.00000),complex(0.227487,-0.00000),complex(-0.255922,0.00000),complex(0.117041,-0.00000),complex(0.0990549,0.00000),complex(-0.271484,0.00000),complex(0.328613,0.00000),complex(-0.277461,0.00000),complex(0.175723,0.00000),complex(-0.0821939,0.00000),complex(0.0251666,0.00000)],$
            [complex(0.00000,0.0503333),complex(0.00000,-0.143839),complex(0.00000,0.260555),complex(0.00000,-0.328925),complex(0.00000,0.271484),complex(0.00000,-0.0786133),complex(0.00000,-0.151189),complex(-0.00000,0.263342),complex(0.00000,-0.170615),complex(0.00000,-0.0568717),complex(0.00000,0.232178),complex(0.00000,-0.205803),0,complex(0.00000,0.205803),complex(0.00000,-0.232178),complex(0.00000,0.0568717),complex(0.00000,0.170615),complex(-0.00000,-0.263342),complex(0.00000,0.151189),complex(0.00000,0.0786133),complex(0.00000,-0.271484),complex(0.00000,0.328925),complex(0.00000,-0.260555),complex(0.00000,0.143839),complex(0.00000,-0.0503333)],$
            [complex(-0.0895687,0.00000),complex(0.219398,0.00000),complex(-0.323484,0.00000),complex(0.294654,0.00000),complex(-0.0990549,0.00000),complex(-0.151189,-0.00000),complex(0.262695,0.00000),complex(-0.137024,-0.00000),complex(-0.111857,0.00000),complex(0.245020,0.00000),complex(-0.130473,0.00000),complex(-0.115652,-0.00000),complex(0.240748,0.00000),complex(-0.115652,0.00000),complex(-0.130473,0.00000),complex(0.245020,-0.00000),complex(-0.111857,0.00000),complex(-0.137024,0.00000),complex(0.262695,0.00000),complex(-0.151189,0.00000),complex(-0.0990549,0.00000),complex(0.294654,-0.00000),complex(-0.323484,0.00000),complex(0.219398,-0.00000),complex(-0.0895687,0.00000)],$
            [complex(0.00000,-0.143629),complex(-0.00000,0.293182),complex(0.00000,-0.328528),complex(-0.00000,0.159628),complex(0.00000,0.117041),complex(0.00000,-0.263342),complex(0.00000,0.137024),complex(-0.00000,0.128418),complex(0.00000,-0.242007),complex(-0.00000,0.0768729),complex(0.00000,0.174351),complex(0.00000,-0.216364),0,complex(0.00000,0.216364),complex(0.00000,-0.174351),complex(-0.00000,-0.0768729),complex(0.00000,0.242007),complex(-0.00000,-0.128418),complex(0.00000,-0.137024),complex(0.00000,0.263342),complex(0.00000,-0.117041),complex(-0.00000,-0.159628),complex(0.00000,0.328528),complex(-0.00000,-0.293182),complex(0.00000,0.143629)],$
            [complex(0.209374,0.00000),complex(-0.341906,-0.00000),complex(0.252057,0.00000),complex(0.0372312,0.00000),complex(-0.255922,0.00000),complex(0.170615,0.00000),complex(0.111857,0.00000),complex(-0.242007,-0.00000),complex(0.0583496,0.00000),complex(0.196289,0.00000),complex(-0.185386,0.00000),complex(-0.0742123,-0.00000),complex(0.231728,0.00000),complex(-0.0742123,0.00000),complex(-0.185386,0.00000),complex(0.196289,-0.00000),complex(0.0583496,0.00000),complex(-0.242007,0.00000),complex(0.111857,0.00000),complex(0.170615,-0.00000),complex(-0.255922,0.00000),complex(0.0372312,-0.00000),complex(0.252057,0.00000),complex(-0.341906,0.00000),complex(0.209374,0.00000)],$
            [complex(0.00000,0.279165),complex(0.00000,-0.341906),complex(0.00000,0.100823),complex(-0.00000,0.210977),complex(0.00000,-0.227487),complex(0.00000,-0.0568717),complex(0.00000,0.245020),complex(0.00000,-0.0768729),complex(0.00000,-0.196289),complex(-0.00000,0.172852),complex(0.00000,0.107644),complex(0.00000,-0.222637),0,complex(0.00000,0.222637),complex(0.00000,-0.107644),complex(-0.00000,-0.172852),complex(0.00000,0.196289),complex(0.00000,0.0768729),complex(0.00000,-0.245020),complex(0.00000,0.0568717),complex(0.00000,0.227487),complex(-0.00000,-0.210977),complex(0.00000,-0.100823),complex(0.00000,0.341906),complex(0.00000,-0.279165)],$
            [complex(-0.341906,0.00000),complex(0.279165,0.00000),complex(0.0823214,0.00000),complex(-0.273593,-0.00000),complex(0.0464355,0.00000),complex(0.232178,0.00000),complex(-0.130473,0.00000),complex(-0.174351,-0.00000),complex(0.185386,0.00000),complex(0.107644,0.00000),complex(-0.216797,0.00000),complex(-0.0363565,-0.00000),complex(0.227046,0.00000),complex(-0.0363565,0.00000),complex(-0.216797,0.00000),complex(0.107644,-0.00000),complex(0.185386,0.00000),complex(-0.174351,0.00000),complex(-0.130473,0.00000),complex(0.232178,-0.00000),complex(0.0464355,0.00000),complex(-0.273593,0.00000),complex(0.0823214,0.00000),complex(0.279165,-0.00000),complex(-0.341906,0.00000)],$
            [complex(0.00000,-0.385722),complex(-0.00000,0.157470),complex(0.00000,0.232178),complex(0.00000,-0.188622),complex(0.00000,-0.164643),complex(-0.00000,0.205803),complex(0.00000,0.115652),complex(0.00000,-0.216364),complex(0.00000,-0.0742123),complex(-0.00000,0.222637),complex(0.00000,0.0363565),complex(0.00000,-0.225586),0,complex(0.00000,0.225586),complex(0.00000,-0.0363565),complex(-0.00000,-0.222637),complex(0.00000,0.0742123),complex(0.00000,0.216364),complex(0.00000,-0.115652),complex(-0.00000,-0.205803),complex(0.00000,0.164643),complex(0.00000,0.188622),complex(0.00000,-0.232178),complex(-0.00000,-0.157470),complex(0.00000,0.385722)],$
            [complex(0.401473,0.00000),0,complex(-0.289990,0.00000),complex(1.55514e-15,0.00000),complex(0.257048,0.00000),0,complex(-0.240748,0.00000),0,complex(0.231728,0.00000),0,complex(-0.227046,0.00000),0,complex(0.225586,0.00000),0,complex(-0.227046,0.00000),0,complex(0.231728,0.00000),0,complex(-0.240748,0.00000),0,complex(0.257048,0.00000),complex(1.55514e-15,-0.00000),complex(-0.289990,0.00000),0,complex(0.401473,0.00000)],$
            [complex(0.00000,0.385722),complex(-0.00000,0.157470),complex(0.00000,-0.232178),complex(0.00000,-0.188622),complex(0.00000,0.164643),complex(-0.00000,0.205803),complex(0.00000,-0.115652),complex(0.00000,-0.216364),complex(0.00000,0.0742123),complex(-0.00000,0.222637),complex(0.00000,-0.0363565),complex(0.00000,-0.225586),0,complex(0.00000,0.225586),complex(0.00000,0.0363565),complex(-0.00000,-0.222637),complex(0.00000,-0.0742123),complex(0.00000,0.216364),complex(0.00000,0.115652),complex(-0.00000,-0.205803),complex(0.00000,-0.164643),complex(0.00000,0.188622),complex(0.00000,0.232178),complex(-0.00000,-0.157470),complex(0.00000,-0.385722)],$
            [complex(-0.341906,0.00000),complex(-0.279165,-0.00000),complex(0.0823214,0.00000),complex(0.273593,0.00000),complex(0.0464355,0.00000),complex(-0.232178,-0.00000),complex(-0.130473,0.00000),complex(0.174351,0.00000),complex(0.185386,0.00000),complex(-0.107644,-0.00000),complex(-0.216797,0.00000),complex(0.0363565,0.00000),complex(0.227046,0.00000),complex(0.0363565,-0.00000),complex(-0.216797,0.00000),complex(-0.107644,0.00000),complex(0.185386,0.00000),complex(0.174351,-0.00000),complex(-0.130473,0.00000),complex(-0.232178,0.00000),complex(0.0464355,0.00000),complex(0.273593,-0.00000),complex(0.0823214,0.00000),complex(-0.279165,0.00000),complex(-0.341906,0.00000)],$
            [complex(0.00000,-0.279165),complex(0.00000,-0.341906),complex(0.00000,-0.100823),complex(-0.00000,0.210977),complex(0.00000,0.227487),complex(0.00000,-0.0568717),complex(0.00000,-0.245020),complex(0.00000,-0.0768729),complex(0.00000,0.196289),complex(-0.00000,0.172852),complex(0.00000,-0.107644),complex(0.00000,-0.222637),0,complex(0.00000,0.222637),complex(0.00000,0.107644),complex(-0.00000,-0.172852),complex(0.00000,-0.196289),complex(0.00000,0.0768729),complex(0.00000,0.245020),complex(0.00000,0.0568717),complex(0.00000,-0.227487),complex(-0.00000,-0.210977),complex(0.00000,0.100823),complex(0.00000,0.341906),complex(0.00000,0.279165)],$
            [complex(0.209374,0.00000),complex(0.341906,0.00000),complex(0.252057,0.00000),complex(-0.0372312,-0.00000),complex(-0.255922,0.00000),complex(-0.170615,-0.00000),complex(0.111857,0.00000),complex(0.242007,0.00000),complex(0.0583496,0.00000),complex(-0.196289,-0.00000),complex(-0.185386,0.00000),complex(0.0742123,0.00000),complex(0.231728,0.00000),complex(0.0742123,-0.00000),complex(-0.185386,0.00000),complex(-0.196289,0.00000),complex(0.0583496,0.00000),complex(0.242007,-0.00000),complex(0.111857,0.00000),complex(-0.170615,0.00000),complex(-0.255922,0.00000),complex(-0.0372312,0.00000),complex(0.252057,0.00000),complex(0.341906,-0.00000),complex(0.209374,0.00000)],$
            [complex(0.00000,0.143629),complex(-0.00000,0.293182),complex(0.00000,0.328528),complex(-0.00000,0.159628),complex(0.00000,-0.117041),complex(0.00000,-0.263342),complex(0.00000,-0.137024),complex(-0.00000,0.128418),complex(0.00000,0.242007),complex(-0.00000,0.0768729),complex(0.00000,-0.174351),complex(0.00000,-0.216364),0,complex(0.00000,0.216364),complex(0.00000,0.174351),complex(-0.00000,-0.0768729),complex(0.00000,-0.242007),complex(-0.00000,-0.128418),complex(0.00000,0.137024),complex(0.00000,0.263342),complex(0.00000,0.117041),complex(-0.00000,-0.159628),complex(0.00000,-0.328528),complex(-0.00000,-0.293182),complex(0.00000,-0.143629)],$
            [complex(-0.0895687,0.00000),complex(-0.219398,-0.00000),complex(-0.323484,0.00000),complex(-0.294654,-0.00000),complex(-0.0990549,0.00000),complex(0.151189,0.00000),complex(0.262695,0.00000),complex(0.137024,0.00000),complex(-0.111857,0.00000),complex(-0.245020,-0.00000),complex(-0.130473,0.00000),complex(0.115652,0.00000),complex(0.240748,0.00000),complex(0.115652,-0.00000),complex(-0.130473,0.00000),complex(-0.245020,0.00000),complex(-0.111857,0.00000),complex(0.137024,-0.00000),complex(0.262695,0.00000),complex(0.151189,-0.00000),complex(-0.0990549,0.00000),complex(-0.294654,0.00000),complex(-0.323484,0.00000),complex(-0.219398,0.00000),complex(-0.0895687,0.00000)],$
            [complex(0.00000,-0.0503333),complex(0.00000,-0.143839),complex(0.00000,-0.260555),complex(0.00000,-0.328925),complex(0.00000,-0.271484),complex(0.00000,-0.0786133),complex(0.00000,0.151189),complex(-0.00000,0.263342),complex(0.00000,0.170615),complex(0.00000,-0.0568717),complex(0.00000,-0.232178),complex(0.00000,-0.205803),0,complex(0.00000,0.205803),complex(0.00000,0.232178),complex(0.00000,0.0568717),complex(0.00000,-0.170615),complex(-0.00000,-0.263342),complex(0.00000,-0.151189),complex(0.00000,0.0786133),complex(0.00000,0.271484),complex(0.00000,0.328925),complex(0.00000,0.260555),complex(0.00000,0.143839),complex(0.00000,0.0503333)],$
            [complex(0.0251666,0.00000),complex(0.0821939,0.00000),complex(0.175723,0.00000),complex(0.277461,0.00000),complex(0.328613,0.00000),complex(0.271484,0.00000),complex(0.0990549,0.00000),complex(-0.117041,-0.00000),complex(-0.255922,0.00000),complex(-0.227487,-0.00000),complex(-0.0464355,0.00000),complex(0.164643,0.00000),complex(0.257048,0.00000),complex(0.164643,-0.00000),complex(-0.0464355,0.00000),complex(-0.227487,0.00000),complex(-0.255922,0.00000),complex(-0.117041,0.00000),complex(0.0990549,0.00000),complex(0.271484,-0.00000),complex(0.328613,0.00000),complex(0.277461,-0.00000),complex(0.175723,0.00000),complex(0.0821939,-0.00000),complex(0.0251666,0.00000)],$
            [complex(0.00000,0.0109836),complex(-0.00000,0.0403564),complex(0.00000,0.0991704),complex(-0.00000,0.186035),complex(0.00000,0.277461),complex(-0.00000,0.328925),complex(0.00000,0.294654),complex(-0.00000,0.159628),complex(0.00000,-0.0372312),complex(0.00000,-0.210977),complex(0.00000,-0.273593),complex(0.00000,-0.188622),0,complex(0.00000,0.188622),complex(0.00000,0.273593),complex(0.00000,0.210977),complex(0.00000,0.0372312),complex(-0.00000,-0.159628),complex(0.00000,-0.294654),complex(-0.00000,-0.328925),complex(0.00000,-0.277461),complex(-0.00000,-0.186035),complex(0.00000,-0.0991704),complex(-0.00000,-0.0403564),complex(0.00000,-0.0109836)],$
            [complex(-0.00405597,0.00000),complex(-0.0165584,-0.00000),complex(-0.0458984,0.00000),complex(-0.0991704,-0.00000),complex(-0.175723,0.00000),complex(-0.260555,-0.00000),complex(-0.323484,0.00000),complex(-0.328528,-0.00000),complex(-0.252057,0.00000),complex(-0.100823,-0.00000),complex(0.0823214,0.00000),complex(0.232178,0.00000),complex(0.289990,0.00000),complex(0.232178,-0.00000),complex(0.0823214,0.00000),complex(-0.100823,0.00000),complex(-0.252057,0.00000),complex(-0.328528,0.00000),complex(-0.323484,0.00000),complex(-0.260555,0.00000),complex(-0.175723,0.00000),complex(-0.0991704,0.00000),complex(-0.0458984,0.00000),complex(-0.0165584,0.00000),complex(-0.00405597,0.00000)],$
            [complex(0.00000,-0.00119604),complex(0.00000,-0.00537109),complex(0.00000,-0.0165584),complex(0.00000,-0.0403564),complex(0.00000,-0.0821939),complex(0.00000,-0.143839),complex(0.00000,-0.219398),complex(0.00000,-0.293182),complex(0.00000,-0.341906),complex(0.00000,-0.341906),complex(0.00000,-0.279165),complex(0.00000,-0.157470),0,complex(0.00000,0.157470),complex(0.00000,0.279165),complex(0.00000,0.341906),complex(0.00000,0.341906),complex(0.00000,0.293182),complex(0.00000,0.219398),complex(0.00000,0.143839),complex(0.00000,0.0821939),complex(0.00000,0.0403564),complex(0.00000,0.0165584),complex(0.00000,0.00537109),complex(0.00000,0.00119604)],$
            [complex(0.000244141,0.00000),complex(0.00119604,0.00000),complex(0.00405597,0.00000),complex(0.0109836,0.00000),complex(0.0251666,0.00000),complex(0.0503333,0.00000),complex(0.0895687,0.00000),complex(0.143629,0.00000),complex(0.209374,0.00000),complex(0.279165,0.00000),complex(0.341906,0.00000),complex(0.385722,0.00000),complex(0.401473,0.00000),complex(0.385722,-0.00000),complex(0.341906,0.00000),complex(0.279165,-0.00000),complex(0.209374,0.00000),complex(0.143629,-0.00000),complex(0.0895687,0.00000),complex(0.0503333,-0.00000),complex(0.0251666,0.00000),complex(0.0109836,-0.00000),complex(0.00405597,0.00000),complex(0.00119604,-0.00000),complex(0.000244141,0.00000)]$
            ]
          end


      23: begin
            matrix[276:299,276:299]=[$
            [complex(-0.00000,0.000345267),complex(0.00000,-0.00165584),complex(-0.00000,0.00549181),complex(0.00000,-0.0145300),complex(-0.00000,0.0324900),complex(0.00000,-0.0633346),complex(-0.00000,0.109699),complex(0.00000,-0.170953),complex(-0.00000,0.241764),complex(0.00000,-0.312116),complex(-0.00000,0.369301),complex(0.00000,-0.401473),complex(0.00000,0.401473),complex(0.00000,-0.369301),complex(0.00000,0.312116),complex(0.00000,-0.241764),complex(0.00000,0.170953),complex(0.00000,-0.109699),complex(0.00000,0.0633346),complex(0.00000,-0.0324900),complex(0.00000,0.0145300),complex(0.00000,-0.00549181),complex(0.00000,0.00165584),complex(0.00000,-0.000345267)],$
            [complex(-0.00165584,0.00000),complex(0.00725061,0.00000),complex(-0.0217573,0.00000),complex(0.0515050,0.00000),complex(-0.101619,0.00000),complex(0.171680,0.00000),complex(-0.251612,0.00000),complex(0.320816,0.00000),complex(-0.352879,0.00000),complex(0.325404,0.00000),complex(-0.231014,0.00000),complex(0.0837128,0.00000),complex(0.0837128,0.00000),complex(-0.231014,0.00000),complex(0.325404,0.00000),complex(-0.352879,0.00000),complex(0.320816,0.00000),complex(-0.251612,0.00000),complex(0.171680,0.00000),complex(-0.101619,0.00000),complex(0.0515050,0.00000),complex(-0.0217573,0.00000),complex(0.00725061,0.00000),complex(-0.00165584,0.00000)],$
            [complex(0.00000,-0.00549181),complex(0.00000,0.0217573),complex(0.00000,-0.0583501),complex(0.00000,0.121494),complex(0.00000,-0.206305),complex(0.00000,0.290672),complex(0.00000,-0.337938),complex(0.00000,0.311684),complex(0.00000,-0.197595),complex(0.00000,0.0196226),complex(-0.00000,0.162524),complex(0.00000,-0.277644),complex(0.00000,0.277644),complex(0.00000,-0.162524),complex(-0.00000,-0.0196226),complex(0.00000,0.197595),complex(-0.00000,-0.311684),complex(0.00000,0.337938),complex(-0.00000,-0.290672),complex(0.00000,0.206305),complex(-0.00000,-0.121494),complex(0.00000,0.0583501),complex(-0.00000,-0.0217573),complex(0.00000,0.00549181)],$
            [complex(0.0145300,0.00000),complex(-0.0515050,-0.00000),complex(0.121494,0.00000),complex(-0.217173,-0.00000),complex(0.304956,0.00000),complex(-0.332601,-0.00000),complex(0.258064,0.00000),complex(-0.0853075,-0.00000),complex(-0.120643,0.00000),complex(0.259582,0.00000),complex(-0.254489,0.00000),complex(0.104940,0.00000),complex(0.104940,0.00000),complex(-0.254489,0.00000),complex(0.259582,0.00000),complex(-0.120643,0.00000),complex(-0.0853075,0.00000),complex(0.258064,-0.00000),complex(-0.332601,0.00000),complex(0.304956,-0.00000),complex(-0.217173,0.00000),complex(0.121494,-0.00000),complex(-0.0515050,0.00000),complex(0.0145300,-0.00000)],$
            [complex(-0.00000,0.0324900),complex(0.00000,-0.101619),complex(-0.00000,0.206305),complex(0.00000,-0.304956),complex(-0.00000,0.328349),complex(0.00000,-0.225471),complex(-0.00000,0.0174863),complex(0.00000,0.190753),complex(0.00000,-0.269766),complex(0.00000,0.162524),complex(-0.00000,0.0588678),complex(0.00000,-0.234652),complex(0.00000,0.234652),complex(0.00000,-0.0588678),complex(-0.00000,-0.162524),complex(0.00000,0.269766),complex(-0.00000,-0.190753),complex(0.00000,-0.0174863),complex(0.00000,0.225471),complex(0.00000,-0.328349),complex(0.00000,0.304956),complex(0.00000,-0.206305),complex(0.00000,0.101619),complex(0.00000,-0.0324900)],$
            [complex(-0.0633346,0.00000),complex(0.171680,0.00000),complex(-0.290672,0.00000),complex(0.332601,0.00000),complex(-0.225471,0.00000),complex(-0.00448847,-0.00000),complex(0.217081,0.00000),complex(-0.254422,-0.00000),complex(0.0830322,0.00000),complex(0.154836,0.00000),complex(-0.251654,0.00000),complex(0.120374,0.00000),complex(0.120374,0.00000),complex(-0.251654,0.00000),complex(0.154836,0.00000),complex(0.0830322,-0.00000),complex(-0.254422,0.00000),complex(0.217081,-0.00000),complex(-0.00448847,0.00000),complex(-0.225471,0.00000),complex(0.332601,0.00000),complex(-0.290672,0.00000),complex(0.171680,0.00000),complex(-0.0633346,0.00000)],$
            [complex(0.00000,-0.109699),complex(0.00000,0.251612),complex(0.00000,-0.337938),complex(0.00000,0.258064),complex(0.00000,-0.0174863),complex(0.00000,-0.217081),complex(-0.00000,0.246175),complex(0.00000,-0.0414306),complex(0.00000,-0.197081),complex(0.00000,0.226925),complex(0.00000,-0.0174351),complex(0.00000,-0.208494),complex(0.00000,0.208494),complex(0.00000,0.0174351),complex(-0.00000,-0.226925),complex(0.00000,0.197081),complex(0.00000,0.0414306),complex(0.00000,-0.246175),complex(0.00000,0.217081),complex(0.00000,0.0174863),complex(-0.00000,-0.258064),complex(0.00000,0.337938),complex(-0.00000,-0.251612),complex(0.00000,0.109699)],$
            [complex(0.170953,0.00000),complex(-0.320816,-0.00000),complex(0.311684,0.00000),complex(-0.0853075,-0.00000),complex(-0.190753,0.00000),complex(0.254422,0.00000),complex(-0.0414306,0.00000),complex(-0.208196,-0.00000),complex(0.205566,0.00000),complex(0.0434954,0.00000),complex(-0.234947,0.00000),complex(0.133788,0.00000),complex(0.133788,0.00000),complex(-0.234947,0.00000),complex(0.0434954,0.00000),complex(0.205566,-0.00000),complex(-0.208196,0.00000),complex(-0.0414306,0.00000),complex(0.254422,0.00000),complex(-0.190753,0.00000),complex(-0.0853075,0.00000),complex(0.311684,-0.00000),complex(-0.320816,0.00000),complex(0.170953,-0.00000)],$
            [complex(-0.00000,0.241764),complex(0.00000,-0.352879),complex(-0.00000,0.197595),complex(0.00000,0.120643),complex(0.00000,-0.269766),complex(0.00000,0.0830322),complex(-0.00000,0.197081),complex(0.00000,-0.205566),complex(0.00000,-0.0628386),complex(0.00000,0.238024),complex(0.00000,-0.0791106),complex(0.00000,-0.189205),complex(0.00000,0.189205),complex(0.00000,0.0791106),complex(-0.00000,-0.238024),complex(0.00000,0.0628386),complex(0.00000,0.205566),complex(0.00000,-0.197081),complex(-0.00000,-0.0830322),complex(0.00000,0.269766),complex(-0.00000,-0.120643),complex(0.00000,-0.197595),complex(0.00000,0.352879),complex(0.00000,-0.241764)],$
            [complex(-0.312116,0.00000),complex(0.325404,0.00000),complex(-0.0196226,0.00000),complex(-0.259582,-0.00000),complex(0.162524,0.00000),complex(0.154836,0.00000),complex(-0.226925,0.00000),complex(-0.0434954,-0.00000),complex(0.238024,0.00000),complex(-0.0600764,-0.00000),complex(-0.208348,0.00000),complex(0.146558,0.00000),complex(0.146558,0.00000),complex(-0.208348,0.00000),complex(-0.0600764,0.00000),complex(0.238024,-0.00000),complex(-0.0434954,0.00000),complex(-0.226925,0.00000),complex(0.154836,0.00000),complex(0.162524,-0.00000),complex(-0.259582,0.00000),complex(-0.0196226,0.00000),complex(0.325404,0.00000),complex(-0.312116,0.00000)],$
            [complex(0.00000,-0.369301),complex(0.00000,0.231014),complex(-0.00000,0.162524),complex(0.00000,-0.254489),complex(0.00000,-0.0588678),complex(0.00000,0.251654),complex(0.00000,-0.0174351),complex(0.00000,-0.234947),complex(-0.00000,0.0791106),complex(0.00000,0.208348),complex(0.00000,-0.130511),complex(0.00000,-0.173409),complex(0.00000,0.173409),complex(0.00000,0.130511),complex(-0.00000,-0.208348),complex(0.00000,-0.0791106),complex(0.00000,0.234947),complex(0.00000,0.0174351),complex(-0.00000,-0.251654),complex(0.00000,0.0588678),complex(0.00000,0.254489),complex(0.00000,-0.162524),complex(-0.00000,-0.231014),complex(0.00000,0.369301)],$
            [complex(0.401473,0.00000),complex(-0.0837128,-0.00000),complex(-0.277644,0.00000),complex(0.104940,0.00000),complex(0.234652,0.00000),complex(-0.120374,-0.00000),complex(-0.208494,0.00000),complex(0.133788,0.00000),complex(0.189205,0.00000),complex(-0.146558,-0.00000),complex(-0.173409,0.00000),complex(0.159513,0.00000),complex(0.159513,0.00000),complex(-0.173409,0.00000),complex(-0.146558,0.00000),complex(0.189205,-0.00000),complex(0.133788,0.00000),complex(-0.208494,0.00000),complex(-0.120374,0.00000),complex(0.234652,-0.00000),complex(0.104940,0.00000),complex(-0.277644,0.00000),complex(-0.0837128,0.00000),complex(0.401473,-0.00000)],$
            [complex(-0.00000,0.401473),complex(0.00000,0.0837128),complex(0.00000,-0.277644),complex(0.00000,-0.104940),complex(-0.00000,0.234652),complex(0.00000,0.120374),complex(0.00000,-0.208494),complex(0.00000,-0.133788),complex(-0.00000,0.189205),complex(0.00000,0.146558),complex(0.00000,-0.173409),complex(0.00000,-0.159513),complex(0.00000,0.159513),complex(0.00000,0.173409),complex(-0.00000,-0.146558),complex(0.00000,-0.189205),complex(0.00000,0.133788),complex(0.00000,0.208494),complex(-0.00000,-0.120374),complex(0.00000,-0.234652),complex(0.00000,0.104940),complex(0.00000,0.277644),complex(-0.00000,-0.0837128),complex(0.00000,-0.401473)],$
            [complex(-0.369301,0.00000),complex(-0.231014,-0.00000),complex(0.162524,0.00000),complex(0.254489,0.00000),complex(-0.0588678,0.00000),complex(-0.251654,-0.00000),complex(-0.0174351,0.00000),complex(0.234947,0.00000),complex(0.0791106,0.00000),complex(-0.208348,-0.00000),complex(-0.130511,0.00000),complex(0.173409,0.00000),complex(0.173409,0.00000),complex(-0.130511,0.00000),complex(-0.208348,0.00000),complex(0.0791106,-0.00000),complex(0.234947,0.00000),complex(-0.0174351,0.00000),complex(-0.251654,0.00000),complex(-0.0588678,0.00000),complex(0.254489,0.00000),complex(0.162524,-0.00000),complex(-0.231014,0.00000),complex(-0.369301,0.00000)],$
            [complex(0.00000,-0.312116),complex(0.00000,-0.325404),complex(0.00000,-0.0196226),complex(0.00000,0.259582),complex(-0.00000,0.162524),complex(0.00000,-0.154836),complex(0.00000,-0.226925),complex(0.00000,0.0434954),complex(-0.00000,0.238024),complex(0.00000,0.0600764),complex(0.00000,-0.208348),complex(0.00000,-0.146558),complex(0.00000,0.146558),complex(0.00000,0.208348),complex(-0.00000,-0.0600764),complex(0.00000,-0.238024),complex(-0.00000,-0.0434954),complex(0.00000,0.226925),complex(0.00000,0.154836),complex(0.00000,-0.162524),complex(-0.00000,-0.259582),complex(0.00000,0.0196226),complex(0.00000,0.325404),complex(0.00000,0.312116)],$
            [complex(0.241764,0.00000),complex(0.352879,0.00000),complex(0.197595,0.00000),complex(-0.120643,-0.00000),complex(-0.269766,0.00000),complex(-0.0830322,-0.00000),complex(0.197081,0.00000),complex(0.205566,0.00000),complex(-0.0628386,0.00000),complex(-0.238024,-0.00000),complex(-0.0791106,0.00000),complex(0.189205,0.00000),complex(0.189205,0.00000),complex(-0.0791106,0.00000),complex(-0.238024,0.00000),complex(-0.0628386,0.00000),complex(0.205566,0.00000),complex(0.197081,-0.00000),complex(-0.0830322,0.00000),complex(-0.269766,0.00000),complex(-0.120643,0.00000),complex(0.197595,-0.00000),complex(0.352879,0.00000),complex(0.241764,-0.00000)],$
            [complex(-0.00000,0.170953),complex(0.00000,0.320816),complex(-0.00000,0.311684),complex(0.00000,0.0853075),complex(0.00000,-0.190753),complex(0.00000,-0.254422),complex(0.00000,-0.0414306),complex(0.00000,0.208196),complex(-0.00000,0.205566),complex(0.00000,-0.0434954),complex(0.00000,-0.234947),complex(0.00000,-0.133788),complex(0.00000,0.133788),complex(0.00000,0.234947),complex(0.00000,0.0434954),complex(0.00000,-0.205566),complex(-0.00000,-0.208196),complex(0.00000,0.0414306),complex(0.00000,0.254422),complex(0.00000,0.190753),complex(-0.00000,-0.0853075),complex(0.00000,-0.311684),complex(-0.00000,-0.320816),complex(0.00000,-0.170953)],$
            [complex(-0.109699,0.00000),complex(-0.251612,-0.00000),complex(-0.337938,0.00000),complex(-0.258064,-0.00000),complex(-0.0174863,0.00000),complex(0.217081,0.00000),complex(0.246175,0.00000),complex(0.0414306,0.00000),complex(-0.197081,0.00000),complex(-0.226925,-0.00000),complex(-0.0174351,0.00000),complex(0.208494,0.00000),complex(0.208494,0.00000),complex(-0.0174351,0.00000),complex(-0.226925,0.00000),complex(-0.197081,0.00000),complex(0.0414306,0.00000),complex(0.246175,-0.00000),complex(0.217081,0.00000),complex(-0.0174863,0.00000),complex(-0.258064,0.00000),complex(-0.337938,0.00000),complex(-0.251612,0.00000),complex(-0.109699,0.00000)],$
            [complex(0.00000,-0.0633346),complex(0.00000,-0.171680),complex(0.00000,-0.290672),complex(0.00000,-0.332601),complex(0.00000,-0.225471),complex(0.00000,0.00448847),complex(-0.00000,0.217081),complex(0.00000,0.254422),complex(-0.00000,0.0830322),complex(0.00000,-0.154836),complex(0.00000,-0.251654),complex(0.00000,-0.120374),complex(0.00000,0.120374),complex(0.00000,0.251654),complex(0.00000,0.154836),complex(0.00000,-0.0830322),complex(-0.00000,-0.254422),complex(0.00000,-0.217081),complex(-0.00000,-0.00448847),complex(0.00000,0.225471),complex(0.00000,0.332601),complex(0.00000,0.290672),complex(0.00000,0.171680),complex(0.00000,0.0633346)],$
            [complex(0.0324900,0.00000),complex(0.101619,0.00000),complex(0.206305,0.00000),complex(0.304956,0.00000),complex(0.328349,0.00000),complex(0.225471,0.00000),complex(0.0174863,0.00000),complex(-0.190753,-0.00000),complex(-0.269766,0.00000),complex(-0.162524,-0.00000),complex(0.0588678,0.00000),complex(0.234652,0.00000),complex(0.234652,0.00000),complex(0.0588678,-0.00000),complex(-0.162524,0.00000),complex(-0.269766,0.00000),complex(-0.190753,0.00000),complex(0.0174863,-0.00000),complex(0.225471,0.00000),complex(0.328349,-0.00000),complex(0.304956,0.00000),complex(0.206305,-0.00000),complex(0.101619,0.00000),complex(0.0324900,-0.00000)],$
            [complex(-0.00000,0.0145300),complex(0.00000,0.0515050),complex(-0.00000,0.121494),complex(0.00000,0.217173),complex(-0.00000,0.304956),complex(0.00000,0.332601),complex(-0.00000,0.258064),complex(0.00000,0.0853075),complex(0.00000,-0.120643),complex(0.00000,-0.259582),complex(0.00000,-0.254489),complex(0.00000,-0.104940),complex(0.00000,0.104940),complex(0.00000,0.254489),complex(0.00000,0.259582),complex(0.00000,0.120643),complex(-0.00000,-0.0853075),complex(0.00000,-0.258064),complex(-0.00000,-0.332601),complex(0.00000,-0.304956),complex(-0.00000,-0.217173),complex(0.00000,-0.121494),complex(-0.00000,-0.0515050),complex(0.00000,-0.0145300)],$
            [complex(-0.00549181,0.00000),complex(-0.0217573,-0.00000),complex(-0.0583501,0.00000),complex(-0.121494,-0.00000),complex(-0.206305,0.00000),complex(-0.290672,-0.00000),complex(-0.337938,0.00000),complex(-0.311684,-0.00000),complex(-0.197595,0.00000),complex(-0.0196226,-0.00000),complex(0.162524,0.00000),complex(0.277644,0.00000),complex(0.277644,0.00000),complex(0.162524,-0.00000),complex(-0.0196226,0.00000),complex(-0.197595,0.00000),complex(-0.311684,0.00000),complex(-0.337938,0.00000),complex(-0.290672,0.00000),complex(-0.206305,0.00000),complex(-0.121494,0.00000),complex(-0.0583501,0.00000),complex(-0.0217573,0.00000),complex(-0.00549181,0.00000)],$
            [complex(0.00000,-0.00165584),complex(0.00000,-0.00725061),complex(0.00000,-0.0217573),complex(0.00000,-0.0515050),complex(0.00000,-0.101619),complex(0.00000,-0.171680),complex(0.00000,-0.251612),complex(0.00000,-0.320816),complex(0.00000,-0.352879),complex(0.00000,-0.325404),complex(0.00000,-0.231014),complex(0.00000,-0.0837128),complex(0.00000,0.0837128),complex(0.00000,0.231014),complex(0.00000,0.325404),complex(0.00000,0.352879),complex(0.00000,0.320816),complex(0.00000,0.251612),complex(0.00000,0.171680),complex(0.00000,0.101619),complex(0.00000,0.0515050),complex(0.00000,0.0217573),complex(0.00000,0.00725061),complex(0.00000,0.00165584)],$
            [complex(0.000345267,0.00000),complex(0.00165584,0.00000),complex(0.00549181,0.00000),complex(0.0145300,0.00000),complex(0.0324900,0.00000),complex(0.0633346,0.00000),complex(0.109699,0.00000),complex(0.170953,0.00000),complex(0.241764,0.00000),complex(0.312116,0.00000),complex(0.369301,0.00000),complex(0.401473,0.00000),complex(0.401473,0.00000),complex(0.369301,-0.00000),complex(0.312116,0.00000),complex(0.241764,-0.00000),complex(0.170953,0.00000),complex(0.109699,-0.00000),complex(0.0633346,0.00000),complex(0.0324900,-0.00000),complex(0.0145300,0.00000),complex(0.00549181,-0.00000),complex(0.00165584,0.00000),complex(0.000345267,-0.00000)]$
            ]
          end

      22: begin
            matrix[253:275,253:275]=[$
            [complex(-0.000488281,-0.00000),complex(0.00229024,0.00000),complex(-0.00742123,-0.00000),complex(0.0191615,0.00000),complex(-0.0417616,-0.00000),complex(0.0792371,0.00000),complex(-0.133376,-0.00000),complex(0.201645,0.00000),complex(-0.276114,-0.00000),complex(0.344375,0.00000),complex(-0.392648,-0.00000),complex(0.410107,0.00000),complex(-0.392648,0.00000),complex(0.344375,0.00000),complex(-0.276114,0.00000),complex(0.201645,0.00000),complex(-0.133376,0.00000),complex(0.0792371,0.00000),complex(-0.0417616,0.00000),complex(0.0191615,0.00000),complex(-0.00742123,0.00000),complex(0.00229024,0.00000),complex(-0.000488281,0.00000)],$
            [complex(0.00000,-0.00229024),complex(0.00000,0.00976562),complex(0.00000,-0.0284798),complex(0.00000,0.0653641),complex(0.00000,-0.124651),complex(0.00000,0.202721),complex(0.00000,-0.284358),complex(0.00000,0.343928),complex(0.00000,-0.353207),complex(0.00000,0.293684),complex(0.00000,-0.167426),0,complex(0.00000,0.167426),complex(0.00000,-0.293684),complex(0.00000,0.353207),complex(0.00000,-0.343928),complex(0.00000,0.284358),complex(0.00000,-0.202721),complex(0.00000,0.124651),complex(0.00000,-0.0653641),complex(0.00000,0.0284798),complex(0.00000,-0.00976562),complex(0.00000,0.00229024)],$
            [complex(0.00742123,0.00000),complex(-0.0284798,0.00000),complex(0.0737305,0.00000),complex(-0.147506,0.00000),complex(0.239051,0.00000),complex(-0.318018,0.00000),complex(0.342244,0.00000),complex(-0.278613,0.00000),complex(0.127169,0.00000),complex(0.0679746,0.00000),complex(-0.232509,-0.00000),complex(0.296814,0.00000),complex(-0.232509,0.00000),complex(0.0679746,0.00000),complex(0.127169,-0.00000),complex(-0.278613,0.00000),complex(0.342244,-0.00000),complex(-0.318018,0.00000),complex(0.239051,-0.00000),complex(-0.147506,0.00000),complex(0.0737305,-0.00000),complex(-0.0284798,0.00000),complex(0.00742123,-0.00000)],$
            [complex(-0.00000,0.0191615),complex(0.00000,-0.0653641),complex(-0.00000,0.147506),complex(0.00000,-0.250000),complex(-0.00000,0.327769),complex(0.00000,-0.323064),complex(-0.00000,0.203924),0,complex(0.00000,-0.197009),complex(0.00000,0.280816),complex(0.00000,-0.200112),0,complex(0.00000,0.200112),complex(0.00000,-0.280816),complex(0.00000,0.197009),0,complex(-0.00000,-0.203924),complex(0.00000,0.323064),complex(-0.00000,-0.327769),complex(0.00000,0.250000),complex(-0.00000,-0.147506),complex(0.00000,0.0653641),complex(-0.00000,-0.0191615)],$
            [complex(-0.0417616,-0.00000),complex(0.124651,0.00000),complex(-0.239051,-0.00000),complex(0.327769,0.00000),complex(-0.313965,-0.00000),complex(0.162128,0.00000),complex(0.0701750,0.00000),complex(-0.247554,0.00000),complex(0.248584,0.00000),complex(-0.0684500,0.00000),complex(-0.160681,-0.00000),complex(0.263726,0.00000),complex(-0.160681,0.00000),complex(-0.0684500,0.00000),complex(0.248584,-0.00000),complex(-0.247554,0.00000),complex(0.0701750,-0.00000),complex(0.162128,0.00000),complex(-0.313965,0.00000),complex(0.327769,0.00000),complex(-0.239051,0.00000),complex(0.124651,0.00000),complex(-0.0417616,0.00000)],$
            [complex(0.00000,-0.0792371),complex(0.00000,0.202721),complex(0.00000,-0.318018),complex(0.00000,0.323064),complex(0.00000,-0.162128),complex(0.00000,-0.0917969),complex(-0.00000,0.261364),complex(0.00000,-0.208756),complex(0.00000,-0.0238209),complex(0.00000,0.229191),complex(0.00000,-0.217765),0,complex(0.00000,0.217765),complex(0.00000,-0.229191),complex(0.00000,0.0238209),complex(0.00000,0.208756),complex(-0.00000,-0.261364),complex(0.00000,0.0917969),complex(0.00000,0.162128),complex(0.00000,-0.323064),complex(0.00000,0.318018),complex(0.00000,-0.202721),complex(0.00000,0.0792371)],$
            [complex(0.133376,0.00000),complex(-0.284358,0.00000),complex(0.342244,0.00000),complex(-0.203924,0.00000),complex(-0.0701750,-0.00000),complex(0.261364,0.00000),complex(-0.192871,-0.00000),complex(-0.0671773,0.00000),complex(0.247655,0.00000),complex(-0.155071,0.00000),complex(-0.107810,-0.00000),complex(0.247728,0.00000),complex(-0.107810,0.00000),complex(-0.155071,0.00000),complex(0.247655,-0.00000),complex(-0.0671773,0.00000),complex(-0.192871,0.00000),complex(0.261364,0.00000),complex(-0.0701750,0.00000),complex(-0.203924,0.00000),complex(0.342244,-0.00000),complex(-0.284358,0.00000),complex(0.133376,-0.00000)],$
            [complex(-0.00000,0.201645),complex(0.00000,-0.343928),complex(-0.00000,0.278613),0,complex(0.00000,-0.247554),complex(0.00000,0.208756),complex(-0.00000,0.0671773),complex(0.00000,-0.250000),complex(-0.00000,0.117675),complex(0.00000,0.160109),complex(0.00000,-0.228190),0,complex(0.00000,0.228190),complex(0.00000,-0.160109),complex(-0.00000,-0.117675),complex(0.00000,0.250000),complex(-0.00000,-0.0671773),complex(0.00000,-0.208756),complex(0.00000,0.247554),0,complex(-0.00000,-0.278613),complex(0.00000,0.343928),complex(-0.00000,-0.201645)],$
            [complex(-0.276114,-0.00000),complex(0.353207,0.00000),complex(-0.127169,-0.00000),complex(-0.197009,0.00000),complex(0.248584,0.00000),complex(0.0238209,0.00000),complex(-0.247655,-0.00000),complex(0.117675,0.00000),complex(0.174805,0.00000),complex(-0.208276,0.00000),complex(-0.0624924,-0.00000),complex(0.239328,0.00000),complex(-0.0624924,0.00000),complex(-0.208276,0.00000),complex(0.174805,-0.00000),complex(0.117675,0.00000),complex(-0.247655,0.00000),complex(0.0238209,0.00000),complex(0.248584,-0.00000),complex(-0.197009,0.00000),complex(-0.127169,0.00000),complex(0.353207,0.00000),complex(-0.276114,0.00000)],$
            [complex(0.00000,-0.344375),complex(0.00000,0.293684),complex(-0.00000,0.0679746),complex(0.00000,-0.280816),complex(-0.00000,0.0684500),complex(0.00000,0.229191),complex(0.00000,-0.155071),complex(0.00000,-0.160109),complex(-0.00000,0.208276),complex(0.00000,0.0820312),complex(0.00000,-0.233825),0,complex(0.00000,0.233825),complex(0.00000,-0.0820312),complex(-0.00000,-0.208276),complex(0.00000,0.160109),complex(0.00000,0.155071),complex(0.00000,-0.229191),complex(-0.00000,-0.0684500),complex(0.00000,0.280816),complex(-0.00000,-0.0679746),complex(0.00000,-0.293684),complex(0.00000,0.344375)],$
            [complex(0.392648,0.00000),complex(-0.167426,0.00000),complex(-0.232509,-0.00000),complex(0.200112,0.00000),complex(0.160681,0.00000),complex(-0.217765,0.00000),complex(-0.107810,-0.00000),complex(0.228190,0.00000),complex(0.0624924,0.00000),complex(-0.233825,0.00000),complex(-0.0205078,-0.00000),complex(0.235617,0.00000),complex(-0.0205078,0.00000),complex(-0.233825,0.00000),complex(0.0624924,-0.00000),complex(0.228190,0.00000),complex(-0.107810,0.00000),complex(-0.217765,0.00000),complex(0.160681,-0.00000),complex(0.200112,0.00000),complex(-0.232509,0.00000),complex(-0.167426,0.00000),complex(0.392648,-0.00000)],$
            [complex(-0.00000,0.410107),complex(0.00000,-1.58721e-16),complex(0.00000,-0.296814),0,complex(-0.00000,0.263726),0,complex(0.00000,-0.247728),0,complex(-0.00000,0.239328),0,complex(0.00000,-0.235617),0,complex(0.00000,0.235617),0,complex(-0.00000,-0.239328),0,complex(0.00000,0.247728),0,complex(-0.00000,-0.263726),0,complex(0.00000,0.296814),complex(0.00000,1.58721e-16),complex(-0.00000,-0.410107)],$
            [complex(-0.392648,-0.00000),complex(-0.167426,0.00000),complex(0.232509,0.00000),complex(0.200112,0.00000),complex(-0.160681,-0.00000),complex(-0.217765,0.00000),complex(0.107810,0.00000),complex(0.228190,0.00000),complex(-0.0624924,-0.00000),complex(-0.233825,0.00000),complex(0.0205078,0.00000),complex(0.235617,0.00000),complex(0.0205078,-0.00000),complex(-0.233825,0.00000),complex(-0.0624924,0.00000),complex(0.228190,0.00000),complex(0.107810,-0.00000),complex(-0.217765,0.00000),complex(-0.160681,0.00000),complex(0.200112,0.00000),complex(0.232509,-0.00000),complex(-0.167426,0.00000),complex(-0.392648,0.00000)],$
            [complex(0.00000,-0.344375),complex(0.00000,-0.293684),complex(-0.00000,0.0679746),complex(0.00000,0.280816),complex(-0.00000,0.0684500),complex(0.00000,-0.229191),complex(0.00000,-0.155071),complex(0.00000,0.160109),complex(-0.00000,0.208276),complex(0.00000,-0.0820312),complex(0.00000,-0.233825),0,complex(0.00000,0.233825),complex(0.00000,0.0820312),complex(-0.00000,-0.208276),complex(0.00000,-0.160109),complex(0.00000,0.155071),complex(0.00000,0.229191),complex(-0.00000,-0.0684500),complex(0.00000,-0.280816),complex(-0.00000,-0.0679746),complex(0.00000,0.293684),complex(0.00000,0.344375)],$
            [complex(0.276114,0.00000),complex(0.353207,0.00000),complex(0.127169,0.00000),complex(-0.197009,0.00000),complex(-0.248584,-0.00000),complex(0.0238209,0.00000),complex(0.247655,0.00000),complex(0.117675,0.00000),complex(-0.174805,-0.00000),complex(-0.208276,0.00000),complex(0.0624924,0.00000),complex(0.239328,0.00000),complex(0.0624924,-0.00000),complex(-0.208276,0.00000),complex(-0.174805,0.00000),complex(0.117675,0.00000),complex(0.247655,-0.00000),complex(0.0238209,0.00000),complex(-0.248584,0.00000),complex(-0.197009,0.00000),complex(0.127169,-0.00000),complex(0.353207,0.00000),complex(0.276114,-0.00000)],$
            [complex(-0.00000,0.201645),complex(0.00000,0.343928),complex(-0.00000,0.278613),0,complex(0.00000,-0.247554),complex(0.00000,-0.208756),complex(-0.00000,0.0671773),complex(0.00000,0.250000),complex(-0.00000,0.117675),complex(0.00000,-0.160109),complex(0.00000,-0.228190),0,complex(0.00000,0.228190),complex(0.00000,0.160109),complex(-0.00000,-0.117675),complex(0.00000,-0.250000),complex(-0.00000,-0.0671773),complex(0.00000,0.208756),complex(0.00000,0.247554),0,complex(-0.00000,-0.278613),complex(0.00000,-0.343928),complex(-0.00000,-0.201645)],$
            [complex(-0.133376,-0.00000),complex(-0.284358,0.00000),complex(-0.342244,-0.00000),complex(-0.203924,0.00000),complex(0.0701750,0.00000),complex(0.261364,0.00000),complex(0.192871,0.00000),complex(-0.0671773,0.00000),complex(-0.247655,-0.00000),complex(-0.155071,0.00000),complex(0.107810,0.00000),complex(0.247728,0.00000),complex(0.107810,-0.00000),complex(-0.155071,0.00000),complex(-0.247655,0.00000),complex(-0.0671773,0.00000),complex(0.192871,-0.00000),complex(0.261364,0.00000),complex(0.0701750,-0.00000),complex(-0.203924,0.00000),complex(-0.342244,0.00000),complex(-0.284358,0.00000),complex(-0.133376,0.00000)],$
            [complex(0.00000,-0.0792371),complex(0.00000,-0.202721),complex(0.00000,-0.318018),complex(0.00000,-0.323064),complex(0.00000,-0.162128),complex(0.00000,0.0917969),complex(-0.00000,0.261364),complex(0.00000,0.208756),complex(0.00000,-0.0238209),complex(0.00000,-0.229191),complex(0.00000,-0.217765),0,complex(0.00000,0.217765),complex(0.00000,0.229191),complex(0.00000,0.0238209),complex(0.00000,-0.208756),complex(-0.00000,-0.261364),complex(0.00000,-0.0917969),complex(0.00000,0.162128),complex(0.00000,0.323064),complex(0.00000,0.318018),complex(0.00000,0.202721),complex(0.00000,0.0792371)],$
            [complex(0.0417616,0.00000),complex(0.124651,0.00000),complex(0.239051,0.00000),complex(0.327769,0.00000),complex(0.313965,0.00000),complex(0.162128,0.00000),complex(-0.0701750,-0.00000),complex(-0.247554,0.00000),complex(-0.248584,-0.00000),complex(-0.0684500,0.00000),complex(0.160681,0.00000),complex(0.263726,0.00000),complex(0.160681,-0.00000),complex(-0.0684500,0.00000),complex(-0.248584,0.00000),complex(-0.247554,0.00000),complex(-0.0701750,0.00000),complex(0.162128,0.00000),complex(0.313965,-0.00000),complex(0.327769,0.00000),complex(0.239051,-0.00000),complex(0.124651,0.00000),complex(0.0417616,-0.00000)],$
            [complex(-0.00000,0.0191615),complex(0.00000,0.0653641),complex(-0.00000,0.147506),complex(0.00000,0.250000),complex(-0.00000,0.327769),complex(0.00000,0.323064),complex(-0.00000,0.203924),0,complex(0.00000,-0.197009),complex(0.00000,-0.280816),complex(0.00000,-0.200112),0,complex(0.00000,0.200112),complex(0.00000,0.280816),complex(0.00000,0.197009),0,complex(-0.00000,-0.203924),complex(0.00000,-0.323064),complex(-0.00000,-0.327769),complex(0.00000,-0.250000),complex(-0.00000,-0.147506),complex(0.00000,-0.0653641),complex(-0.00000,-0.0191615)],$
            [complex(-0.00742123,-0.00000),complex(-0.0284798,0.00000),complex(-0.0737305,-0.00000),complex(-0.147506,0.00000),complex(-0.239051,-0.00000),complex(-0.318018,0.00000),complex(-0.342244,-0.00000),complex(-0.278613,0.00000),complex(-0.127169,-0.00000),complex(0.0679746,0.00000),complex(0.232509,0.00000),complex(0.296814,0.00000),complex(0.232509,-0.00000),complex(0.0679746,0.00000),complex(-0.127169,0.00000),complex(-0.278613,0.00000),complex(-0.342244,0.00000),complex(-0.318018,0.00000),complex(-0.239051,0.00000),complex(-0.147506,0.00000),complex(-0.0737305,0.00000),complex(-0.0284798,0.00000),complex(-0.00742123,0.00000)],$
            [complex(0.00000,-0.00229024),complex(0.00000,-0.00976562),complex(0.00000,-0.0284798),complex(0.00000,-0.0653641),complex(0.00000,-0.124651),complex(0.00000,-0.202721),complex(0.00000,-0.284358),complex(0.00000,-0.343928),complex(0.00000,-0.353207),complex(0.00000,-0.293684),complex(0.00000,-0.167426),0,complex(0.00000,0.167426),complex(0.00000,0.293684),complex(0.00000,0.353207),complex(0.00000,0.343928),complex(0.00000,0.284358),complex(0.00000,0.202721),complex(0.00000,0.124651),complex(0.00000,0.0653641),complex(0.00000,0.0284798),complex(0.00000,0.00976562),complex(0.00000,0.00229024)],$
            [complex(0.000488281,0.00000),complex(0.00229024,0.00000),complex(0.00742123,0.00000),complex(0.0191615,0.00000),complex(0.0417616,0.00000),complex(0.0792371,0.00000),complex(0.133376,0.00000),complex(0.201645,0.00000),complex(0.276114,0.00000),complex(0.344375,0.00000),complex(0.392648,0.00000),complex(0.410107,0.00000),complex(0.392648,-0.00000),complex(0.344375,0.00000),complex(0.276114,-0.00000),complex(0.201645,0.00000),complex(0.133376,-0.00000),complex(0.0792371,0.00000),complex(0.0417616,-0.00000),complex(0.0191615,0.00000),complex(0.00742123,-0.00000),complex(0.00229024,0.00000),complex(0.000488281,-0.00000)]$
            ]
          end


      21: begin
            matrix[231:252,231:252]=[$
            [complex(0.00000,-0.000690534),complex(-0.00000,0.00316442),complex(0.00000,-0.0100068),complex(-0.00000,0.0251832),complex(0.00000,-0.0534216),complex(-0.00000,0.0985046),complex(0.00000,-0.160857),complex(-0.00000,0.235471),complex(0.00000,-0.311499),complex(-0.00000,0.374375),complex(0.00000,-0.410107),complex(0.00000,0.410107),complex(0.00000,-0.374375),complex(0.00000,0.311499),complex(0.00000,-0.235471),complex(0.00000,0.160857),complex(0.00000,-0.0985046),complex(0.00000,0.0534216),complex(0.00000,-0.0251832),complex(0.00000,0.0100068),complex(0.00000,-0.00316442),complex(0.00000,0.000690534)],$
            [complex(0.00316442,0.00000),complex(-0.0131202,0.00000),complex(0.0371222,0.00000),complex(-0.0824314,0.00000),complex(0.151548,0.00000),complex(-0.236450,0.00000),complex(0.315918,0.00000),complex(-0.359688,0.00000),complex(0.339873,0.00000),complex(-0.245086,0.00000),complex(0.0894928,0.00000),complex(0.0894928,0.00000),complex(-0.245086,0.00000),complex(0.339873,0.00000),complex(-0.359688,0.00000),complex(0.315918,0.00000),complex(-0.236450,0.00000),complex(0.151548,0.00000),complex(-0.0824314,0.00000),complex(0.0371222,0.00000),complex(-0.0131202,0.00000),complex(0.00316442,0.00000)],$
            [complex(0.00000,0.0100068),complex(0.00000,-0.0371222),complex(0.00000,0.0925315),complex(0.00000,-0.177256),complex(0.00000,0.272797),complex(0.00000,-0.339873),complex(0.00000,0.333006),complex(0.00000,-0.227487),complex(0.00000,0.0429909),complex(-0.00000,0.155006),complex(0.00000,-0.283001),complex(0.00000,0.283001),complex(0.00000,-0.155006),complex(-0.00000,-0.0429909),complex(0.00000,0.227487),complex(-0.00000,-0.333006),complex(0.00000,0.339873),complex(-0.00000,-0.272797),complex(0.00000,0.177256),complex(-0.00000,-0.0925315),complex(0.00000,0.0371222),complex(-0.00000,-0.0100068)],$
            [complex(-0.0251832,-0.00000),complex(0.0824314,0.00000),complex(-0.177256,-0.00000),complex(0.283119,0.00000),complex(-0.342773,-0.00000),complex(0.297114,0.00000),complex(-0.132323,-0.00000),complex(-0.0903940,0.00000),complex(0.256243,0.00000),complex(-0.266904,0.00000),complex(0.112453,0.00000),complex(0.112453,0.00000),complex(-0.266904,0.00000),complex(0.256243,0.00000),complex(-0.0903940,0.00000),complex(-0.132323,0.00000),complex(0.297114,-0.00000),complex(-0.342773,0.00000),complex(0.283119,-0.00000),complex(-0.177256,0.00000),complex(0.0824314,-0.00000),complex(-0.0251832,0.00000)],$
            [complex(0.00000,-0.0534216),complex(-0.00000,0.151548),complex(0.00000,-0.272797),complex(-0.00000,0.342773),complex(0.00000,-0.282428),complex(-0.00000,0.0827633),complex(0.00000,0.155944),complex(0.00000,-0.276979),complex(0.00000,0.189244),complex(-0.00000,0.0435529),complex(0.00000,-0.238549),complex(0.00000,0.238549),complex(0.00000,-0.0435529),complex(-0.00000,-0.189244),complex(0.00000,0.276979),complex(-0.00000,-0.155944),complex(0.00000,-0.0827633),complex(0.00000,0.282428),complex(0.00000,-0.342773),complex(0.00000,0.272797),complex(0.00000,-0.151548),complex(0.00000,0.0534216)],$
            [complex(0.0985046,0.00000),complex(-0.236450,0.00000),complex(0.339873,0.00000),complex(-0.297114,0.00000),complex(0.0827633,0.00000),complex(0.174705,0.00000),complex(-0.274016,-0.00000),complex(0.127103,0.00000),complex(0.133203,0.00000),complex(-0.259819,0.00000),complex(0.129371,0.00000),complex(0.129371,0.00000),complex(-0.259819,0.00000),complex(0.133203,0.00000),complex(0.127103,-0.00000),complex(-0.274016,0.00000),complex(0.174705,-0.00000),complex(0.0827633,0.00000),complex(-0.297114,0.00000),complex(0.339873,0.00000),complex(-0.236450,0.00000),complex(0.0985046,0.00000)],$
            [complex(0.00000,0.160857),complex(0.00000,-0.315918),complex(0.00000,0.333006),complex(0.00000,-0.132323),complex(0.00000,-0.155944),complex(-0.00000,0.274016),complex(0.00000,-0.104961),complex(0.00000,-0.169821),complex(0.00000,0.246047),complex(0.00000,-0.0385711),complex(0.00000,-0.211263),complex(0.00000,0.211263),complex(0.00000,0.0385711),complex(-0.00000,-0.246047),complex(0.00000,0.169821),complex(0.00000,0.104961),complex(0.00000,-0.274016),complex(0.00000,0.155944),complex(0.00000,0.132323),complex(-0.00000,-0.333006),complex(0.00000,0.315918),complex(-0.00000,-0.160857)],$
            [complex(-0.235471,-0.00000),complex(0.359688,0.00000),complex(-0.227487,-0.00000),complex(-0.0903940,0.00000),complex(0.276979,0.00000),complex(-0.127103,0.00000),complex(-0.169821,-0.00000),complex(0.237544,0.00000),complex(0.00730792,0.00000),complex(-0.237142,0.00000),complex(0.144320,0.00000),complex(0.144320,0.00000),complex(-0.237142,0.00000),complex(0.00730792,0.00000),complex(0.237544,-0.00000),complex(-0.169821,0.00000),complex(-0.127103,0.00000),complex(0.276979,0.00000),complex(-0.0903940,0.00000),complex(-0.227487,0.00000),complex(0.359688,-0.00000),complex(-0.235471,0.00000)],$
            [complex(0.00000,-0.311499),complex(-0.00000,0.339873),complex(0.00000,-0.0429909),complex(0.00000,-0.256243),complex(0.00000,0.189244),complex(-0.00000,0.133203),complex(0.00000,-0.246047),complex(0.00000,-0.00730792),complex(0.00000,0.241687),complex(0.00000,-0.104570),complex(0.00000,-0.190917),complex(0.00000,0.190917),complex(0.00000,0.104570),complex(-0.00000,-0.241687),complex(0.00000,0.00730792),complex(0.00000,0.246047),complex(0.00000,-0.133203),complex(-0.00000,-0.189244),complex(0.00000,0.256243),complex(0.00000,0.0429909),complex(0.00000,-0.339873),complex(0.00000,0.311499)],$
            [complex(0.374375,0.00000),complex(-0.245086,0.00000),complex(-0.155006,-0.00000),complex(0.266904,0.00000),complex(0.0435529,0.00000),complex(-0.259819,0.00000),complex(0.0385711,0.00000),complex(0.237142,0.00000),complex(-0.104570,-0.00000),complex(-0.203017,0.00000),complex(0.158853,0.00000),complex(0.158853,0.00000),complex(-0.203017,0.00000),complex(-0.104570,0.00000),complex(0.237142,-0.00000),complex(0.0385711,0.00000),complex(-0.259819,0.00000),complex(0.0435529,0.00000),complex(0.266904,-0.00000),complex(-0.155006,0.00000),complex(-0.245086,0.00000),complex(0.374375,0.00000)],$
            [complex(0.00000,0.410107),complex(0.00000,-0.0894928),complex(0.00000,-0.283001),complex(-0.00000,0.112453),complex(0.00000,0.238549),complex(0.00000,-0.129371),complex(0.00000,-0.211263),complex(-0.00000,0.144320),complex(0.00000,0.190917),complex(0.00000,-0.158853),complex(0.00000,-0.174015),complex(0.00000,0.174015),complex(0.00000,0.158853),complex(-0.00000,-0.190917),complex(0.00000,-0.144320),complex(0.00000,0.211263),complex(0.00000,0.129371),complex(-0.00000,-0.238549),complex(0.00000,-0.112453),complex(0.00000,0.283001),complex(0.00000,0.0894928),complex(-0.00000,-0.410107)],$
            [complex(-0.410107,-0.00000),complex(-0.0894928,0.00000),complex(0.283001,0.00000),complex(0.112453,0.00000),complex(-0.238549,-0.00000),complex(-0.129371,0.00000),complex(0.211263,0.00000),complex(0.144320,0.00000),complex(-0.190917,-0.00000),complex(-0.158853,0.00000),complex(0.174015,0.00000),complex(0.174015,0.00000),complex(-0.158853,0.00000),complex(-0.190917,0.00000),complex(0.144320,-0.00000),complex(0.211263,0.00000),complex(-0.129371,0.00000),complex(-0.238549,0.00000),complex(0.112453,-0.00000),complex(0.283001,0.00000),complex(-0.0894928,0.00000),complex(-0.410107,0.00000)],$
            [complex(0.00000,-0.374375),complex(0.00000,-0.245086),complex(0.00000,0.155006),complex(-0.00000,0.266904),complex(0.00000,-0.0435529),complex(0.00000,-0.259819),complex(0.00000,-0.0385711),complex(-0.00000,0.237142),complex(0.00000,0.104570),complex(0.00000,-0.203017),complex(0.00000,-0.158853),complex(0.00000,0.158853),complex(0.00000,0.203017),complex(-0.00000,-0.104570),complex(0.00000,-0.237142),complex(0.00000,0.0385711),complex(0.00000,0.259819),complex(0.00000,0.0435529),complex(0.00000,-0.266904),complex(-0.00000,-0.155006),complex(0.00000,0.245086),complex(0.00000,0.374375)],$
            [complex(0.311499,0.00000),complex(0.339873,0.00000),complex(0.0429909,0.00000),complex(-0.256243,0.00000),complex(-0.189244,-0.00000),complex(0.133203,0.00000),complex(0.246047,0.00000),complex(-0.00730792,0.00000),complex(-0.241687,-0.00000),complex(-0.104570,0.00000),complex(0.190917,0.00000),complex(0.190917,0.00000),complex(-0.104570,0.00000),complex(-0.241687,0.00000),complex(-0.00730792,0.00000),complex(0.246047,0.00000),complex(0.133203,-0.00000),complex(-0.189244,0.00000),complex(-0.256243,0.00000),complex(0.0429909,0.00000),complex(0.339873,-0.00000),complex(0.311499,0.00000)],$
            [complex(0.00000,0.235471),complex(-0.00000,0.359688),complex(0.00000,0.227487),complex(0.00000,-0.0903940),complex(0.00000,-0.276979),complex(0.00000,-0.127103),complex(0.00000,0.169821),complex(-0.00000,0.237544),complex(0.00000,-0.00730792),complex(0.00000,-0.237142),complex(0.00000,-0.144320),complex(0.00000,0.144320),complex(0.00000,0.237142),complex(0.00000,0.00730792),complex(0.00000,-0.237544),complex(-0.00000,-0.169821),complex(0.00000,0.127103),complex(0.00000,0.276979),complex(0.00000,0.0903940),complex(-0.00000,-0.227487),complex(0.00000,-0.359688),complex(-0.00000,-0.235471)],$
            [complex(-0.160857,-0.00000),complex(-0.315918,0.00000),complex(-0.333006,-0.00000),complex(-0.132323,0.00000),complex(0.155944,0.00000),complex(0.274016,0.00000),complex(0.104961,0.00000),complex(-0.169821,0.00000),complex(-0.246047,-0.00000),complex(-0.0385711,0.00000),complex(0.211263,0.00000),complex(0.211263,0.00000),complex(-0.0385711,0.00000),complex(-0.246047,0.00000),complex(-0.169821,0.00000),complex(0.104961,0.00000),complex(0.274016,-0.00000),complex(0.155944,0.00000),complex(-0.132323,0.00000),complex(-0.333006,0.00000),complex(-0.315918,0.00000),complex(-0.160857,0.00000)],$
            [complex(0.00000,-0.0985046),complex(0.00000,-0.236450),complex(0.00000,-0.339873),complex(0.00000,-0.297114),complex(0.00000,-0.0827633),complex(-0.00000,0.174705),complex(0.00000,0.274016),complex(-0.00000,0.127103),complex(0.00000,-0.133203),complex(0.00000,-0.259819),complex(0.00000,-0.129371),complex(0.00000,0.129371),complex(0.00000,0.259819),complex(0.00000,0.133203),complex(0.00000,-0.127103),complex(-0.00000,-0.274016),complex(0.00000,-0.174705),complex(0.00000,0.0827633),complex(0.00000,0.297114),complex(0.00000,0.339873),complex(0.00000,0.236450),complex(0.00000,0.0985046)],$
            [complex(0.0534216,0.00000),complex(0.151548,0.00000),complex(0.272797,0.00000),complex(0.342773,0.00000),complex(0.282428,0.00000),complex(0.0827633,0.00000),complex(-0.155944,-0.00000),complex(-0.276979,0.00000),complex(-0.189244,-0.00000),complex(0.0435529,0.00000),complex(0.238549,0.00000),complex(0.238549,0.00000),complex(0.0435529,-0.00000),complex(-0.189244,0.00000),complex(-0.276979,0.00000),complex(-0.155944,0.00000),complex(0.0827633,-0.00000),complex(0.282428,0.00000),complex(0.342773,-0.00000),complex(0.272797,0.00000),complex(0.151548,-0.00000),complex(0.0534216,0.00000)],$
            [complex(0.00000,0.0251832),complex(-0.00000,0.0824314),complex(0.00000,0.177256),complex(-0.00000,0.283119),complex(0.00000,0.342773),complex(-0.00000,0.297114),complex(0.00000,0.132323),complex(0.00000,-0.0903940),complex(0.00000,-0.256243),complex(0.00000,-0.266904),complex(0.00000,-0.112453),complex(0.00000,0.112453),complex(0.00000,0.266904),complex(0.00000,0.256243),complex(0.00000,0.0903940),complex(-0.00000,-0.132323),complex(0.00000,-0.297114),complex(-0.00000,-0.342773),complex(0.00000,-0.283119),complex(-0.00000,-0.177256),complex(0.00000,-0.0824314),complex(-0.00000,-0.0251832)],$
            [complex(-0.0100068,-0.00000),complex(-0.0371222,0.00000),complex(-0.0925315,-0.00000),complex(-0.177256,0.00000),complex(-0.272797,-0.00000),complex(-0.339873,0.00000),complex(-0.333006,-0.00000),complex(-0.227487,0.00000),complex(-0.0429909,-0.00000),complex(0.155006,0.00000),complex(0.283001,0.00000),complex(0.283001,0.00000),complex(0.155006,-0.00000),complex(-0.0429909,0.00000),complex(-0.227487,0.00000),complex(-0.333006,0.00000),complex(-0.339873,0.00000),complex(-0.272797,0.00000),complex(-0.177256,0.00000),complex(-0.0925315,0.00000),complex(-0.0371222,0.00000),complex(-0.0100068,0.00000)],$
            [complex(0.00000,-0.00316442),complex(0.00000,-0.0131202),complex(0.00000,-0.0371222),complex(0.00000,-0.0824314),complex(0.00000,-0.151548),complex(0.00000,-0.236450),complex(0.00000,-0.315918),complex(0.00000,-0.359688),complex(0.00000,-0.339873),complex(0.00000,-0.245086),complex(0.00000,-0.0894928),complex(0.00000,0.0894928),complex(0.00000,0.245086),complex(0.00000,0.339873),complex(0.00000,0.359688),complex(0.00000,0.315918),complex(0.00000,0.236450),complex(0.00000,0.151548),complex(0.00000,0.0824314),complex(0.00000,0.0371222),complex(0.00000,0.0131202),complex(0.00000,0.00316442)],$
            [complex(0.000690534,0.00000),complex(0.00316442,0.00000),complex(0.0100068,0.00000),complex(0.0251832,0.00000),complex(0.0534216,0.00000),complex(0.0985046,0.00000),complex(0.160857,0.00000),complex(0.235471,0.00000),complex(0.311499,0.00000),complex(0.374375,0.00000),complex(0.410107,0.00000),complex(0.410107,0.00000),complex(0.374375,-0.00000),complex(0.311499,0.00000),complex(0.235471,-0.00000),complex(0.160857,0.00000),complex(0.0985046,-0.00000),complex(0.0534216,0.00000),complex(0.0251832,-0.00000),complex(0.0100068,0.00000),complex(0.00316442,-0.00000),complex(0.000690534,0.00000)]$
            ]
          end

      20: begin
            matrix[210:230,210:230]=[$
            [complex(+9.765625E-04,+0.000000E+00),complex(-4.367320E-03,-0.000000E+00),complex(+1.346099E-02,+0.000000E+00),complex(-3.297254E-02,-0.000000E+00),complex(+6.797464E-02,+0.000000E+00),complex(-1.215967E-01,-0.000000E+00),complex(+1.922613E-01,+0.000000E+00),complex(-2.718986E-01,-0.000000E+00),complex(+3.466040E-01,+0.000000E+00),complex(-4.002239E-01,-0.000000E+00),complex(+4.197583E-01,+0.000000E+00),complex(-4.002239E-01,+0.000000E+00),complex(+3.466040E-01,+0.000000E+00),complex(-2.718986E-01,+0.000000E+00),complex(+1.922613E-01,+0.000000E+00),complex(-1.215967E-01,+0.000000E+00),complex(+6.797464E-02,+0.000000E+00),complex(-3.297254E-02,+0.000000E+00),complex(+1.346099E-02,+0.000000E+00),complex(-4.367320E-03,+0.000000E+00),complex(+9.765625E-04,+0.000000E+00)],$
            [complex(+0.000000E+00,+4.367320E-03),complex(+0.000000E+00,-1.757812E-02),complex(+0.000000E+00,+4.815948E-02),complex(+0.000000E+00,-1.032204E-01),complex(+0.000000E+00,+1.823951E-01),complex(+0.000000E+00,-2.718986E-01),complex(+0.000000E+00,+3.439275E-01),complex(+0.000000E+00,-3.647902E-01),complex(+0.000000E+00,+3.100121E-01),complex(+0.000000E+00,-1.789856E-01),0,complex(+0.000000E+00,+1.789856E-01),complex(+0.000000E+00,-3.100121E-01),complex(+0.000000E+00,+3.647902E-01),complex(+0.000000E+00,-3.439275E-01),complex(+0.000000E+00,+2.718986E-01),complex(+0.000000E+00,-1.823951E-01),complex(+0.000000E+00,+1.032204E-01),complex(+0.000000E+00,-4.815948E-02),complex(+0.000000E+00,+1.757812E-02),complex(+0.000000E+00,-4.367320E-03)],$
            [complex(-1.346099E-02,+0.000000E+00),complex(+4.815948E-02,+0.000000E+00),complex(-1.152344E-01,+0.000000E+00),complex(+2.105030E-01,+0.000000E+00),complex(-3.057467E-01,+0.000000E+00),complex(+3.528622E-01,+0.000000E+00),complex(-3.068582E-01,+0.000000E+00),complex(+1.578048E-01,+0.000000E+00),complex(+5.029060E-02,+0.000000E+00),complex(-2.322823E-01,-0.000000E+00),complex(+3.045247E-01,+0.000000E+00),complex(-2.322823E-01,+0.000000E+00),complex(+5.029060E-02,+0.000000E+00),complex(+1.578048E-01,-0.000000E+00),complex(-3.068582E-01,+0.000000E+00),complex(+3.528622E-01,-0.000000E+00),complex(-3.057467E-01,+0.000000E+00),complex(+2.105030E-01,-0.000000E+00),complex(-1.152344E-01,+0.000000E+00),complex(+4.815948E-02,-0.000000E+00),complex(-1.346099E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.297254E-02),complex(-0.000000E+00,+1.032204E-01),complex(+0.000000E+00,-2.105030E-01),complex(-0.000000E+00,+3.144531E-01),complex(+0.000000E+00,-3.462765E-01),complex(-0.000000E+00,+2.520969E-01),complex(+0.000000E+00,-4.555431E-02),complex(+0.000000E+00,-1.771647E-01),complex(+0.000000E+00,+2.874347E-01),complex(+0.000000E+00,-2.133649E-01),0,complex(+0.000000E+00,+2.133649E-01),complex(+0.000000E+00,-2.874347E-01),complex(+0.000000E+00,+1.771647E-01),complex(+0.000000E+00,+4.555431E-02),complex(-0.000000E+00,-2.520969E-01),complex(+0.000000E+00,+3.462765E-01),complex(-0.000000E+00,-3.144531E-01),complex(+0.000000E+00,+2.105030E-01),complex(-0.000000E+00,-1.032204E-01),complex(+0.000000E+00,+3.297254E-02)],$
            [complex(+6.797464E-02,+0.000000E+00),complex(-1.823951E-01,-0.000000E+00),complex(+3.057467E-01,+0.000000E+00),complex(-3.462765E-01,-0.000000E+00),complex(+2.314453E-01,+0.000000E+00),complex(+8.734641E-03,+0.000000E+00),complex(-2.292573E-01,+0.000000E+00),complex(+2.695312E-01,+0.000000E+00),complex(-9.461071E-02,+0.000000E+00),complex(-1.552458E-01,-0.000000E+00),complex(+2.713719E-01,+0.000000E+00),complex(-1.552458E-01,+0.000000E+00),complex(-9.461071E-02,+0.000000E+00),complex(+2.695312E-01,-0.000000E+00),complex(-2.292573E-01,+0.000000E+00),complex(+8.734641E-03,-0.000000E+00),complex(+2.314453E-01,+0.000000E+00),complex(-3.462765E-01,+0.000000E+00),complex(+3.057467E-01,+0.000000E+00),complex(-1.823951E-01,+0.000000E+00),complex(+6.797464E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,+1.215967E-01),complex(+0.000000E+00,-2.718986E-01),complex(+0.000000E+00,+3.528622E-01),complex(+0.000000E+00,-2.520969E-01),complex(+0.000000E+00,-8.734641E-03),complex(-0.000000E+00,+2.421875E-01),complex(+0.000000E+00,-2.470529E-01),complex(-0.000000E+00,+1.746928E-02),complex(+0.000000E+00,+2.226905E-01),complex(+0.000000E+00,-2.314268E-01),0,complex(+0.000000E+00,+2.314268E-01),complex(+0.000000E+00,-2.226905E-01),complex(-0.000000E+00,-1.746928E-02),complex(+0.000000E+00,+2.470529E-01),complex(-0.000000E+00,-2.421875E-01),complex(+0.000000E+00,+8.734641E-03),complex(+0.000000E+00,+2.520969E-01),complex(+0.000000E+00,-3.528622E-01),complex(+0.000000E+00,+2.718986E-01),complex(+0.000000E+00,-1.215967E-01)],$
            [complex(-1.922613E-01,+0.000000E+00),complex(+3.439275E-01,+0.000000E+00),complex(-3.068582E-01,+0.000000E+00),complex(+4.555431E-02,+0.000000E+00),complex(+2.292573E-01,+0.000000E+00),complex(-2.470529E-01,-0.000000E+00),complex(-7.812500E-03,+0.000000E+00),complex(+2.430680E-01,+0.000000E+00),complex(-1.830944E-01,+0.000000E+00),complex(-9.757809E-02,-0.000000E+00),complex(+2.558519E-01,+0.000000E+00),complex(-9.757809E-02,+0.000000E+00),complex(-1.830944E-01,+0.000000E+00),complex(+2.430680E-01,-0.000000E+00),complex(-7.812500E-03,+0.000000E+00),complex(-2.470529E-01,+0.000000E+00),complex(+2.292573E-01,+0.000000E+00),complex(+4.555431E-02,-0.000000E+00),complex(-3.068582E-01,+0.000000E+00),complex(+3.439275E-01,-0.000000E+00),complex(-1.922613E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.718986E-01),complex(-0.000000E+00,+3.647902E-01),complex(+0.000000E+00,-1.578048E-01),complex(+0.000000E+00,-1.771647E-01),complex(+0.000000E+00,+2.695312E-01),complex(+0.000000E+00,-1.746928E-02),complex(+0.000000E+00,-2.430680E-01),complex(-0.000000E+00,+1.640625E-01),complex(+0.000000E+00,+1.394263E-01),complex(+0.000000E+00,-2.414935E-01),0,complex(+0.000000E+00,+2.414935E-01),complex(+0.000000E+00,-1.394263E-01),complex(-0.000000E+00,-1.640625E-01),complex(+0.000000E+00,+2.430680E-01),complex(+0.000000E+00,+1.746928E-02),complex(+0.000000E+00,-2.695312E-01),complex(+0.000000E+00,+1.771647E-01),complex(+0.000000E+00,+1.578048E-01),complex(-0.000000E+00,-3.647902E-01),complex(+0.000000E+00,+2.718986E-01)],$
            [complex(+3.466040E-01,+0.000000E+00),complex(-3.100121E-01,-0.000000E+00),complex(-5.029060E-02,+0.000000E+00),complex(+2.874347E-01,+0.000000E+00),complex(-9.461071E-02,+0.000000E+00),complex(-2.226905E-01,-0.000000E+00),complex(+1.830944E-01,+0.000000E+00),complex(+1.394263E-01,+0.000000E+00),complex(-2.324219E-01,+0.000000E+00),complex(-4.736076E-02,-0.000000E+00),complex(+2.483619E-01,+0.000000E+00),complex(-4.736076E-02,+0.000000E+00),complex(-2.324219E-01,+0.000000E+00),complex(+1.394263E-01,-0.000000E+00),complex(+1.830944E-01,+0.000000E+00),complex(-2.226905E-01,+0.000000E+00),complex(-9.461071E-02,+0.000000E+00),complex(+2.874347E-01,-0.000000E+00),complex(-5.029060E-02,+0.000000E+00),complex(-3.100121E-01,+0.000000E+00),complex(+3.466040E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+4.002239E-01),complex(+0.000000E+00,-1.789856E-01),complex(+0.000000E+00,-2.322823E-01),complex(-0.000000E+00,+2.133649E-01),complex(+0.000000E+00,+1.552458E-01),complex(+0.000000E+00,-2.314268E-01),complex(+0.000000E+00,-9.757809E-02),complex(-0.000000E+00,+2.414935E-01),complex(+0.000000E+00,+4.736076E-02),complex(+0.000000E+00,-2.460938E-01),0,complex(+0.000000E+00,+2.460938E-01),complex(+0.000000E+00,-4.736076E-02),complex(-0.000000E+00,-2.414935E-01),complex(+0.000000E+00,+9.757809E-02),complex(+0.000000E+00,+2.314268E-01),complex(+0.000000E+00,-1.552458E-01),complex(-0.000000E+00,-2.133649E-01),complex(+0.000000E+00,+2.322823E-01),complex(+0.000000E+00,+1.789856E-01),complex(+0.000000E+00,-4.002239E-01)],$
            [complex(-4.197583E-01,+0.000000E+00),0,complex(+3.045247E-01,+0.000000E+00),0,complex(-2.713719E-01,+0.000000E+00),0,complex(+2.558519E-01,+0.000000E+00),0,complex(-2.483619E-01,+0.000000E+00),0,complex(+2.460938E-01,+0.000000E+00),0,complex(-2.483619E-01,+0.000000E+00),0,complex(+2.558519E-01,+0.000000E+00),0,complex(-2.713719E-01,+0.000000E+00),0,complex(+3.045247E-01,+0.000000E+00),0,complex(-4.197583E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.002239E-01),complex(+0.000000E+00,-1.789856E-01),complex(+0.000000E+00,+2.322823E-01),complex(-0.000000E+00,+2.133649E-01),complex(+0.000000E+00,-1.552458E-01),complex(+0.000000E+00,-2.314268E-01),complex(+0.000000E+00,+9.757809E-02),complex(-0.000000E+00,+2.414935E-01),complex(+0.000000E+00,-4.736076E-02),complex(+0.000000E+00,-2.460938E-01),0,complex(+0.000000E+00,+2.460938E-01),complex(+0.000000E+00,+4.736076E-02),complex(-0.000000E+00,-2.414935E-01),complex(+0.000000E+00,-9.757809E-02),complex(+0.000000E+00,+2.314268E-01),complex(+0.000000E+00,+1.552458E-01),complex(-0.000000E+00,-2.133649E-01),complex(+0.000000E+00,-2.322823E-01),complex(+0.000000E+00,+1.789856E-01),complex(+0.000000E+00,+4.002239E-01)],$
            [complex(+3.466040E-01,+0.000000E+00),complex(+3.100121E-01,+0.000000E+00),complex(-5.029060E-02,+0.000000E+00),complex(-2.874347E-01,-0.000000E+00),complex(-9.461071E-02,+0.000000E+00),complex(+2.226905E-01,+0.000000E+00),complex(+1.830944E-01,+0.000000E+00),complex(-1.394263E-01,-0.000000E+00),complex(-2.324219E-01,+0.000000E+00),complex(+4.736076E-02,+0.000000E+00),complex(+2.483619E-01,+0.000000E+00),complex(+4.736076E-02,-0.000000E+00),complex(-2.324219E-01,+0.000000E+00),complex(-1.394263E-01,+0.000000E+00),complex(+1.830944E-01,+0.000000E+00),complex(+2.226905E-01,-0.000000E+00),complex(-9.461071E-02,+0.000000E+00),complex(-2.874347E-01,+0.000000E+00),complex(-5.029060E-02,+0.000000E+00),complex(+3.100121E-01,-0.000000E+00),complex(+3.466040E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+2.718986E-01),complex(-0.000000E+00,+3.647902E-01),complex(+0.000000E+00,+1.578048E-01),complex(+0.000000E+00,-1.771647E-01),complex(+0.000000E+00,-2.695312E-01),complex(+0.000000E+00,-1.746928E-02),complex(+0.000000E+00,+2.430680E-01),complex(-0.000000E+00,+1.640625E-01),complex(+0.000000E+00,-1.394263E-01),complex(+0.000000E+00,-2.414935E-01),0,complex(+0.000000E+00,+2.414935E-01),complex(+0.000000E+00,+1.394263E-01),complex(-0.000000E+00,-1.640625E-01),complex(+0.000000E+00,-2.430680E-01),complex(+0.000000E+00,+1.746928E-02),complex(+0.000000E+00,+2.695312E-01),complex(+0.000000E+00,+1.771647E-01),complex(+0.000000E+00,-1.578048E-01),complex(-0.000000E+00,-3.647902E-01),complex(+0.000000E+00,-2.718986E-01)],$
            [complex(-1.922613E-01,+0.000000E+00),complex(-3.439275E-01,-0.000000E+00),complex(-3.068582E-01,+0.000000E+00),complex(-4.555431E-02,-0.000000E+00),complex(+2.292573E-01,+0.000000E+00),complex(+2.470529E-01,+0.000000E+00),complex(-7.812500E-03,+0.000000E+00),complex(-2.430680E-01,-0.000000E+00),complex(-1.830944E-01,+0.000000E+00),complex(+9.757809E-02,+0.000000E+00),complex(+2.558519E-01,+0.000000E+00),complex(+9.757809E-02,-0.000000E+00),complex(-1.830944E-01,+0.000000E+00),complex(-2.430680E-01,+0.000000E+00),complex(-7.812500E-03,+0.000000E+00),complex(+2.470529E-01,-0.000000E+00),complex(+2.292573E-01,+0.000000E+00),complex(-4.555431E-02,+0.000000E+00),complex(-3.068582E-01,+0.000000E+00),complex(-3.439275E-01,+0.000000E+00),complex(-1.922613E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.215967E-01),complex(+0.000000E+00,-2.718986E-01),complex(+0.000000E+00,-3.528622E-01),complex(+0.000000E+00,-2.520969E-01),complex(+0.000000E+00,+8.734641E-03),complex(-0.000000E+00,+2.421875E-01),complex(+0.000000E+00,+2.470529E-01),complex(-0.000000E+00,+1.746928E-02),complex(+0.000000E+00,-2.226905E-01),complex(+0.000000E+00,-2.314268E-01),0,complex(+0.000000E+00,+2.314268E-01),complex(+0.000000E+00,+2.226905E-01),complex(-0.000000E+00,-1.746928E-02),complex(+0.000000E+00,-2.470529E-01),complex(-0.000000E+00,-2.421875E-01),complex(+0.000000E+00,-8.734641E-03),complex(+0.000000E+00,+2.520969E-01),complex(+0.000000E+00,+3.528622E-01),complex(+0.000000E+00,+2.718986E-01),complex(+0.000000E+00,+1.215967E-01)],$
            [complex(+6.797464E-02,+0.000000E+00),complex(+1.823951E-01,+0.000000E+00),complex(+3.057467E-01,+0.000000E+00),complex(+3.462765E-01,+0.000000E+00),complex(+2.314453E-01,+0.000000E+00),complex(-8.734641E-03,-0.000000E+00),complex(-2.292573E-01,+0.000000E+00),complex(-2.695312E-01,-0.000000E+00),complex(-9.461071E-02,+0.000000E+00),complex(+1.552458E-01,+0.000000E+00),complex(+2.713719E-01,+0.000000E+00),complex(+1.552458E-01,-0.000000E+00),complex(-9.461071E-02,+0.000000E+00),complex(-2.695312E-01,+0.000000E+00),complex(-2.292573E-01,+0.000000E+00),complex(-8.734641E-03,+0.000000E+00),complex(+2.314453E-01,+0.000000E+00),complex(+3.462765E-01,-0.000000E+00),complex(+3.057467E-01,+0.000000E+00),complex(+1.823951E-01,-0.000000E+00),complex(+6.797464E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,+3.297254E-02),complex(-0.000000E+00,+1.032204E-01),complex(+0.000000E+00,+2.105030E-01),complex(-0.000000E+00,+3.144531E-01),complex(+0.000000E+00,+3.462765E-01),complex(-0.000000E+00,+2.520969E-01),complex(+0.000000E+00,+4.555431E-02),complex(+0.000000E+00,-1.771647E-01),complex(+0.000000E+00,-2.874347E-01),complex(+0.000000E+00,-2.133649E-01),0,complex(+0.000000E+00,+2.133649E-01),complex(+0.000000E+00,+2.874347E-01),complex(+0.000000E+00,+1.771647E-01),complex(+0.000000E+00,-4.555431E-02),complex(-0.000000E+00,-2.520969E-01),complex(+0.000000E+00,-3.462765E-01),complex(-0.000000E+00,-3.144531E-01),complex(+0.000000E+00,-2.105030E-01),complex(-0.000000E+00,-1.032204E-01),complex(+0.000000E+00,-3.297254E-02)],$
            [complex(-1.346099E-02,+0.000000E+00),complex(-4.815948E-02,-0.000000E+00),complex(-1.152344E-01,+0.000000E+00),complex(-2.105030E-01,-0.000000E+00),complex(-3.057467E-01,+0.000000E+00),complex(-3.528622E-01,-0.000000E+00),complex(-3.068582E-01,+0.000000E+00),complex(-1.578048E-01,-0.000000E+00),complex(+5.029060E-02,+0.000000E+00),complex(+2.322823E-01,+0.000000E+00),complex(+3.045247E-01,+0.000000E+00),complex(+2.322823E-01,-0.000000E+00),complex(+5.029060E-02,+0.000000E+00),complex(-1.578048E-01,+0.000000E+00),complex(-3.068582E-01,+0.000000E+00),complex(-3.528622E-01,+0.000000E+00),complex(-3.057467E-01,+0.000000E+00),complex(-2.105030E-01,+0.000000E+00),complex(-1.152344E-01,+0.000000E+00),complex(-4.815948E-02,+0.000000E+00),complex(-1.346099E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.367320E-03),complex(+0.000000E+00,-1.757812E-02),complex(+0.000000E+00,-4.815948E-02),complex(+0.000000E+00,-1.032204E-01),complex(+0.000000E+00,-1.823951E-01),complex(+0.000000E+00,-2.718986E-01),complex(+0.000000E+00,-3.439275E-01),complex(+0.000000E+00,-3.647902E-01),complex(+0.000000E+00,-3.100121E-01),complex(+0.000000E+00,-1.789856E-01),0,complex(+0.000000E+00,+1.789856E-01),complex(+0.000000E+00,+3.100121E-01),complex(+0.000000E+00,+3.647902E-01),complex(+0.000000E+00,+3.439275E-01),complex(+0.000000E+00,+2.718986E-01),complex(+0.000000E+00,+1.823951E-01),complex(+0.000000E+00,+1.032204E-01),complex(+0.000000E+00,+4.815948E-02),complex(+0.000000E+00,+1.757812E-02),complex(+0.000000E+00,+4.367320E-03)],$
            [complex(+9.765625E-04,+0.000000E+00),complex(+4.367320E-03,+0.000000E+00),complex(+1.346099E-02,+0.000000E+00),complex(+3.297254E-02,+0.000000E+00),complex(+6.797464E-02,+0.000000E+00),complex(+1.215967E-01,+0.000000E+00),complex(+1.922613E-01,+0.000000E+00),complex(+2.718986E-01,+0.000000E+00),complex(+3.466040E-01,+0.000000E+00),complex(+4.002239E-01,+0.000000E+00),complex(+4.197583E-01,+0.000000E+00),complex(+4.002239E-01,-0.000000E+00),complex(+3.466040E-01,+0.000000E+00),complex(+2.718986E-01,-0.000000E+00),complex(+1.922613E-01,+0.000000E+00),complex(+1.215967E-01,-0.000000E+00),complex(+6.797464E-02,+0.000000E+00),complex(+3.297254E-02,-0.000000E+00),complex(+1.346099E-02,+0.000000E+00),complex(+4.367320E-03,-0.000000E+00),complex(+9.765625E-04,+0.000000E+00)]$
            ]
          end

      19: begin
            matrix[190:209,190:209]=[$
            [complex(-0.000000E+00,+1.381068E-03),complex(+0.000000E+00,-6.019935E-03),complex(-0.000000E+00,+1.805981E-02),complex(+0.000000E+00,-4.299094E-02),complex(-0.000000E+00,+8.598188E-02),complex(+0.000000E+00,-1.489250E-01),complex(-0.000000E+00,+2.274867E-01),complex(+0.000000E+00,-3.100121E-01),complex(-0.000000E+00,+3.796857E-01),complex(+0.000000E+00,-4.197583E-01),complex(+0.000000E+00,+4.197583E-01),complex(+0.000000E+00,-3.796857E-01),complex(+0.000000E+00,+3.100121E-01),complex(+0.000000E+00,-2.274867E-01),complex(+0.000000E+00,+1.489250E-01),complex(+0.000000E+00,-8.598188E-02),complex(+0.000000E+00,+4.299094E-02),complex(+0.000000E+00,-1.805981E-02),complex(+0.000000E+00,+6.019935E-03),complex(+0.000000E+00,-1.381068E-03)],$
            [complex(-6.019935E-03,+0.000000E+00),complex(+2.347815E-02,+0.000000E+00),complex(-6.214806E-02,+0.000000E+00),complex(+1.282164E-01,+0.000000E+00),complex(-2.169815E-01,+0.000000E+00),complex(+3.074916E-01,+0.000000E+00),complex(-3.653231E-01,+0.000000E+00),complex(+3.556082E-01,+0.000000E+00),complex(-2.613176E-01,+0.000000E+00),complex(+9.629916E-02,+0.000000E+00),complex(+9.629916E-02,+0.000000E+00),complex(-2.613176E-01,+0.000000E+00),complex(+3.556082E-01,+0.000000E+00),complex(-3.653231E-01,+0.000000E+00),complex(+3.074916E-01,+0.000000E+00),complex(-2.169815E-01,+0.000000E+00),complex(+1.282164E-01,+0.000000E+00),complex(-6.214806E-02,+0.000000E+00),complex(+2.347815E-02,+0.000000E+00),complex(-6.019935E-03,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.805981E-02),complex(+0.000000E+00,+6.214806E-02),complex(+0.000000E+00,-1.422500E-01),complex(+0.000000E+00,+2.465699E-01),complex(+0.000000E+00,-3.353351E-01),complex(+0.000000E+00,+3.530459E-01),complex(+0.000000E+00,-2.609451E-01),complex(+0.000000E+00,+7.112165E-02),complex(-0.000000E+00,+1.451765E-01),complex(+0.000000E+00,-2.888975E-01),complex(+0.000000E+00,+2.888975E-01),complex(+0.000000E+00,-1.451765E-01),complex(-0.000000E+00,-7.112165E-02),complex(+0.000000E+00,+2.609451E-01),complex(-0.000000E+00,-3.530459E-01),complex(+0.000000E+00,+3.353351E-01),complex(-0.000000E+00,-2.465699E-01),complex(+0.000000E+00,+1.422500E-01),complex(-0.000000E+00,-6.214806E-02),complex(+0.000000E+00,+1.805981E-02)],$
            [complex(+4.299094E-02,+0.000000E+00),complex(-1.282164E-01,-0.000000E+00),complex(+2.465699E-01,+0.000000E+00),complex(-3.411238E-01,-0.000000E+00),complex(+3.342184E-01,+0.000000E+00),complex(-1.865822E-01,-0.000000E+00),complex(-5.115547E-02,+0.000000E+00),complex(+2.489756E-01,+0.000000E+00),complex(-2.805370E-01,+0.000000E+00),complex(+1.213612E-01,+0.000000E+00),complex(+1.213612E-01,+0.000000E+00),complex(-2.805370E-01,+0.000000E+00),complex(+2.489756E-01,+0.000000E+00),complex(-5.115547E-02,+0.000000E+00),complex(-1.865822E-01,+0.000000E+00),complex(+3.342184E-01,-0.000000E+00),complex(-3.411238E-01,+0.000000E+00),complex(+2.465699E-01,-0.000000E+00),complex(-1.282164E-01,+0.000000E+00),complex(+4.299094E-02,-0.000000E+00)],$
            [complex(-0.000000E+00,+8.598188E-02),complex(+0.000000E+00,-2.169815E-01),complex(-0.000000E+00,+3.353351E-01),complex(+0.000000E+00,-3.342184E-01),complex(-0.000000E+00,+1.602039E-01),complex(+0.000000E+00,+1.052515E-01),complex(+0.000000E+00,-2.777011E-01),complex(+0.000000E+00,+2.190985E-01),complex(-0.000000E+00,+2.439452E-02),complex(+0.000000E+00,-2.427224E-01),complex(+0.000000E+00,+2.427224E-01),complex(+0.000000E+00,-2.439452E-02),complex(-0.000000E+00,-2.190985E-01),complex(+0.000000E+00,+2.777011E-01),complex(-0.000000E+00,-1.052515E-01),complex(+0.000000E+00,-1.602039E-01),complex(+0.000000E+00,+3.342184E-01),complex(+0.000000E+00,-3.353351E-01),complex(+0.000000E+00,+2.169815E-01),complex(+0.000000E+00,-8.598188E-02)],$
            [complex(-1.489250E-01,+0.000000E+00),complex(+3.074916E-01,+0.000000E+00),complex(-3.530459E-01,+0.000000E+00),complex(+1.865822E-01,+0.000000E+00),complex(+1.052515E-01,+0.000000E+00),complex(-2.817379E-01,-0.000000E+00),complex(+1.772078E-01,+0.000000E+00),complex(+1.034972E-01,+0.000000E+00),complex(-2.675995E-01,+0.000000E+00),complex(+1.401359E-01,+0.000000E+00),complex(+1.401359E-01,+0.000000E+00),complex(-2.675995E-01,+0.000000E+00),complex(+1.034972E-01,+0.000000E+00),complex(+1.772078E-01,-0.000000E+00),complex(-2.817379E-01,+0.000000E+00),complex(+1.052515E-01,-0.000000E+00),complex(+1.865822E-01,+0.000000E+00),complex(-3.530459E-01,+0.000000E+00),complex(+3.074916E-01,+0.000000E+00),complex(-1.489250E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.274867E-01),complex(+0.000000E+00,+3.653231E-01),complex(+0.000000E+00,-2.609451E-01),complex(+0.000000E+00,-5.115547E-02),complex(-0.000000E+00,+2.777011E-01),complex(+0.000000E+00,-1.772078E-01),complex(+0.000000E+00,-1.270583E-01),complex(+0.000000E+00,+2.634910E-01),complex(+0.000000E+00,-6.454184E-02),complex(+0.000000E+00,-2.140611E-01),complex(+0.000000E+00,+2.140611E-01),complex(+0.000000E+00,+6.454184E-02),complex(-0.000000E+00,-2.634910E-01),complex(+0.000000E+00,+1.270583E-01),complex(+0.000000E+00,+1.772078E-01),complex(+0.000000E+00,-2.777011E-01),complex(+0.000000E+00,+5.115547E-02),complex(+0.000000E+00,+2.609451E-01),complex(-0.000000E+00,-3.653231E-01),complex(+0.000000E+00,+2.274867E-01)],$
            [complex(+3.100121E-01,+0.000000E+00),complex(-3.556082E-01,-0.000000E+00),complex(+7.112165E-02,+0.000000E+00),complex(+2.489756E-01,+0.000000E+00),complex(-2.190985E-01,+0.000000E+00),complex(-1.034972E-01,-0.000000E+00),complex(+2.634910E-01,+0.000000E+00),complex(-3.866990E-02,-0.000000E+00),complex(-2.368038E-01,+0.000000E+00),complex(+1.570779E-01,+0.000000E+00),complex(+1.570779E-01,+0.000000E+00),complex(-2.368038E-01,+0.000000E+00),complex(-3.866990E-02,+0.000000E+00),complex(+2.634910E-01,-0.000000E+00),complex(-1.034972E-01,+0.000000E+00),complex(-2.190985E-01,+0.000000E+00),complex(+2.489756E-01,+0.000000E+00),complex(+7.112165E-02,-0.000000E+00),complex(-3.556082E-01,+0.000000E+00),complex(+3.100121E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+3.796857E-01),complex(+0.000000E+00,-2.613176E-01),complex(+0.000000E+00,-1.451765E-01),complex(+0.000000E+00,+2.805370E-01),complex(-0.000000E+00,+2.439452E-02),complex(+0.000000E+00,-2.675995E-01),complex(-0.000000E+00,+6.454184E-02),complex(+0.000000E+00,+2.368038E-01),complex(+0.000000E+00,-1.353447E-01),complex(+0.000000E+00,-1.923803E-01),complex(+0.000000E+00,+1.923803E-01),complex(+0.000000E+00,+1.353447E-01),complex(-0.000000E+00,-2.368038E-01),complex(+0.000000E+00,-6.454184E-02),complex(+0.000000E+00,+2.675995E-01),complex(+0.000000E+00,-2.439452E-02),complex(-0.000000E+00,-2.805370E-01),complex(+0.000000E+00,+1.451765E-01),complex(+0.000000E+00,+2.613176E-01),complex(+0.000000E+00,-3.796857E-01)],$
            [complex(-4.197583E-01,+0.000000E+00),complex(+9.629916E-02,+0.000000E+00),complex(+2.888975E-01,+0.000000E+00),complex(-1.213612E-01,-0.000000E+00),complex(-2.427224E-01,+0.000000E+00),complex(+1.401359E-01,+0.000000E+00),complex(+2.140611E-01,+0.000000E+00),complex(-1.570779E-01,-0.000000E+00),complex(-1.923803E-01,+0.000000E+00),complex(+1.740146E-01,+0.000000E+00),complex(+1.740146E-01,+0.000000E+00),complex(-1.923803E-01,+0.000000E+00),complex(-1.570779E-01,+0.000000E+00),complex(+2.140611E-01,-0.000000E+00),complex(+1.401359E-01,+0.000000E+00),complex(-2.427224E-01,+0.000000E+00),complex(-1.213612E-01,+0.000000E+00),complex(+2.888975E-01,-0.000000E+00),complex(+9.629916E-02,+0.000000E+00),complex(-4.197583E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.197583E-01),complex(+0.000000E+00,-9.629916E-02),complex(-0.000000E+00,+2.888975E-01),complex(+0.000000E+00,+1.213612E-01),complex(+0.000000E+00,-2.427224E-01),complex(+0.000000E+00,-1.401359E-01),complex(-0.000000E+00,+2.140611E-01),complex(+0.000000E+00,+1.570779E-01),complex(+0.000000E+00,-1.923803E-01),complex(+0.000000E+00,-1.740146E-01),complex(+0.000000E+00,+1.740146E-01),complex(+0.000000E+00,+1.923803E-01),complex(-0.000000E+00,-1.570779E-01),complex(+0.000000E+00,-2.140611E-01),complex(+0.000000E+00,+1.401359E-01),complex(+0.000000E+00,+2.427224E-01),complex(-0.000000E+00,-1.213612E-01),complex(+0.000000E+00,-2.888975E-01),complex(+0.000000E+00,+9.629916E-02),complex(+0.000000E+00,+4.197583E-01)],$
            [complex(+3.796857E-01,+0.000000E+00),complex(+2.613176E-01,+0.000000E+00),complex(-1.451765E-01,+0.000000E+00),complex(-2.805370E-01,-0.000000E+00),complex(+2.439452E-02,+0.000000E+00),complex(+2.675995E-01,+0.000000E+00),complex(+6.454184E-02,+0.000000E+00),complex(-2.368038E-01,-0.000000E+00),complex(-1.353447E-01,+0.000000E+00),complex(+1.923803E-01,+0.000000E+00),complex(+1.923803E-01,+0.000000E+00),complex(-1.353447E-01,+0.000000E+00),complex(-2.368038E-01,+0.000000E+00),complex(+6.454184E-02,-0.000000E+00),complex(+2.675995E-01,+0.000000E+00),complex(+2.439452E-02,-0.000000E+00),complex(-2.805370E-01,+0.000000E+00),complex(-1.451765E-01,+0.000000E+00),complex(+2.613176E-01,+0.000000E+00),complex(+3.796857E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+3.100121E-01),complex(+0.000000E+00,+3.556082E-01),complex(-0.000000E+00,+7.112165E-02),complex(+0.000000E+00,-2.489756E-01),complex(+0.000000E+00,-2.190985E-01),complex(+0.000000E+00,+1.034972E-01),complex(-0.000000E+00,+2.634910E-01),complex(+0.000000E+00,+3.866990E-02),complex(+0.000000E+00,-2.368038E-01),complex(+0.000000E+00,-1.570779E-01),complex(+0.000000E+00,+1.570779E-01),complex(+0.000000E+00,+2.368038E-01),complex(-0.000000E+00,-3.866990E-02),complex(+0.000000E+00,-2.634910E-01),complex(-0.000000E+00,-1.034972E-01),complex(+0.000000E+00,+2.190985E-01),complex(+0.000000E+00,+2.489756E-01),complex(+0.000000E+00,-7.112165E-02),complex(-0.000000E+00,-3.556082E-01),complex(+0.000000E+00,-3.100121E-01)],$
            [complex(-2.274867E-01,+0.000000E+00),complex(-3.653231E-01,-0.000000E+00),complex(-2.609451E-01,+0.000000E+00),complex(+5.115547E-02,+0.000000E+00),complex(+2.777011E-01,+0.000000E+00),complex(+1.772078E-01,+0.000000E+00),complex(-1.270583E-01,+0.000000E+00),complex(-2.634910E-01,-0.000000E+00),complex(-6.454184E-02,+0.000000E+00),complex(+2.140611E-01,+0.000000E+00),complex(+2.140611E-01,+0.000000E+00),complex(-6.454184E-02,+0.000000E+00),complex(-2.634910E-01,+0.000000E+00),complex(-1.270583E-01,+0.000000E+00),complex(+1.772078E-01,+0.000000E+00),complex(+2.777011E-01,-0.000000E+00),complex(+5.115547E-02,+0.000000E+00),complex(-2.609451E-01,+0.000000E+00),complex(-3.653231E-01,+0.000000E+00),complex(-2.274867E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.489250E-01),complex(+0.000000E+00,-3.074916E-01),complex(+0.000000E+00,-3.530459E-01),complex(+0.000000E+00,-1.865822E-01),complex(-0.000000E+00,+1.052515E-01),complex(+0.000000E+00,+2.817379E-01),complex(-0.000000E+00,+1.772078E-01),complex(+0.000000E+00,-1.034972E-01),complex(+0.000000E+00,-2.675995E-01),complex(+0.000000E+00,-1.401359E-01),complex(+0.000000E+00,+1.401359E-01),complex(+0.000000E+00,+2.675995E-01),complex(+0.000000E+00,+1.034972E-01),complex(+0.000000E+00,-1.772078E-01),complex(-0.000000E+00,-2.817379E-01),complex(+0.000000E+00,-1.052515E-01),complex(+0.000000E+00,+1.865822E-01),complex(+0.000000E+00,+3.530459E-01),complex(+0.000000E+00,+3.074916E-01),complex(+0.000000E+00,+1.489250E-01)],$
            [complex(+8.598188E-02,+0.000000E+00),complex(+2.169815E-01,+0.000000E+00),complex(+3.353351E-01,+0.000000E+00),complex(+3.342184E-01,+0.000000E+00),complex(+1.602039E-01,+0.000000E+00),complex(-1.052515E-01,-0.000000E+00),complex(-2.777011E-01,+0.000000E+00),complex(-2.190985E-01,-0.000000E+00),complex(+2.439452E-02,+0.000000E+00),complex(+2.427224E-01,+0.000000E+00),complex(+2.427224E-01,+0.000000E+00),complex(+2.439452E-02,-0.000000E+00),complex(-2.190985E-01,+0.000000E+00),complex(-2.777011E-01,+0.000000E+00),complex(-1.052515E-01,+0.000000E+00),complex(+1.602039E-01,-0.000000E+00),complex(+3.342184E-01,+0.000000E+00),complex(+3.353351E-01,-0.000000E+00),complex(+2.169815E-01,+0.000000E+00),complex(+8.598188E-02,-0.000000E+00)],$
            [complex(-0.000000E+00,+4.299094E-02),complex(+0.000000E+00,+1.282164E-01),complex(-0.000000E+00,+2.465699E-01),complex(+0.000000E+00,+3.411238E-01),complex(-0.000000E+00,+3.342184E-01),complex(+0.000000E+00,+1.865822E-01),complex(+0.000000E+00,-5.115547E-02),complex(+0.000000E+00,-2.489756E-01),complex(+0.000000E+00,-2.805370E-01),complex(+0.000000E+00,-1.213612E-01),complex(+0.000000E+00,+1.213612E-01),complex(+0.000000E+00,+2.805370E-01),complex(+0.000000E+00,+2.489756E-01),complex(+0.000000E+00,+5.115547E-02),complex(-0.000000E+00,-1.865822E-01),complex(+0.000000E+00,-3.342184E-01),complex(-0.000000E+00,-3.411238E-01),complex(+0.000000E+00,-2.465699E-01),complex(-0.000000E+00,-1.282164E-01),complex(+0.000000E+00,-4.299094E-02)],$
            [complex(-1.805981E-02,+0.000000E+00),complex(-6.214806E-02,-0.000000E+00),complex(-1.422500E-01,+0.000000E+00),complex(-2.465699E-01,-0.000000E+00),complex(-3.353351E-01,+0.000000E+00),complex(-3.530459E-01,-0.000000E+00),complex(-2.609451E-01,+0.000000E+00),complex(-7.112165E-02,-0.000000E+00),complex(+1.451765E-01,+0.000000E+00),complex(+2.888975E-01,+0.000000E+00),complex(+2.888975E-01,+0.000000E+00),complex(+1.451765E-01,-0.000000E+00),complex(-7.112165E-02,+0.000000E+00),complex(-2.609451E-01,+0.000000E+00),complex(-3.530459E-01,+0.000000E+00),complex(-3.353351E-01,+0.000000E+00),complex(-2.465699E-01,+0.000000E+00),complex(-1.422500E-01,+0.000000E+00),complex(-6.214806E-02,+0.000000E+00),complex(-1.805981E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-6.019935E-03),complex(+0.000000E+00,-2.347815E-02),complex(+0.000000E+00,-6.214806E-02),complex(+0.000000E+00,-1.282164E-01),complex(+0.000000E+00,-2.169815E-01),complex(+0.000000E+00,-3.074916E-01),complex(+0.000000E+00,-3.653231E-01),complex(+0.000000E+00,-3.556082E-01),complex(+0.000000E+00,-2.613176E-01),complex(+0.000000E+00,-9.629916E-02),complex(+0.000000E+00,+9.629916E-02),complex(+0.000000E+00,+2.613176E-01),complex(+0.000000E+00,+3.556082E-01),complex(+0.000000E+00,+3.653231E-01),complex(+0.000000E+00,+3.074916E-01),complex(+0.000000E+00,+2.169815E-01),complex(+0.000000E+00,+1.282164E-01),complex(+0.000000E+00,+6.214806E-02),complex(+0.000000E+00,+2.347815E-02),complex(+0.000000E+00,+6.019935E-03)],$
            [complex(+1.381068E-03,+0.000000E+00),complex(+6.019935E-03,+0.000000E+00),complex(+1.805981E-02,+0.000000E+00),complex(+4.299094E-02,+0.000000E+00),complex(+8.598188E-02,+0.000000E+00),complex(+1.489250E-01,+0.000000E+00),complex(+2.274867E-01,+0.000000E+00),complex(+3.100121E-01,+0.000000E+00),complex(+3.796857E-01,+0.000000E+00),complex(+4.197583E-01,+0.000000E+00),complex(+4.197583E-01,+0.000000E+00),complex(+3.796857E-01,-0.000000E+00),complex(+3.100121E-01,+0.000000E+00),complex(+2.274867E-01,-0.000000E+00),complex(+1.489250E-01,+0.000000E+00),complex(+8.598188E-02,-0.000000E+00),complex(+4.299094E-02,+0.000000E+00),complex(+1.805981E-02,-0.000000E+00),complex(+6.019935E-03,+0.000000E+00),complex(+1.381068E-03,-0.000000E+00)]$
            ]
          end

      18: begin
            matrix[171:189,171:189]=[$
            [complex(-1.953125E-03,-0.000000E+00),complex(+8.286407E-03,+0.000000E+00),complex(-2.415882E-02,-0.000000E+00),complex(+5.579241E-02,+0.000000E+00),complex(-1.080415E-01,-0.000000E+00),complex(+1.807881E-01,+0.000000E+00),complex(-2.661128E-01,-0.000000E+00),complex(+3.484235E-01,+0.000000E+00),complex(-4.085627E-01,-0.000000E+00),complex(+4.306630E-01,+0.000000E+00),complex(-4.085627E-01,+0.000000E+00),complex(+3.484235E-01,+0.000000E+00),complex(-2.661128E-01,+0.000000E+00),complex(+1.807881E-01,+0.000000E+00),complex(-1.080415E-01,+0.000000E+00),complex(+5.579241E-02,+0.000000E+00),complex(-2.415882E-02,+0.000000E+00),complex(+8.286407E-03,+0.000000E+00),complex(-1.953125E-03,+0.000000E+00)],$
            [complex(+0.000000E+00,-8.286407E-03),complex(+0.000000E+00,+3.125000E-02),complex(+0.000000E+00,-7.972004E-02),complex(+0.000000E+00,+1.578048E-01),complex(+0.000000E+00,-2.546563E-01),complex(+0.000000E+00,+3.408973E-01),complex(+0.000000E+00,-3.763404E-01),complex(+0.000000E+00,+3.284968E-01),complex(+0.000000E+00,-1.925983E-01),0,complex(+0.000000E+00,+1.925983E-01),complex(+0.000000E+00,-3.284968E-01),complex(+0.000000E+00,+3.763404E-01),complex(+0.000000E+00,-3.408973E-01),complex(+0.000000E+00,+2.546563E-01),complex(+0.000000E+00,-1.578048E-01),complex(+0.000000E+00,+7.972004E-02),complex(+0.000000E+00,-3.125000E-02),complex(+0.000000E+00,+8.286407E-03)],$
            [complex(+2.415882E-02,+0.000000E+00),complex(-7.972004E-02,+0.000000E+00),complex(+1.738281E-01,+0.000000E+00),complex(-2.841646E-01,+0.000000E+00),complex(+3.581203E-01,+0.000000E+00),complex(-3.361645E-01,+0.000000E+00),complex(+1.936255E-01,+0.000000E+00),complex(+2.816837E-02,+0.000000E+00),complex(-2.312124E-01,-0.000000E+00),complex(+3.133533E-01,+0.000000E+00),complex(-2.312124E-01,+0.000000E+00),complex(+2.816837E-02,+0.000000E+00),complex(+1.936255E-01,-0.000000E+00),complex(-3.361645E-01,+0.000000E+00),complex(+3.581203E-01,-0.000000E+00),complex(-2.841646E-01,+0.000000E+00),complex(+1.738281E-01,-0.000000E+00),complex(-7.972004E-02,+0.000000E+00),complex(+2.415882E-02,-0.000000E+00)],$
            [complex(-0.000000E+00,+5.579241E-02),complex(+0.000000E+00,-1.578048E-01),complex(-0.000000E+00,+2.841646E-01),complex(+0.000000E+00,-3.593750E-01),complex(-0.000000E+00,+3.025768E-01),complex(+0.000000E+00,-1.012616E-01),complex(+0.000000E+00,-1.490530E-01),complex(+0.000000E+00,+2.927343E-01),complex(+0.000000E+00,-2.288409E-01),0,complex(+0.000000E+00,+2.288409E-01),complex(+0.000000E+00,-2.927343E-01),complex(+0.000000E+00,+1.490530E-01),complex(+0.000000E+00,+1.012616E-01),complex(-0.000000E+00,-3.025768E-01),complex(+0.000000E+00,+3.593750E-01),complex(-0.000000E+00,-2.841646E-01),complex(+0.000000E+00,+1.578048E-01),complex(-0.000000E+00,-5.579241E-02)],$
            [complex(-1.080415E-01,-0.000000E+00),complex(+2.546563E-01,+0.000000E+00),complex(-3.581203E-01,-0.000000E+00),complex(+3.025768E-01,+0.000000E+00),complex(-7.031250E-02,-0.000000E+00),complex(-1.960922E-01,+0.000000E+00),complex(+2.886399E-01,+0.000000E+00),complex(-1.259728E-01,+0.000000E+00),complex(-1.477162E-01,-0.000000E+00),complex(+2.802717E-01,+0.000000E+00),complex(-1.477162E-01,+0.000000E+00),complex(-1.259728E-01,+0.000000E+00),complex(+2.886399E-01,-0.000000E+00),complex(-1.960922E-01,+0.000000E+00),complex(-7.031250E-02,+0.000000E+00),complex(+3.025768E-01,+0.000000E+00),complex(-3.581203E-01,+0.000000E+00),complex(+2.546563E-01,+0.000000E+00),complex(-1.080415E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.807881E-01),complex(+0.000000E+00,+3.408973E-01),complex(+0.000000E+00,-3.361645E-01),complex(+0.000000E+00,+1.012616E-01),complex(-0.000000E+00,+1.960922E-01),complex(+0.000000E+00,-2.812500E-01),complex(-0.000000E+00,+6.899813E-02),complex(+0.000000E+00,+2.107928E-01),complex(+0.000000E+00,-2.471764E-01),0,complex(+0.000000E+00,+2.471764E-01),complex(+0.000000E+00,-2.107928E-01),complex(-0.000000E+00,-6.899813E-02),complex(+0.000000E+00,+2.812500E-01),complex(-0.000000E+00,-1.960922E-01),complex(+0.000000E+00,-1.012616E-01),complex(+0.000000E+00,+3.361645E-01),complex(+0.000000E+00,-3.408973E-01),complex(+0.000000E+00,+1.807881E-01)],$
            [complex(+2.661128E-01,+0.000000E+00),complex(-3.763404E-01,+0.000000E+00),complex(+1.936255E-01,+0.000000E+00),complex(+1.490530E-01,+0.000000E+00),complex(-2.886399E-01,-0.000000E+00),complex(+6.899813E-02,+0.000000E+00),complex(+2.265625E-01,+0.000000E+00),complex(-2.148082E-01,+0.000000E+00),complex(-8.396166E-02,-0.000000E+00),complex(+2.655101E-01,+0.000000E+00),complex(-8.396166E-02,+0.000000E+00),complex(-2.148082E-01,+0.000000E+00),complex(+2.265625E-01,-0.000000E+00),complex(+6.899813E-02,+0.000000E+00),complex(-2.886399E-01,+0.000000E+00),complex(+1.490530E-01,+0.000000E+00),complex(+1.936255E-01,-0.000000E+00),complex(-3.763404E-01,+0.000000E+00),complex(+2.661128E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+3.484235E-01),complex(+0.000000E+00,-3.284968E-01),complex(+0.000000E+00,-2.816837E-02),complex(+0.000000E+00,+2.927343E-01),complex(+0.000000E+00,-1.259728E-01),complex(+0.000000E+00,-2.107928E-01),complex(-0.000000E+00,+2.148082E-01),complex(+0.000000E+00,+1.093750E-01),complex(+0.000000E+00,-2.565071E-01),0,complex(+0.000000E+00,+2.565071E-01),complex(+0.000000E+00,-1.093750E-01),complex(-0.000000E+00,-2.148082E-01),complex(+0.000000E+00,+2.107928E-01),complex(+0.000000E+00,+1.259728E-01),complex(+0.000000E+00,-2.927343E-01),complex(+0.000000E+00,+2.816837E-02),complex(+0.000000E+00,+3.284968E-01),complex(-0.000000E+00,-3.484235E-01)],$
            [complex(-4.085627E-01,-0.000000E+00),complex(+1.925983E-01,+0.000000E+00),complex(+2.312124E-01,+0.000000E+00),complex(-2.288409E-01,+0.000000E+00),complex(-1.477162E-01,-0.000000E+00),complex(+2.471764E-01,+0.000000E+00),complex(+8.396166E-02,+0.000000E+00),complex(-2.565071E-01,+0.000000E+00),complex(-2.734375E-02,-0.000000E+00),complex(+2.594056E-01,+0.000000E+00),complex(-2.734375E-02,+0.000000E+00),complex(-2.565071E-01,+0.000000E+00),complex(+8.396166E-02,-0.000000E+00),complex(+2.471764E-01,+0.000000E+00),complex(-1.477162E-01,+0.000000E+00),complex(-2.288409E-01,+0.000000E+00),complex(+2.312124E-01,-0.000000E+00),complex(+1.925983E-01,+0.000000E+00),complex(-4.085627E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.306630E-01),0,complex(-0.000000E+00,+3.133533E-01),0,complex(+0.000000E+00,-2.802717E-01),0,complex(-0.000000E+00,+2.655101E-01),0,complex(+0.000000E+00,-2.594056E-01),0,complex(+0.000000E+00,+2.594056E-01),0,complex(-0.000000E+00,-2.655101E-01),0,complex(+0.000000E+00,+2.802717E-01),0,complex(-0.000000E+00,-3.133533E-01),0,complex(+0.000000E+00,+4.306630E-01)],$
            [complex(+4.085627E-01,+0.000000E+00),complex(+1.925983E-01,+0.000000E+00),complex(-2.312124E-01,-0.000000E+00),complex(-2.288409E-01,+0.000000E+00),complex(+1.477162E-01,+0.000000E+00),complex(+2.471764E-01,+0.000000E+00),complex(-8.396166E-02,-0.000000E+00),complex(-2.565071E-01,+0.000000E+00),complex(+2.734375E-02,+0.000000E+00),complex(+2.594056E-01,+0.000000E+00),complex(+2.734375E-02,-0.000000E+00),complex(-2.565071E-01,+0.000000E+00),complex(-8.396166E-02,+0.000000E+00),complex(+2.471764E-01,+0.000000E+00),complex(+1.477162E-01,-0.000000E+00),complex(-2.288409E-01,+0.000000E+00),complex(-2.312124E-01,+0.000000E+00),complex(+1.925983E-01,+0.000000E+00),complex(+4.085627E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+3.484235E-01),complex(+0.000000E+00,+3.284968E-01),complex(+0.000000E+00,-2.816837E-02),complex(+0.000000E+00,-2.927343E-01),complex(+0.000000E+00,-1.259728E-01),complex(+0.000000E+00,+2.107928E-01),complex(-0.000000E+00,+2.148082E-01),complex(+0.000000E+00,-1.093750E-01),complex(+0.000000E+00,-2.565071E-01),0,complex(+0.000000E+00,+2.565071E-01),complex(+0.000000E+00,+1.093750E-01),complex(-0.000000E+00,-2.148082E-01),complex(+0.000000E+00,-2.107928E-01),complex(+0.000000E+00,+1.259728E-01),complex(+0.000000E+00,+2.927343E-01),complex(+0.000000E+00,+2.816837E-02),complex(+0.000000E+00,-3.284968E-01),complex(-0.000000E+00,-3.484235E-01)],$
            [complex(-2.661128E-01,-0.000000E+00),complex(-3.763404E-01,+0.000000E+00),complex(-1.936255E-01,-0.000000E+00),complex(+1.490530E-01,+0.000000E+00),complex(+2.886399E-01,+0.000000E+00),complex(+6.899813E-02,+0.000000E+00),complex(-2.265625E-01,-0.000000E+00),complex(-2.148082E-01,+0.000000E+00),complex(+8.396166E-02,+0.000000E+00),complex(+2.655101E-01,+0.000000E+00),complex(+8.396166E-02,-0.000000E+00),complex(-2.148082E-01,+0.000000E+00),complex(-2.265625E-01,+0.000000E+00),complex(+6.899813E-02,+0.000000E+00),complex(+2.886399E-01,-0.000000E+00),complex(+1.490530E-01,+0.000000E+00),complex(-1.936255E-01,+0.000000E+00),complex(-3.763404E-01,+0.000000E+00),complex(-2.661128E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.807881E-01),complex(+0.000000E+00,-3.408973E-01),complex(+0.000000E+00,-3.361645E-01),complex(+0.000000E+00,-1.012616E-01),complex(-0.000000E+00,+1.960922E-01),complex(+0.000000E+00,+2.812500E-01),complex(-0.000000E+00,+6.899813E-02),complex(+0.000000E+00,-2.107928E-01),complex(+0.000000E+00,-2.471764E-01),0,complex(+0.000000E+00,+2.471764E-01),complex(+0.000000E+00,+2.107928E-01),complex(-0.000000E+00,-6.899813E-02),complex(+0.000000E+00,-2.812500E-01),complex(-0.000000E+00,-1.960922E-01),complex(+0.000000E+00,+1.012616E-01),complex(+0.000000E+00,+3.361645E-01),complex(+0.000000E+00,+3.408973E-01),complex(+0.000000E+00,+1.807881E-01)],$
            [complex(+1.080415E-01,+0.000000E+00),complex(+2.546563E-01,+0.000000E+00),complex(+3.581203E-01,+0.000000E+00),complex(+3.025768E-01,+0.000000E+00),complex(+7.031250E-02,+0.000000E+00),complex(-1.960922E-01,+0.000000E+00),complex(-2.886399E-01,-0.000000E+00),complex(-1.259728E-01,+0.000000E+00),complex(+1.477162E-01,+0.000000E+00),complex(+2.802717E-01,+0.000000E+00),complex(+1.477162E-01,-0.000000E+00),complex(-1.259728E-01,+0.000000E+00),complex(-2.886399E-01,+0.000000E+00),complex(-1.960922E-01,+0.000000E+00),complex(+7.031250E-02,-0.000000E+00),complex(+3.025768E-01,+0.000000E+00),complex(+3.581203E-01,-0.000000E+00),complex(+2.546563E-01,+0.000000E+00),complex(+1.080415E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+5.579241E-02),complex(+0.000000E+00,+1.578048E-01),complex(-0.000000E+00,+2.841646E-01),complex(+0.000000E+00,+3.593750E-01),complex(-0.000000E+00,+3.025768E-01),complex(+0.000000E+00,+1.012616E-01),complex(+0.000000E+00,-1.490530E-01),complex(+0.000000E+00,-2.927343E-01),complex(+0.000000E+00,-2.288409E-01),0,complex(+0.000000E+00,+2.288409E-01),complex(+0.000000E+00,+2.927343E-01),complex(+0.000000E+00,+1.490530E-01),complex(+0.000000E+00,-1.012616E-01),complex(-0.000000E+00,-3.025768E-01),complex(+0.000000E+00,-3.593750E-01),complex(-0.000000E+00,-2.841646E-01),complex(+0.000000E+00,-1.578048E-01),complex(-0.000000E+00,-5.579241E-02)],$
            [complex(-2.415882E-02,-0.000000E+00),complex(-7.972004E-02,+0.000000E+00),complex(-1.738281E-01,-0.000000E+00),complex(-2.841646E-01,+0.000000E+00),complex(-3.581203E-01,-0.000000E+00),complex(-3.361645E-01,+0.000000E+00),complex(-1.936255E-01,-0.000000E+00),complex(+2.816837E-02,+0.000000E+00),complex(+2.312124E-01,+0.000000E+00),complex(+3.133533E-01,+0.000000E+00),complex(+2.312124E-01,-0.000000E+00),complex(+2.816837E-02,+0.000000E+00),complex(-1.936255E-01,+0.000000E+00),complex(-3.361645E-01,+0.000000E+00),complex(-3.581203E-01,+0.000000E+00),complex(-2.841646E-01,+0.000000E+00),complex(-1.738281E-01,+0.000000E+00),complex(-7.972004E-02,+0.000000E+00),complex(-2.415882E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-8.286407E-03),complex(+0.000000E+00,-3.125000E-02),complex(+0.000000E+00,-7.972004E-02),complex(+0.000000E+00,-1.578048E-01),complex(+0.000000E+00,-2.546563E-01),complex(+0.000000E+00,-3.408973E-01),complex(+0.000000E+00,-3.763404E-01),complex(+0.000000E+00,-3.284968E-01),complex(+0.000000E+00,-1.925983E-01),0,complex(+0.000000E+00,+1.925983E-01),complex(+0.000000E+00,+3.284968E-01),complex(+0.000000E+00,+3.763404E-01),complex(+0.000000E+00,+3.408973E-01),complex(+0.000000E+00,+2.546563E-01),complex(+0.000000E+00,+1.578048E-01),complex(+0.000000E+00,+7.972004E-02),complex(+0.000000E+00,+3.125000E-02),complex(+0.000000E+00,+8.286407E-03)],$
            [complex(+1.953125E-03,+0.000000E+00),complex(+8.286407E-03,+0.000000E+00),complex(+2.415882E-02,+0.000000E+00),complex(+5.579241E-02,+0.000000E+00),complex(+1.080415E-01,+0.000000E+00),complex(+1.807881E-01,+0.000000E+00),complex(+2.661128E-01,+0.000000E+00),complex(+3.484235E-01,+0.000000E+00),complex(+4.085627E-01,+0.000000E+00),complex(+4.306630E-01,+0.000000E+00),complex(+4.085627E-01,-0.000000E+00),complex(+3.484235E-01,+0.000000E+00),complex(+2.661128E-01,-0.000000E+00),complex(+1.807881E-01,+0.000000E+00),complex(+1.080415E-01,-0.000000E+00),complex(+5.579241E-02,+0.000000E+00),complex(+2.415882E-02,-0.000000E+00),complex(+8.286407E-03,+0.000000E+00),complex(+1.953125E-03,-0.000000E+00)]$
            ]
          end

      17: begin
            matrix[153:170,153:170]=[$
            [complex(+0.000000E+00,-2.762136E-03),complex(-0.000000E+00,+1.138858E-02),complex(+0.000000E+00,-3.221176E-02),complex(-0.000000E+00,+7.202769E-02),complex(+0.000000E+00,-1.347515E-01),complex(-0.000000E+00,+2.172802E-01),complex(+0.000000E+00,-3.072806E-01),complex(-0.000000E+00,+3.851967E-01),complex(+0.000000E+00,-4.306630E-01),complex(+0.000000E+00,+4.306630E-01),complex(+0.000000E+00,-3.851967E-01),complex(+0.000000E+00,+3.072806E-01),complex(+0.000000E+00,-2.172802E-01),complex(+0.000000E+00,+1.347515E-01),complex(+0.000000E+00,-7.202769E-02),complex(+0.000000E+00,+3.221176E-02),complex(+0.000000E+00,-1.138858E-02),complex(+0.000000E+00,+2.762136E-03)],$
            [complex(+1.138858E-02,+0.000000E+00),complex(-4.143204E-02,+0.000000E+00),complex(+1.015625E-01,+0.000000E+00),complex(-1.921621E-01,+0.000000E+00),complex(+2.941383E-01,+0.000000E+00),complex(-3.688873E-01,+0.000000E+00),complex(+3.726325E-01,+0.000000E+00),complex(-2.802717E-01,+0.000000E+00),complex(+1.044511E-01,+0.000000E+00),complex(+1.044511E-01,+0.000000E+00),complex(-2.802717E-01,+0.000000E+00),complex(+3.726325E-01,+0.000000E+00),complex(-3.688873E-01,+0.000000E+00),complex(+2.941383E-01,+0.000000E+00),complex(-1.921621E-01,+0.000000E+00),complex(+1.015625E-01,+0.000000E+00),complex(-4.143204E-02,+0.000000E+00),complex(+1.138858E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,+3.221176E-02),complex(+0.000000E+00,-1.015625E-01),complex(+0.000000E+00,+2.099223E-01),complex(+0.000000E+00,-3.211688E-01),complex(+0.000000E+00,+3.697550E-01),complex(+0.000000E+00,-2.981060E-01),complex(+0.000000E+00,+1.053964E-01),complex(-0.000000E+00,+1.321214E-01),complex(+0.000000E+00,-2.954324E-01),complex(+0.000000E+00,+2.954324E-01),complex(+0.000000E+00,-1.321214E-01),complex(-0.000000E+00,-1.053964E-01),complex(+0.000000E+00,+2.981060E-01),complex(-0.000000E+00,-3.697550E-01),complex(+0.000000E+00,+3.211688E-01),complex(-0.000000E+00,-2.099223E-01),complex(+0.000000E+00,+1.015625E-01),complex(-0.000000E+00,-3.221176E-02)],$
            [complex(-7.202769E-02,-0.000000E+00),complex(+1.921621E-01,+0.000000E+00),complex(-3.211688E-01,-0.000000E+00),complex(+3.646019E-01,+0.000000E+00),complex(-2.480392E-01,-0.000000E+00),0,complex(+2.356735E-01,+0.000000E+00),complex(-2.954324E-01,+0.000000E+00),complex(+1.321214E-01,+0.000000E+00),complex(+1.321214E-01,+0.000000E+00),complex(-2.954324E-01,+0.000000E+00),complex(+2.356735E-01,+0.000000E+00),0,complex(-2.480392E-01,+0.000000E+00),complex(+3.646019E-01,-0.000000E+00),complex(-3.211688E-01,+0.000000E+00),complex(+1.921621E-01,-0.000000E+00),complex(-7.202769E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.347515E-01),complex(-0.000000E+00,+2.941383E-01),complex(+0.000000E+00,-3.697550E-01),complex(-0.000000E+00,+2.480392E-01),complex(+0.000000E+00,+3.314563E-02),complex(+0.000000E+00,-2.672286E-01),complex(+0.000000E+00,+2.519456E-01),0,complex(+0.000000E+00,-2.471764E-01),complex(+0.000000E+00,+2.471764E-01),0,complex(-0.000000E+00,-2.519456E-01),complex(+0.000000E+00,+2.672286E-01),complex(-0.000000E+00,-3.314563E-02),complex(+0.000000E+00,-2.480392E-01),complex(+0.000000E+00,+3.697550E-01),complex(+0.000000E+00,-2.941383E-01),complex(+0.000000E+00,+1.347515E-01)],$
            [complex(+2.172802E-01,+0.000000E+00),complex(-3.688873E-01,+0.000000E+00),complex(+2.981060E-01,+0.000000E+00),0,complex(-2.672286E-01,-0.000000E+00),complex(+2.320194E-01,+0.000000E+00),complex(+6.250000E-02,+0.000000E+00),complex(-2.742176E-01,+0.000000E+00),complex(+1.532923E-01,+0.000000E+00),complex(+1.532923E-01,+0.000000E+00),complex(-2.742176E-01,+0.000000E+00),complex(+6.250000E-02,+0.000000E+00),complex(+2.320194E-01,-0.000000E+00),complex(-2.672286E-01,+0.000000E+00),0,complex(+2.981060E-01,+0.000000E+00),complex(-3.688873E-01,+0.000000E+00),complex(+2.172802E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+3.072806E-01),complex(+0.000000E+00,-3.726325E-01),complex(+0.000000E+00,+1.053964E-01),complex(-0.000000E+00,+2.356735E-01),complex(+0.000000E+00,-2.519456E-01),complex(+0.000000E+00,-6.250000E-02),complex(+0.000000E+00,+2.762136E-01),complex(+0.000000E+00,-9.695058E-02),complex(+0.000000E+00,-2.167881E-01),complex(+0.000000E+00,+2.167881E-01),complex(+0.000000E+00,+9.695058E-02),complex(-0.000000E+00,-2.762136E-01),complex(+0.000000E+00,+6.250000E-02),complex(+0.000000E+00,+2.519456E-01),complex(+0.000000E+00,-2.356735E-01),complex(-0.000000E+00,-1.053964E-01),complex(+0.000000E+00,+3.726325E-01),complex(-0.000000E+00,-3.072806E-01)],$
            [complex(-3.851967E-01,-0.000000E+00),complex(+2.802717E-01,+0.000000E+00),complex(+1.321214E-01,+0.000000E+00),complex(-2.954324E-01,+0.000000E+00),0,complex(+2.742176E-01,+0.000000E+00),complex(-9.695058E-02,-0.000000E+00),complex(-2.320194E-01,+0.000000E+00),complex(+1.729371E-01,+0.000000E+00),complex(+1.729371E-01,+0.000000E+00),complex(-2.320194E-01,+0.000000E+00),complex(-9.695058E-02,+0.000000E+00),complex(+2.742176E-01,-0.000000E+00),0,complex(-2.954324E-01,+0.000000E+00),complex(+1.321214E-01,+0.000000E+00),complex(+2.802717E-01,-0.000000E+00),complex(-3.851967E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.306630E-01),complex(-0.000000E+00,+1.044511E-01),complex(+0.000000E+00,+2.954324E-01),complex(+0.000000E+00,-1.321214E-01),complex(+0.000000E+00,-2.471764E-01),complex(-0.000000E+00,+1.532923E-01),complex(+0.000000E+00,+2.167881E-01),complex(+0.000000E+00,-1.729371E-01),complex(+0.000000E+00,-1.933495E-01),complex(+0.000000E+00,+1.933495E-01),complex(+0.000000E+00,+1.729371E-01),complex(-0.000000E+00,-2.167881E-01),complex(+0.000000E+00,-1.532923E-01),complex(+0.000000E+00,+2.471764E-01),complex(+0.000000E+00,+1.321214E-01),complex(-0.000000E+00,-2.954324E-01),complex(+0.000000E+00,-1.044511E-01),complex(+0.000000E+00,+4.306630E-01)],$
            [complex(+4.306630E-01,+0.000000E+00),complex(+1.044511E-01,+0.000000E+00),complex(-2.954324E-01,-0.000000E+00),complex(-1.321214E-01,+0.000000E+00),complex(+2.471764E-01,+0.000000E+00),complex(+1.532923E-01,+0.000000E+00),complex(-2.167881E-01,-0.000000E+00),complex(-1.729371E-01,+0.000000E+00),complex(+1.933495E-01,+0.000000E+00),complex(+1.933495E-01,+0.000000E+00),complex(-1.729371E-01,+0.000000E+00),complex(-2.167881E-01,+0.000000E+00),complex(+1.532923E-01,-0.000000E+00),complex(+2.471764E-01,+0.000000E+00),complex(-1.321214E-01,+0.000000E+00),complex(-2.954324E-01,+0.000000E+00),complex(+1.044511E-01,-0.000000E+00),complex(+4.306630E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+3.851967E-01),complex(-0.000000E+00,+2.802717E-01),complex(+0.000000E+00,-1.321214E-01),complex(+0.000000E+00,-2.954324E-01),0,complex(-0.000000E+00,+2.742176E-01),complex(+0.000000E+00,+9.695058E-02),complex(+0.000000E+00,-2.320194E-01),complex(+0.000000E+00,-1.729371E-01),complex(+0.000000E+00,+1.729371E-01),complex(+0.000000E+00,+2.320194E-01),complex(-0.000000E+00,-9.695058E-02),complex(+0.000000E+00,-2.742176E-01),0,complex(+0.000000E+00,+2.954324E-01),complex(+0.000000E+00,+1.321214E-01),complex(+0.000000E+00,-2.802717E-01),complex(-0.000000E+00,-3.851967E-01)],$
            [complex(-3.072806E-01,-0.000000E+00),complex(-3.726325E-01,+0.000000E+00),complex(-1.053964E-01,-0.000000E+00),complex(+2.356735E-01,+0.000000E+00),complex(+2.519456E-01,+0.000000E+00),complex(-6.250000E-02,+0.000000E+00),complex(-2.762136E-01,-0.000000E+00),complex(-9.695058E-02,+0.000000E+00),complex(+2.167881E-01,+0.000000E+00),complex(+2.167881E-01,+0.000000E+00),complex(-9.695058E-02,+0.000000E+00),complex(-2.762136E-01,+0.000000E+00),complex(-6.250000E-02,+0.000000E+00),complex(+2.519456E-01,+0.000000E+00),complex(+2.356735E-01,-0.000000E+00),complex(-1.053964E-01,+0.000000E+00),complex(-3.726325E-01,+0.000000E+00),complex(-3.072806E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.172802E-01),complex(+0.000000E+00,-3.688873E-01),complex(+0.000000E+00,-2.981060E-01),0,complex(+0.000000E+00,+2.672286E-01),complex(-0.000000E+00,+2.320194E-01),complex(+0.000000E+00,-6.250000E-02),complex(+0.000000E+00,-2.742176E-01),complex(+0.000000E+00,-1.532923E-01),complex(+0.000000E+00,+1.532923E-01),complex(+0.000000E+00,+2.742176E-01),complex(+0.000000E+00,+6.250000E-02),complex(+0.000000E+00,-2.320194E-01),complex(-0.000000E+00,-2.672286E-01),0,complex(+0.000000E+00,+2.981060E-01),complex(+0.000000E+00,+3.688873E-01),complex(+0.000000E+00,+2.172802E-01)],$
            [complex(+1.347515E-01,+0.000000E+00),complex(+2.941383E-01,+0.000000E+00),complex(+3.697550E-01,+0.000000E+00),complex(+2.480392E-01,+0.000000E+00),complex(-3.314563E-02,-0.000000E+00),complex(-2.672286E-01,+0.000000E+00),complex(-2.519456E-01,-0.000000E+00),0,complex(+2.471764E-01,+0.000000E+00),complex(+2.471764E-01,+0.000000E+00),0,complex(-2.519456E-01,+0.000000E+00),complex(-2.672286E-01,+0.000000E+00),complex(-3.314563E-02,+0.000000E+00),complex(+2.480392E-01,-0.000000E+00),complex(+3.697550E-01,+0.000000E+00),complex(+2.941383E-01,-0.000000E+00),complex(+1.347515E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+7.202769E-02),complex(-0.000000E+00,+1.921621E-01),complex(+0.000000E+00,+3.211688E-01),complex(-0.000000E+00,+3.646019E-01),complex(+0.000000E+00,+2.480392E-01),0,complex(+0.000000E+00,-2.356735E-01),complex(+0.000000E+00,-2.954324E-01),complex(+0.000000E+00,-1.321214E-01),complex(+0.000000E+00,+1.321214E-01),complex(+0.000000E+00,+2.954324E-01),complex(+0.000000E+00,+2.356735E-01),0,complex(-0.000000E+00,-2.480392E-01),complex(+0.000000E+00,-3.646019E-01),complex(-0.000000E+00,-3.211688E-01),complex(+0.000000E+00,-1.921621E-01),complex(-0.000000E+00,-7.202769E-02)],$
            [complex(-3.221176E-02,-0.000000E+00),complex(-1.015625E-01,+0.000000E+00),complex(-2.099223E-01,-0.000000E+00),complex(-3.211688E-01,+0.000000E+00),complex(-3.697550E-01,-0.000000E+00),complex(-2.981060E-01,+0.000000E+00),complex(-1.053964E-01,-0.000000E+00),complex(+1.321214E-01,+0.000000E+00),complex(+2.954324E-01,+0.000000E+00),complex(+2.954324E-01,+0.000000E+00),complex(+1.321214E-01,-0.000000E+00),complex(-1.053964E-01,+0.000000E+00),complex(-2.981060E-01,+0.000000E+00),complex(-3.697550E-01,+0.000000E+00),complex(-3.211688E-01,+0.000000E+00),complex(-2.099223E-01,+0.000000E+00),complex(-1.015625E-01,+0.000000E+00),complex(-3.221176E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.138858E-02),complex(+0.000000E+00,-4.143204E-02),complex(+0.000000E+00,-1.015625E-01),complex(+0.000000E+00,-1.921621E-01),complex(+0.000000E+00,-2.941383E-01),complex(+0.000000E+00,-3.688873E-01),complex(+0.000000E+00,-3.726325E-01),complex(+0.000000E+00,-2.802717E-01),complex(+0.000000E+00,-1.044511E-01),complex(+0.000000E+00,+1.044511E-01),complex(+0.000000E+00,+2.802717E-01),complex(+0.000000E+00,+3.726325E-01),complex(+0.000000E+00,+3.688873E-01),complex(+0.000000E+00,+2.941383E-01),complex(+0.000000E+00,+1.921621E-01),complex(+0.000000E+00,+1.015625E-01),complex(+0.000000E+00,+4.143204E-02),complex(+0.000000E+00,+1.138858E-02)],$
            [complex(+2.762136E-03,+0.000000E+00),complex(+1.138858E-02,+0.000000E+00),complex(+3.221176E-02,+0.000000E+00),complex(+7.202769E-02,+0.000000E+00),complex(+1.347515E-01,+0.000000E+00),complex(+2.172802E-01,+0.000000E+00),complex(+3.072806E-01,+0.000000E+00),complex(+3.851967E-01,+0.000000E+00),complex(+4.306630E-01,+0.000000E+00),complex(+4.306630E-01,+0.000000E+00),complex(+3.851967E-01,-0.000000E+00),complex(+3.072806E-01,+0.000000E+00),complex(+2.172802E-01,-0.000000E+00),complex(+1.347515E-01,+0.000000E+00),complex(+7.202769E-02,-0.000000E+00),complex(+3.221176E-02,+0.000000E+00),complex(+1.138858E-02,-0.000000E+00),complex(+2.762136E-03,+0.000000E+00)]$
            ]
          end

      16: begin
            matrix[136:152,136:152]=[$
            [complex(+3.906250E-03,+0.000000E+00),complex(-1.562500E-02,-0.000000E+00),complex(+4.279083E-02,+0.000000E+00),complex(-9.243875E-02,-0.000000E+00),complex(+1.666463E-01,+0.000000E+00),complex(-2.581674E-01,-0.000000E+00),complex(+3.495603E-01,+0.000000E+00),complex(-4.178044E-01,-0.000000E+00),complex(+4.431485E-01,+0.000000E+00),complex(-4.178044E-01,+0.000000E+00),complex(+3.495603E-01,+0.000000E+00),complex(-2.581674E-01,+0.000000E+00),complex(+1.666463E-01,+0.000000E+00),complex(-9.243875E-02,+0.000000E+00),complex(+4.279083E-02,+0.000000E+00),complex(-1.562500E-02,+0.000000E+00),complex(+3.906250E-03,+0.000000E+00)],$
            [complex(+0.000000E+00,+1.562500E-02),complex(+0.000000E+00,-5.468750E-02),complex(+0.000000E+00,+1.283725E-01),complex(+0.000000E+00,-2.310969E-01),complex(+0.000000E+00,+3.332926E-01),complex(+0.000000E+00,-3.872510E-01),complex(+0.000000E+00,+3.495603E-01),complex(+0.000000E+00,-2.089022E-01),0,complex(+0.000000E+00,+2.089022E-01),complex(+0.000000E+00,-3.495603E-01),complex(+0.000000E+00,+3.872510E-01),complex(+0.000000E+00,-3.332926E-01),complex(+0.000000E+00,+2.310969E-01),complex(+0.000000E+00,-1.283725E-01),complex(+0.000000E+00,+5.468750E-02),complex(+0.000000E+00,-1.562500E-02)],$
            [complex(-4.279083E-02,+0.000000E+00),complex(+1.283725E-01,+0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(+3.544155E-01,+0.000000E+00),complex(-3.651038E-01,+0.000000E+00),complex(+2.356735E-01,+0.000000E+00),0,complex(-2.288409E-01,-0.000000E+00),complex(+3.236299E-01,+0.000000E+00),complex(-2.288409E-01,+0.000000E+00),0,complex(+2.356735E-01,-0.000000E+00),complex(-3.651038E-01,+0.000000E+00),complex(+3.544155E-01,-0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(+1.283725E-01,-0.000000E+00),complex(-4.279083E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-9.243875E-02),complex(-0.000000E+00,+2.310969E-01),complex(+0.000000E+00,-3.544155E-01),complex(-0.000000E+00,+3.515625E-01),complex(+0.000000E+00,-1.690102E-01),complex(+0.000000E+00,-1.090956E-01),complex(+0.000000E+00,+2.954324E-01),complex(+0.000000E+00,-2.471764E-01),0,complex(+0.000000E+00,+2.471764E-01),complex(+0.000000E+00,-2.954324E-01),complex(+0.000000E+00,+1.090956E-01),complex(+0.000000E+00,+1.690102E-01),complex(-0.000000E+00,-3.515625E-01),complex(+0.000000E+00,+3.544155E-01),complex(-0.000000E+00,-2.310969E-01),complex(+0.000000E+00,+9.243875E-02)],$
            [complex(+1.666463E-01,+0.000000E+00),complex(-3.332926E-01,-0.000000E+00),complex(+3.651038E-01,+0.000000E+00),complex(-1.690102E-01,-0.000000E+00),complex(-1.406250E-01,+0.000000E+00),complex(+3.025768E-01,+0.000000E+00),complex(-1.638764E-01,+0.000000E+00),complex(-1.371088E-01,-0.000000E+00),complex(+2.908517E-01,+0.000000E+00),complex(-1.371088E-01,+0.000000E+00),complex(-1.638764E-01,+0.000000E+00),complex(+3.025768E-01,-0.000000E+00),complex(-1.406250E-01,+0.000000E+00),complex(-1.690102E-01,+0.000000E+00),complex(+3.651038E-01,+0.000000E+00),complex(-3.332926E-01,+0.000000E+00),complex(+1.666463E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+2.581674E-01),complex(+0.000000E+00,-3.872510E-01),complex(+0.000000E+00,+2.356735E-01),complex(-0.000000E+00,+1.090956E-01),complex(+0.000000E+00,-3.025768E-01),complex(-0.000000E+00,+1.328125E-01),complex(+0.000000E+00,+1.904072E-01),complex(+0.000000E+00,-2.655101E-01),0,complex(+0.000000E+00,+2.655101E-01),complex(+0.000000E+00,-1.904072E-01),complex(-0.000000E+00,-1.328125E-01),complex(+0.000000E+00,+3.025768E-01),complex(-0.000000E+00,-1.090956E-01),complex(+0.000000E+00,-2.356735E-01),complex(+0.000000E+00,+3.872510E-01),complex(+0.000000E+00,-2.581674E-01)],$
            [complex(-3.495603E-01,+0.000000E+00),complex(+3.495603E-01,+0.000000E+00),0,complex(-2.954324E-01,-0.000000E+00),complex(+1.638764E-01,+0.000000E+00),complex(+1.904072E-01,+0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(-6.536406E-02,-0.000000E+00),complex(+2.773162E-01,+0.000000E+00),complex(-6.536406E-02,+0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(+1.904072E-01,-0.000000E+00),complex(+1.638764E-01,+0.000000E+00),complex(-2.954324E-01,+0.000000E+00),0,complex(+3.495603E-01,-0.000000E+00),complex(-3.495603E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.178044E-01),complex(-0.000000E+00,+2.089022E-01),complex(+0.000000E+00,+2.288409E-01),complex(+0.000000E+00,-2.471764E-01),complex(+0.000000E+00,-1.371088E-01),complex(-0.000000E+00,+2.655101E-01),complex(+0.000000E+00,+6.536406E-02),complex(+0.000000E+00,-2.734375E-01),0,complex(+0.000000E+00,+2.734375E-01),complex(+0.000000E+00,-6.536406E-02),complex(-0.000000E+00,-2.655101E-01),complex(+0.000000E+00,+1.371088E-01),complex(+0.000000E+00,+2.471764E-01),complex(+0.000000E+00,-2.288409E-01),complex(-0.000000E+00,-2.089022E-01),complex(+0.000000E+00,+4.178044E-01)],$
            [complex(+4.431485E-01,+0.000000E+00),0,complex(-3.236299E-01,+0.000000E+00),0,complex(+2.908517E-01,+0.000000E+00),0,complex(-2.773162E-01,+0.000000E+00),0,complex(+2.734375E-01,+0.000000E+00),0,complex(-2.773162E-01,+0.000000E+00),0,complex(+2.908517E-01,+0.000000E+00),0,complex(-3.236299E-01,+0.000000E+00),0,complex(+4.431485E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+4.178044E-01),complex(-0.000000E+00,+2.089022E-01),complex(+0.000000E+00,-2.288409E-01),complex(+0.000000E+00,-2.471764E-01),complex(+0.000000E+00,+1.371088E-01),complex(-0.000000E+00,+2.655101E-01),complex(+0.000000E+00,-6.536406E-02),complex(+0.000000E+00,-2.734375E-01),0,complex(+0.000000E+00,+2.734375E-01),complex(+0.000000E+00,+6.536406E-02),complex(-0.000000E+00,-2.655101E-01),complex(+0.000000E+00,-1.371088E-01),complex(+0.000000E+00,+2.471764E-01),complex(+0.000000E+00,+2.288409E-01),complex(-0.000000E+00,-2.089022E-01),complex(+0.000000E+00,-4.178044E-01)],$
            [complex(-3.495603E-01,+0.000000E+00),complex(-3.495603E-01,-0.000000E+00),0,complex(+2.954324E-01,+0.000000E+00),complex(+1.638764E-01,+0.000000E+00),complex(-1.904072E-01,-0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(+6.536406E-02,+0.000000E+00),complex(+2.773162E-01,+0.000000E+00),complex(+6.536406E-02,-0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(-1.904072E-01,+0.000000E+00),complex(+1.638764E-01,+0.000000E+00),complex(+2.954324E-01,-0.000000E+00),0,complex(-3.495603E-01,+0.000000E+00),complex(-3.495603E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.581674E-01),complex(+0.000000E+00,-3.872510E-01),complex(+0.000000E+00,-2.356735E-01),complex(-0.000000E+00,+1.090956E-01),complex(+0.000000E+00,+3.025768E-01),complex(-0.000000E+00,+1.328125E-01),complex(+0.000000E+00,-1.904072E-01),complex(+0.000000E+00,-2.655101E-01),0,complex(+0.000000E+00,+2.655101E-01),complex(+0.000000E+00,+1.904072E-01),complex(-0.000000E+00,-1.328125E-01),complex(+0.000000E+00,-3.025768E-01),complex(-0.000000E+00,-1.090956E-01),complex(+0.000000E+00,+2.356735E-01),complex(+0.000000E+00,+3.872510E-01),complex(+0.000000E+00,+2.581674E-01)],$
            [complex(+1.666463E-01,+0.000000E+00),complex(+3.332926E-01,+0.000000E+00),complex(+3.651038E-01,+0.000000E+00),complex(+1.690102E-01,+0.000000E+00),complex(-1.406250E-01,+0.000000E+00),complex(-3.025768E-01,-0.000000E+00),complex(-1.638764E-01,+0.000000E+00),complex(+1.371088E-01,+0.000000E+00),complex(+2.908517E-01,+0.000000E+00),complex(+1.371088E-01,-0.000000E+00),complex(-1.638764E-01,+0.000000E+00),complex(-3.025768E-01,+0.000000E+00),complex(-1.406250E-01,+0.000000E+00),complex(+1.690102E-01,-0.000000E+00),complex(+3.651038E-01,+0.000000E+00),complex(+3.332926E-01,-0.000000E+00),complex(+1.666463E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+9.243875E-02),complex(-0.000000E+00,+2.310969E-01),complex(+0.000000E+00,+3.544155E-01),complex(-0.000000E+00,+3.515625E-01),complex(+0.000000E+00,+1.690102E-01),complex(+0.000000E+00,-1.090956E-01),complex(+0.000000E+00,-2.954324E-01),complex(+0.000000E+00,-2.471764E-01),0,complex(+0.000000E+00,+2.471764E-01),complex(+0.000000E+00,+2.954324E-01),complex(+0.000000E+00,+1.090956E-01),complex(+0.000000E+00,-1.690102E-01),complex(-0.000000E+00,-3.515625E-01),complex(+0.000000E+00,-3.544155E-01),complex(-0.000000E+00,-2.310969E-01),complex(+0.000000E+00,-9.243875E-02)],$
            [complex(-4.279083E-02,+0.000000E+00),complex(-1.283725E-01,-0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(-3.544155E-01,-0.000000E+00),complex(-3.651038E-01,+0.000000E+00),complex(-2.356735E-01,-0.000000E+00),0,complex(+2.288409E-01,+0.000000E+00),complex(+3.236299E-01,+0.000000E+00),complex(+2.288409E-01,-0.000000E+00),0,complex(-2.356735E-01,+0.000000E+00),complex(-3.651038E-01,+0.000000E+00),complex(-3.544155E-01,+0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(-1.283725E-01,+0.000000E+00),complex(-4.279083E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.562500E-02),complex(+0.000000E+00,-5.468750E-02),complex(+0.000000E+00,-1.283725E-01),complex(+0.000000E+00,-2.310969E-01),complex(+0.000000E+00,-3.332926E-01),complex(+0.000000E+00,-3.872510E-01),complex(+0.000000E+00,-3.495603E-01),complex(+0.000000E+00,-2.089022E-01),0,complex(+0.000000E+00,+2.089022E-01),complex(+0.000000E+00,+3.495603E-01),complex(+0.000000E+00,+3.872510E-01),complex(+0.000000E+00,+3.332926E-01),complex(+0.000000E+00,+2.310969E-01),complex(+0.000000E+00,+1.283725E-01),complex(+0.000000E+00,+5.468750E-02),complex(+0.000000E+00,+1.562500E-02)],$
            [complex(+3.906250E-03,+0.000000E+00),complex(+1.562500E-02,+0.000000E+00),complex(+4.279083E-02,+0.000000E+00),complex(+9.243875E-02,+0.000000E+00),complex(+1.666463E-01,+0.000000E+00),complex(+2.581674E-01,+0.000000E+00),complex(+3.495603E-01,+0.000000E+00),complex(+4.178044E-01,+0.000000E+00),complex(+4.431485E-01,+0.000000E+00),complex(+4.178044E-01,-0.000000E+00),complex(+3.495603E-01,+0.000000E+00),complex(+2.581674E-01,-0.000000E+00),complex(+1.666463E-01,+0.000000E+00),complex(+9.243875E-02,-0.000000E+00),complex(+4.279083E-02,+0.000000E+00),complex(+1.562500E-02,-0.000000E+00),complex(+3.906250E-03,+0.000000E+00)]$
            ]
          end

      15: begin
            matrix[120:135,120:135]=[$
            [complex(-0.000000E+00,+5.524272E-03),complex(+0.000000E+00,-2.139541E-02),complex(-0.000000E+00,+5.660694E-02),complex(+0.000000E+00,-1.178367E-01),complex(-0.000000E+00,+2.040992E-01),complex(+0.000000E+00,-3.027281E-01),complex(-0.000000E+00,+3.908203E-01),complex(+0.000000E+00,-4.431485E-01),complex(+0.000000E+00,+4.431485E-01),complex(+0.000000E+00,-3.908203E-01),complex(+0.000000E+00,+3.027281E-01),complex(+0.000000E+00,-2.040992E-01),complex(+0.000000E+00,+1.178367E-01),complex(+0.000000E+00,-5.660694E-02),complex(+0.000000E+00,+2.139541E-02),complex(+0.000000E+00,-5.524272E-03)],$
            [complex(-2.139541E-02,+0.000000E+00),complex(+7.181553E-02,+0.000000E+00),complex(-1.607743E-01,+0.000000E+00),complex(+2.738279E-01,+0.000000E+00),complex(-3.688873E-01,+0.000000E+00),complex(+3.908203E-01,+0.000000E+00),complex(-3.027281E-01,+0.000000E+00),complex(+1.144205E-01,+0.000000E+00),complex(+1.144205E-01,+0.000000E+00),complex(-3.027281E-01,+0.000000E+00),complex(+3.908203E-01,+0.000000E+00),complex(-3.688873E-01,+0.000000E+00),complex(+2.738279E-01,+0.000000E+00),complex(-1.607743E-01,+0.000000E+00),complex(+7.181553E-02,+0.000000E+00),complex(-2.139541E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-5.660694E-02),complex(+0.000000E+00,+1.607743E-01),complex(+0.000000E+00,-2.927864E-01),complex(+0.000000E+00,+3.794897E-01),complex(+0.000000E+00,-3.386067E-01),complex(+0.000000E+00,+1.477162E-01),complex(-0.000000E+00,+1.144205E-01),complex(+0.000000E+00,-3.027281E-01),complex(+0.000000E+00,+3.027281E-01),complex(+0.000000E+00,-1.144205E-01),complex(-0.000000E+00,-1.477162E-01),complex(+0.000000E+00,+3.386067E-01),complex(-0.000000E+00,-3.794897E-01),complex(+0.000000E+00,+2.927864E-01),complex(-0.000000E+00,-1.607743E-01),complex(+0.000000E+00,+5.660694E-02)],$
            [complex(+1.178367E-01,+0.000000E+00),complex(-2.738279E-01,-0.000000E+00),complex(+3.794897E-01,+0.000000E+00),complex(-3.148835E-01,-0.000000E+00),complex(+6.697823E-02,+0.000000E+00),complex(+2.128817E-01,+0.000000E+00),complex(-3.114729E-01,+0.000000E+00),complex(+1.454259E-01,+0.000000E+00),complex(+1.454259E-01,+0.000000E+00),complex(-3.114729E-01,+0.000000E+00),complex(+2.128817E-01,+0.000000E+00),complex(+6.697823E-02,-0.000000E+00),complex(-3.148835E-01,+0.000000E+00),complex(+3.794897E-01,-0.000000E+00),complex(-2.738279E-01,+0.000000E+00),complex(+1.178367E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+2.040992E-01),complex(+0.000000E+00,-3.688873E-01),complex(-0.000000E+00,+3.386067E-01),complex(+0.000000E+00,-6.697823E-02),complex(+0.000000E+00,-2.375437E-01),complex(+0.000000E+00,+2.867837E-01),complex(+0.000000E+00,-3.173453E-02),complex(+0.000000E+00,-2.518850E-01),complex(+0.000000E+00,+2.518850E-01),complex(+0.000000E+00,+3.173453E-02),complex(-0.000000E+00,-2.867837E-01),complex(+0.000000E+00,+2.375437E-01),complex(+0.000000E+00,+6.697823E-02),complex(+0.000000E+00,-3.386067E-01),complex(+0.000000E+00,+3.688873E-01),complex(+0.000000E+00,-2.040992E-01)],$
            [complex(-3.027281E-01,+0.000000E+00),complex(+3.908203E-01,+0.000000E+00),complex(-1.477162E-01,+0.000000E+00),complex(-2.128817E-01,-0.000000E+00),complex(+2.867837E-01,+0.000000E+00),complex(+5.524272E-03,+0.000000E+00),complex(-2.781404E-01,+0.000000E+00),complex(+1.698208E-01,+0.000000E+00),complex(+1.698208E-01,+0.000000E+00),complex(-2.781404E-01,+0.000000E+00),complex(+5.524272E-03,+0.000000E+00),complex(+2.867837E-01,-0.000000E+00),complex(-2.128817E-01,+0.000000E+00),complex(-1.477162E-01,+0.000000E+00),complex(+3.908203E-01,+0.000000E+00),complex(-3.027281E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.908203E-01),complex(+0.000000E+00,+3.027281E-01),complex(-0.000000E+00,+1.144205E-01),complex(+0.000000E+00,-3.114729E-01),complex(-0.000000E+00,+3.173453E-02),complex(+0.000000E+00,+2.781404E-01),complex(+0.000000E+00,-1.381068E-01),complex(+0.000000E+00,-2.192377E-01),complex(+0.000000E+00,+2.192377E-01),complex(+0.000000E+00,+1.381068E-01),complex(-0.000000E+00,-2.781404E-01),complex(+0.000000E+00,-3.173453E-02),complex(+0.000000E+00,+3.114729E-01),complex(+0.000000E+00,-1.144205E-01),complex(-0.000000E+00,-3.027281E-01),complex(+0.000000E+00,+3.908203E-01)],$
            [complex(+4.431485E-01,+0.000000E+00),complex(-1.144205E-01,-0.000000E+00),complex(-3.027281E-01,+0.000000E+00),complex(+1.454259E-01,+0.000000E+00),complex(+2.518850E-01,+0.000000E+00),complex(-1.698208E-01,-0.000000E+00),complex(-2.192377E-01,+0.000000E+00),complex(+1.933495E-01,+0.000000E+00),complex(+1.933495E-01,+0.000000E+00),complex(-2.192377E-01,+0.000000E+00),complex(-1.698208E-01,+0.000000E+00),complex(+2.518850E-01,-0.000000E+00),complex(+1.454259E-01,+0.000000E+00),complex(-3.027281E-01,+0.000000E+00),complex(-1.144205E-01,+0.000000E+00),complex(+4.431485E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+4.431485E-01),complex(+0.000000E+00,+1.144205E-01),complex(+0.000000E+00,-3.027281E-01),complex(+0.000000E+00,-1.454259E-01),complex(-0.000000E+00,+2.518850E-01),complex(+0.000000E+00,+1.698208E-01),complex(+0.000000E+00,-2.192377E-01),complex(+0.000000E+00,-1.933495E-01),complex(+0.000000E+00,+1.933495E-01),complex(+0.000000E+00,+2.192377E-01),complex(-0.000000E+00,-1.698208E-01),complex(+0.000000E+00,-2.518850E-01),complex(+0.000000E+00,+1.454259E-01),complex(+0.000000E+00,+3.027281E-01),complex(-0.000000E+00,-1.144205E-01),complex(+0.000000E+00,-4.431485E-01)],$
            [complex(-3.908203E-01,+0.000000E+00),complex(-3.027281E-01,-0.000000E+00),complex(+1.144205E-01,+0.000000E+00),complex(+3.114729E-01,+0.000000E+00),complex(+3.173453E-02,+0.000000E+00),complex(-2.781404E-01,-0.000000E+00),complex(-1.381068E-01,+0.000000E+00),complex(+2.192377E-01,+0.000000E+00),complex(+2.192377E-01,+0.000000E+00),complex(-1.381068E-01,+0.000000E+00),complex(-2.781404E-01,+0.000000E+00),complex(+3.173453E-02,-0.000000E+00),complex(+3.114729E-01,+0.000000E+00),complex(+1.144205E-01,-0.000000E+00),complex(-3.027281E-01,+0.000000E+00),complex(-3.908203E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.027281E-01),complex(+0.000000E+00,-3.908203E-01),complex(+0.000000E+00,-1.477162E-01),complex(+0.000000E+00,+2.128817E-01),complex(-0.000000E+00,+2.867837E-01),complex(+0.000000E+00,-5.524272E-03),complex(+0.000000E+00,-2.781404E-01),complex(+0.000000E+00,-1.698208E-01),complex(+0.000000E+00,+1.698208E-01),complex(+0.000000E+00,+2.781404E-01),complex(+0.000000E+00,+5.524272E-03),complex(+0.000000E+00,-2.867837E-01),complex(-0.000000E+00,-2.128817E-01),complex(+0.000000E+00,+1.477162E-01),complex(+0.000000E+00,+3.908203E-01),complex(+0.000000E+00,+3.027281E-01)],$
            [complex(+2.040992E-01,+0.000000E+00),complex(+3.688873E-01,+0.000000E+00),complex(+3.386067E-01,+0.000000E+00),complex(+6.697823E-02,+0.000000E+00),complex(-2.375437E-01,+0.000000E+00),complex(-2.867837E-01,-0.000000E+00),complex(-3.173453E-02,+0.000000E+00),complex(+2.518850E-01,+0.000000E+00),complex(+2.518850E-01,+0.000000E+00),complex(-3.173453E-02,+0.000000E+00),complex(-2.867837E-01,+0.000000E+00),complex(-2.375437E-01,+0.000000E+00),complex(+6.697823E-02,+0.000000E+00),complex(+3.386067E-01,-0.000000E+00),complex(+3.688873E-01,+0.000000E+00),complex(+2.040992E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+1.178367E-01),complex(+0.000000E+00,+2.738279E-01),complex(-0.000000E+00,+3.794897E-01),complex(+0.000000E+00,+3.148835E-01),complex(-0.000000E+00,+6.697823E-02),complex(+0.000000E+00,-2.128817E-01),complex(+0.000000E+00,-3.114729E-01),complex(+0.000000E+00,-1.454259E-01),complex(+0.000000E+00,+1.454259E-01),complex(+0.000000E+00,+3.114729E-01),complex(+0.000000E+00,+2.128817E-01),complex(+0.000000E+00,-6.697823E-02),complex(-0.000000E+00,-3.148835E-01),complex(+0.000000E+00,-3.794897E-01),complex(-0.000000E+00,-2.738279E-01),complex(+0.000000E+00,-1.178367E-01)],$
            [complex(-5.660694E-02,+0.000000E+00),complex(-1.607743E-01,-0.000000E+00),complex(-2.927864E-01,+0.000000E+00),complex(-3.794897E-01,-0.000000E+00),complex(-3.386067E-01,+0.000000E+00),complex(-1.477162E-01,-0.000000E+00),complex(+1.144205E-01,+0.000000E+00),complex(+3.027281E-01,+0.000000E+00),complex(+3.027281E-01,+0.000000E+00),complex(+1.144205E-01,-0.000000E+00),complex(-1.477162E-01,+0.000000E+00),complex(-3.386067E-01,+0.000000E+00),complex(-3.794897E-01,+0.000000E+00),complex(-2.927864E-01,+0.000000E+00),complex(-1.607743E-01,+0.000000E+00),complex(-5.660694E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.139541E-02),complex(+0.000000E+00,-7.181553E-02),complex(+0.000000E+00,-1.607743E-01),complex(+0.000000E+00,-2.738279E-01),complex(+0.000000E+00,-3.688873E-01),complex(+0.000000E+00,-3.908203E-01),complex(+0.000000E+00,-3.027281E-01),complex(+0.000000E+00,-1.144205E-01),complex(+0.000000E+00,+1.144205E-01),complex(+0.000000E+00,+3.027281E-01),complex(+0.000000E+00,+3.908203E-01),complex(+0.000000E+00,+3.688873E-01),complex(+0.000000E+00,+2.738279E-01),complex(+0.000000E+00,+1.607743E-01),complex(+0.000000E+00,+7.181553E-02),complex(+0.000000E+00,+2.139541E-02)],$
            [complex(+5.524272E-03,+0.000000E+00),complex(+2.139541E-02,+0.000000E+00),complex(+5.660694E-02,+0.000000E+00),complex(+1.178367E-01,+0.000000E+00),complex(+2.040992E-01,+0.000000E+00),complex(+3.027281E-01,+0.000000E+00),complex(+3.908203E-01,+0.000000E+00),complex(+4.431485E-01,+0.000000E+00),complex(+4.431485E-01,+0.000000E+00),complex(+3.908203E-01,-0.000000E+00),complex(+3.027281E-01,+0.000000E+00),complex(+2.040992E-01,-0.000000E+00),complex(+1.178367E-01,+0.000000E+00),complex(+5.660694E-02,-0.000000E+00),complex(+2.139541E-02,+0.000000E+00),complex(+5.524272E-03,-0.000000E+00)]$
            ]
          end

      14: begin
            matrix[105:119,105:119]=[$
            [complex(-7.812500E-03,-0.000000E+00),complex(+2.923170E-02,+0.000000E+00),complex(-7.452650E-02,-0.000000E+00),complex(+1.490530E-01,+0.000000E+00),complex(-2.471764E-01,-0.000000E+00),complex(+3.495603E-01,+0.000000E+00),complex(-4.281222E-01,-0.000000E+00),complex(+4.576818E-01,+0.000000E+00),complex(-4.281222E-01,+0.000000E+00),complex(+3.495603E-01,+0.000000E+00),complex(-2.471764E-01,+0.000000E+00),complex(+1.490530E-01,+0.000000E+00),complex(-7.452650E-02,+0.000000E+00),complex(+2.923170E-02,+0.000000E+00),complex(-7.812500E-03,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.923170E-02),complex(+0.000000E+00,+9.375000E-02),complex(+0.000000E+00,-1.991805E-01),complex(+0.000000E+00,+3.186887E-01),complex(+0.000000E+00,-3.963641E-01),complex(+0.000000E+00,+3.736956E-01),complex(+0.000000E+00,-2.288409E-01),0,complex(+0.000000E+00,+2.288409E-01),complex(+0.000000E+00,-3.736956E-01),complex(+0.000000E+00,+3.963641E-01),complex(+0.000000E+00,-3.186887E-01),complex(+0.000000E+00,+1.991805E-01),complex(+0.000000E+00,-9.375000E-02),complex(+0.000000E+00,+2.923170E-02)],$
            [complex(+7.452650E-02,+0.000000E+00),complex(-1.991805E-01,+0.000000E+00),complex(+3.359375E-01,+0.000000E+00),complex(-3.906250E-01,+0.000000E+00),complex(+2.850224E-01,+0.000000E+00),complex(-3.664387E-02,+0.000000E+00),complex(-2.243970E-01,-0.000000E+00),complex(+3.358466E-01,+0.000000E+00),complex(-2.243970E-01,+0.000000E+00),complex(-3.664387E-02,+0.000000E+00),complex(+2.850224E-01,-0.000000E+00),complex(-3.906250E-01,+0.000000E+00),complex(+3.359375E-01,-0.000000E+00),complex(-1.991805E-01,+0.000000E+00),complex(+7.452650E-02,-0.000000E+00)],$
            [complex(-0.000000E+00,+1.490530E-01),complex(+0.000000E+00,-3.186887E-01),complex(-0.000000E+00,+3.906250E-01),complex(+0.000000E+00,-2.500000E-01),complex(+0.000000E+00,-5.182226E-02),complex(+0.000000E+00,+2.931510E-01),complex(+0.000000E+00,-2.692764E-01),0,complex(+0.000000E+00,+2.692764E-01),complex(+0.000000E+00,-2.931510E-01),complex(+0.000000E+00,+5.182226E-02),complex(+0.000000E+00,+2.500000E-01),complex(-0.000000E+00,-3.906250E-01),complex(+0.000000E+00,+3.186887E-01),complex(-0.000000E+00,-1.490530E-01)],$
            [complex(-2.471764E-01,-0.000000E+00),complex(+3.963641E-01,+0.000000E+00),complex(-2.850224E-01,-0.000000E+00),complex(-5.182226E-02,+0.000000E+00),complex(+3.046875E-01,+0.000000E+00),complex(-2.099223E-01,+0.000000E+00),complex(-1.217848E-01,-0.000000E+00),complex(+3.037847E-01,+0.000000E+00),complex(-1.217848E-01,+0.000000E+00),complex(-2.099223E-01,+0.000000E+00),complex(+3.046875E-01,-0.000000E+00),complex(-5.182226E-02,+0.000000E+00),complex(-2.850224E-01,+0.000000E+00),complex(+3.963641E-01,+0.000000E+00),complex(-2.471764E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.495603E-01),complex(+0.000000E+00,+3.736956E-01),complex(+0.000000E+00,-3.664387E-02),complex(+0.000000E+00,-2.931510E-01),complex(-0.000000E+00,+2.099223E-01),complex(+0.000000E+00,+1.562500E-01),complex(+0.000000E+00,-2.870496E-01),0,complex(+0.000000E+00,+2.870496E-01),complex(+0.000000E+00,-1.562500E-01),complex(-0.000000E+00,-2.099223E-01),complex(+0.000000E+00,+2.931510E-01),complex(+0.000000E+00,+3.664387E-02),complex(+0.000000E+00,-3.736956E-01),complex(+0.000000E+00,+3.495603E-01)],$
            [complex(+4.281222E-01,+0.000000E+00),complex(-2.288409E-01,+0.000000E+00),complex(-2.243970E-01,-0.000000E+00),complex(+2.692764E-01,+0.000000E+00),complex(+1.217848E-01,+0.000000E+00),complex(-2.870496E-01,+0.000000E+00),complex(-3.906250E-02,-0.000000E+00),complex(+2.923170E-01,+0.000000E+00),complex(-3.906250E-02,+0.000000E+00),complex(-2.870496E-01,+0.000000E+00),complex(+1.217848E-01,-0.000000E+00),complex(+2.692764E-01,+0.000000E+00),complex(-2.243970E-01,+0.000000E+00),complex(-2.288409E-01,+0.000000E+00),complex(+4.281222E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+4.576818E-01),0,complex(+0.000000E+00,-3.358466E-01),0,complex(-0.000000E+00,+3.037847E-01),0,complex(+0.000000E+00,-2.923170E-01),0,complex(+0.000000E+00,+2.923170E-01),0,complex(-0.000000E+00,-3.037847E-01),0,complex(+0.000000E+00,+3.358466E-01),0,complex(-0.000000E+00,-4.576818E-01)],$
            [complex(-4.281222E-01,-0.000000E+00),complex(-2.288409E-01,+0.000000E+00),complex(+2.243970E-01,+0.000000E+00),complex(+2.692764E-01,+0.000000E+00),complex(-1.217848E-01,-0.000000E+00),complex(-2.870496E-01,+0.000000E+00),complex(+3.906250E-02,+0.000000E+00),complex(+2.923170E-01,+0.000000E+00),complex(+3.906250E-02,-0.000000E+00),complex(-2.870496E-01,+0.000000E+00),complex(-1.217848E-01,+0.000000E+00),complex(+2.692764E-01,+0.000000E+00),complex(+2.243970E-01,-0.000000E+00),complex(-2.288409E-01,+0.000000E+00),complex(-4.281222E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.495603E-01),complex(+0.000000E+00,-3.736956E-01),complex(+0.000000E+00,-3.664387E-02),complex(+0.000000E+00,+2.931510E-01),complex(-0.000000E+00,+2.099223E-01),complex(+0.000000E+00,-1.562500E-01),complex(+0.000000E+00,-2.870496E-01),0,complex(+0.000000E+00,+2.870496E-01),complex(+0.000000E+00,+1.562500E-01),complex(-0.000000E+00,-2.099223E-01),complex(+0.000000E+00,-2.931510E-01),complex(+0.000000E+00,+3.664387E-02),complex(+0.000000E+00,+3.736956E-01),complex(+0.000000E+00,+3.495603E-01)],$
            [complex(+2.471764E-01,+0.000000E+00),complex(+3.963641E-01,+0.000000E+00),complex(+2.850224E-01,+0.000000E+00),complex(-5.182226E-02,+0.000000E+00),complex(-3.046875E-01,-0.000000E+00),complex(-2.099223E-01,+0.000000E+00),complex(+1.217848E-01,+0.000000E+00),complex(+3.037847E-01,+0.000000E+00),complex(+1.217848E-01,-0.000000E+00),complex(-2.099223E-01,+0.000000E+00),complex(-3.046875E-01,+0.000000E+00),complex(-5.182226E-02,+0.000000E+00),complex(+2.850224E-01,-0.000000E+00),complex(+3.963641E-01,+0.000000E+00),complex(+2.471764E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+1.490530E-01),complex(+0.000000E+00,+3.186887E-01),complex(-0.000000E+00,+3.906250E-01),complex(+0.000000E+00,+2.500000E-01),complex(+0.000000E+00,-5.182226E-02),complex(+0.000000E+00,-2.931510E-01),complex(+0.000000E+00,-2.692764E-01),0,complex(+0.000000E+00,+2.692764E-01),complex(+0.000000E+00,+2.931510E-01),complex(+0.000000E+00,+5.182226E-02),complex(+0.000000E+00,-2.500000E-01),complex(-0.000000E+00,-3.906250E-01),complex(+0.000000E+00,-3.186887E-01),complex(-0.000000E+00,-1.490530E-01)],$
            [complex(-7.452650E-02,-0.000000E+00),complex(-1.991805E-01,+0.000000E+00),complex(-3.359375E-01,-0.000000E+00),complex(-3.906250E-01,+0.000000E+00),complex(-2.850224E-01,-0.000000E+00),complex(-3.664387E-02,+0.000000E+00),complex(+2.243970E-01,+0.000000E+00),complex(+3.358466E-01,+0.000000E+00),complex(+2.243970E-01,-0.000000E+00),complex(-3.664387E-02,+0.000000E+00),complex(-2.850224E-01,+0.000000E+00),complex(-3.906250E-01,+0.000000E+00),complex(-3.359375E-01,+0.000000E+00),complex(-1.991805E-01,+0.000000E+00),complex(-7.452650E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.923170E-02),complex(+0.000000E+00,-9.375000E-02),complex(+0.000000E+00,-1.991805E-01),complex(+0.000000E+00,-3.186887E-01),complex(+0.000000E+00,-3.963641E-01),complex(+0.000000E+00,-3.736956E-01),complex(+0.000000E+00,-2.288409E-01),0,complex(+0.000000E+00,+2.288409E-01),complex(+0.000000E+00,+3.736956E-01),complex(+0.000000E+00,+3.963641E-01),complex(+0.000000E+00,+3.186887E-01),complex(+0.000000E+00,+1.991805E-01),complex(+0.000000E+00,+9.375000E-02),complex(+0.000000E+00,+2.923170E-02)],$
            [complex(+7.812500E-03,+0.000000E+00),complex(+2.923170E-02,+0.000000E+00),complex(+7.452650E-02,+0.000000E+00),complex(+1.490530E-01,+0.000000E+00),complex(+2.471764E-01,+0.000000E+00),complex(+3.495603E-01,+0.000000E+00),complex(+4.281222E-01,+0.000000E+00),complex(+4.576818E-01,+0.000000E+00),complex(+4.281222E-01,-0.000000E+00),complex(+3.495603E-01,+0.000000E+00),complex(+2.471764E-01,-0.000000E+00),complex(+1.490530E-01,+0.000000E+00),complex(+7.452650E-02,-0.000000E+00),complex(+2.923170E-02,+0.000000E+00),complex(+7.812500E-03,-0.000000E+00)]$
            ]
          end

      13: begin
            matrix[91:104,91:104]=[$
            [complex(+0.000000E+00,-1.104854E-02),complex(-0.000000E+00,+3.983609E-02),complex(+0.000000E+00,-9.757809E-02),complex(-0.000000E+00,+1.868478E-01),complex(+0.000000E+00,-2.954324E-01),complex(-0.000000E+00,+3.963641E-01),complex(+0.000000E+00,-4.576818E-01),complex(+0.000000E+00,+4.576818E-01),complex(+0.000000E+00,-3.963641E-01),complex(+0.000000E+00,+2.954324E-01),complex(+0.000000E+00,-1.868478E-01),complex(+0.000000E+00,+9.757809E-02),complex(+0.000000E+00,-3.983609E-02),complex(+0.000000E+00,+1.104854E-02)],$
            [complex(+3.983609E-02,+0.000000E+00),complex(-1.215340E-01,+0.000000E+00),complex(+2.435696E-01,+0.000000E+00),complex(-3.627558E-01,+0.000000E+00),complex(+4.096909E-01,+0.000000E+00),complex(-3.297949E-01,+0.000000E+00),complex(+1.269381E-01,+0.000000E+00),complex(+1.269381E-01,+0.000000E+00),complex(-3.297949E-01,+0.000000E+00),complex(+4.096909E-01,+0.000000E+00),complex(-3.627558E-01,+0.000000E+00),complex(+2.435696E-01,+0.000000E+00),complex(-1.215340E-01,+0.000000E+00),complex(+3.983609E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,+9.757809E-02),complex(+0.000000E+00,-2.435696E-01),complex(+0.000000E+00,+3.756505E-01),complex(+0.000000E+00,-3.808143E-01),complex(+0.000000E+00,+2.007068E-01),complex(-0.000000E+00,+8.975879E-02),complex(+0.000000E+00,-3.109336E-01),complex(+0.000000E+00,+3.109336E-01),complex(+0.000000E+00,-8.975879E-02),complex(-0.000000E+00,-2.007068E-01),complex(+0.000000E+00,+3.808143E-01),complex(-0.000000E+00,-3.756505E-01),complex(+0.000000E+00,+2.435696E-01),complex(-0.000000E+00,-9.757809E-02)],$
            [complex(-1.868478E-01,-0.000000E+00),complex(+3.627558E-01,+0.000000E+00),complex(-3.808143E-01,-0.000000E+00),complex(+1.546796E-01,+0.000000E+00),complex(+1.746928E-01,+0.000000E+00),complex(-3.281250E-01,+0.000000E+00),complex(+1.623798E-01,+0.000000E+00),complex(+1.623798E-01,+0.000000E+00),complex(-3.281250E-01,+0.000000E+00),complex(+1.746928E-01,+0.000000E+00),complex(+1.546796E-01,-0.000000E+00),complex(-3.808143E-01,+0.000000E+00),complex(+3.627558E-01,-0.000000E+00),complex(-1.868478E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.954324E-01),complex(-0.000000E+00,+4.096909E-01),complex(+0.000000E+00,-2.007068E-01),complex(+0.000000E+00,-1.746928E-01),complex(+0.000000E+00,+3.204077E-01),complex(+0.000000E+00,-7.411588E-02),complex(+0.000000E+00,-2.567450E-01),complex(+0.000000E+00,+2.567450E-01),complex(+0.000000E+00,+7.411588E-02),complex(-0.000000E+00,-3.204077E-01),complex(+0.000000E+00,+1.746928E-01),complex(+0.000000E+00,+2.007068E-01),complex(+0.000000E+00,-4.096909E-01),complex(+0.000000E+00,+2.954324E-01)],$
            [complex(+3.963641E-01,+0.000000E+00),complex(-3.297949E-01,+0.000000E+00),complex(-8.975879E-02,-0.000000E+00),complex(+3.281250E-01,+0.000000E+00),complex(-7.411588E-02,-0.000000E+00),complex(-2.762136E-01,+0.000000E+00),complex(+1.913664E-01,+0.000000E+00),complex(+1.913664E-01,+0.000000E+00),complex(-2.762136E-01,+0.000000E+00),complex(-7.411588E-02,+0.000000E+00),complex(+3.281250E-01,-0.000000E+00),complex(-8.975879E-02,+0.000000E+00),complex(-3.297949E-01,+0.000000E+00),complex(+3.963641E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+4.576818E-01),complex(+0.000000E+00,-1.269381E-01),complex(+0.000000E+00,-3.109336E-01),complex(-0.000000E+00,+1.623798E-01),complex(+0.000000E+00,+2.567450E-01),complex(+0.000000E+00,-1.913664E-01),complex(+0.000000E+00,-2.209709E-01),complex(+0.000000E+00,+2.209709E-01),complex(+0.000000E+00,+1.913664E-01),complex(-0.000000E+00,-2.567450E-01),complex(+0.000000E+00,-1.623798E-01),complex(+0.000000E+00,+3.109336E-01),complex(+0.000000E+00,+1.269381E-01),complex(-0.000000E+00,-4.576818E-01)],$
            [complex(-4.576818E-01,-0.000000E+00),complex(-1.269381E-01,+0.000000E+00),complex(+3.109336E-01,+0.000000E+00),complex(+1.623798E-01,+0.000000E+00),complex(-2.567450E-01,-0.000000E+00),complex(-1.913664E-01,+0.000000E+00),complex(+2.209709E-01,+0.000000E+00),complex(+2.209709E-01,+0.000000E+00),complex(-1.913664E-01,+0.000000E+00),complex(-2.567450E-01,+0.000000E+00),complex(+1.623798E-01,-0.000000E+00),complex(+3.109336E-01,+0.000000E+00),complex(-1.269381E-01,+0.000000E+00),complex(-4.576818E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.963641E-01),complex(+0.000000E+00,-3.297949E-01),complex(+0.000000E+00,+8.975879E-02),complex(-0.000000E+00,+3.281250E-01),complex(+0.000000E+00,+7.411588E-02),complex(+0.000000E+00,-2.762136E-01),complex(+0.000000E+00,-1.913664E-01),complex(+0.000000E+00,+1.913664E-01),complex(+0.000000E+00,+2.762136E-01),complex(-0.000000E+00,-7.411588E-02),complex(+0.000000E+00,-3.281250E-01),complex(-0.000000E+00,-8.975879E-02),complex(+0.000000E+00,+3.297949E-01),complex(+0.000000E+00,+3.963641E-01)],$
            [complex(+2.954324E-01,+0.000000E+00),complex(+4.096909E-01,+0.000000E+00),complex(+2.007068E-01,+0.000000E+00),complex(-1.746928E-01,+0.000000E+00),complex(-3.204077E-01,-0.000000E+00),complex(-7.411588E-02,+0.000000E+00),complex(+2.567450E-01,+0.000000E+00),complex(+2.567450E-01,+0.000000E+00),complex(-7.411588E-02,+0.000000E+00),complex(-3.204077E-01,+0.000000E+00),complex(-1.746928E-01,+0.000000E+00),complex(+2.007068E-01,+0.000000E+00),complex(+4.096909E-01,-0.000000E+00),complex(+2.954324E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+1.868478E-01),complex(-0.000000E+00,+3.627558E-01),complex(+0.000000E+00,+3.808143E-01),complex(-0.000000E+00,+1.546796E-01),complex(+0.000000E+00,-1.746928E-01),complex(+0.000000E+00,-3.281250E-01),complex(+0.000000E+00,-1.623798E-01),complex(+0.000000E+00,+1.623798E-01),complex(+0.000000E+00,+3.281250E-01),complex(+0.000000E+00,+1.746928E-01),complex(+0.000000E+00,-1.546796E-01),complex(-0.000000E+00,-3.808143E-01),complex(+0.000000E+00,-3.627558E-01),complex(-0.000000E+00,-1.868478E-01)],$
            [complex(-9.757809E-02,-0.000000E+00),complex(-2.435696E-01,+0.000000E+00),complex(-3.756505E-01,-0.000000E+00),complex(-3.808143E-01,+0.000000E+00),complex(-2.007068E-01,-0.000000E+00),complex(+8.975879E-02,+0.000000E+00),complex(+3.109336E-01,+0.000000E+00),complex(+3.109336E-01,+0.000000E+00),complex(+8.975879E-02,-0.000000E+00),complex(-2.007068E-01,+0.000000E+00),complex(-3.808143E-01,+0.000000E+00),complex(-3.756505E-01,+0.000000E+00),complex(-2.435696E-01,+0.000000E+00),complex(-9.757809E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.983609E-02),complex(+0.000000E+00,-1.215340E-01),complex(+0.000000E+00,-2.435696E-01),complex(+0.000000E+00,-3.627558E-01),complex(+0.000000E+00,-4.096909E-01),complex(+0.000000E+00,-3.297949E-01),complex(+0.000000E+00,-1.269381E-01),complex(+0.000000E+00,+1.269381E-01),complex(+0.000000E+00,+3.297949E-01),complex(+0.000000E+00,+4.096909E-01),complex(+0.000000E+00,+3.627558E-01),complex(+0.000000E+00,+2.435696E-01),complex(+0.000000E+00,+1.215340E-01),complex(+0.000000E+00,+3.983609E-02)],$
            [complex(+1.104854E-02,+0.000000E+00),complex(+3.983609E-02,+0.000000E+00),complex(+9.757809E-02,+0.000000E+00),complex(+1.868478E-01,+0.000000E+00),complex(+2.954324E-01,+0.000000E+00),complex(+3.963641E-01,+0.000000E+00),complex(+4.576818E-01,+0.000000E+00),complex(+4.576818E-01,+0.000000E+00),complex(+3.963641E-01,-0.000000E+00),complex(+2.954324E-01,+0.000000E+00),complex(+1.868478E-01,-0.000000E+00),complex(+9.757809E-02,+0.000000E+00),complex(+3.983609E-02,-0.000000E+00),complex(+1.104854E-02,+0.000000E+00)]$
            ]
          end

      12: begin
            matrix[78:90,78:90]=[$
            [complex(+1.562500E-02,+0.000000E+00),complex(-5.412659E-02,-0.000000E+00),complex(+1.269381E-01,+0.000000E+00),complex(-2.317562E-01,-0.000000E+00),complex(+3.476343E-01,+0.000000E+00),complex(-4.397265E-01,-0.000000E+00),complex(+4.749589E-01,+0.000000E+00),complex(-4.397265E-01,+0.000000E+00),complex(+3.476343E-01,+0.000000E+00),complex(-2.317562E-01,+0.000000E+00),complex(+1.269381E-01,+0.000000E+00),complex(-5.412659E-02,+0.000000E+00),complex(+1.562500E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,+5.412659E-02),complex(+0.000000E+00,-1.562500E-01),complex(+0.000000E+00,+2.931510E-01),complex(+0.000000E+00,-4.014135E-01),complex(+0.000000E+00,+4.014135E-01),complex(+0.000000E+00,-2.538762E-01),0,complex(+0.000000E+00,+2.538762E-01),complex(+0.000000E+00,-4.014135E-01),complex(+0.000000E+00,+4.014135E-01),complex(+0.000000E+00,-2.931510E-01),complex(+0.000000E+00,+1.562500E-01),complex(+0.000000E+00,-5.412659E-02)],$
            [complex(-1.269381E-01,+0.000000E+00),complex(+2.931510E-01,+0.000000E+00),complex(-4.062500E-01,+0.000000E+00),complex(+3.423266E-01,+0.000000E+00),complex(-8.558165E-02,+0.000000E+00),complex(-2.165063E-01,-0.000000E+00),complex(+3.507804E-01,+0.000000E+00),complex(-2.165063E-01,+0.000000E+00),complex(-8.558165E-02,+0.000000E+00),complex(+3.423266E-01,-0.000000E+00),complex(-4.062500E-01,+0.000000E+00),complex(+2.931510E-01,-0.000000E+00),complex(-1.269381E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.317562E-01),complex(-0.000000E+00,+4.014135E-01),complex(+0.000000E+00,-3.423266E-01),complex(-0.000000E+00,+3.125000E-02),complex(+0.000000E+00,+2.812500E-01),complex(+0.000000E+00,-2.964635E-01),0,complex(+0.000000E+00,+2.964635E-01),complex(+0.000000E+00,-2.812500E-01),complex(-0.000000E+00,-3.125000E-02),complex(+0.000000E+00,+3.423266E-01),complex(-0.000000E+00,-4.014135E-01),complex(+0.000000E+00,+2.317562E-01)],$
            [complex(+3.476343E-01,+0.000000E+00),complex(-4.014135E-01,-0.000000E+00),complex(+8.558165E-02,+0.000000E+00),complex(+2.812500E-01,+0.000000E+00),complex(-2.656250E-01,+0.000000E+00),complex(-9.882118E-02,-0.000000E+00),complex(+3.202172E-01,+0.000000E+00),complex(-9.882118E-02,+0.000000E+00),complex(-2.656250E-01,+0.000000E+00),complex(+2.812500E-01,-0.000000E+00),complex(+8.558165E-02,+0.000000E+00),complex(-4.014135E-01,+0.000000E+00),complex(+3.476343E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+4.397265E-01),complex(+0.000000E+00,-2.538762E-01),complex(+0.000000E+00,-2.165063E-01),complex(-0.000000E+00,+2.964635E-01),complex(+0.000000E+00,+9.882118E-02),complex(+0.000000E+00,-3.125000E-01),0,complex(+0.000000E+00,+3.125000E-01),complex(+0.000000E+00,-9.882118E-02),complex(-0.000000E+00,-2.964635E-01),complex(+0.000000E+00,+2.165063E-01),complex(+0.000000E+00,+2.538762E-01),complex(+0.000000E+00,-4.397265E-01)],$
            [complex(-4.749589E-01,+0.000000E+00),0,complex(+3.507804E-01,+0.000000E+00),0,complex(-3.202172E-01,+0.000000E+00),0,complex(+3.125000E-01,+0.000000E+00),0,complex(-3.202172E-01,+0.000000E+00),0,complex(+3.507804E-01,+0.000000E+00),0,complex(-4.749589E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.397265E-01),complex(+0.000000E+00,-2.538762E-01),complex(+0.000000E+00,+2.165063E-01),complex(-0.000000E+00,+2.964635E-01),complex(+0.000000E+00,-9.882118E-02),complex(+0.000000E+00,-3.125000E-01),0,complex(+0.000000E+00,+3.125000E-01),complex(+0.000000E+00,+9.882118E-02),complex(-0.000000E+00,-2.964635E-01),complex(+0.000000E+00,-2.165063E-01),complex(+0.000000E+00,+2.538762E-01),complex(+0.000000E+00,+4.397265E-01)],$
            [complex(+3.476343E-01,+0.000000E+00),complex(+4.014135E-01,+0.000000E+00),complex(+8.558165E-02,+0.000000E+00),complex(-2.812500E-01,-0.000000E+00),complex(-2.656250E-01,+0.000000E+00),complex(+9.882118E-02,+0.000000E+00),complex(+3.202172E-01,+0.000000E+00),complex(+9.882118E-02,-0.000000E+00),complex(-2.656250E-01,+0.000000E+00),complex(-2.812500E-01,+0.000000E+00),complex(+8.558165E-02,+0.000000E+00),complex(+4.014135E-01,-0.000000E+00),complex(+3.476343E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+2.317562E-01),complex(-0.000000E+00,+4.014135E-01),complex(+0.000000E+00,+3.423266E-01),complex(-0.000000E+00,+3.125000E-02),complex(+0.000000E+00,-2.812500E-01),complex(+0.000000E+00,-2.964635E-01),0,complex(+0.000000E+00,+2.964635E-01),complex(+0.000000E+00,+2.812500E-01),complex(-0.000000E+00,-3.125000E-02),complex(+0.000000E+00,-3.423266E-01),complex(-0.000000E+00,-4.014135E-01),complex(+0.000000E+00,-2.317562E-01)],$
            [complex(-1.269381E-01,+0.000000E+00),complex(-2.931510E-01,-0.000000E+00),complex(-4.062500E-01,+0.000000E+00),complex(-3.423266E-01,-0.000000E+00),complex(-8.558165E-02,+0.000000E+00),complex(+2.165063E-01,+0.000000E+00),complex(+3.507804E-01,+0.000000E+00),complex(+2.165063E-01,-0.000000E+00),complex(-8.558165E-02,+0.000000E+00),complex(-3.423266E-01,+0.000000E+00),complex(-4.062500E-01,+0.000000E+00),complex(-2.931510E-01,+0.000000E+00),complex(-1.269381E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-5.412659E-02),complex(+0.000000E+00,-1.562500E-01),complex(+0.000000E+00,-2.931510E-01),complex(+0.000000E+00,-4.014135E-01),complex(+0.000000E+00,-4.014135E-01),complex(+0.000000E+00,-2.538762E-01),0,complex(+0.000000E+00,+2.538762E-01),complex(+0.000000E+00,+4.014135E-01),complex(+0.000000E+00,+4.014135E-01),complex(+0.000000E+00,+2.931510E-01),complex(+0.000000E+00,+1.562500E-01),complex(+0.000000E+00,+5.412659E-02)],$
            [complex(+1.562500E-02,+0.000000E+00),complex(+5.412659E-02,+0.000000E+00),complex(+1.269381E-01,+0.000000E+00),complex(+2.317562E-01,+0.000000E+00),complex(+3.476343E-01,+0.000000E+00),complex(+4.397265E-01,+0.000000E+00),complex(+4.749589E-01,+0.000000E+00),complex(+4.397265E-01,-0.000000E+00),complex(+3.476343E-01,+0.000000E+00),complex(+2.317562E-01,-0.000000E+00),complex(+1.269381E-01,+0.000000E+00),complex(+5.412659E-02,-0.000000E+00),complex(+1.562500E-02,+0.000000E+00)]$
            ]
          end

      11: begin
            matrix[66:77,66:77]=[$
            [complex(-0.000000E+00,+2.209709E-02),complex(+0.000000E+00,-7.328775E-02),complex(-0.000000E+00,+1.638764E-01),complex(+0.000000E+00,-2.838422E-01),complex(-0.000000E+00,+4.014135E-01),complex(+0.000000E+00,-4.749589E-01),complex(+0.000000E+00,+4.749589E-01),complex(+0.000000E+00,-4.014135E-01),complex(+0.000000E+00,+2.838422E-01),complex(+0.000000E+00,-1.638764E-01),complex(+0.000000E+00,+7.328775E-02),complex(+0.000000E+00,-2.209709E-02)],$
            [complex(-7.328775E-02,+0.000000E+00),complex(+1.988738E-01,+0.000000E+00),complex(-3.458741E-01,+0.000000E+00),complex(+4.279082E-01,+0.000000E+00),complex(-3.630922E-01,+0.000000E+00),complex(+1.432055E-01,+0.000000E+00),complex(+1.432055E-01,+0.000000E+00),complex(-3.630922E-01,+0.000000E+00),complex(+4.279082E-01,+0.000000E+00),complex(-3.458741E-01,+0.000000E+00),complex(+1.988738E-01,+0.000000E+00),complex(-7.328775E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.638764E-01),complex(+0.000000E+00,+3.458741E-01),complex(+0.000000E+00,-4.198447E-01),complex(+0.000000E+00,+2.679129E-01),complex(-0.000000E+00,+5.412659E-02),complex(+0.000000E+00,-3.202172E-01),complex(+0.000000E+00,+3.202172E-01),complex(+0.000000E+00,-5.412659E-02),complex(-0.000000E+00,-2.679129E-01),complex(+0.000000E+00,+4.198447E-01),complex(-0.000000E+00,-3.458741E-01),complex(+0.000000E+00,+1.638764E-01)],$
            [complex(+2.838422E-01,+0.000000E+00),complex(-4.279082E-01,-0.000000E+00),complex(+2.679129E-01,+0.000000E+00),complex(+1.104854E-01,+0.000000E+00),complex(-3.437500E-01,+0.000000E+00),complex(+1.848775E-01,+0.000000E+00),complex(+1.848775E-01,+0.000000E+00),complex(-3.437500E-01,+0.000000E+00),complex(+1.104854E-01,+0.000000E+00),complex(+2.679129E-01,-0.000000E+00),complex(-4.279082E-01,+0.000000E+00),complex(+2.838422E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+4.014135E-01),complex(+0.000000E+00,-3.630922E-01),complex(+0.000000E+00,-5.412659E-02),complex(+0.000000E+00,+3.437500E-01),complex(+0.000000E+00,-1.325825E-01),complex(+0.000000E+00,-2.614563E-01),complex(+0.000000E+00,+2.614563E-01),complex(+0.000000E+00,+1.325825E-01),complex(-0.000000E+00,-3.437500E-01),complex(+0.000000E+00,+5.412659E-02),complex(+0.000000E+00,+3.630922E-01),complex(+0.000000E+00,-4.014135E-01)],$
            [complex(-4.749589E-01,+0.000000E+00),complex(+1.432055E-01,+0.000000E+00),complex(+3.202172E-01,+0.000000E+00),complex(-1.848775E-01,-0.000000E+00),complex(-2.614563E-01,+0.000000E+00),complex(+2.209709E-01,+0.000000E+00),complex(+2.209709E-01,+0.000000E+00),complex(-2.614563E-01,+0.000000E+00),complex(-1.848775E-01,+0.000000E+00),complex(+3.202172E-01,-0.000000E+00),complex(+1.432055E-01,+0.000000E+00),complex(-4.749589E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.749589E-01),complex(+0.000000E+00,-1.432055E-01),complex(-0.000000E+00,+3.202172E-01),complex(+0.000000E+00,+1.848775E-01),complex(+0.000000E+00,-2.614563E-01),complex(+0.000000E+00,-2.209709E-01),complex(+0.000000E+00,+2.209709E-01),complex(+0.000000E+00,+2.614563E-01),complex(-0.000000E+00,-1.848775E-01),complex(+0.000000E+00,-3.202172E-01),complex(+0.000000E+00,+1.432055E-01),complex(+0.000000E+00,+4.749589E-01)],$
            [complex(+4.014135E-01,+0.000000E+00),complex(+3.630922E-01,+0.000000E+00),complex(-5.412659E-02,+0.000000E+00),complex(-3.437500E-01,-0.000000E+00),complex(-1.325825E-01,+0.000000E+00),complex(+2.614563E-01,+0.000000E+00),complex(+2.614563E-01,+0.000000E+00),complex(-1.325825E-01,+0.000000E+00),complex(-3.437500E-01,+0.000000E+00),complex(-5.412659E-02,+0.000000E+00),complex(+3.630922E-01,+0.000000E+00),complex(+4.014135E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+2.838422E-01),complex(+0.000000E+00,+4.279082E-01),complex(-0.000000E+00,+2.679129E-01),complex(+0.000000E+00,-1.104854E-01),complex(+0.000000E+00,-3.437500E-01),complex(+0.000000E+00,-1.848775E-01),complex(+0.000000E+00,+1.848775E-01),complex(+0.000000E+00,+3.437500E-01),complex(+0.000000E+00,+1.104854E-01),complex(+0.000000E+00,-2.679129E-01),complex(-0.000000E+00,-4.279082E-01),complex(+0.000000E+00,-2.838422E-01)],$
            [complex(-1.638764E-01,+0.000000E+00),complex(-3.458741E-01,-0.000000E+00),complex(-4.198447E-01,+0.000000E+00),complex(-2.679129E-01,-0.000000E+00),complex(+5.412659E-02,+0.000000E+00),complex(+3.202172E-01,+0.000000E+00),complex(+3.202172E-01,+0.000000E+00),complex(+5.412659E-02,-0.000000E+00),complex(-2.679129E-01,+0.000000E+00),complex(-4.198447E-01,+0.000000E+00),complex(-3.458741E-01,+0.000000E+00),complex(-1.638764E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-7.328775E-02),complex(+0.000000E+00,-1.988738E-01),complex(+0.000000E+00,-3.458741E-01),complex(+0.000000E+00,-4.279082E-01),complex(+0.000000E+00,-3.630922E-01),complex(+0.000000E+00,-1.432055E-01),complex(+0.000000E+00,+1.432055E-01),complex(+0.000000E+00,+3.630922E-01),complex(+0.000000E+00,+4.279082E-01),complex(+0.000000E+00,+3.458741E-01),complex(+0.000000E+00,+1.988738E-01),complex(+0.000000E+00,+7.328775E-02)],$
            [complex(+2.209709E-02,+0.000000E+00),complex(+7.328775E-02,+0.000000E+00),complex(+1.638764E-01,+0.000000E+00),complex(+2.838422E-01,+0.000000E+00),complex(+4.014135E-01,+0.000000E+00),complex(+4.749589E-01,+0.000000E+00),complex(+4.749589E-01,+0.000000E+00),complex(+4.014135E-01,-0.000000E+00),complex(+2.838422E-01,+0.000000E+00),complex(+1.638764E-01,-0.000000E+00),complex(+7.328775E-02,+0.000000E+00),complex(+2.209709E-02,-0.000000E+00)]$
            ]
          end

      10: begin
            matrix[55:65,55:65]=[$
            [complex(-3.125000E-02,-0.000000E+00),complex(+9.882118E-02,+0.000000E+00),complex(-2.096314E-01,-0.000000E+00),complex(+3.423266E-01,+0.000000E+00),complex(-4.528555E-01,-0.000000E+00),complex(+4.960784E-01,+0.000000E+00),complex(-4.528555E-01,+0.000000E+00),complex(+3.423266E-01,+0.000000E+00),complex(-2.096314E-01,+0.000000E+00),complex(+9.882118E-02,+0.000000E+00),complex(-3.125000E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-9.882118E-02),complex(+0.000000E+00,+2.500000E-01),complex(+0.000000E+00,-3.977476E-01),complex(+0.000000E+00,+4.330127E-01),complex(+0.000000E+00,-2.864110E-01),0,complex(+0.000000E+00,+2.864110E-01),complex(+0.000000E+00,-4.330127E-01),complex(+0.000000E+00,+3.977476E-01),complex(+0.000000E+00,-2.500000E-01),complex(+0.000000E+00,+9.882118E-02)],$
            [complex(+2.096314E-01,+0.000000E+00),complex(-3.977476E-01,+0.000000E+00),complex(+4.062500E-01,+0.000000E+00),complex(-1.530931E-01,+0.000000E+00),complex(-2.025231E-01,-0.000000E+00),complex(+3.697550E-01,+0.000000E+00),complex(-2.025231E-01,+0.000000E+00),complex(-1.530931E-01,+0.000000E+00),complex(+4.062500E-01,-0.000000E+00),complex(-3.977476E-01,+0.000000E+00),complex(+2.096314E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+3.423266E-01),complex(+0.000000E+00,-4.330127E-01),complex(-0.000000E+00,+1.530931E-01),complex(+0.000000E+00,+2.500000E-01),complex(+0.000000E+00,-3.307189E-01),0,complex(+0.000000E+00,+3.307189E-01),complex(+0.000000E+00,-2.500000E-01),complex(-0.000000E+00,-1.530931E-01),complex(+0.000000E+00,+4.330127E-01),complex(-0.000000E+00,-3.423266E-01)],$
            [complex(-4.528555E-01,-0.000000E+00),complex(+2.864110E-01,+0.000000E+00),complex(+2.025231E-01,+0.000000E+00),complex(-3.307189E-01,+0.000000E+00),complex(-6.250000E-02,-0.000000E+00),complex(+3.423266E-01,+0.000000E+00),complex(-6.250000E-02,+0.000000E+00),complex(-3.307189E-01,+0.000000E+00),complex(+2.025231E-01,-0.000000E+00),complex(+2.864110E-01,+0.000000E+00),complex(-4.528555E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.960784E-01),0,complex(-0.000000E+00,+3.697550E-01),0,complex(+0.000000E+00,-3.423266E-01),0,complex(+0.000000E+00,+3.423266E-01),0,complex(-0.000000E+00,-3.697550E-01),0,complex(+0.000000E+00,+4.960784E-01)],$
            [complex(+4.528555E-01,+0.000000E+00),complex(+2.864110E-01,+0.000000E+00),complex(-2.025231E-01,-0.000000E+00),complex(-3.307189E-01,+0.000000E+00),complex(+6.250000E-02,+0.000000E+00),complex(+3.423266E-01,+0.000000E+00),complex(+6.250000E-02,-0.000000E+00),complex(-3.307189E-01,+0.000000E+00),complex(-2.025231E-01,+0.000000E+00),complex(+2.864110E-01,+0.000000E+00),complex(+4.528555E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+3.423266E-01),complex(+0.000000E+00,+4.330127E-01),complex(-0.000000E+00,+1.530931E-01),complex(+0.000000E+00,-2.500000E-01),complex(+0.000000E+00,-3.307189E-01),0,complex(+0.000000E+00,+3.307189E-01),complex(+0.000000E+00,+2.500000E-01),complex(-0.000000E+00,-1.530931E-01),complex(+0.000000E+00,-4.330127E-01),complex(-0.000000E+00,-3.423266E-01)],$
            [complex(-2.096314E-01,-0.000000E+00),complex(-3.977476E-01,+0.000000E+00),complex(-4.062500E-01,-0.000000E+00),complex(-1.530931E-01,+0.000000E+00),complex(+2.025231E-01,+0.000000E+00),complex(+3.697550E-01,+0.000000E+00),complex(+2.025231E-01,-0.000000E+00),complex(-1.530931E-01,+0.000000E+00),complex(-4.062500E-01,+0.000000E+00),complex(-3.977476E-01,+0.000000E+00),complex(-2.096314E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-9.882118E-02),complex(+0.000000E+00,-2.500000E-01),complex(+0.000000E+00,-3.977476E-01),complex(+0.000000E+00,-4.330127E-01),complex(+0.000000E+00,-2.864110E-01),0,complex(+0.000000E+00,+2.864110E-01),complex(+0.000000E+00,+4.330127E-01),complex(+0.000000E+00,+3.977476E-01),complex(+0.000000E+00,+2.500000E-01),complex(+0.000000E+00,+9.882118E-02)],$
            [complex(+3.125000E-02,+0.000000E+00),complex(+9.882118E-02,+0.000000E+00),complex(+2.096314E-01,+0.000000E+00),complex(+3.423266E-01,+0.000000E+00),complex(+4.528555E-01,+0.000000E+00),complex(+4.960784E-01,+0.000000E+00),complex(+4.528555E-01,-0.000000E+00),complex(+3.423266E-01,+0.000000E+00),complex(+2.096314E-01,-0.000000E+00),complex(+9.882118E-02,+0.000000E+00),complex(+3.125000E-02,-0.000000E+00)]$
            ]
          end

       9: begin
            matrix[45:54,45:54]=[$
            [complex(+0.000000E+00,-4.419417E-02),complex(-0.000000E+00,+1.325825E-01),complex(+0.000000E+00,-2.651650E-01),complex(-0.000000E+00,+4.050463E-01),complex(+0.000000E+00,-4.960784E-01),complex(+0.000000E+00,+4.960784E-01),complex(+0.000000E+00,-4.050463E-01),complex(+0.000000E+00,+2.651650E-01),complex(+0.000000E+00,-1.325825E-01),complex(+0.000000E+00,+4.419417E-02)],$
            [complex(+1.325825E-01,+0.000000E+00),complex(-3.093592E-01,+0.000000E+00),complex(+4.419417E-01,+0.000000E+00),complex(-4.050463E-01,+0.000000E+00),complex(+1.653595E-01,+0.000000E+00),complex(+1.653595E-01,+0.000000E+00),complex(-4.050463E-01,+0.000000E+00),complex(+4.419417E-01,+0.000000E+00),complex(-3.093592E-01,+0.000000E+00),complex(+1.325825E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+2.651650E-01),complex(+0.000000E+00,-4.419417E-01),complex(+0.000000E+00,+3.535534E-01),0,complex(+0.000000E+00,-3.307189E-01),complex(+0.000000E+00,+3.307189E-01),0,complex(-0.000000E+00,-3.535534E-01),complex(+0.000000E+00,+4.419417E-01),complex(-0.000000E+00,-2.651650E-01)],$
            [complex(-4.050463E-01,-0.000000E+00),complex(+4.050463E-01,+0.000000E+00),0,complex(-3.535534E-01,+0.000000E+00),complex(+2.165063E-01,+0.000000E+00),complex(+2.165063E-01,+0.000000E+00),complex(-3.535534E-01,+0.000000E+00),0,complex(+4.050463E-01,-0.000000E+00),complex(-4.050463E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.960784E-01),complex(-0.000000E+00,+1.653595E-01),complex(+0.000000E+00,+3.307189E-01),complex(+0.000000E+00,-2.165063E-01),complex(+0.000000E+00,-2.651650E-01),complex(+0.000000E+00,+2.651650E-01),complex(+0.000000E+00,+2.165063E-01),complex(-0.000000E+00,-3.307189E-01),complex(+0.000000E+00,-1.653595E-01),complex(+0.000000E+00,+4.960784E-01)],$
            [complex(+4.960784E-01,+0.000000E+00),complex(+1.653595E-01,+0.000000E+00),complex(-3.307189E-01,-0.000000E+00),complex(-2.165063E-01,+0.000000E+00),complex(+2.651650E-01,+0.000000E+00),complex(+2.651650E-01,+0.000000E+00),complex(-2.165063E-01,+0.000000E+00),complex(-3.307189E-01,+0.000000E+00),complex(+1.653595E-01,-0.000000E+00),complex(+4.960784E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+4.050463E-01),complex(-0.000000E+00,+4.050463E-01),0,complex(+0.000000E+00,-3.535534E-01),complex(+0.000000E+00,-2.165063E-01),complex(+0.000000E+00,+2.165063E-01),complex(+0.000000E+00,+3.535534E-01),0,complex(+0.000000E+00,-4.050463E-01),complex(-0.000000E+00,-4.050463E-01)],$
            [complex(-2.651650E-01,-0.000000E+00),complex(-4.419417E-01,+0.000000E+00),complex(-3.535534E-01,-0.000000E+00),0,complex(+3.307189E-01,+0.000000E+00),complex(+3.307189E-01,+0.000000E+00),0,complex(-3.535534E-01,+0.000000E+00),complex(-4.419417E-01,+0.000000E+00),complex(-2.651650E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.325825E-01),complex(+0.000000E+00,-3.093592E-01),complex(+0.000000E+00,-4.419417E-01),complex(+0.000000E+00,-4.050463E-01),complex(+0.000000E+00,-1.653595E-01),complex(+0.000000E+00,+1.653595E-01),complex(+0.000000E+00,+4.050463E-01),complex(+0.000000E+00,+4.419417E-01),complex(+0.000000E+00,+3.093592E-01),complex(+0.000000E+00,+1.325825E-01)],$
            [complex(+4.419417E-02,+0.000000E+00),complex(+1.325825E-01,+0.000000E+00),complex(+2.651650E-01,+0.000000E+00),complex(+4.050463E-01,+0.000000E+00),complex(+4.960784E-01,+0.000000E+00),complex(+4.960784E-01,+0.000000E+00),complex(+4.050463E-01,-0.000000E+00),complex(+2.651650E-01,+0.000000E+00),complex(+1.325825E-01,-0.000000E+00),complex(+4.419417E-02,+0.000000E+00)]$
            ]
          end

       8: begin
            matrix[36:44,36:44]=[$
            [complex(+6.250000E-02,+0.000000E+00),complex(-1.767767E-01,-0.000000E+00),complex(+3.307189E-01,+0.000000E+00),complex(-4.677072E-01,-0.000000E+00),complex(+5.229125E-01,+0.000000E+00),complex(-4.677072E-01,+0.000000E+00),complex(+3.307189E-01,+0.000000E+00),complex(-1.767767E-01,+0.000000E+00),complex(+6.250000E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,+1.767767E-01),complex(+0.000000E+00,-3.750000E-01),complex(+0.000000E+00,+4.677072E-01),complex(+0.000000E+00,-3.307189E-01),0,complex(+0.000000E+00,+3.307189E-01),complex(+0.000000E+00,-4.677072E-01),complex(+0.000000E+00,+3.750000E-01),complex(+0.000000E+00,-1.767767E-01)],$
            [complex(-3.307189E-01,+0.000000E+00),complex(+4.677072E-01,+0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(-1.767767E-01,-0.000000E+00),complex(+3.952847E-01,+0.000000E+00),complex(-1.767767E-01,+0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(+4.677072E-01,-0.000000E+00),complex(-3.307189E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.677072E-01),complex(-0.000000E+00,+3.307189E-01),complex(+0.000000E+00,+1.767767E-01),complex(+0.000000E+00,-3.750000E-01),0,complex(+0.000000E+00,+3.750000E-01),complex(+0.000000E+00,-1.767767E-01),complex(-0.000000E+00,-3.307189E-01),complex(+0.000000E+00,+4.677072E-01)],$
            [complex(+5.229125E-01,+0.000000E+00),0,complex(-3.952847E-01,+0.000000E+00),0,complex(+3.750000E-01,+0.000000E+00),0,complex(-3.952847E-01,+0.000000E+00),0,complex(+5.229125E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+4.677072E-01),complex(-0.000000E+00,+3.307189E-01),complex(+0.000000E+00,-1.767767E-01),complex(+0.000000E+00,-3.750000E-01),0,complex(+0.000000E+00,+3.750000E-01),complex(+0.000000E+00,+1.767767E-01),complex(-0.000000E+00,-3.307189E-01),complex(+0.000000E+00,-4.677072E-01)],$
            [complex(-3.307189E-01,+0.000000E+00),complex(-4.677072E-01,-0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(+1.767767E-01,+0.000000E+00),complex(+3.952847E-01,+0.000000E+00),complex(+1.767767E-01,-0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(-4.677072E-01,+0.000000E+00),complex(-3.307189E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.767767E-01),complex(+0.000000E+00,-3.750000E-01),complex(+0.000000E+00,-4.677072E-01),complex(+0.000000E+00,-3.307189E-01),0,complex(+0.000000E+00,+3.307189E-01),complex(+0.000000E+00,+4.677072E-01),complex(+0.000000E+00,+3.750000E-01),complex(+0.000000E+00,+1.767767E-01)],$
            [complex(+6.250000E-02,+0.000000E+00),complex(+1.767767E-01,+0.000000E+00),complex(+3.307189E-01,+0.000000E+00),complex(+4.677072E-01,+0.000000E+00),complex(+5.229125E-01,+0.000000E+00),complex(+4.677072E-01,-0.000000E+00),complex(+3.307189E-01,+0.000000E+00),complex(+1.767767E-01,-0.000000E+00),complex(+6.250000E-02,+0.000000E+00)]$
            ]
          end

       7: begin
            matrix[28:35,28:35]=[$
            [complex(-0.000000E+00,+8.838835E-02),complex(+0.000000E+00,-2.338536E-01),complex(-0.000000E+00,+4.050463E-01),complex(+0.000000E+00,-5.229125E-01),complex(+0.000000E+00,+5.229125E-01),complex(+0.000000E+00,-4.050463E-01),complex(+0.000000E+00,+2.338536E-01),complex(+0.000000E+00,-8.838835E-02)],$
            [complex(-2.338536E-01,+0.000000E+00),complex(+4.419417E-01,+0.000000E+00),complex(-4.592793E-01,+0.000000E+00),complex(+1.976424E-01,+0.000000E+00),complex(+1.976424E-01,+0.000000E+00),complex(-4.592793E-01,+0.000000E+00),complex(+4.419417E-01,+0.000000E+00),complex(-2.338536E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-4.050463E-01),complex(+0.000000E+00,+4.592793E-01),complex(+0.000000E+00,-8.838835E-02),complex(+0.000000E+00,-3.423266E-01),complex(+0.000000E+00,+3.423266E-01),complex(+0.000000E+00,+8.838835E-02),complex(-0.000000E+00,-4.592793E-01),complex(+0.000000E+00,+4.050463E-01)],$
            [complex(+5.229125E-01,+0.000000E+00),complex(-1.976424E-01,-0.000000E+00),complex(-3.423266E-01,+0.000000E+00),complex(+2.651650E-01,+0.000000E+00),complex(+2.651650E-01,+0.000000E+00),complex(-3.423266E-01,+0.000000E+00),complex(-1.976424E-01,+0.000000E+00),complex(+5.229125E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+5.229125E-01),complex(+0.000000E+00,+1.976424E-01),complex(+0.000000E+00,-3.423266E-01),complex(+0.000000E+00,-2.651650E-01),complex(+0.000000E+00,+2.651650E-01),complex(+0.000000E+00,+3.423266E-01),complex(-0.000000E+00,-1.976424E-01),complex(+0.000000E+00,-5.229125E-01)],$
            [complex(-4.050463E-01,+0.000000E+00),complex(-4.592793E-01,-0.000000E+00),complex(-8.838835E-02,+0.000000E+00),complex(+3.423266E-01,+0.000000E+00),complex(+3.423266E-01,+0.000000E+00),complex(-8.838835E-02,+0.000000E+00),complex(-4.592793E-01,+0.000000E+00),complex(-4.050463E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.338536E-01),complex(+0.000000E+00,-4.419417E-01),complex(+0.000000E+00,-4.592793E-01),complex(+0.000000E+00,-1.976424E-01),complex(+0.000000E+00,+1.976424E-01),complex(+0.000000E+00,+4.592793E-01),complex(+0.000000E+00,+4.419417E-01),complex(+0.000000E+00,+2.338536E-01)],$
            [complex(+8.838835E-02,+0.000000E+00),complex(+2.338536E-01,+0.000000E+00),complex(+4.050463E-01,+0.000000E+00),complex(+5.229125E-01,+0.000000E+00),complex(+5.229125E-01,+0.000000E+00),complex(+4.050463E-01,-0.000000E+00),complex(+2.338536E-01,+0.000000E+00),complex(+8.838835E-02,-0.000000E+00)]$
            ]
          end

       6: begin
            matrix[21:27,21:27]=[$
            [complex(-1.250000E-01,-0.000000E+00),complex(+3.061862E-01,+0.000000E+00),complex(-4.841229E-01,-0.000000E+00),complex(+5.590170E-01,+0.000000E+00),complex(-4.841229E-01,+0.000000E+00),complex(+3.061862E-01,+0.000000E+00),complex(-1.250000E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.061862E-01),complex(+0.000000E+00,+5.000000E-01),complex(+0.000000E+00,-3.952847E-01),0,complex(+0.000000E+00,+3.952847E-01),complex(+0.000000E+00,-5.000000E-01),complex(+0.000000E+00,+3.061862E-01)],$
            [complex(+4.841229E-01,+0.000000E+00),complex(-3.952847E-01,+0.000000E+00),complex(-1.250000E-01,-0.000000E+00),complex(+4.330127E-01,+0.000000E+00),complex(-1.250000E-01,+0.000000E+00),complex(-3.952847E-01,+0.000000E+00),complex(+4.841229E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+5.590170E-01),0,complex(+0.000000E+00,-4.330127E-01),0,complex(+0.000000E+00,+4.330127E-01),0,complex(-0.000000E+00,-5.590170E-01)],$
            [complex(-4.841229E-01,-0.000000E+00),complex(-3.952847E-01,+0.000000E+00),complex(+1.250000E-01,+0.000000E+00),complex(+4.330127E-01,+0.000000E+00),complex(+1.250000E-01,-0.000000E+00),complex(-3.952847E-01,+0.000000E+00),complex(-4.841229E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.061862E-01),complex(+0.000000E+00,-5.000000E-01),complex(+0.000000E+00,-3.952847E-01),0,complex(+0.000000E+00,+3.952847E-01),complex(+0.000000E+00,+5.000000E-01),complex(+0.000000E+00,+3.061862E-01)],$
            [complex(+1.250000E-01,+0.000000E+00),complex(+3.061862E-01,+0.000000E+00),complex(+4.841229E-01,+0.000000E+00),complex(+5.590170E-01,+0.000000E+00),complex(+4.841229E-01,-0.000000E+00),complex(+3.061862E-01,+0.000000E+00),complex(+1.250000E-01,-0.000000E+00)]$
            ]
          end

       5: begin
            matrix[15:20,15:20]=[$
            [complex(+0.000000E+00,-1.767767E-01),complex(-0.000000E+00,+3.952847E-01),complex(+0.000000E+00,-5.590170E-01),complex(+0.000000E+00,+5.590170E-01),complex(+0.000000E+00,-3.952847E-01),complex(+0.000000E+00,+1.767767E-01)],$
            [complex(+3.952847E-01,+0.000000E+00),complex(-5.303301E-01,+0.000000E+00),complex(+2.500000E-01,+0.000000E+00),complex(+2.500000E-01,+0.000000E+00),complex(-5.303301E-01,+0.000000E+00),complex(+3.952847E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+5.590170E-01),complex(+0.000000E+00,-2.500000E-01),complex(+0.000000E+00,-3.535534E-01),complex(+0.000000E+00,+3.535534E-01),complex(+0.000000E+00,+2.500000E-01),complex(-0.000000E+00,-5.590170E-01)],$
            [complex(-5.590170E-01,-0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(+3.535534E-01,+0.000000E+00),complex(+3.535534E-01,+0.000000E+00),complex(-2.500000E-01,+0.000000E+00),complex(-5.590170E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.952847E-01),complex(+0.000000E+00,-5.303301E-01),complex(+0.000000E+00,-2.500000E-01),complex(+0.000000E+00,+2.500000E-01),complex(+0.000000E+00,+5.303301E-01),complex(+0.000000E+00,+3.952847E-01)],$
            [complex(+1.767767E-01,+0.000000E+00),complex(+3.952847E-01,+0.000000E+00),complex(+5.590170E-01,+0.000000E+00),complex(+5.590170E-01,+0.000000E+00),complex(+3.952847E-01,-0.000000E+00),complex(+1.767767E-01,+0.000000E+00)]$
            ]
          end

       4: begin
            matrix[10:14,10:14]=[$
            [complex(+2.500000E-01,+0.000000E+00),complex(-5.000000E-01,-0.000000E+00),complex(+6.123725E-01,+0.000000E+00),complex(-5.000000E-01,+0.000000E+00),complex(+2.500000E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+5.000000E-01),complex(+0.000000E+00,-5.000000E-01),0,complex(+0.000000E+00,+5.000000E-01),complex(+0.000000E+00,-5.000000E-01)],$
            [complex(-6.123725E-01,+0.000000E+00),0,complex(+5.000000E-01,+0.000000E+00),0,complex(-6.123725E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-5.000000E-01),complex(+0.000000E+00,-5.000000E-01),0,complex(+0.000000E+00,+5.000000E-01),complex(+0.000000E+00,+5.000000E-01)],$
            [complex(+2.500000E-01,+0.000000E+00),complex(+5.000000E-01,+0.000000E+00),complex(+6.123725E-01,+0.000000E+00),complex(+5.000000E-01,-0.000000E+00),complex(+2.500000E-01,+0.000000E+00)]$
            ]
          end

       3: begin
            matrix[6:9,6:9]=[$
            [complex(-0.000000E+00,+3.535534E-01),complex(+0.000000E+00,-6.123724E-01),complex(+0.000000E+00,+6.123724E-01),complex(+0.000000E+00,-3.535534E-01)],$
            [complex(-6.123724E-01,+0.000000E+00),complex(+3.535534E-01,+0.000000E+00),complex(+3.535534E-01,+0.000000E+00),complex(-6.123724E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-6.123724E-01),complex(+0.000000E+00,-3.535534E-01),complex(+0.000000E+00,+3.535534E-01),complex(+0.000000E+00,+6.123724E-01)],$
            [complex(+3.535534E-01,+0.000000E+00),complex(+6.123724E-01,+0.000000E+00),complex(+6.123724E-01,+0.000000E+00),complex(+3.535534E-01,-0.000000E+00)]$
            ]
          end

       2: begin
            matrix[3:5,3:5]=[$
            [complex(-5.000000E-01,-0.000000E+00),complex(+7.071068E-01,+0.000000E+00),complex(-5.000000E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-7.071068E-01),0,complex(+0.000000E+00,+7.071068E-01)],$
            [complex(+5.000000E-01,+0.000000E+00),complex(+7.071068E-01,+0.000000E+00),complex(+5.000000E-01,-0.000000E+00)]$
            ]
          end

       1: begin
            matrix[1:2,1:2]=[$
            [complex(+0.000000E+00,-7.071068E-01),complex(+0.000000E+00,+7.071068E-01)],$
            [complex(+7.071068E-01,+0.000000E+00),complex(+7.071068E-01,+0.000000E+00)]$
            ]
          end

       0: matrix[0]=complex(1.,0.)

    endswitch
  endelse
  
endif else begin
  
  ; Calculate inverse matrix, to transfer from polar shapelets to Cartesians.
  ; First, initialise an empty array
  matrix=complexarr(n_coeffs,n_coeffs)
  
  if n_max gt 30 or keyword_set(calculate) then begin
    ; Calculate matrix. It would be nice to know how to calculate this
    ; by itself, especially since it seems a simpler matrix than the
    ; forward transform. Oh well, at least this minimises the amount of
    ; independent code that has to be managed.
    matrix=la_invert(shapelets_polar_matrix(n_max,calculate=calculate))
  endif else begin
    ; Use look-up table    
    switch n_max of
    
      30: begin
            matrix[465:495,465:495]=[$
            [complex(-3.05436e-05,0.00000),complex(0.00000,0.000167131),complex(0.000636542,0.00000),complex(0.00000,-0.00194459),complex(-0.00505209,0.00000),complex(0.00000,0.0115203),complex(0.0235159,0.00000),complex(0.00000,-0.0435430),complex(-0.0738307,0.00000),complex(0.00000,0.115432),complex(0.167277,0.00000),complex(0.00000,-0.225556),complex(-0.283819,0.00000),complex(0.00000,0.333969),complex(0.368016,0.00000),complex(0.00000,-0.380085),complex(-0.368016,0.00000),complex(0.00000,0.333969),complex(0.283819,0.00000),complex(0.00000,-0.225556),complex(-0.167277,0.00000),complex(0.00000,0.115432),complex(0.0738307,0.00000),complex(0.00000,-0.0435430),complex(-0.0235159,0.00000),complex(0.00000,0.0115203),complex(0.00505201,0.00000),complex(0.00000,-0.00194455),complex(-0.000636485,0.00000),complex(0.00000,0.000167141),complex(3.05293e-05,-0.00000)],$
            [complex(0.000167176,0.00000),complex(0.00000,-0.000854541),complex(-0.00302134,0.00000),complex(0.00000,0.00852039),complex(0.0202921,0.00000),complex(0.00000,-0.0420664),complex(-0.0772811,0.00000),complex(0.00000,0.127197),complex(0.188714,0.00000),complex(0.00000,-0.252899),complex(-0.305405,0.00000),complex(0.00000,0.329446),complex(0.310908,0.00000),complex(0.00000,-0.243896),complex(-0.134380,0.00000),complex(0.00000,-8.59145e-08),complex(-0.134380,0.00000),complex(0.00000,0.243896),complex(0.310908,0.00000),complex(0.00000,-0.329446),complex(-0.305405,0.00000),complex(0.00000,0.252899),complex(0.188714,0.00000),complex(0.00000,-0.127197),complex(-0.0772811,0.00000),complex(0.00000,0.0420664),complex(0.0202921,0.00000),complex(0.00000,-0.00852056),complex(-0.00302138,0.00000),complex(0.00000,0.000854489),complex(0.000167119,-0.00000)],$
            [complex(-0.000636473,0.00000),complex(0.00000,0.00302137),complex(0.00985718,0.00000),complex(0.00000,-0.0254526),complex(-0.0549853,0.00000),complex(0.00000,0.102187),complex(0.165742,0.00000),complex(0.00000,-0.235913),complex(-0.293812,0.00000),complex(0.00000,0.315469),complex(0.280711,0.00000),complex(0.00000,-0.183848),complex(-0.0408242,0.00000),complex(0.00000,-0.112088),complex(-0.229385,0.00000),complex(0.00000,0.273355),complex(0.229385,0.00000),complex(0.00000,-0.112088),complex(0.0408242,0.00000),complex(0.00000,-0.183848),complex(-0.280711,0.00000),complex(0.00000,0.315469),complex(0.293812,0.00000),complex(0.00000,-0.235913),complex(-0.165742,0.00000),complex(0.00000,0.102187),complex(0.0549852,0.00000),complex(0.00000,-0.0254525),complex(-0.00985714,0.00000),complex(0.00000,0.00302139),complex(0.000636469,-0.00000)],$
            [complex(0.00194456,0.00000),complex(0.00000,-0.00852059),complex(-0.0254525,0.00000),complex(0.00000,0.0595700),complex(0.115125,0.00000),complex(0.00000,-0.188034),complex(-0.261295,0.00000),complex(0.00000,0.306149),complex(0.291994,0.00000),complex(0.00000,-0.202900),complex(-0.0525053,0.00000),complex(0.00000,-0.113277),complex(-0.231623,0.00000),complex(0.00000,0.251585),complex(0.161719,0.00000),complex(0.00000,5.57629e-08),complex(0.161719,0.00000),complex(0.00000,-0.251585),complex(-0.231623,0.00000),complex(0.00000,0.113277),complex(-0.0525053,0.00000),complex(0.00000,0.202900),complex(0.291994,0.00000),complex(0.00000,-0.306149),complex(-0.261295,0.00000),complex(0.00000,0.188034),complex(0.115125,0.00000),complex(0.00000,-0.0595703),complex(-0.0254525,0.00000),complex(0.00000,0.00852047),complex(0.00194453,-0.00000)],$
            [complex(-0.00505203,0.00000),complex(0.00000,0.0202921),complex(0.0549852,0.00000),complex(0.00000,-0.115125),complex(-0.195221,0.00000),complex(0.00000,0.271752),complex(0.306405,0.00000),complex(0.00000,-0.263291),complex(-0.134242,0.00000),complex(0.00000,-0.0439291),complex(-0.197041,0.00000),complex(0.00000,0.249339),complex(0.169731,0.00000),complex(0.00000,0.00201736),complex(0.171176,0.00000),complex(0.00000,-0.241077),complex(-0.171176,0.00000),complex(0.00000,0.00201740),complex(-0.169731,0.00000),complex(0.00000,0.249340),complex(0.197041,0.00000),complex(0.00000,-0.0439291),complex(0.134242,0.00000),complex(0.00000,-0.263291),complex(-0.306405,0.00000),complex(0.00000,0.271752),complex(0.195221,0.00000),complex(0.00000,-0.115125),complex(-0.0549852,0.00000),complex(0.00000,0.0202921),complex(0.00505203,-0.00000)],$
            [complex(0.0115205,0.00000),complex(0.00000,-0.0420665),complex(-0.102187,0.00000),complex(0.00000,0.188034),complex(0.271752,0.00000),complex(0.00000,-0.305298),complex(-0.245562,0.00000),complex(0.00000,0.0904309),complex(-0.101309,0.00000),complex(0.00000,0.231170),complex(0.220673,0.00000),complex(0.00000,-0.0717001),complex(0.121798,0.00000),complex(0.00000,-0.230018),complex(-0.177428,0.00000),complex(0.00000,-5.21541e-08),complex(-0.177428,0.00000),complex(0.00000,0.230018),complex(0.121798,0.00000),complex(0.00000,0.0717001),complex(0.220673,0.00000),complex(0.00000,-0.231170),complex(-0.101309,0.00000),complex(0.00000,-0.0904310),complex(-0.245562,0.00000),complex(0.00000,0.305298),complex(0.271752,0.00000),complex(0.00000,-0.188034),complex(-0.102187,0.00000),complex(0.00000,0.0420665),complex(0.0115203,-0.00000)],$
            [complex(-0.0235159,0.00000),complex(0.00000,0.0772810),complex(0.165742,0.00000),complex(0.00000,-0.261295),complex(-0.306405,0.00000),complex(0.00000,0.245562),complex(0.0756530,0.00000),complex(0.00000,0.126973),complex(0.240779,0.00000),complex(0.00000,-0.185604),complex(0.00325625,0.00000),complex(0.00000,-0.185288),complex(-0.217679,0.00000),complex(0.00000,0.0732456),complex(-0.130382,0.00000),complex(0.00000,0.224430),complex(0.130382,0.00000),complex(0.00000,0.0732456),complex(0.217680,0.00000),complex(0.00000,-0.185288),complex(-0.00325620,0.00000),complex(0.00000,-0.185604),complex(-0.240779,0.00000),complex(0.00000,0.126972),complex(-0.0756531,0.00000),complex(0.00000,0.245562),complex(0.306405,0.00000),complex(0.00000,-0.261295),complex(-0.165742,0.00000),complex(0.00000,0.0772811),complex(0.0235159,-0.00000)],$
            [complex(0.0435430,-0.00000),complex(0.00000,-0.127197),complex(-0.235913,0.00000),complex(0.00000,0.306149),complex(0.263291,0.00000),complex(0.00000,-0.0904309),complex(0.126972,0.00000),complex(0.00000,-0.242188),complex(-0.164342,0.00000),complex(0.00000,-0.0465996),complex(-0.211029,0.00000),complex(0.00000,0.182113),complex(-0.0143221,0.00000),complex(0.00000,0.194743),complex(0.187772,0.00000),complex(0.00000,9.68575e-08),complex(0.187772,0.00000),complex(0.00000,-0.194743),complex(-0.0143221,0.00000),complex(0.00000,-0.182113),complex(-0.211029,-0.00000),complex(0.00000,0.0465995),complex(-0.164342,0.00000),complex(0.00000,0.242188),complex(0.126973,0.00000),complex(0.00000,0.0904309),complex(0.263291,0.00000),complex(0.00000,-0.306149),complex(-0.235913,0.00000),complex(0.00000,0.127197),complex(0.0435431,0.00000)],$
            [complex(-0.0738307,0.00000),complex(0.00000,0.188714),complex(0.293812,0.00000),complex(0.00000,-0.291994),complex(-0.134242,0.00000),complex(0.00000,-0.101309),complex(-0.240779,0.00000),complex(0.00000,0.164342),complex(-0.0604552,0.00000),complex(0.00000,0.218575),complex(0.152461,0.00000),complex(0.00000,0.0696448),complex(0.214335,0.00000),complex(0.00000,-0.127415),complex(0.0968991,0.00000),complex(0.00000,-0.214450),complex(-0.0968990,0.00000),complex(0.00000,-0.127415),complex(-0.214335,0.00000),complex(0.00000,0.0696448),complex(-0.152461,0.00000),complex(0.00000,0.218575),complex(0.0604553,0.00000),complex(0.00000,0.164342),complex(0.240779,0.00000),complex(0.00000,-0.101309),complex(0.134242,0.00000),complex(0.00000,-0.291994),complex(-0.293812,0.00000),complex(0.00000,0.188714),complex(0.0738307,-0.00000)],$
            [complex(0.115432,0.00000),complex(0.00000,-0.252899),complex(-0.315469,0.00000),complex(0.00000,0.202900),complex(-0.0439291,0.00000),complex(0.00000,0.231170),complex(0.185604,0.00000),complex(0.00000,0.0465996),complex(0.218575,0.00000),complex(0.00000,-0.141479),complex(0.0950819,0.00000),complex(0.00000,-0.215152),complex(-0.0775863,0.00000),complex(0.00000,-0.151512),complex(-0.194784,0.00000),complex(0.00000,-3.72529e-08),complex(-0.194784,0.00000),complex(0.00000,0.151512),complex(-0.0775864,0.00000),complex(0.00000,0.215152),complex(0.0950820,0.00000),complex(0.00000,0.141480),complex(0.218575,0.00000),complex(0.00000,-0.0465995),complex(0.185604,0.00000),complex(0.00000,-0.231170),complex(-0.0439291,0.00000),complex(0.00000,-0.202900),complex(-0.315469,0.00000),complex(0.00000,0.252899),complex(0.115432,-0.00000)],$
            [complex(-0.167277,-0.00000),complex(0.00000,0.305405),complex(0.280711,0.00000),complex(0.00000,-0.0525052),complex(0.197041,0.00000),complex(0.00000,-0.220673),complex(0.00325623,0.00000),complex(0.00000,-0.211029),complex(-0.152461,0.00000),complex(0.00000,-0.0950819),complex(-0.213654,0.00000),complex(0.00000,0.0511493),complex(-0.175997,0.00000),complex(0.00000,0.165542),complex(-0.0672069,0.00000),complex(0.00000,0.208233),complex(0.0672070,0.00000),complex(0.00000,0.165542),complex(0.175997,0.00000),complex(0.00000,0.0511493),complex(0.213654,0.00000),complex(0.00000,-0.0950819),complex(0.152461,0.00000),complex(0.00000,-0.211029),complex(-0.00325617,0.00000),complex(0.00000,-0.220673),complex(-0.197041,0.00000),complex(0.00000,-0.0525053),complex(-0.280711,0.00000),complex(0.00000,0.305405),complex(0.167277,0.00000)],$
            [complex(0.225556,0.00000),complex(0.00000,-0.329446),complex(-0.183848,0.00000),complex(0.00000,-0.113277),complex(-0.249339,0.00000),complex(0.00000,0.0717002),complex(-0.185288,0.00000),complex(0.00000,0.182113),complex(-0.0696448,0.00000),complex(0.00000,0.215152),complex(0.0511493,0.00000),complex(0.00000,0.182617),complex(0.146997,0.00000),complex(0.00000,0.103385),complex(0.199368,0.00000),complex(0.00000,2.98023e-08),complex(0.199368,0.00000),complex(0.00000,-0.103385),complex(0.146997,0.00000),complex(0.00000,-0.182617),complex(0.0511493,0.00000),complex(0.00000,-0.215152),complex(-0.0696449,0.00000),complex(0.00000,-0.182113),complex(-0.185288,0.00000),complex(0.00000,-0.0717001),complex(-0.249339,0.00000),complex(0.00000,0.113277),complex(-0.183848,0.00000),complex(0.00000,0.329446),complex(0.225556,-0.00000)],$
            [complex(-0.283819,0.00000),complex(0.00000,0.310908),complex(0.0408243,0.00000),complex(0.00000,0.231623),complex(0.169731,0.00000),complex(0.00000,0.121798),complex(0.217680,0.00000),complex(0.00000,0.0143221),complex(0.214335,0.00000),complex(0.00000,-0.0775862),complex(0.175997,0.00000),complex(0.00000,-0.146997),complex(0.114471,0.00000),complex(0.00000,-0.190000),complex(0.0396103,0.00000),complex(0.00000,-0.204547),complex(-0.0396104,0.00000),complex(0.00000,-0.189999),complex(-0.114471,0.00000),complex(0.00000,-0.146997),complex(-0.175997,0.00000),complex(0.00000,-0.0775863),complex(-0.214335,0.00000),complex(0.00000,0.0143221),complex(-0.217680,0.00000),complex(0.00000,0.121798),complex(-0.169731,0.00000),complex(0.00000,0.231623),complex(-0.0408242,0.00000),complex(0.00000,0.310908),complex(0.283819,-0.00000)],$
            [complex(0.333969,0.00000),complex(0.00000,-0.243896),complex(0.112088,0.00000),complex(0.00000,-0.251585),complex(0.00201750,0.00000),complex(0.00000,-0.230018),complex(-0.0732456,0.00000),complex(0.00000,-0.194743),complex(-0.127415,0.00000),complex(0.00000,-0.151512),complex(-0.165542,0.00000),complex(0.00000,-0.103385),complex(-0.190000,0.00000),complex(0.00000,-0.0523681),complex(-0.201974,0.00000),complex(0.00000,-2.98023e-08),complex(-0.201974,0.00000),complex(0.00000,0.0523682),complex(-0.190000,0.00000),complex(0.00000,0.103385),complex(-0.165542,0.00000),complex(0.00000,0.151512),complex(-0.127415,0.00000),complex(0.00000,0.194743),complex(-0.0732455,0.00000),complex(0.00000,0.230018),complex(0.00201757,0.00000),complex(0.00000,0.251585),complex(0.112088,0.00000),complex(0.00000,0.243896),complex(0.333969,-0.00000)],$
            [complex(-0.368016,0.00000),complex(0.00000,0.134380),complex(-0.229385,0.00000),complex(0.00000,0.161719),complex(-0.171176,0.00000),complex(0.00000,0.177428),complex(-0.130382,0.00000),complex(0.00000,0.187772),complex(-0.0968991,0.00000),complex(0.00000,0.194784),complex(-0.0672070,0.00000),complex(0.00000,0.199368),complex(-0.0396105,0.00000),complex(0.00000,0.201974),complex(-0.0130920,0.00000),complex(0.00000,0.202821),complex(0.0130921,0.00000),complex(0.00000,0.201974),complex(0.0396104,0.00000),complex(0.00000,0.199368),complex(0.0672070,0.00000),complex(0.00000,0.194784),complex(0.0968990,0.00000),complex(0.00000,0.187772),complex(0.130382,0.00000),complex(0.00000,0.177428),complex(0.171176,0.00000),complex(0.00000,0.161719),complex(0.229385,0.00000),complex(0.00000,0.134380),complex(0.368015,-0.00000)],$
            [complex(0.380085,0.00000),complex(0.00000,-2.38419e-07),complex(0.273355,0.00000),complex(0.00000,-5.96046e-08),complex(0.241076,0.00000),complex(0.00000,-4.47035e-08),complex(0.224430,0.00000),complex(0.00000,2.79397e-08),complex(0.214450,0.00000),complex(0.00000,-2.98023e-08),complex(0.208233,0.00000),complex(0.00000,-7.45058e-08),complex(0.204547,0.00000),complex(0.00000,-5.96046e-08),complex(0.202821,0.00000),0,complex(0.202821,0.00000),0,complex(0.204547,0.00000),complex(0.00000,5.96046e-08),complex(0.208233,0.00000),complex(0.00000,-5.21541e-08),complex(0.214450,0.00000),complex(0.00000,-4.47035e-08),complex(0.224430,0.00000),complex(0.00000,2.98023e-08),complex(0.241077,0.00000),0,complex(0.273355,0.00000),complex(0.00000,-5.96046e-08),complex(0.380085,-0.00000)],$
            [complex(-0.368016,0.00000),complex(0.00000,-0.134380),complex(-0.229385,0.00000),complex(0.00000,-0.161719),complex(-0.171176,0.00000),complex(0.00000,-0.177428),complex(-0.130382,0.00000),complex(0.00000,-0.187772),complex(-0.0968990,0.00000),complex(0.00000,-0.194784),complex(-0.0672070,0.00000),complex(0.00000,-0.199368),complex(-0.0396104,0.00000),complex(0.00000,-0.201974),complex(-0.0130920,0.00000),complex(0.00000,-0.202821),complex(0.0130921,0.00000),complex(0.00000,-0.201974),complex(0.0396104,0.00000),complex(0.00000,-0.199368),complex(0.0672069,0.00000),complex(0.00000,-0.194784),complex(0.0968989,0.00000),complex(0.00000,-0.187772),complex(0.130382,0.00000),complex(0.00000,-0.177428),complex(0.171176,0.00000),complex(0.00000,-0.161719),complex(0.229385,0.00000),complex(0.00000,-0.134380),complex(0.368016,-0.00000)],$
            [complex(0.333969,0.00000),complex(0.00000,0.243896),complex(0.112088,0.00000),complex(0.00000,0.251585),complex(0.00201739,0.00000),complex(0.00000,0.230018),complex(-0.0732456,0.00000),complex(0.00000,0.194743),complex(-0.127415,0.00000),complex(0.00000,0.151512),complex(-0.165542,0.00000),complex(0.00000,0.103385),complex(-0.189999,0.00000),complex(0.00000,0.0523681),complex(-0.201974,0.00000),complex(0.00000,2.98023e-08),complex(-0.201974,0.00000),complex(0.00000,-0.0523681),complex(-0.189999,0.00000),complex(0.00000,-0.103385),complex(-0.165542,0.00000),complex(0.00000,-0.151512),complex(-0.127415,0.00000),complex(0.00000,-0.194743),complex(-0.0732456,0.00000),complex(0.00000,-0.230018),complex(0.00201738,0.00000),complex(0.00000,-0.251585),complex(0.112088,0.00000),complex(0.00000,-0.243896),complex(0.333969,-0.00000)],$
            [complex(-0.283819,0.00000),complex(0.00000,-0.310908),complex(0.0408243,0.00000),complex(0.00000,-0.231623),complex(0.169731,0.00000),complex(0.00000,-0.121798),complex(0.217680,0.00000),complex(0.00000,-0.0143221),complex(0.214335,0.00000),complex(0.00000,0.0775864),complex(0.175997,0.00000),complex(0.00000,0.146997),complex(0.114471,0.00000),complex(0.00000,0.189999),complex(0.0396104,0.00000),complex(0.00000,0.204547),complex(-0.0396105,0.00000),complex(0.00000,0.190000),complex(-0.114471,0.00000),complex(0.00000,0.146997),complex(-0.175997,0.00000),complex(0.00000,0.0775862),complex(-0.214335,0.00000),complex(0.00000,-0.0143221),complex(-0.217679,0.00000),complex(0.00000,-0.121798),complex(-0.169731,0.00000),complex(0.00000,-0.231623),complex(-0.0408242,0.00000),complex(0.00000,-0.310908),complex(0.283819,-0.00000)],$
            [complex(0.225556,0.00000),complex(0.00000,0.329446),complex(-0.183848,0.00000),complex(0.00000,0.113277),complex(-0.249340,0.00000),complex(0.00000,-0.0717000),complex(-0.185288,0.00000),complex(0.00000,-0.182113),complex(-0.0696448,0.00000),complex(0.00000,-0.215152),complex(0.0511493,0.00000),complex(0.00000,-0.182617),complex(0.146997,0.00000),complex(0.00000,-0.103385),complex(0.199368,0.00000),complex(0.00000,-5.96046e-08),complex(0.199368,0.00000),complex(0.00000,0.103385),complex(0.146997,0.00000),complex(0.00000,0.182617),complex(0.0511493,0.00000),complex(0.00000,0.215152),complex(-0.0696449,0.00000),complex(0.00000,0.182113),complex(-0.185288,0.00000),complex(0.00000,0.0717001),complex(-0.249340,0.00000),complex(0.00000,-0.113277),complex(-0.183848,0.00000),complex(0.00000,-0.329446),complex(0.225556,0.00000)],$
            [complex(-0.167277,0.00000),complex(0.00000,-0.305405),complex(0.280711,0.00000),complex(0.00000,0.0525054),complex(0.197041,0.00000),complex(0.00000,0.220673),complex(0.00325635,0.00000),complex(0.00000,0.211029),complex(-0.152461,0.00000),complex(0.00000,0.0950821),complex(-0.213654,0.00000),complex(0.00000,-0.0511492),complex(-0.175997,0.00000),complex(0.00000,-0.165542),complex(-0.0672070,0.00000),complex(0.00000,-0.208233),complex(0.0672069,0.00000),complex(0.00000,-0.165542),complex(0.175997,0.00000),complex(0.00000,-0.0511493),complex(0.213654,0.00000),complex(0.00000,0.0950819),complex(0.152461,0.00000),complex(0.00000,0.211029),complex(-0.00325610,0.00000),complex(0.00000,0.220673),complex(-0.197041,0.00000),complex(0.00000,0.0525054),complex(-0.280711,0.00000),complex(0.00000,-0.305405),complex(0.167277,-0.00000)],$
            [complex(0.115432,0.00000),complex(0.00000,0.252899),complex(-0.315469,0.00000),complex(0.00000,-0.202900),complex(-0.0439291,0.00000),complex(0.00000,-0.231170),complex(0.185604,0.00000),complex(0.00000,-0.0465996),complex(0.218575,0.00000),complex(0.00000,0.141480),complex(0.0950820,0.00000),complex(0.00000,0.215152),complex(-0.0775863,0.00000),complex(0.00000,0.151512),complex(-0.194784,0.00000),complex(0.00000,2.23517e-08),complex(-0.194784,0.00000),complex(0.00000,-0.151512),complex(-0.0775863,0.00000),complex(0.00000,-0.215152),complex(0.0950819,0.00000),complex(0.00000,-0.141479),complex(0.218575,0.00000),complex(0.00000,0.0465995),complex(0.185604,0.00000),complex(0.00000,0.231170),complex(-0.0439291,0.00000),complex(0.00000,0.202900),complex(-0.315469,0.00000),complex(0.00000,-0.252899),complex(0.115432,-0.00000)],$
            [complex(-0.0738307,0.00000),complex(0.00000,-0.188714),complex(0.293812,0.00000),complex(0.00000,0.291994),complex(-0.134242,0.00000),complex(0.00000,0.101309),complex(-0.240779,0.00000),complex(0.00000,-0.164342),complex(-0.0604552,0.00000),complex(0.00000,-0.218575),complex(0.152461,0.00000),complex(0.00000,-0.0696450),complex(0.214335,0.00000),complex(0.00000,0.127415),complex(0.0968991,0.00000),complex(0.00000,0.214450),complex(-0.0968990,0.00000),complex(0.00000,0.127415),complex(-0.214335,0.00000),complex(0.00000,-0.0696449),complex(-0.152461,0.00000),complex(0.00000,-0.218575),complex(0.0604553,0.00000),complex(0.00000,-0.164342),complex(0.240779,0.00000),complex(0.00000,0.101309),complex(0.134242,0.00000),complex(0.00000,0.291994),complex(-0.293812,0.00000),complex(0.00000,-0.188714),complex(0.0738306,-0.00000)],$
            [complex(0.0435430,0.00000),complex(0.00000,0.127197),complex(-0.235913,0.00000),complex(0.00000,-0.306149),complex(0.263291,0.00000),complex(0.00000,0.0904308),complex(0.126973,0.00000),complex(0.00000,0.242188),complex(-0.164342,0.00000),complex(0.00000,0.0465996),complex(-0.211028,0.00000),complex(0.00000,-0.182113),complex(-0.0143221,0.00000),complex(0.00000,-0.194743),complex(0.187772,0.00000),complex(0.00000,4.09782e-08),complex(0.187772,0.00000),complex(0.00000,0.194743),complex(-0.0143222,0.00000),complex(0.00000,0.182113),complex(-0.211029,0.00000),complex(0.00000,-0.0465996),complex(-0.164342,0.00000),complex(0.00000,-0.242188),complex(0.126973,0.00000),complex(0.00000,-0.0904309),complex(0.263291,0.00000),complex(0.00000,0.306149),complex(-0.235913,0.00000),complex(0.00000,-0.127197),complex(0.0435430,-0.00000)],$
            [complex(-0.0235159,0.00000),complex(0.00000,-0.0772811),complex(0.165742,0.00000),complex(0.00000,0.261295),complex(-0.306405,0.00000),complex(0.00000,-0.245562),complex(0.0756531,0.00000),complex(0.00000,-0.126972),complex(0.240779,0.00000),complex(0.00000,0.185604),complex(0.00325623,0.00000),complex(0.00000,0.185288),complex(-0.217680,0.00000),complex(0.00000,-0.0732456),complex(-0.130382,0.00000),complex(0.00000,-0.224430),complex(0.130382,0.00000),complex(0.00000,-0.0732455),complex(0.217680,0.00000),complex(0.00000,0.185288),complex(-0.00325620,0.00000),complex(0.00000,0.185604),complex(-0.240779,0.00000),complex(0.00000,-0.126973),complex(-0.0756530,0.00000),complex(0.00000,-0.245562),complex(0.306405,0.00000),complex(0.00000,0.261295),complex(-0.165742,0.00000),complex(0.00000,-0.0772811),complex(0.0235159,-0.00000)],$
            [complex(0.0115204,0.00000),complex(0.00000,0.0420665),complex(-0.102187,0.00000),complex(0.00000,-0.188034),complex(0.271752,0.00000),complex(0.00000,0.305298),complex(-0.245562,0.00000),complex(0.00000,-0.0904308),complex(-0.101309,0.00000),complex(0.00000,-0.231170),complex(0.220673,0.00000),complex(0.00000,0.0717000),complex(0.121798,0.00000),complex(0.00000,0.230018),complex(-0.177428,0.00000),complex(0.00000,-1.10827e-07),complex(-0.177428,0.00000),complex(0.00000,-0.230018),complex(0.121798,0.00000),complex(0.00000,-0.0717002),complex(0.220673,0.00000),complex(0.00000,0.231170),complex(-0.101309,0.00000),complex(0.00000,0.0904309),complex(-0.245562,0.00000),complex(0.00000,-0.305298),complex(0.271752,0.00000),complex(0.00000,0.188034),complex(-0.102187,0.00000),complex(0.00000,-0.0420665),complex(0.0115204,-0.00000)],$
            [complex(-0.00505205,0.00000),complex(0.00000,-0.0202921),complex(0.0549852,0.00000),complex(0.00000,0.115125),complex(-0.195221,0.00000),complex(0.00000,-0.271752),complex(0.306405,0.00000),complex(0.00000,0.263292),complex(-0.134242,0.00000),complex(0.00000,0.0439292),complex(-0.197041,0.00000),complex(0.00000,-0.249340),complex(0.169731,0.00000),complex(0.00000,-0.00201735),complex(0.171176,0.00000),complex(0.00000,0.241076),complex(-0.171176,0.00000),complex(0.00000,-0.00201734),complex(-0.169731,0.00000),complex(0.00000,-0.249340),complex(0.197041,0.00000),complex(0.00000,0.0439291),complex(0.134242,0.00000),complex(0.00000,0.263291),complex(-0.306405,0.00000),complex(0.00000,-0.271752),complex(0.195221,0.00000),complex(0.00000,0.115125),complex(-0.0549852,0.00000),complex(0.00000,-0.0202921),complex(0.00505205,-0.00000)],$
            [complex(0.00194448,0.00000),complex(0.00000,0.00852057),complex(-0.0254525,0.00000),complex(0.00000,-0.0595704),complex(0.115125,0.00000),complex(0.00000,0.188034),complex(-0.261295,0.00000),complex(0.00000,-0.306149),complex(0.291994,0.00000),complex(0.00000,0.202900),complex(-0.0525054,0.00000),complex(0.00000,0.113277),complex(-0.231623,0.00000),complex(0.00000,-0.251585),complex(0.161719,0.00000),complex(0.00000,2.40980e-08),complex(0.161719,0.00000),complex(0.00000,0.251585),complex(-0.231623,0.00000),complex(0.00000,-0.113277),complex(-0.0525053,0.00000),complex(0.00000,-0.202900),complex(0.291994,0.00000),complex(0.00000,0.306149),complex(-0.261295,0.00000),complex(0.00000,-0.188034),complex(0.115125,0.00000),complex(0.00000,0.0595703),complex(-0.0254525,0.00000),complex(0.00000,-0.00852046),complex(0.00194459,-0.00000)],$
            [complex(-0.000636563,0.00000),complex(0.00000,-0.00302135),complex(0.00985716,0.00000),complex(0.00000,0.0254524),complex(-0.0549852,0.00000),complex(0.00000,-0.102187),complex(0.165742,0.00000),complex(0.00000,0.235913),complex(-0.293812,0.00000),complex(0.00000,-0.315469),complex(0.280711,0.00000),complex(0.00000,0.183848),complex(-0.0408243,0.00000),complex(0.00000,0.112088),complex(-0.229385,0.00000),complex(0.00000,-0.273355),complex(0.229385,0.00000),complex(0.00000,0.112088),complex(0.0408241,0.00000),complex(0.00000,0.183848),complex(-0.280711,0.00000),complex(0.00000,-0.315469),complex(0.293812,0.00000),complex(0.00000,0.235913),complex(-0.165742,0.00000),complex(0.00000,-0.102187),complex(0.0549852,0.00000),complex(0.00000,0.0254525),complex(-0.00985712,0.00000),complex(0.00000,-0.00302133),complex(0.000636527,-0.00000)],$
            [complex(0.000167154,0.00000),complex(0.00000,0.000854558),complex(-0.00302140,0.00000),complex(0.00000,-0.00852054),complex(0.0202921,0.00000),complex(0.00000,0.0420664),complex(-0.0772810,0.00000),complex(0.00000,-0.127197),complex(0.188714,0.00000),complex(0.00000,0.252899),complex(-0.305405,0.00000),complex(0.00000,-0.329446),complex(0.310908,0.00000),complex(0.00000,0.243896),complex(-0.134380,0.00000),complex(0.00000,-3.30328e-09),complex(-0.134380,0.00000),complex(0.00000,-0.243896),complex(0.310908,0.00000),complex(0.00000,0.329446),complex(-0.305405,0.00000),complex(0.00000,-0.252899),complex(0.188714,0.00000),complex(0.00000,0.127197),complex(-0.0772811,0.00000),complex(0.00000,-0.0420665),complex(0.0202921,0.00000),complex(0.00000,0.00852052),complex(-0.00302137,0.00000),complex(0.00000,-0.000854483),complex(0.000167187,-0.00000)],$
            [complex(-3.04708e-05,0.00000),complex(0.00000,-0.000167209),complex(0.000636484,0.00000),complex(0.00000,0.00194458),complex(-0.00505207,0.00000),complex(0.00000,-0.0115203),complex(0.0235159,0.00000),complex(0.00000,0.0435430),complex(-0.0738307,0.00000),complex(0.00000,-0.115432),complex(0.167277,0.00000),complex(0.00000,0.225556),complex(-0.283819,0.00000),complex(0.00000,-0.333969),complex(0.368015,0.00000),complex(0.00000,0.380085),complex(-0.368016,0.00000),complex(0.00000,-0.333969),complex(0.283819,0.00000),complex(0.00000,0.225556),complex(-0.167277,0.00000),complex(0.00000,-0.115432),complex(0.0738306,0.00000),complex(0.00000,0.0435430),complex(-0.0235159,0.00000),complex(0.00000,-0.0115204),complex(0.00505205,0.00000),complex(0.00000,0.00194455),complex(-0.000636513,0.00000),complex(0.00000,-0.000167163),complex(3.05183e-05,-0.00000)]$
            ]
          end

      29: begin
            matrix[435:464,435:464]=[$
            [complex(0.00000,4.32038e-05),complex(0.000232352,0.00000),complex(0.00000,-0.000869593),complex(-0.00260888,0.00000),complex(0.00000,0.00665136),complex(0.0148726,0.00000),complex(0.00000,-0.0297455),complex(-0.0539183,0.00000),complex(0.00000,0.0894135),complex(0.136581,0.00000),complex(0.00000,-0.193155),complex(-0.253855,0.00000),complex(0.00000,0.310908),complex(0.355537,0.00000),complex(0.00000,-0.380085),complex(-0.380085,0.00000),complex(0.00000,0.355537),complex(0.310908,0.00000),complex(0.00000,-0.253855),complex(-0.193155,0.00000),complex(0.00000,0.136581),complex(0.0894134,0.00000),complex(0.00000,-0.0539183),complex(-0.0297455,0.00000),complex(0.00000,0.0148727),complex(0.00665132,0.00000),complex(0.00000,-0.00260885),complex(-0.000869678,0.00000),complex(0.00000,0.000232422),complex(4.31273e-05,-0.00000)],$
            [complex(0.00000,-0.000232324),complex(-0.00116521,0.00000),complex(0.00000,0.00403714),complex(0.0111424,0.00000),complex(0.00000,-0.0259373),complex(-0.0524741,0.00000),complex(0.00000,0.0939011),complex(0.150186,0.00000),complex(0.00000,-0.215847),complex(-0.278987,0.00000),complex(0.00000,0.322812),complex(0.329978,0.00000),complex(0.00000,-0.288671),complex(-0.198065,0.00000),complex(0.00000,0.0705800),complex(-0.0705799,0.00000),complex(0.00000,0.198065),complex(0.288671,0.00000),complex(0.00000,-0.329978),complex(-0.322812,0.00000),complex(0.00000,0.278987),complex(0.215847,0.00000),complex(0.00000,-0.150186),complex(-0.0939011,0.00000),complex(0.00000,0.0524741),complex(0.0259374,0.00000),complex(0.00000,-0.0111424),complex(-0.00403707,0.00000),complex(0.00000,0.00116525),complex(0.000232327,-0.00000)],$
            [complex(0.00000,0.000869673),complex(0.00403706,0.00000),complex(0.00000,-0.0128612),complex(-0.0323687,0.00000),complex(0.00000,0.0680002),complex(0.122528,0.00000),complex(0.00000,-0.191912),complex(-0.262240,0.00000),complex(0.00000,0.310626),complex(0.311807,0.00000),complex(0.00000,-0.249239),complex(-0.125986,0.00000),complex(0.00000,-0.0308603),complex(-0.176450,0.00000),complex(0.00000,0.264086),complex(0.264086,0.00000),complex(0.00000,-0.176450),complex(-0.0308602,0.00000),complex(0.00000,-0.125986),complex(-0.249239,0.00000),complex(0.00000,0.311807),complex(0.310626,0.00000),complex(0.00000,-0.262240),complex(-0.191912,0.00000),complex(0.00000,0.122528),complex(0.0680002,0.00000),complex(0.00000,-0.0323688),complex(-0.0128612,0.00000),complex(0.00000,0.00403715),complex(0.000869621,-0.00000)],$
            [complex(0.00000,-0.00260887),complex(-0.0111425,0.00000),complex(0.00000,0.0323688),complex(0.0734556,0.00000),complex(0.00000,-0.137101),complex(-0.215040,0.00000),complex(0.00000,0.284423),complex(0.312191,0.00000),complex(0.00000,-0.269209),complex(-0.149125,0.00000),complex(0.00000,-0.0191722),complex(-0.176381,0.00000),complex(0.00000,0.257168),complex(0.223503,0.00000),complex(0.00000,-0.0880287),complex(0.0880287,0.00000),complex(0.00000,-0.223503),complex(-0.257168,0.00000),complex(0.00000,0.176381),complex(0.0191722,0.00000),complex(0.00000,0.149125),complex(0.269209,0.00000),complex(0.00000,-0.312191),complex(-0.284423,0.00000),complex(0.00000,0.215039),complex(0.137101,0.00000),complex(0.00000,-0.0734555),complex(-0.0323688,0.00000),complex(0.00000,0.0111424),complex(0.00260884,-0.00000)],$
            [complex(0.00000,0.00665129),complex(0.0259373,0.00000),complex(0.00000,-0.0680001),complex(-0.137101,0.00000),complex(0.00000,0.222309),complex(0.292507,0.00000),complex(0.00000,-0.304763),complex(-0.227759,0.00000),complex(0.00000,0.0690412),complex(-0.114324,0.00000),complex(0.00000,0.236879),complex(0.232255,0.00000),complex(0.00000,-0.0988523),complex(0.0899722,0.00000),complex(0.00000,-0.224430),complex(-0.224430,0.00000),complex(0.00000,0.0899722),complex(-0.0988522,0.00000),complex(0.00000,0.232255),complex(0.236879,0.00000),complex(0.00000,-0.114325),complex(0.0690411,0.00000),complex(0.00000,-0.227759),complex(-0.304763,0.00000),complex(0.00000,0.292507),complex(0.222309,0.00000),complex(0.00000,-0.137101),complex(-0.0680002,0.00000),complex(0.00000,0.0259374),complex(0.00665135,-0.00000)],$
            [complex(0.00000,-0.0148727),complex(-0.0524741,0.00000),complex(0.00000,0.122528),complex(0.215040,0.00000),complex(0.00000,-0.292507),complex(-0.300943,0.00000),complex(0.00000,0.203966),complex(0.0208095,0.00000),complex(0.00000,0.165278),complex(0.248503,0.00000),complex(0.00000,-0.173196),complex(0.0154697,0.00000),complex(0.00000,-0.190365),complex(-0.228008,0.00000),complex(0.00000,0.100368),complex(-0.100368,0.00000),complex(0.00000,0.228008),complex(0.190366,0.00000),complex(0.00000,-0.0154696),complex(0.173196,0.00000),complex(0.00000,-0.248503),complex(-0.165278,0.00000),complex(0.00000,-0.0208095),complex(-0.203967,0.00000),complex(0.00000,0.300943),complex(0.292507,0.00000),complex(0.00000,-0.215039),complex(-0.122528,0.00000),complex(0.00000,0.0524741),complex(0.0148727,-0.00000)],$
            [complex(0.00000,0.0297455),complex(0.0939013,0.00000),complex(0.00000,-0.191912),complex(-0.284423,0.00000),complex(0.00000,0.304763),complex(0.203966,0.00000),complex(0.00000,-0.00500646),complex(0.186190,0.00000),complex(0.00000,-0.243377),complex(-0.121279,0.00000),complex(0.00000,-0.0908017),complex(-0.225414,0.00000),complex(0.00000,0.171419),complex(-0.0268247,0.00000),complex(0.00000,0.200736),complex(0.200736,0.00000),complex(0.00000,-0.0268246),complex(0.171419,0.00000),complex(0.00000,-0.225414),complex(-0.0908018,0.00000),complex(0.00000,-0.121279),complex(-0.243377,0.00000),complex(0.00000,0.186190),complex(-0.00500639,0.00000),complex(0.00000,0.203966),complex(0.304763,0.00000),complex(0.00000,-0.284423),complex(-0.191912,0.00000),complex(0.00000,0.0939011),complex(0.0297455,-0.00000)],$
            [complex(0.00000,-0.0539183),complex(-0.150186,0.00000),complex(0.00000,0.262240),complex(0.312191,0.00000),complex(0.00000,-0.227759),complex(-0.0208095,0.00000),complex(0.00000,-0.186190),complex(-0.239788,0.00000),complex(0.00000,0.0930412),complex(-0.129878,0.00000),complex(0.00000,0.228203),complex(0.109726,0.00000),complex(0.00000,0.112486),complex(0.221977,0.00000),complex(0.00000,-0.110742),complex(0.110742,0.00000),complex(0.00000,-0.221977),complex(-0.112486,0.00000),complex(0.00000,-0.109726),complex(-0.228203,0.00000),complex(0.00000,0.129878),complex(-0.0930412,0.00000),complex(0.00000,0.239788),complex(0.186190,0.00000),complex(0.00000,0.0208095),complex(0.227759,0.00000),complex(0.00000,-0.312190),complex(-0.262240,0.00000),complex(0.00000,0.150186),complex(0.0539183,-0.00000)],$
            [complex(0.00000,0.0894133),complex(0.215847,0.00000),complex(0.00000,-0.310626),complex(-0.269209,-0.00000),complex(0.00000,0.0690413),complex(-0.165278,-0.00000),complex(0.00000,0.243377),complex(0.0930410,0.00000),complex(0.00000,0.141603),complex(0.223685,0.00000),complex(0.00000,-0.0679667),complex(0.157698,0.00000),complex(0.00000,-0.206347),complex(-0.0245405,0.00000),complex(0.00000,-0.183644),complex(-0.183644,0.00000),complex(0.00000,-0.0245405),complex(-0.206347,0.00000),complex(0.00000,0.157699),complex(-0.0679666,0.00000),complex(0.00000,0.223685),complex(0.141603,0.00000),complex(0.00000,0.0930410),complex(0.243377,0.00000),complex(0.00000,-0.165278),complex(0.0690411,0.00000),complex(0.00000,-0.269209),complex(-0.310626,0.00000),complex(0.00000,0.215847),complex(0.0894134,0.00000)],$
            [complex(0.00000,-0.136581),complex(-0.278987,0.00000),complex(0.00000,0.311807),complex(0.149125,0.00000),complex(0.00000,0.114325),complex(0.248503,0.00000),complex(0.00000,-0.121279),complex(0.129878,0.00000),complex(0.00000,-0.223685),complex(-0.0536458,0.00000),complex(0.00000,-0.175720),complex(-0.186181,0.00000),complex(0.00000,-0.0335012),complex(-0.208852,0.00000),complex(0.00000,0.120223),complex(-0.120223,0.00000),complex(0.00000,0.208852),complex(0.0335012,0.00000),complex(0.00000,0.186181),complex(0.175720,0.00000),complex(0.00000,0.0536458),complex(0.223685,0.00000),complex(0.00000,-0.129878),complex(0.121279,0.00000),complex(0.00000,-0.248503),complex(-0.114325,0.00000),complex(0.00000,-0.149125),complex(-0.311807,0.00000),complex(0.00000,0.278987),complex(0.136581,-0.00000)],$
            [complex(0.00000,0.193155),complex(0.322812,0.00000),complex(0.00000,-0.249239),complex(0.0191723,0.00000),complex(0.00000,-0.236879),complex(-0.173196,0.00000),complex(0.00000,-0.0908019),complex(-0.228203,0.00000),complex(0.00000,0.0679667),complex(-0.175720,0.00000),complex(0.00000,0.177899),complex(-0.0611455,0.00000),complex(0.00000,0.212436),complex(0.0681603,0.00000),complex(0.00000,0.170022),complex(0.170022,0.00000),complex(0.00000,0.0681603),complex(0.212436,0.00000),complex(0.00000,-0.0611455),complex(0.177899,0.00000),complex(0.00000,-0.175720),complex(0.0679666,0.00000),complex(0.00000,-0.228203),complex(-0.0908018,0.00000),complex(0.00000,-0.173196),complex(-0.236879,0.00000),complex(0.00000,0.0191723),complex(-0.249239,0.00000),complex(0.00000,0.322812),complex(0.193155,-0.00000)],$
            [complex(0.00000,-0.253855),complex(-0.329978,0.00000),complex(0.00000,0.125986),complex(-0.176381,0.00000),complex(0.00000,0.232255),complex(-0.0154696,0.00000),complex(0.00000,0.225414),complex(0.109726,0.00000),complex(0.00000,0.157699),complex(0.186181,0.00000),complex(0.00000,0.0611454),complex(0.211735,0.00000),complex(0.00000,-0.0407007),complex(0.190161,0.00000),complex(0.00000,-0.129367),complex(0.129367,0.00000),complex(0.00000,-0.190161),complex(0.0407006,0.00000),complex(0.00000,-0.211735),complex(-0.0611454,0.00000),complex(0.00000,-0.186181),complex(-0.157698,0.00000),complex(0.00000,-0.109726),complex(-0.225414,0.00000),complex(0.00000,0.0154696),complex(-0.232255,0.00000),complex(0.00000,0.176381),complex(-0.125986,0.00000),complex(0.00000,0.329978),complex(0.253855,-0.00000)],$
            [complex(0.00000,0.310908),complex(0.288671,0.00000),complex(0.00000,0.0308602),complex(0.257169,0.00000),complex(0.00000,-0.0988523),complex(0.190366,0.00000),complex(0.00000,-0.171419),complex(0.112486,0.00000),complex(0.00000,-0.206346),complex(0.0335013,0.00000),complex(0.00000,-0.212436),complex(-0.0407005,0.00000),complex(0.00000,-0.195119),complex(-0.105863,0.00000),complex(0.00000,-0.158442),complex(-0.158442,0.00000),complex(0.00000,-0.105863),complex(-0.195119,0.00000),complex(0.00000,-0.0407006),complex(-0.212436,0.00000),complex(0.00000,0.0335012),complex(-0.206347,0.00000),complex(0.00000,0.112486),complex(-0.171419,0.00000),complex(0.00000,0.190366),complex(-0.0988523,0.00000),complex(0.00000,0.257168),complex(0.0308602,0.00000),complex(0.00000,0.288671),complex(0.310908,-0.00000)],$
            [complex(0.00000,-0.355537),complex(-0.198065,0.00000),complex(0.00000,-0.176450),complex(-0.223503,0.00000),complex(0.00000,-0.0899722),complex(-0.228008,0.00000),complex(0.00000,-0.0268245),complex(-0.221977,0.00000),complex(0.00000,0.0245405),complex(-0.208852,0.00000),complex(0.00000,0.0681602),complex(-0.190161,0.00000),complex(0.00000,0.105863),complex(-0.166635,0.00000),complex(0.00000,0.138553),complex(-0.138553,0.00000),complex(0.00000,0.166635),complex(-0.105863,0.00000),complex(0.00000,0.190161),complex(-0.0681602,0.00000),complex(0.00000,0.208852),complex(-0.0245405,0.00000),complex(0.00000,0.221977),complex(0.0268245,0.00000),complex(0.00000,0.228008),complex(0.0899723,0.00000),complex(0.00000,0.223503),complex(0.176450,0.00000),complex(0.00000,0.198065),complex(0.355537,-0.00000)],$
            [complex(0.00000,0.380085),complex(0.0705798,0.00000),complex(0.00000,0.264086),complex(0.0880288,0.00000),complex(0.00000,0.224430),complex(0.100368,0.00000),complex(0.00000,0.200736),complex(0.110742,0.00000),complex(0.00000,0.183644),complex(0.120223,0.00000),complex(0.00000,0.170021),complex(0.129367,0.00000),complex(0.00000,0.158442),complex(0.138553,0.00000),complex(0.00000,0.148120),complex(0.148120,0.00000),complex(0.00000,0.138553),complex(0.158442,0.00000),complex(0.00000,0.129367),complex(0.170022,0.00000),complex(0.00000,0.120223),complex(0.183644,0.00000),complex(0.00000,0.110742),complex(0.200736,0.00000),complex(0.00000,0.100368),complex(0.224430,0.00000),complex(0.00000,0.0880288),complex(0.264086,0.00000),complex(0.00000,0.0705800),complex(0.380085,-0.00000)],$
            [complex(0.00000,-0.380085),complex(0.0705799,0.00000),complex(0.00000,-0.264086),complex(0.0880287,0.00000),complex(0.00000,-0.224430),complex(0.100368,0.00000),complex(0.00000,-0.200736),complex(0.110742,0.00000),complex(0.00000,-0.183644),complex(0.120223,0.00000),complex(0.00000,-0.170022),complex(0.129367,0.00000),complex(0.00000,-0.158442),complex(0.138553,0.00000),complex(0.00000,-0.148120),complex(0.148120,0.00000),complex(0.00000,-0.138553),complex(0.158442,0.00000),complex(0.00000,-0.129367),complex(0.170022,0.00000),complex(0.00000,-0.120223),complex(0.183644,0.00000),complex(0.00000,-0.110742),complex(0.200736,0.00000),complex(0.00000,-0.100368),complex(0.224430,0.00000),complex(0.00000,-0.0880287),complex(0.264086,0.00000),complex(0.00000,-0.0705800),complex(0.380085,-0.00000)],$
            [complex(0.00000,0.355537),complex(-0.198065,0.00000),complex(0.00000,0.176450),complex(-0.223503,0.00000),complex(0.00000,0.0899723),complex(-0.228008,0.00000),complex(0.00000,0.0268243),complex(-0.221977,-0.00000),complex(0.00000,-0.0245405),complex(-0.208852,0.00000),complex(0.00000,-0.0681603),complex(-0.190161,0.00000),complex(0.00000,-0.105863),complex(-0.166635,0.00000),complex(0.00000,-0.138553),complex(-0.138553,0.00000),complex(0.00000,-0.166635),complex(-0.105863,0.00000),complex(0.00000,-0.190161),complex(-0.0681603,0.00000),complex(0.00000,-0.208852),complex(-0.0245405,0.00000),complex(0.00000,-0.221977),complex(0.0268245,0.00000),complex(0.00000,-0.228008),complex(0.0899721,0.00000),complex(0.00000,-0.223503),complex(0.176450,0.00000),complex(0.00000,-0.198065),complex(0.355537,-0.00000)],$
            [complex(0.00000,-0.310908),complex(0.288671,0.00000),complex(0.00000,-0.0308602),complex(0.257168,0.00000),complex(0.00000,0.0988522),complex(0.190365,0.00000),complex(0.00000,0.171419),complex(0.112486,0.00000),complex(0.00000,0.206347),complex(0.0335012,0.00000),complex(0.00000,0.212436),complex(-0.0407006,0.00000),complex(0.00000,0.195119),complex(-0.105863,0.00000),complex(0.00000,0.158442),complex(-0.158442,0.00000),complex(0.00000,0.105863),complex(-0.195119,0.00000),complex(0.00000,0.0407007),complex(-0.212436,0.00000),complex(0.00000,-0.0335013),complex(-0.206347,0.00000),complex(0.00000,-0.112486),complex(-0.171419,0.00000),complex(0.00000,-0.190366),complex(-0.0988523,0.00000),complex(0.00000,-0.257168),complex(0.0308603,0.00000),complex(0.00000,-0.288671),complex(0.310908,0.00000)],$
            [complex(0.00000,0.253855),complex(-0.329978,0.00000),complex(0.00000,-0.125986),complex(-0.176381,0.00000),complex(0.00000,-0.232255),complex(-0.0154697,0.00000),complex(0.00000,-0.225414),complex(0.109726,0.00000),complex(0.00000,-0.157699),complex(0.186181,0.00000),complex(0.00000,-0.0611454),complex(0.211735,0.00000),complex(0.00000,0.0407007),complex(0.190161,0.00000),complex(0.00000,0.129367),complex(0.129367,0.00000),complex(0.00000,0.190161),complex(0.0407006,0.00000),complex(0.00000,0.211735),complex(-0.0611455,0.00000),complex(0.00000,0.186181),complex(-0.157698,0.00000),complex(0.00000,0.109726),complex(-0.225414,0.00000),complex(0.00000,-0.0154696),complex(-0.232255,0.00000),complex(0.00000,-0.176381),complex(-0.125986,0.00000),complex(0.00000,-0.329978),complex(0.253855,-0.00000)],$
            [complex(0.00000,-0.193155),complex(0.322812,0.00000),complex(0.00000,0.249239),complex(0.0191722,0.00000),complex(0.00000,0.236879),complex(-0.173196,-0.00000),complex(0.00000,0.0908019),complex(-0.228203,0.00000),complex(0.00000,-0.0679668),complex(-0.175720,0.00000),complex(0.00000,-0.177899),complex(-0.0611455,0.00000),complex(0.00000,-0.212436),complex(0.0681602,0.00000),complex(0.00000,-0.170022),complex(0.170022,0.00000),complex(0.00000,-0.0681602),complex(0.212436,0.00000),complex(0.00000,0.0611455),complex(0.177899,0.00000),complex(0.00000,0.175720),complex(0.0679667,0.00000),complex(0.00000,0.228203),complex(-0.0908018,0.00000),complex(0.00000,0.173196),complex(-0.236879,0.00000),complex(0.00000,-0.0191723),complex(-0.249239,0.00000),complex(0.00000,-0.322812),complex(0.193155,0.00000)],$
            [complex(0.00000,0.136581),complex(-0.278987,0.00000),complex(0.00000,-0.311807),complex(0.149125,0.00000),complex(0.00000,-0.114325),complex(0.248503,0.00000),complex(0.00000,0.121279),complex(0.129878,0.00000),complex(0.00000,0.223685),complex(-0.0536458,0.00000),complex(0.00000,0.175720),complex(-0.186181,0.00000),complex(0.00000,0.0335012),complex(-0.208852,0.00000),complex(0.00000,-0.120223),complex(-0.120223,0.00000),complex(0.00000,-0.208852),complex(0.0335013,0.00000),complex(0.00000,-0.186181),complex(0.175720,0.00000),complex(0.00000,-0.0536458),complex(0.223685,0.00000),complex(0.00000,0.129878),complex(0.121279,0.00000),complex(0.00000,0.248503),complex(-0.114324,0.00000),complex(0.00000,0.149125),complex(-0.311807,0.00000),complex(0.00000,-0.278987),complex(0.136581,-0.00000)],$
            [complex(0.00000,-0.0894134),complex(0.215847,0.00000),complex(0.00000,0.310626),complex(-0.269209,0.00000),complex(0.00000,-0.0690412),complex(-0.165278,0.00000),complex(0.00000,-0.243377),complex(0.0930411,0.00000),complex(0.00000,-0.141602),complex(0.223685,0.00000),complex(0.00000,0.0679666),complex(0.157699,0.00000),complex(0.00000,0.206346),complex(-0.0245405,0.00000),complex(0.00000,0.183644),complex(-0.183644,0.00000),complex(0.00000,0.0245405),complex(-0.206346,0.00000),complex(0.00000,-0.157698),complex(-0.0679667,0.00000),complex(0.00000,-0.223685),complex(0.141603,0.00000),complex(0.00000,-0.0930410),complex(0.243377,0.00000),complex(0.00000,0.165278),complex(0.0690412,0.00000),complex(0.00000,0.269209),complex(-0.310626,0.00000),complex(0.00000,-0.215847),complex(0.0894133,-0.00000)],$
            [complex(0.00000,0.0539184),complex(-0.150186,0.00000),complex(0.00000,-0.262240),complex(0.312191,0.00000),complex(0.00000,0.227759),complex(-0.0208096,0.00000),complex(0.00000,0.186190),complex(-0.239788,0.00000),complex(0.00000,-0.0930410),complex(-0.129878,0.00000),complex(0.00000,-0.228203),complex(0.109726,0.00000),complex(0.00000,-0.112486),complex(0.221977,0.00000),complex(0.00000,0.110742),complex(0.110742,0.00000),complex(0.00000,0.221977),complex(-0.112486,0.00000),complex(0.00000,0.109726),complex(-0.228203,0.00000),complex(0.00000,-0.129878),complex(-0.0930411,0.00000),complex(0.00000,-0.239788),complex(0.186190,0.00000),complex(0.00000,-0.0208096),complex(0.227759,0.00000),complex(0.00000,0.312191),complex(-0.262240,0.00000),complex(0.00000,-0.150186),complex(0.0539182,-0.00000)],$
            [complex(0.00000,-0.0297455),complex(0.0939012,0.00000),complex(0.00000,0.191912),complex(-0.284423,0.00000),complex(0.00000,-0.304763),complex(0.203966,0.00000),complex(0.00000,0.00500641),complex(0.186190,0.00000),complex(0.00000,0.243377),complex(-0.121279,0.00000),complex(0.00000,0.0908018),complex(-0.225414,0.00000),complex(0.00000,-0.171419),complex(-0.0268244,0.00000),complex(0.00000,-0.200736),complex(0.200736,0.00000),complex(0.00000,0.0268244),complex(0.171419,0.00000),complex(0.00000,0.225414),complex(-0.0908018,0.00000),complex(0.00000,0.121279),complex(-0.243377,0.00000),complex(0.00000,-0.186190),complex(-0.00500641,0.00000),complex(0.00000,-0.203966),complex(0.304763,0.00000),complex(0.00000,0.284423),complex(-0.191912,0.00000),complex(0.00000,-0.0939012),complex(0.0297455,-0.00000)],$
            [complex(0.00000,0.0148726),complex(-0.0524740,0.00000),complex(0.00000,-0.122528),complex(0.215040,0.00000),complex(0.00000,0.292507),complex(-0.300943,0.00000),complex(0.00000,-0.203966),complex(0.0208095,0.00000),complex(0.00000,-0.165278),complex(0.248503,0.00000),complex(0.00000,0.173196),complex(0.0154697,0.00000),complex(0.00000,0.190365),complex(-0.228008,0.00000),complex(0.00000,-0.100368),complex(-0.100368,0.00000),complex(0.00000,-0.228008),complex(0.190365,0.00000),complex(0.00000,0.0154696),complex(0.173196,0.00000),complex(0.00000,0.248503),complex(-0.165278,0.00000),complex(0.00000,0.0208094),complex(-0.203967,0.00000),complex(0.00000,-0.300943),complex(0.292507,0.00000),complex(0.00000,0.215039),complex(-0.122528,0.00000),complex(0.00000,-0.0524741),complex(0.0148727,-0.00000)],$
            [complex(0.00000,-0.00665136),complex(0.0259374,0.00000),complex(0.00000,0.0680000),complex(-0.137101,0.00000),complex(0.00000,-0.222309),complex(0.292507,0.00000),complex(0.00000,0.304763),complex(-0.227759,0.00000),complex(0.00000,-0.0690412),complex(-0.114325,0.00000),complex(0.00000,-0.236879),complex(0.232255,0.00000),complex(0.00000,0.0988523),complex(0.0899723,0.00000),complex(0.00000,0.224430),complex(-0.224430,0.00000),complex(0.00000,-0.0899722),complex(-0.0988523,0.00000),complex(0.00000,-0.232255),complex(0.236879,0.00000),complex(0.00000,0.114325),complex(0.0690413,0.00000),complex(0.00000,0.227759),complex(-0.304763,0.00000),complex(0.00000,-0.292507),complex(0.222309,0.00000),complex(0.00000,0.137101),complex(-0.0680002,0.00000),complex(0.00000,-0.0259373),complex(0.00665133,-0.00000)],$
            [complex(0.00000,0.00260881),complex(-0.0111423,0.00000),complex(0.00000,-0.0323688),complex(0.0734555,0.00000),complex(0.00000,0.137101),complex(-0.215039,0.00000),complex(0.00000,-0.284423),complex(0.312191,0.00000),complex(0.00000,0.269209),complex(-0.149125,0.00000),complex(0.00000,0.0191722),complex(-0.176381,0.00000),complex(0.00000,-0.257168),complex(0.223503,0.00000),complex(0.00000,0.0880287),complex(0.0880287,0.00000),complex(0.00000,0.223503),complex(-0.257168,0.00000),complex(0.00000,-0.176381),complex(0.0191722,0.00000),complex(0.00000,-0.149125),complex(0.269209,0.00000),complex(0.00000,0.312191),complex(-0.284423,0.00000),complex(0.00000,-0.215039),complex(0.137101,0.00000),complex(0.00000,0.0734555),complex(-0.0323688,0.00000),complex(0.00000,-0.0111424),complex(0.00260888,-0.00000)],$
            [complex(0.00000,-0.000869667),complex(0.00403707,0.00000),complex(0.00000,0.0128612),complex(-0.0323688,0.00000),complex(0.00000,-0.0680002),complex(0.122528,0.00000),complex(0.00000,0.191912),complex(-0.262240,0.00000),complex(0.00000,-0.310626),complex(0.311807,0.00000),complex(0.00000,0.249239),complex(-0.125986,0.00000),complex(0.00000,0.0308601),complex(-0.176450,0.00000),complex(0.00000,-0.264086),complex(0.264086,0.00000),complex(0.00000,0.176450),complex(-0.0308603,0.00000),complex(0.00000,0.125986),complex(-0.249239,0.00000),complex(0.00000,-0.311807),complex(0.310626,0.00000),complex(0.00000,0.262240),complex(-0.191912,0.00000),complex(0.00000,-0.122528),complex(0.0680002,0.00000),complex(0.00000,0.0323687),complex(-0.0128612,0.00000),complex(0.00000,-0.00403706),complex(0.000869701,-0.00000)],$
            [complex(0.00000,0.000232296),complex(-0.00116521,0.00000),complex(0.00000,-0.00403709),complex(0.0111423,0.00000),complex(0.00000,0.0259373),complex(-0.0524741,0.00000),complex(0.00000,-0.0939012),complex(0.150185,0.00000),complex(0.00000,0.215847),complex(-0.278987,0.00000),complex(0.00000,-0.322812),complex(0.329978,0.00000),complex(0.00000,0.288671),complex(-0.198065,0.00000),complex(0.00000,-0.0705801),complex(-0.0705800,0.00000),complex(0.00000,-0.198065),complex(0.288671,0.00000),complex(0.00000,0.329978),complex(-0.322812,0.00000),complex(0.00000,-0.278987),complex(0.215847,0.00000),complex(0.00000,0.150186),complex(-0.0939011,0.00000),complex(0.00000,-0.0524742),complex(0.0259374,0.00000),complex(0.00000,0.0111424),complex(-0.00403709,0.00000),complex(0.00000,-0.00116517),complex(0.000232405,-0.00000)],$
            [complex(0.00000,-4.31552e-05),complex(0.000232398,0.00000),complex(0.00000,0.000869587),complex(-0.00260886,0.00000),complex(0.00000,-0.00665139),complex(0.0148727,0.00000),complex(0.00000,0.0297455),complex(-0.0539183,0.00000),complex(0.00000,-0.0894133),complex(0.136581,0.00000),complex(0.00000,0.193155),complex(-0.253855,0.00000),complex(0.00000,-0.310908),complex(0.355537,0.00000),complex(0.00000,0.380085),complex(-0.380085,0.00000),complex(0.00000,-0.355537),complex(0.310908,0.00000),complex(0.00000,0.253855),complex(-0.193155,0.00000),complex(0.00000,-0.136581),complex(0.0894133,0.00000),complex(0.00000,0.0539183),complex(-0.0297455,0.00000),complex(0.00000,-0.0148727),complex(0.00665132,0.00000),complex(0.00000,0.00260889),complex(-0.000869619,0.00000),complex(0.00000,-0.000232399),complex(4.31484e-05,-0.00000)]$
            ]
          end

      28: begin
            matrix[406:434,406:434]=[$
            [complex(6.10770e-05,0.00000),complex(0.00000,-0.000322981),complex(-0.00118652,0.00000),complex(0.00000,0.00349348),complex(0.00873356,0.00000),complex(0.00000,-0.0191343),complex(-0.0374628,0.00000),complex(0.00000,0.0664145),complex(0.107604,0.00000),complex(0.00000,-0.160406),complex(-0.221105,0.00000),complex(0.00000,0.282838),complex(0.336645,0.00000),complex(0.00000,-0.373474),complex(-0.386582,0.00000),complex(0.00000,0.373474),complex(0.336645,0.00000),complex(0.00000,-0.282838),complex(-0.221105,0.00000),complex(0.00000,0.160406),complex(0.107604,0.00000),complex(0.00000,-0.0664146),complex(-0.0374628,0.00000),complex(0.00000,0.0191343),complex(0.00873361,0.00000),complex(0.00000,-0.00349341),complex(-0.00118666,0.00000),complex(0.00000,0.000323000),complex(6.09718e-05,-0.00000)],$
            [complex(-0.000322969,0.00000),complex(0.00000,0.00158687),complex(0.00538228,0.00000),complex(0.00000,-0.0145243),complex(-0.0330099,0.00000),complex(0.00000,0.0650888),complex(0.113277,0.00000),complex(0.00000,-0.175716),complex(-0.244023,0.00000),complex(0.00000,0.303140),complex(0.334279,0.00000),complex(0.00000,-0.320709),complex(-0.254480,0.00000),complex(0.00000,0.141160),complex(2.13420e-08,0.00000),complex(0.00000,0.141160),complex(0.254480,0.00000),complex(0.00000,-0.320709),complex(-0.334279,0.00000),complex(0.00000,0.303140),complex(0.244023,0.00000),complex(0.00000,-0.175716),complex(-0.113277,0.00000),complex(0.00000,0.0650888),complex(0.0330099,0.00000),complex(0.00000,-0.0145243),complex(-0.00538217,0.00000),complex(0.00000,0.00158694),complex(0.000322904,-0.00000)],$
            [complex(0.00118661,0.00000),complex(0.00000,-0.00538223),complex(-0.0167236,0.00000),complex(0.00000,0.0409676),complex(0.0835524,0.00000),complex(0.00000,-0.145656),complex(-0.219665,0.00000),complex(0.00000,0.286944),complex(0.321004,0.00000),complex(0.00000,-0.297015),complex(-0.204703,0.00000),complex(0.00000,0.0581906),complex(-0.103891,0.00000),complex(0.00000,0.230513),complex(0.278371,0.00000),complex(0.00000,-0.230513),complex(-0.103891,0.00000),complex(0.00000,-0.0581905),complex(-0.204703,0.00000),complex(0.00000,0.297015),complex(0.321004,0.00000),complex(0.00000,-0.286944),complex(-0.219664,0.00000),complex(0.00000,0.145656),complex(0.0835524,0.00000),complex(0.00000,-0.0409676),complex(-0.0167237,0.00000),complex(0.00000,0.00538218),complex(0.00118662,-0.00000)],$
            [complex(-0.00349353,0.00000),complex(0.00000,0.0145243),complex(0.0409676,0.00000),complex(0.00000,-0.0899658),complex(-0.161743,0.00000),complex(0.00000,0.242704),complex(0.303702,0.00000),complex(0.00000,-0.308655),complex(-0.233119,0.00000),complex(0.00000,0.0840757),complex(-0.0927124,0.00000),complex(0.00000,0.227313),complex(0.258793,0.00000),complex(0.00000,-0.169653),complex(-3.06647e-08,0.00000),complex(0.00000,-0.169653),complex(-0.258793,0.00000),complex(0.00000,0.227313),complex(0.0927124,0.00000),complex(0.00000,0.0840759),complex(0.233119,0.00000),complex(0.00000,-0.308655),complex(-0.303701,0.00000),complex(0.00000,0.242704),complex(0.161743,0.00000),complex(0.00000,-0.0899658),complex(-0.0409677,0.00000),complex(0.00000,0.0145243),complex(0.00349350,-0.00000)],$
            [complex(0.00873351,0.00000),complex(0.00000,-0.0330099),complex(-0.0835525,0.00000),complex(0.00000,0.161743),complex(0.249695,0.00000),complex(0.00000,-0.308228),complex(-0.291920,0.00000),complex(0.00000,0.178695),complex(-0.00375987,0.00000),complex(0.00000,0.178241),complex(0.254959,0.00000),complex(0.00000,-0.187780),complex(-0.0117633,0.00000),complex(0.00000,-0.169653),complex(-0.245851,0.00000),complex(0.00000,0.169653),complex(-0.0117634,0.00000),complex(0.00000,0.187780),complex(0.254959,0.00000),complex(0.00000,-0.178241),complex(-0.00375999,0.00000),complex(0.00000,-0.178695),complex(-0.291920,0.00000),complex(0.00000,0.308228),complex(0.249695,0.00000),complex(0.00000,-0.161743),complex(-0.0835524,0.00000),complex(0.00000,0.0330097),complex(0.00873359,-0.00000)],$
            [complex(-0.0191343,0.00000),complex(0.00000,0.0650887),complex(0.145656,0.00000),complex(0.00000,-0.242704),complex(-0.308228,0.00000),complex(0.00000,0.284912),complex(0.149136,0.00000),complex(0.00000,0.0533864),complex(0.216926,0.00000),complex(0.00000,-0.239461),complex(-0.101561,0.00000),complex(-0.00000,-0.104656),complex(-0.231950,0.00000),complex(0.00000,0.185846),complex(-3.48768e-08,0.00000),complex(0.00000,0.185846),complex(0.231950,0.00000),complex(0.00000,-0.104656),complex(0.101561,0.00000),complex(0.00000,-0.239461),complex(-0.216926,0.00000),complex(0.00000,0.0533864),complex(-0.149136,0.00000),complex(0.00000,0.284912),complex(0.308228,0.00000),complex(0.00000,-0.242704),complex(-0.145656,0.00000),complex(0.00000,0.0650888),complex(0.0191343,-0.00000)],$
            [complex(0.0374628,0.00000),complex(0.00000,-0.113277),complex(-0.219664,0.00000),complex(0.00000,0.303702),complex(0.291920,0.00000),complex(0.00000,-0.149136),complex(0.0690918,0.00000),complex(0.00000,-0.230258),complex(-0.218086,0.00000),complex(0.00000,0.0376324),complex(-0.168587,0.00000),complex(0.00000,0.228559),complex(0.0899486,0.00000),complex(0.00000,0.126562),complex(0.229257,0.00000),complex(0.00000,-0.126562),complex(0.0899488,0.00000),complex(0.00000,-0.228560),complex(-0.168587,0.00000),complex(0.00000,-0.0376326),complex(-0.218086,0.00000),complex(0.00000,0.230258),complex(0.0690918,0.00000),complex(0.00000,0.149136),complex(0.291920,0.00000),complex(0.00000,-0.303701),complex(-0.219665,0.00000),complex(0.00000,0.113277),complex(0.0374629,-0.00000)],$
            [complex(-0.0664145,0.00000),complex(0.00000,0.175716),complex(0.286944,0.00000),complex(0.00000,-0.308655),complex(-0.178695,0.00000),complex(0.00000,-0.0533863),complex(-0.230258,0.00000),complex(0.00000,0.209228),complex(0.00553771,0.00000),complex(0.00000,0.196355),complex(0.204822,0.00000),complex(0.00000,-0.0114369),complex(0.190576,0.00000),complex(0.00000,-0.196324),complex(-1.11262e-08,0.00000),complex(0.00000,-0.196324),complex(-0.190577,0.00000),complex(0.00000,-0.0114369),complex(-0.204822,0.00000),complex(0.00000,0.196355),complex(-0.00553767,0.00000),complex(0.00000,0.209229),complex(0.230258,0.00000),complex(0.00000,-0.0533866),complex(0.178695,0.00000),complex(0.00000,-0.308655),complex(-0.286944,0.00000),complex(0.00000,0.175716),complex(0.0664145,-0.00000)],$
            [complex(0.107604,0.00000),complex(0.00000,-0.244023),complex(-0.321003,0.00000),complex(0.00000,0.233119),complex(-0.00376001,0.00000),complex(0.00000,0.216926),complex(0.218086,0.00000),complex(0.00000,-0.00553767),complex(0.203674,0.00000),complex(0.00000,-0.187522),complex(0.0349909,0.00000),complex(0.00000,-0.213535),complex(-0.144933,0.00000),complex(0.00000,-0.0908803),complex(-0.219497,0.00000),complex(0.00000,0.0908804),complex(-0.144933,0.00000),complex(0.00000,0.213535),complex(0.0349908,0.00000),complex(0.00000,0.187522),complex(0.203674,0.00000),complex(0.00000,0.00553771),complex(0.218086,0.00000),complex(0.00000,-0.216926),complex(-0.00375997,0.00000),complex(0.00000,-0.233119),complex(-0.321004,0.00000),complex(0.00000,0.244023),complex(0.107604,-0.00000)],$
            [complex(-0.160406,0.00000),complex(0.00000,0.303140),complex(0.297015,0.00000),complex(0.00000,-0.0840758),complex(0.178241,0.00000),complex(0.00000,-0.239461),complex(-0.0376326,0.00000),complex(0.00000,-0.196355),complex(-0.187522,0.00000),complex(0.00000,-0.0499268),complex(-0.218741,0.00000),complex(0.00000,0.106545),complex(-0.140904,0.00000),complex(0.00000,0.203215),complex(-2.50657e-08,0.00000),complex(0.00000,0.203215),complex(0.140904,0.00000),complex(0.00000,0.106545),complex(0.218741,0.00000),complex(0.00000,-0.0499269),complex(0.187522,0.00000),complex(0.00000,-0.196355),complex(0.0376326,0.00000),complex(0.00000,-0.239461),complex(-0.178241,0.00000),complex(0.00000,-0.0840758),complex(-0.297015,0.00000),complex(0.00000,0.303140),complex(0.160406,-0.00000)],$
            [complex(0.221105,0.00000),complex(0.00000,-0.334279),complex(-0.204704,0.00000),complex(0.00000,-0.0927124),complex(-0.254959,0.00000),complex(0.00000,0.101561),complex(-0.168587,0.00000),complex(0.00000,0.204822),complex(-0.0349909,0.00000),complex(0.00000,0.218741),complex(0.0928955,0.00000),complex(0.00000,0.161462),complex(0.181956,0.00000),complex(0.00000,0.0589711),complex(0.213643,0.00000),complex(0.00000,-0.0589710),complex(0.181956,0.00000),complex(0.00000,-0.161462),complex(0.0928954,0.00000),complex(0.00000,-0.218741),complex(-0.0349909,0.00000),complex(0.00000,-0.204822),complex(-0.168587,0.00000),complex(0.00000,-0.101561),complex(-0.254959,0.00000),complex(0.00000,0.0927125),complex(-0.204703,0.00000),complex(0.00000,0.334279),complex(0.221105,-0.00000)],$
            [complex(-0.282838,0.00000),complex(0.00000,0.320709),complex(0.0581905,0.00000),complex(0.00000,0.227313),complex(0.187780,0.00000),complex(0.00000,0.104656),complex(0.228559,0.00000),complex(0.00000,-0.0114370),complex(0.213535,0.00000),complex(0.00000,-0.106545),complex(0.161462,0.00000),complex(-0.00000,-0.173218),complex(0.0863039,0.00000),complex(-0.00000,-0.207449),complex(1.05282e-08,0.00000),complex(-0.00000,-0.207449),complex(-0.0863038,0.00000),complex(0.00000,-0.173218),complex(-0.161462,0.00000),complex(0.00000,-0.106545),complex(-0.213535,0.00000),complex(0.00000,-0.0114369),complex(-0.228559,0.00000),complex(0.00000,0.104656),complex(-0.187780,0.00000),complex(0.00000,0.227313),complex(-0.0581906,0.00000),complex(0.00000,0.320709),complex(0.282838,-0.00000)],$
            [complex(0.336645,0.00000),complex(0.00000,-0.254480),complex(0.103891,0.00000),complex(0.00000,-0.258793),complex(-0.0117634,0.00000),complex(0.00000,-0.231950),complex(-0.0899488,0.00000),complex(0.00000,-0.190577),complex(-0.144933,0.00000),complex(0.00000,-0.140904),complex(-0.181956,0.00000),complex(0.00000,-0.0863038),complex(-0.203430,0.00000),complex(0.00000,-0.0290486),complex(-0.210477,0.00000),complex(0.00000,0.0290487),complex(-0.203430,0.00000),complex(0.00000,0.0863039),complex(-0.181956,0.00000),complex(0.00000,0.140904),complex(-0.144933,0.00000),complex(0.00000,0.190577),complex(-0.0899487,0.00000),complex(0.00000,0.231949),complex(-0.0117633,0.00000),complex(0.00000,0.258793),complex(0.103891,0.00000),complex(0.00000,0.254480),complex(0.336645,-0.00000)],$
            [complex(-0.373474,0.00000),complex(0.00000,0.141160),complex(-0.230513,0.00000),complex(0.00000,0.169653),complex(-0.169653,0.00000),complex(0.00000,0.185846),complex(-0.126562,0.00000),complex(0.00000,0.196324),complex(-0.0908803,0.00000),complex(0.00000,0.203214),complex(-0.0589710,0.00000),complex(0.00000,0.207449),complex(-0.0290487,0.00000),complex(0.00000,0.209473),complex(-2.02418e-08,0.00000),complex(0.00000,0.209473),complex(0.0290486,0.00000),complex(0.00000,0.207449),complex(0.0589710,0.00000),complex(0.00000,0.203215),complex(0.0908803,0.00000),complex(0.00000,0.196324),complex(0.126562,0.00000),complex(0.00000,0.185846),complex(0.169653,0.00000),complex(0.00000,0.169653),complex(0.230513,0.00000),complex(0.00000,0.141160),complex(0.373474,-0.00000)],$
            [complex(0.386582,0.00000),complex(0.00000,4.47035e-08),complex(0.278371,0.00000),complex(0.00000,-4.47035e-08),complex(0.245851,0.00000),complex(0.00000,-4.47035e-08),complex(0.229257,0.00000),complex(0.00000,7.45058e-08),complex(0.219497,0.00000),complex(0.00000,-8.94070e-08),complex(0.213643,0.00000),complex(0.00000,-1.04308e-07),complex(0.210477,0.00000),complex(0.00000,-1.49012e-08),complex(0.209473,0.00000),0,complex(0.210477,0.00000),complex(0.00000,-2.98023e-08),complex(0.213643,0.00000),0,complex(0.219497,0.00000),complex(0.00000,1.49012e-08),complex(0.229257,0.00000),complex(0.00000,-1.49012e-08),complex(0.245851,0.00000),complex(0.00000,-4.47035e-08),complex(0.278371,0.00000),complex(0.00000,-1.49012e-08),complex(0.386582,-0.00000)],$
            [complex(-0.373474,0.00000),complex(0.00000,-0.141160),complex(-0.230513,0.00000),complex(0.00000,-0.169653),complex(-0.169653,0.00000),complex(0.00000,-0.185846),complex(-0.126562,0.00000),complex(0.00000,-0.196324),complex(-0.0908803,0.00000),complex(0.00000,-0.203215),complex(-0.0589710,0.00000),complex(0.00000,-0.207449),complex(-0.0290486,0.00000),complex(0.00000,-0.209473),complex(3.26985e-08,0.00000),complex(0.00000,-0.209473),complex(0.0290486,0.00000),complex(0.00000,-0.207449),complex(0.0589710,0.00000),complex(0.00000,-0.203215),complex(0.0908803,0.00000),complex(0.00000,-0.196324),complex(0.126562,0.00000),complex(0.00000,-0.185846),complex(0.169653,0.00000),complex(0.00000,-0.169653),complex(0.230513,0.00000),complex(0.00000,-0.141160),complex(0.373474,-0.00000)],$
            [complex(0.336645,0.00000),complex(0.00000,0.254480),complex(0.103891,0.00000),complex(0.00000,0.258793),complex(-0.0117634,0.00000),complex(0.00000,0.231949),complex(-0.0899488,0.00000),complex(0.00000,0.190577),complex(-0.144933,0.00000),complex(0.00000,0.140904),complex(-0.181956,0.00000),complex(0.00000,0.0863039),complex(-0.203430,0.00000),complex(0.00000,0.0290486),complex(-0.210477,0.00000),complex(0.00000,-0.0290487),complex(-0.203430,0.00000),complex(0.00000,-0.0863039),complex(-0.181956,0.00000),complex(0.00000,-0.140904),complex(-0.144933,0.00000),complex(0.00000,-0.190577),complex(-0.0899488,0.00000),complex(0.00000,-0.231949),complex(-0.0117634,0.00000),complex(0.00000,-0.258793),complex(0.103891,0.00000),complex(0.00000,-0.254480),complex(0.336645,-0.00000)],$
            [complex(-0.282838,0.00000),complex(-0.00000,-0.320708),complex(0.0581906,0.00000),complex(-0.00000,-0.227313),complex(0.187780,0.00000),complex(0.00000,-0.104656),complex(0.228559,0.00000),complex(0.00000,0.0114369),complex(0.213535,0.00000),complex(0.00000,0.106545),complex(0.161462,0.00000),complex(0.00000,0.173218),complex(0.0863038,0.00000),complex(0.00000,0.207449),complex(4.07447e-08,0.00000),complex(0.00000,0.207449),complex(-0.0863038,0.00000),complex(0.00000,0.173218),complex(-0.161462,0.00000),complex(0.00000,0.106545),complex(-0.213535,0.00000),complex(0.00000,0.0114368),complex(-0.228560,0.00000),complex(0.00000,-0.104656),complex(-0.187780,0.00000),complex(0.00000,-0.227313),complex(-0.0581905,0.00000),complex(0.00000,-0.320709),complex(0.282838,-0.00000)],$
            [complex(0.221105,0.00000),complex(0.00000,0.334279),complex(-0.204703,0.00000),complex(0.00000,0.0927123),complex(-0.254959,0.00000),complex(0.00000,-0.101561),complex(-0.168587,0.00000),complex(0.00000,-0.204822),complex(-0.0349908,0.00000),complex(0.00000,-0.218741),complex(0.0928955,0.00000),complex(0.00000,-0.161462),complex(0.181956,0.00000),complex(0.00000,-0.0589710),complex(0.213643,0.00000),complex(0.00000,0.0589709),complex(0.181956,0.00000),complex(-0.00000,0.161462),complex(0.0928955,0.00000),complex(0.00000,0.218741),complex(-0.0349909,0.00000),complex(0.00000,0.204822),complex(-0.168587,0.00000),complex(0.00000,0.101561),complex(-0.254959,0.00000),complex(0.00000,-0.0927125),complex(-0.204703,-0.00000),complex(0.00000,-0.334279),complex(0.221105,-0.00000)],$
            [complex(-0.160406,0.00000),complex(0.00000,-0.303140),complex(0.297015,0.00000),complex(0.00000,0.0840758),complex(0.178241,0.00000),complex(0.00000,0.239461),complex(-0.0376324,0.00000),complex(0.00000,0.196355),complex(-0.187522,0.00000),complex(0.00000,0.0499267),complex(-0.218741,0.00000),complex(0.00000,-0.106545),complex(-0.140904,0.00000),complex(0.00000,-0.203215),complex(-6.47842e-08,0.00000),complex(0.00000,-0.203215),complex(0.140904,0.00000),complex(0.00000,-0.106545),complex(0.218741,0.00000),complex(0.00000,0.0499268),complex(0.187522,0.00000),complex(0.00000,0.196355),complex(0.0376325,0.00000),complex(0.00000,0.239461),complex(-0.178241,0.00000),complex(0.00000,0.0840758),complex(-0.297015,0.00000),complex(0.00000,-0.303140),complex(0.160406,-0.00000)],$
            [complex(0.107604,0.00000),complex(0.00000,0.244023),complex(-0.321004,0.00000),complex(0.00000,-0.233119),complex(-0.00376002,0.00000),complex(0.00000,-0.216926),complex(0.218086,0.00000),complex(0.00000,0.00553777),complex(0.203674,0.00000),complex(0.00000,0.187522),complex(0.0349909,0.00000),complex(0.00000,0.213535),complex(-0.144933,0.00000),complex(0.00000,0.0908802),complex(-0.219497,0.00000),complex(0.00000,-0.0908805),complex(-0.144933,0.00000),complex(0.00000,-0.213535),complex(0.0349910,0.00000),complex(0.00000,-0.187522),complex(0.203674,0.00000),complex(0.00000,-0.00553772),complex(0.218086,0.00000),complex(0.00000,0.216926),complex(-0.00375994,0.00000),complex(0.00000,0.233119),complex(-0.321004,0.00000),complex(0.00000,-0.244023),complex(0.107604,-0.00000)],$
            [complex(-0.0664146,-0.00000),complex(0.00000,-0.175716),complex(0.286944,0.00000),complex(0.00000,0.308655),complex(-0.178695,-0.00000),complex(0.00000,0.0533864),complex(-0.230258,-0.00000),complex(0.00000,-0.209229),complex(0.00553768,0.00000),complex(0.00000,-0.196355),complex(0.204822,0.00000),complex(0.00000,0.0114368),complex(0.190577,0.00000),complex(0.00000,0.196324),complex(4.74782e-08,0.00000),complex(0.00000,0.196324),complex(-0.190577,0.00000),complex(0.00000,0.0114369),complex(-0.204822,0.00000),complex(0.00000,-0.196355),complex(-0.00553771,0.00000),complex(0.00000,-0.209229),complex(0.230258,0.00000),complex(0.00000,0.0533864),complex(0.178695,0.00000),complex(0.00000,0.308655),complex(-0.286944,0.00000),complex(0.00000,-0.175716),complex(0.0664145,0.00000)],$
            [complex(0.0374628,0.00000),complex(0.00000,0.113277),complex(-0.219664,0.00000),complex(0.00000,-0.303702),complex(0.291920,0.00000),complex(0.00000,0.149136),complex(0.0690918,0.00000),complex(0.00000,0.230258),complex(-0.218086,0.00000),complex(0.00000,-0.0376325),complex(-0.168587,0.00000),complex(0.00000,-0.228560),complex(0.0899487,0.00000),complex(0.00000,-0.126562),complex(0.229257,0.00000),complex(0.00000,0.126562),complex(0.0899488,0.00000),complex(0.00000,0.228559),complex(-0.168587,0.00000),complex(0.00000,0.0376325),complex(-0.218086,0.00000),complex(0.00000,-0.230258),complex(0.0690918,0.00000),complex(0.00000,-0.149136),complex(0.291920,0.00000),complex(-0.00000,0.303702),complex(-0.219664,0.00000),complex(0.00000,-0.113277),complex(0.0374629,-0.00000)],$
            [complex(-0.0191345,0.00000),complex(0.00000,-0.0650888),complex(0.145656,0.00000),complex(0.00000,0.242704),complex(-0.308228,0.00000),complex(0.00000,-0.284912),complex(0.149136,0.00000),complex(0.00000,-0.0533864),complex(0.216926,0.00000),complex(0.00000,0.239461),complex(-0.101561,0.00000),complex(-0.00000,0.104656),complex(-0.231949,0.00000),complex(0.00000,-0.185846),complex(3.73173e-09,0.00000),complex(0.00000,-0.185846),complex(0.231949,0.00000),complex(0.00000,0.104656),complex(0.101561,0.00000),complex(0.00000,0.239461),complex(-0.216926,0.00000),complex(0.00000,-0.0533865),complex(-0.149136,0.00000),complex(0.00000,-0.284912),complex(0.308228,0.00000),complex(-0.00000,0.242704),complex(-0.145656,0.00000),complex(0.00000,-0.0650888),complex(0.0191344,-0.00000)],$
            [complex(0.00873350,0.00000),complex(0.00000,0.0330098),complex(-0.0835525,0.00000),complex(0.00000,-0.161743),complex(0.249695,0.00000),complex(0.00000,0.308228),complex(-0.291920,0.00000),complex(0.00000,-0.178695),complex(-0.00375998,0.00000),complex(0.00000,-0.178241),complex(0.254959,0.00000),complex(0.00000,0.187780),complex(-0.0117634,0.00000),complex(0.00000,0.169653),complex(-0.245851,0.00000),complex(0.00000,-0.169653),complex(-0.0117633,0.00000),complex(0.00000,-0.187780),complex(0.254959,0.00000),complex(0.00000,0.178241),complex(-0.00376002,0.00000),complex(0.00000,0.178695),complex(-0.291920,0.00000),complex(0.00000,-0.308228),complex(0.249695,0.00000),complex(-0.00000,0.161743),complex(-0.0835524,0.00000),complex(0.00000,-0.0330098),complex(0.00873364,-0.00000)],$
            [complex(-0.00349340,0.00000),complex(0.00000,-0.0145243),complex(0.0409676,0.00000),complex(0.00000,0.0899659),complex(-0.161743,0.00000),complex(0.00000,-0.242704),complex(0.303702,0.00000),complex(0.00000,0.308655),complex(-0.233119,0.00000),complex(0.00000,-0.0840758),complex(-0.0927125,0.00000),complex(0.00000,-0.227313),complex(0.258793,0.00000),complex(0.00000,0.169653),complex(-2.54901e-08,0.00000),complex(0.00000,0.169653),complex(-0.258793,0.00000),complex(0.00000,-0.227313),complex(0.0927125,0.00000),complex(0.00000,-0.0840758),complex(0.233119,0.00000),complex(0.00000,0.308655),complex(-0.303702,0.00000),complex(0.00000,-0.242704),complex(0.161743,0.00000),complex(-0.00000,0.0899659),complex(-0.0409677,0.00000),complex(0.00000,-0.0145244),complex(0.00349349,-0.00000)],$
            [complex(0.00118663,0.00000),complex(0.00000,0.00538220),complex(-0.0167236,0.00000),complex(0.00000,-0.0409675),complex(0.0835524,0.00000),complex(0.00000,0.145656),complex(-0.219664,0.00000),complex(0.00000,-0.286944),complex(0.321004,0.00000),complex(0.00000,0.297015),complex(-0.204703,0.00000),complex(0.00000,-0.0581905),complex(-0.103891,0.00000),complex(0.00000,-0.230513),complex(0.278371,0.00000),complex(0.00000,0.230513),complex(-0.103891,0.00000),complex(0.00000,0.0581906),complex(-0.204703,0.00000),complex(0.00000,-0.297015),complex(0.321004,0.00000),complex(0.00000,0.286944),complex(-0.219665,0.00000),complex(0.00000,-0.145656),complex(0.0835524,0.00000),complex(-0.00000,0.0409676),complex(-0.0167237,0.00000),complex(0.00000,-0.00538219),complex(0.00118666,-0.00000)],$
            [complex(-0.000322896,0.00000),complex(0.00000,-0.00158694),complex(0.00538222,0.00000),complex(0.00000,0.0145244),complex(-0.0330099,0.00000),complex(0.00000,-0.0650888),complex(0.113277,0.00000),complex(0.00000,0.175716),complex(-0.244023,0.00000),complex(0.00000,-0.303140),complex(0.334279,0.00000),complex(0.00000,0.320709),complex(-0.254480,0.00000),complex(0.00000,-0.141160),complex(1.81097e-08,0.00000),complex(0.00000,-0.141160),complex(0.254480,0.00000),complex(0.00000,0.320709),complex(-0.334279,0.00000),complex(0.00000,-0.303140),complex(0.244023,0.00000),complex(0.00000,0.175716),complex(-0.113277,0.00000),complex(0.00000,-0.0650888),complex(0.0330098,0.00000),complex(-0.00000,0.0145244),complex(-0.00538224,0.00000),complex(0.00000,-0.00158687),complex(0.000322963,-0.00000)],$
            [complex(6.10088e-05,0.00000),complex(0.00000,0.000322975),complex(-0.00118664,0.00000),complex(0.00000,-0.00349354),complex(0.00873359,0.00000),complex(0.00000,0.0191343),complex(-0.0374628,0.00000),complex(0.00000,-0.0664145),complex(0.107604,0.00000),complex(0.00000,0.160406),complex(-0.221105,0.00000),complex(0.00000,-0.282838),complex(0.336645,0.00000),complex(0.00000,0.373474),complex(-0.386582,0.00000),complex(0.00000,-0.373474),complex(0.336645,0.00000),complex(0.00000,0.282838),complex(-0.221105,0.00000),complex(0.00000,-0.160406),complex(0.107604,0.00000),complex(0.00000,0.0664145),complex(-0.0374628,0.00000),complex(0.00000,-0.0191344),complex(0.00873356,0.00000),complex(-0.00000,0.00349347),complex(-0.00118663,0.00000),complex(0.00000,-0.000323044),complex(6.10005e-05,-0.00000)]$
            ]
          end

      27: begin
            matrix[378:405,378:405]=[$
            [complex(0.00000,-8.63138e-05),complex(-0.000448512,0.00000),complex(0.00000,0.00161714),complex(0.00466831,0.00000),complex(0.00000,-0.0114349),complex(-0.0245253,0.00000),complex(0.00000,0.0469621),complex(0.0813408,0.00000),complex(0.00000,-0.128611),complex(-0.186868,0.00000),complex(0.00000,0.250709),complex(0.311673,0.00000),complex(0.00000,-0.359889),complex(-0.386582,0.00000),complex(0.00000,0.386582),complex(0.359889,0.00000),complex(0.00000,-0.311673),complex(-0.250709,0.00000),complex(0.00000,0.186868),complex(0.128611,0.00000),complex(0.00000,-0.0813409),complex(-0.0469622,0.00000),complex(0.00000,0.0245253),complex(0.0114349,0.00000),complex(0.00000,-0.00466830),complex(-0.00161713,0.00000),complex(0.00000,0.000448462),complex(8.63028e-05,-0.00000)],$
            [complex(0.00000,0.000448464),complex(0.00215787,0.00000),complex(0.00000,-0.00715806),complex(-0.0188666,0.00000),complex(0.00000,0.0418124),complex(0.0802380,0.00000),complex(0.00000,-0.135568),complex(-0.203503,0.00000),complex(0.00000,0.272264),complex(0.323665,0.00000),complex(0.00000,-0.337743),complex(-0.299907,0.00000),complex(0.00000,0.207782),complex(0.0743978,0.00000),complex(0.00000,0.0743978),complex(0.207782,0.00000),complex(0.00000,-0.299907),complex(-0.337743,0.00000),complex(0.00000,0.323664),complex(0.272264,0.00000),complex(0.00000,-0.203503),complex(-0.135568,0.00000),complex(0.00000,0.0802380),complex(0.0418124,0.00000),complex(0.00000,-0.0188667),complex(-0.00715797,0.00000),complex(0.00000,0.00215788),complex(0.000448447,-0.00000)],$
            [complex(0.00000,-0.00161713),complex(-0.00715805,0.00000),complex(0.00000,0.0216655),complex(0.0515792,0.00000),complex(0.00000,-0.101929),complex(-0.171487,0.00000),complex(0.00000,0.248159),complex(0.308257,0.00000),complex(0.00000,-0.322644),complex(-0.269305,0.00000),complex(0.00000,0.147201),complex(-0.0166359,0.00000),complex(0.00000,0.172885),complex(0.268245,0.00000),complex(0.00000,-0.268245),complex(-0.172885,0.00000),complex(0.00000,0.0166359),complex(-0.147201,0.00000),complex(0.00000,0.269305),complex(0.322644,0.00000),complex(0.00000,-0.308258),complex(-0.248159,0.00000),complex(0.00000,0.171487),complex(0.101929,0.00000),complex(0.00000,-0.0515792),complex(-0.0216655,0.00000),complex(0.00000,0.00715803),complex(0.00161717,-0.00000)],$
            [complex(0.00000,0.00466824),complex(0.0188667,0.00000),complex(0.00000,-0.0515792),complex(-0.109363,0.00000),complex(0.00000,0.188809),complex(0.269815,0.00000),complex(0.00000,-0.316941),complex(-0.293279,-0.00000),complex(0.00000,0.183108),complex(0.0103655,0.00000),complex(0.00000,0.162247),complex(0.259327,0.00000),complex(0.00000,-0.232902),complex(-0.0929229,0.00000),complex(0.00000,-0.0929229),complex(-0.232902,0.00000),complex(0.00000,0.259327),complex(0.162247,0.00000),complex(0.00000,0.0103656),complex(0.183108,0.00000),complex(0.00000,-0.293279),complex(-0.316941,0.00000),complex(0.00000,0.269815),complex(0.188809,0.00000),complex(0.00000,-0.109363),complex(-0.0515792,0.00000),complex(0.00000,0.0188667),complex(0.00466827,-0.00000)],$
            [complex(0.00000,-0.0114349),complex(-0.0418124,0.00000),complex(0.00000,0.101929),complex(0.188809,0.00000),complex(0.00000,-0.276041),complex(-0.316570,0.00000),complex(0.00000,0.265871),complex(0.116661,0.00000),complex(0.00000,0.0796076),complex(0.228513,0.00000),complex(0.00000,-0.246023),complex(-0.117633,0.00000),complex(0.00000,-0.0814990),complex(-0.227614,0.00000),complex(0.00000,0.227614),complex(0.0814987,0.00000),complex(0.00000,0.117633),complex(0.246023,0.00000),complex(0.00000,-0.228513),complex(-0.0796074,0.00000),complex(0.00000,-0.116661),complex(-0.265871,0.00000),complex(0.00000,0.316570),complex(0.276041,0.00000),complex(0.00000,-0.188809),complex(-0.101929,0.00000),complex(0.00000,0.0418124),complex(0.0114350,-0.00000)],$
            [complex(0.00000,0.0245252),complex(0.0802379,0.00000),complex(0.00000,-0.171487),complex(-0.269815,0.00000),complex(0.00000,0.316570),complex(0.255325,0.00000),complex(0.00000,-0.0823115),complex(0.126536,0.00000),complex(0.00000,-0.248957),complex(-0.201251,0.00000),complex(0.00000,0.0123532),complex(-0.182091,0.00000),complex(0.00000,0.235594),complex(0.106125,0.00000),complex(0.00000,0.106125),complex(0.235594,0.00000),complex(0.00000,-0.182091),complex(0.0123533,0.00000),complex(0.00000,-0.201251),complex(-0.248957,0.00000),complex(0.00000,0.126536),complex(-0.0823115,0.00000),complex(0.00000,0.255325),complex(0.316570,0.00000),complex(0.00000,-0.269815),complex(-0.171487,0.00000),complex(0.00000,0.0802380),complex(0.0245253,-0.00000)],$
            [complex(0.00000,-0.0469622),complex(-0.135568,0.00000),complex(0.00000,0.248159),complex(0.316941,0.00000),complex(0.00000,-0.265871),complex(-0.0823114,0.00000),complex(0.00000,-0.140696),complex(-0.252065,0.00000),complex(0.00000,0.164053),complex(-0.0556412,0.00000),complex(0.00000,0.222108),complex(0.189043,0.00000),complex(0.00000,0.0145526),complex(0.203215,0.00000),complex(0.00000,-0.203215),complex(-0.0145525,0.00000),complex(0.00000,-0.189043),complex(-0.222108,0.00000),complex(0.00000,0.0556411),complex(-0.164053,0.00000),complex(0.00000,0.252065),complex(0.140696,0.00000),complex(0.00000,0.0823115),complex(0.265871,0.00000),complex(0.00000,-0.316941),complex(-0.248159,0.00000),complex(0.00000,0.135568),complex(0.0469621,-0.00000)],$
            [complex(0.00000,0.0813409),complex(0.203503,0.00000),complex(0.00000,-0.308257),complex(-0.293279,0.00000),complex(0.00000,0.116660),complex(-0.126536,0.00000),complex(0.00000,0.252065),complex(0.150364,0.00000),complex(0.00000,0.0870734),complex(0.232010,0.00000),complex(0.00000,-0.139940),complex(0.0945913,0.00000),complex(0.00000,-0.226851),complex(-0.117326,-0.00000),complex(0.00000,-0.117326),complex(-0.226851,0.00000),complex(0.00000,0.0945912),complex(-0.139940,0.00000),complex(0.00000,0.232010),complex(0.0870735,0.00000),complex(0.00000,0.150364),complex(0.252065,0.00000),complex(0.00000,-0.126536),complex(0.116661,0.00000),complex(0.00000,-0.293279),complex(-0.308258,0.00000),complex(0.00000,0.203503),complex(0.0813409,-0.00000)],$
            [complex(0.00000,-0.128611),complex(-0.272264,0.00000),complex(0.00000,0.322644),complex(0.183108,0.00000),complex(0.00000,0.0796075),complex(0.248957,0.00000),complex(0.00000,-0.164053),complex(0.0870734,0.00000),complex(0.00000,-0.232969),complex(-0.111745,0.00000),complex(0.00000,-0.135451),complex(-0.218590,0.00000),complex(0.00000,0.0398537),complex(-0.185509,0.00000),complex(0.00000,0.185509),complex(-0.0398536,0.00000),complex(0.00000,0.218590),complex(0.135451,0.00000),complex(0.00000,0.111745),complex(0.232969,0.00000),complex(0.00000,-0.0870734),complex(0.164053,0.00000),complex(0.00000,-0.248957),complex(-0.0796076,0.00000),complex(0.00000,-0.183108),complex(-0.322644,0.00000),complex(0.00000,0.272264),complex(0.128611,-0.00000)],$
            [complex(0.00000,0.186868),complex(0.323665,0.00000),complex(0.00000,-0.269305),complex(-0.0103656,0.00000),complex(0.00000,-0.228513),complex(-0.201251,0.00000),complex(0.00000,-0.0556411),complex(-0.232010,0.00000),complex(0.00000,0.111745),complex(-0.147515,0.00000),complex(0.00000,0.207872),complex(-0.00791818,0.00000),complex(0.00000,0.210290),complex(0.127676,0.00000),complex(0.00000,0.127676),complex(0.210290,0.00000),complex(0.00000,-0.00791807),complex(0.207872,0.00000),complex(0.00000,-0.147515),complex(0.111745,0.00000),complex(0.00000,-0.232010),complex(-0.0556412,0.00000),complex(0.00000,-0.201251),complex(-0.228513,0.00000),complex(0.00000,-0.0103655),complex(-0.269305,0.00000),complex(0.00000,0.323665),complex(0.186868,-0.00000)],$
            [complex(-0.00000,-0.250709),complex(-0.337743,0.00000),complex(0.00000,0.147201),complex(-0.162247,0.00000),complex(0.00000,0.246023),complex(0.0123532,0.00000),complex(0.00000,0.222108),complex(0.139940,0.00000),complex(0.00000,0.135451),complex(0.207872,0.00000),complex(0.00000,0.0235645),complex(0.216007,0.00000),complex(0.00000,-0.0858668),complex(0.171295,0.00000),complex(-0.00000,-0.171295),complex(0.0858668,0.00000),complex(-0.00000,-0.216006),complex(-0.0235644,0.00000),complex(-0.00000,-0.207872),complex(-0.135451,0.00000),complex(-0.00000,-0.139940),complex(-0.222108,0.00000),complex(0.00000,-0.0123532),complex(-0.246023,0.00000),complex(0.00000,0.162247),complex(-0.147201,0.00000),complex(0.00000,0.337743),complex(0.250709,-0.00000)],$
            [complex(0.00000,0.311673),complex(0.299907,0.00000),complex(0.00000,0.0166359),complex(0.259327,0.00000),complex(0.00000,-0.117633),complex(0.182091,0.00000),complex(0.00000,-0.189042),complex(0.0945912,0.00000),complex(0.00000,-0.218590),complex(0.00791816,0.00000),complex(0.00000,-0.216007),complex(-0.0712113,0.00000),complex(0.00000,-0.187479),complex(-0.137790,0.00000),complex(0.00000,-0.137790),complex(-0.187479,0.00000),complex(0.00000,-0.0712113),complex(-0.216007,0.00000),complex(0.00000,0.00791808),complex(-0.218590,0.00000),complex(0.00000,0.0945913),complex(-0.189043,0.00000),complex(0.00000,0.182091),complex(-0.117633,0.00000),complex(0.00000,0.259327),complex(0.0166358,0.00000),complex(0.00000,0.299907),complex(0.311673,-0.00000)],$
            [complex(0.00000,-0.359889),complex(-0.207782,0.00000),complex(0.00000,-0.172885),complex(-0.232902,0.00000),complex(0.00000,-0.0814988),complex(-0.235594,0.00000),complex(0.00000,-0.0145525),complex(-0.226851,0.00000),complex(0.00000,0.0398537),complex(-0.210290,0.00000),complex(0.00000,0.0858669),complex(-0.187479,0.00000),complex(0.00000,0.125332),complex(-0.159106,0.00000),complex(0.00000,0.159106),complex(-0.125332,0.00000),complex(0.00000,0.187479),complex(-0.0858668,0.00000),complex(0.00000,0.210290),complex(-0.0398537,0.00000),complex(0.00000,0.226851),complex(0.0145525,0.00000),complex(0.00000,0.235594),complex(0.0814988,0.00000),complex(0.00000,0.232902),complex(0.172885,0.00000),complex(0.00000,0.207782),complex(0.359889,-0.00000)],$
            [complex(0.00000,0.386583),complex(0.0743979,0.00000),complex(0.00000,0.268245),complex(0.0929230,0.00000),complex(0.00000,0.227614),complex(0.106125,0.00000),complex(0.00000,0.203215),complex(0.117326,0.00000),complex(0.00000,0.185509),complex(0.127676,0.00000),complex(0.00000,0.171295),complex(0.137790,0.00000),complex(0.00000,0.159106),complex(0.148120,0.00000),complex(0.00000,0.148119),complex(0.159106,0.00000),complex(0.00000,0.137790),complex(0.171295,0.00000),complex(0.00000,0.127676),complex(0.185509,0.00000),complex(0.00000,0.117326),complex(0.203215,0.00000),complex(0.00000,0.106125),complex(0.227614,0.00000),complex(0.00000,0.0929229),complex(0.268245,0.00000),complex(0.00000,0.0743978),complex(0.386583,-0.00000)],$
            [complex(0.00000,-0.386583),complex(0.0743978,0.00000),complex(0.00000,-0.268245),complex(0.0929229,0.00000),complex(0.00000,-0.227614),complex(0.106125,0.00000),complex(0.00000,-0.203215),complex(0.117326,0.00000),complex(0.00000,-0.185509),complex(0.127676,0.00000),complex(0.00000,-0.171295),complex(0.137790,0.00000),complex(0.00000,-0.159106),complex(0.148120,0.00000),complex(0.00000,-0.148120),complex(0.159106,0.00000),complex(0.00000,-0.137790),complex(0.171295,0.00000),complex(0.00000,-0.127676),complex(0.185509,0.00000),complex(0.00000,-0.117326),complex(0.203215,0.00000),complex(0.00000,-0.106125),complex(0.227614,-0.00000),complex(0.00000,-0.0929229),complex(0.268245,0.00000),complex(0.00000,-0.0743978),complex(0.386583,-0.00000)],$
            [complex(0.00000,0.359889),complex(-0.207782,0.00000),complex(0.00000,0.172885),complex(-0.232902,0.00000),complex(0.00000,0.0814989),complex(-0.235594,0.00000),complex(0.00000,0.0145525),complex(-0.226851,0.00000),complex(0.00000,-0.0398536),complex(-0.210290,0.00000),complex(0.00000,-0.0858669),complex(-0.187479,0.00000),complex(0.00000,-0.125332),complex(-0.159106,0.00000),complex(0.00000,-0.159106),complex(-0.125332,0.00000),complex(0.00000,-0.187479),complex(-0.0858669,0.00000),complex(0.00000,-0.210290),complex(-0.0398537,0.00000),complex(0.00000,-0.226851),complex(0.0145525,0.00000),complex(0.00000,-0.235594),complex(0.0814988,0.00000),complex(0.00000,-0.232902),complex(0.172885,0.00000),complex(0.00000,-0.207782),complex(0.359889,-0.00000)],$
            [complex(0.00000,-0.311673),complex(0.299907,0.00000),complex(0.00000,-0.0166358),complex(0.259327,0.00000),complex(0.00000,0.117633),complex(0.182091,0.00000),complex(0.00000,0.189043),complex(0.0945912,0.00000),complex(0.00000,0.218591),complex(0.00791814,0.00000),complex(0.00000,0.216006),complex(-0.0712112,0.00000),complex(0.00000,0.187479),complex(-0.137790,0.00000),complex(0.00000,0.137790),complex(-0.187479,0.00000),complex(0.00000,0.0712112),complex(-0.216007,-0.00000),complex(0.00000,-0.00791816),complex(-0.218590,0.00000),complex(0.00000,-0.0945913),complex(-0.189043,0.00000),complex(0.00000,-0.182091),complex(-0.117633,0.00000),complex(0.00000,-0.259327),complex(0.0166359,-0.00000),complex(0.00000,-0.299907),complex(0.311673,-0.00000)],$
            [complex(0.00000,0.250709),complex(-0.337743,0.00000),complex(0.00000,-0.147201),complex(-0.162246,0.00000),complex(0.00000,-0.246023),complex(0.0123534,0.00000),complex(0.00000,-0.222108),complex(0.139940,0.00000),complex(0.00000,-0.135451),complex(0.207872,0.00000),complex(0.00000,-0.0235644),complex(0.216007,0.00000),complex(0.00000,0.0858670),complex(0.171295,0.00000),complex(0.00000,0.171295),complex(0.0858669,0.00000),complex(0.00000,0.216006),complex(-0.0235645,0.00000),complex(0.00000,0.207872),complex(-0.135451,0.00000),complex(0.00000,0.139940),complex(-0.222108,0.00000),complex(0.00000,0.0123532),complex(-0.246023,0.00000),complex(0.00000,-0.162247),complex(-0.147201,0.00000),complex(0.00000,-0.337743),complex(0.250710,-0.00000)],$
            [complex(0.00000,-0.186868),complex(0.323664,0.00000),complex(0.00000,0.269305),complex(-0.0103656,0.00000),complex(0.00000,0.228513),complex(-0.201251,0.00000),complex(0.00000,0.0556412),complex(-0.232010,0.00000),complex(0.00000,-0.111745),complex(-0.147515,0.00000),complex(0.00000,-0.207872),complex(-0.00791813,0.00000),complex(0.00000,-0.210290),complex(0.127676,0.00000),complex(0.00000,-0.127676),complex(0.210290,0.00000),complex(0.00000,0.00791802),complex(0.207872,0.00000),complex(0.00000,0.147515),complex(0.111745,0.00000),complex(0.00000,0.232010),complex(-0.0556413,0.00000),complex(0.00000,0.201251),complex(-0.228513,0.00000),complex(0.00000,0.0103655),complex(-0.269305,0.00000),complex(0.00000,-0.323664),complex(0.186868,-0.00000)],$
            [complex(0.00000,0.128611),complex(-0.272264,0.00000),complex(0.00000,-0.322644),complex(0.183108,0.00000),complex(0.00000,-0.0796075),complex(0.248957,0.00000),complex(0.00000,0.164053),complex(0.0870734,0.00000),complex(0.00000,0.232969),complex(-0.111745,0.00000),complex(0.00000,0.135451),complex(-0.218591,0.00000),complex(0.00000,-0.0398537),complex(-0.185509,0.00000),complex(0.00000,-0.185509),complex(-0.0398537,0.00000),complex(0.00000,-0.218590),complex(0.135451,0.00000),complex(0.00000,-0.111745),complex(0.232969,0.00000),complex(0.00000,0.0870734),complex(0.164053,0.00000),complex(0.00000,0.248957),complex(-0.0796075,0.00000),complex(0.00000,0.183108),complex(-0.322644,0.00000),complex(0.00000,-0.272264),complex(0.128611,0.00000)],$
            [complex(0.00000,-0.0813408),complex(0.203503,0.00000),complex(0.00000,0.308258),complex(-0.293279,0.00000),complex(0.00000,-0.116660),complex(-0.126536,0.00000),complex(0.00000,-0.252065),complex(0.150364,0.00000),complex(0.00000,-0.0870734),complex(0.232010,0.00000),complex(0.00000,0.139940),complex(0.0945913,0.00000),complex(0.00000,0.226851),complex(-0.117326,0.00000),complex(0.00000,0.117326),complex(-0.226851,0.00000),complex(0.00000,-0.0945911),complex(-0.139940,0.00000),complex(0.00000,-0.232010),complex(0.0870735,0.00000),complex(0.00000,-0.150364),complex(0.252065,0.00000),complex(-0.00000,0.126536),complex(0.116661,0.00000),complex(-0.00000,0.293279),complex(-0.308258,0.00000),complex(0.00000,-0.203503),complex(0.0813409,-0.00000)],$
            [complex(0.00000,0.0469622),complex(-0.135568,0.00000),complex(0.00000,-0.248159),complex(0.316941,0.00000),complex(0.00000,0.265871),complex(-0.0823115,0.00000),complex(0.00000,0.140696),complex(-0.252066,0.00000),complex(0.00000,-0.164053),complex(-0.0556412,0.00000),complex(0.00000,-0.222108),complex(0.189042,0.00000),complex(0.00000,-0.0145525),complex(0.203215,0.00000),complex(0.00000,0.203215),complex(-0.0145525,0.00000),complex(0.00000,0.189043),complex(-0.222108,0.00000),complex(0.00000,-0.0556413),complex(-0.164053,0.00000),complex(0.00000,-0.252065),complex(0.140696,0.00000),complex(-0.00000,-0.0823114),complex(0.265871,0.00000),complex(-0.00000,0.316941),complex(-0.248159,0.00000),complex(0.00000,-0.135568),complex(0.0469622,-0.00000)],$
            [complex(0.00000,-0.0245253),complex(0.0802379,0.00000),complex(0.00000,0.171487),complex(-0.269815,0.00000),complex(0.00000,-0.316570),complex(0.255325,0.00000),complex(0.00000,0.0823114),complex(0.126536,0.00000),complex(0.00000,0.248957),complex(-0.201251,0.00000),complex(0.00000,-0.0123532),complex(-0.182091,0.00000),complex(0.00000,-0.235594),complex(0.106125,0.00000),complex(0.00000,-0.106125),complex(0.235594,0.00000),complex(0.00000,0.182091),complex(0.0123533,0.00000),complex(0.00000,0.201251),complex(-0.248957,0.00000),complex(0.00000,-0.126536),complex(-0.0823115,0.00000),complex(0.00000,-0.255325),complex(0.316570,0.00000),complex(-0.00000,0.269815),complex(-0.171487,0.00000),complex(0.00000,-0.0802380),complex(0.0245252,-0.00000)],$
            [complex(0.00000,0.0114349),complex(-0.0418124,0.00000),complex(0.00000,-0.101929),complex(0.188809,0.00000),complex(0.00000,0.276041),complex(-0.316570,0.00000),complex(0.00000,-0.265871),complex(0.116661,0.00000),complex(0.00000,-0.0796075),complex(0.228513,0.00000),complex(0.00000,0.246023),complex(-0.117633,0.00000),complex(0.00000,0.0814989),complex(-0.227614,0.00000),complex(0.00000,-0.227613),complex(0.0814987,0.00000),complex(0.00000,-0.117633),complex(0.246023,0.00000),complex(0.00000,0.228513),complex(-0.0796075,0.00000),complex(0.00000,0.116661),complex(-0.265871,0.00000),complex(0.00000,-0.316570),complex(0.276041,0.00000),complex(-0.00000,0.188809),complex(-0.101929,0.00000),complex(0.00000,-0.0418124),complex(0.0114349,-0.00000)],$
            [complex(0.00000,-0.00466829),complex(0.0188667,0.00000),complex(0.00000,0.0515793),complex(-0.109363,0.00000),complex(0.00000,-0.188809),complex(0.269815,0.00000),complex(0.00000,0.316941),complex(-0.293279,0.00000),complex(0.00000,-0.183108),complex(0.0103655,0.00000),complex(0.00000,-0.162247),complex(0.259327,0.00000),complex(0.00000,0.232902),complex(-0.0929229,0.00000),complex(0.00000,0.0929229),complex(-0.232902,0.00000),complex(0.00000,-0.259327),complex(0.162247,0.00000),complex(0.00000,-0.0103655),complex(0.183108,0.00000),complex(0.00000,0.293279),complex(-0.316941,0.00000),complex(0.00000,-0.269815),complex(0.188809,0.00000),complex(-0.00000,0.109363),complex(-0.0515793,0.00000),complex(0.00000,-0.0188667),complex(0.00466829,-0.00000)],$
            [complex(0.00000,0.00161720),complex(-0.00715801,0.00000),complex(0.00000,-0.0216655),complex(0.0515792,0.00000),complex(0.00000,0.101929),complex(-0.171487,0.00000),complex(0.00000,-0.248159),complex(0.308258,0.00000),complex(0.00000,0.322644),complex(-0.269305,0.00000),complex(0.00000,-0.147201),complex(-0.0166359,0.00000),complex(0.00000,-0.172885),complex(0.268245,0.00000),complex(0.00000,0.268245),complex(-0.172885,0.00000),complex(0.00000,-0.0166360),complex(-0.147201,0.00000),complex(0.00000,-0.269305),complex(0.322644,0.00000),complex(0.00000,0.308258),complex(-0.248159,0.00000),complex(0.00000,-0.171487),complex(0.101929,0.00000),complex(-0.00000,0.0515792),complex(-0.0216655,0.00000),complex(0.00000,-0.00715810),complex(0.00161714,-0.00000)],$
            [complex(0.00000,-0.000448538),complex(0.00215799,0.00000),complex(0.00000,0.00715807),complex(-0.0188668,0.00000),complex(0.00000,-0.0418124),complex(0.0802379,0.00000),complex(0.00000,0.135568),complex(-0.203503,-0.00000),complex(0.00000,-0.272264),complex(0.323665,0.00000),complex(0.00000,0.337743),complex(-0.299907,0.00000),complex(0.00000,-0.207782),complex(0.0743978,0.00000),complex(0.00000,-0.0743978),complex(0.207782,0.00000),complex(0.00000,0.299907),complex(-0.337743,0.00000),complex(0.00000,-0.323665),complex(0.272264,0.00000),complex(0.00000,0.203503),complex(-0.135568,0.00000),complex(0.00000,-0.0802379),complex(0.0418124,0.00000),complex(-0.00000,0.0188667),complex(-0.00715808,0.00000),complex(0.00000,-0.00215787),complex(0.000448468,-0.00000)],$
            [complex(0.00000,8.62540e-05),complex(-0.000448487,0.00000),complex(0.00000,-0.00161715),complex(0.00466820,0.00000),complex(0.00000,0.0114349),complex(-0.0245252,0.00000),complex(0.00000,-0.0469621),complex(0.0813408,0.00000),complex(0.00000,0.128611),complex(-0.186868,0.00000),complex(0.00000,-0.250709),complex(0.311673,0.00000),complex(0.00000,0.359889),complex(-0.386582,0.00000),complex(0.00000,-0.386583),complex(0.359889,0.00000),complex(0.00000,0.311673),complex(-0.250709,0.00000),complex(0.00000,-0.186868),complex(0.128611,0.00000),complex(0.00000,0.0813409),complex(-0.0469622,0.00000),complex(0.00000,-0.0245252),complex(0.0114349,0.00000),complex(-0.00000,0.00466836),complex(-0.00161717,0.00000),complex(0.00000,-0.000448499),complex(8.63237e-05,-0.00000)]$
            ]
          end

      26: begin
            matrix[351:377,351:377]=[$
            [complex(-0.000122061,0.00000),complex(0.00000,0.000622524),complex(0.00220060,0.00000),complex(0.00000,-0.00622441),complex(-0.0149256,0.00000),complex(0.00000,0.0313081),complex(0.0585720,0.00000),complex(0.00000,-0.0990050),complex(-0.152577,0.00000),complex(0.00000,0.215776),complex(0.281338,0.00000),complex(0.00000,-0.339306),complex(-0.379356,0.00000),complex(0.00000,0.393676),complex(0.379356,0.00000),complex(0.00000,-0.339306),complex(-0.281338,0.00000),complex(0.00000,0.215776),complex(0.152577,0.00000),complex(0.00000,-0.0990050),complex(-0.0585721,0.00000),complex(0.00000,0.0313081),complex(0.0149256,0.00000),complex(0.00000,-0.00622439),complex(-0.00220059,0.00000),complex(0.00000,0.000622431),complex(0.000122062,-0.00000)],$
            [complex(0.000622593,0.00000),complex(0.00000,-0.00292983),complex(-0.00949485,0.00000),complex(0.00000,0.0244139),complex(0.0526886,0.00000),complex(0.00000,-0.0982404),complex(-0.160817,0.00000),complex(0.00000,0.232998),complex(0.299228,0.00000),complex(0.00000,-0.338538),complex(-0.331049,0.00000),complex(0.00000,0.266174),complex(0.148796,0.00000),complex(0.00000,-3.97849e-08),complex(0.148796,0.00000),complex(0.00000,-0.266174),complex(-0.331049,0.00000),complex(0.00000,0.338538),complex(0.299228,0.00000),complex(0.00000,-0.232998),complex(-0.160817,0.00000),complex(0.00000,0.0982405),complex(0.0526886,0.00000),complex(0.00000,-0.0244140),complex(-0.00949479,0.00000),complex(0.00000,0.00292961),complex(0.000622467,-0.00000)],$
            [complex(-0.00220063,0.00000),complex(0.00000,0.00949480),complex(0.0279541,0.00000),complex(0.00000,-0.0645649),complex(-0.123360,0.00000),complex(0.00000,0.199716),complex(0.276165,0.00000),complex(0.00000,-0.324017),complex(-0.313147,0.00000),complex(0.00000,0.227413),complex(0.0780291,0.00000),complex(0.00000,0.0941066),complex(0.231472,0.00000),complex(0.00000,-0.283884),complex(-0.231472,0.00000),complex(0.00000,0.0941066),complex(-0.0780291,0.00000),complex(0.00000,0.227413),complex(0.313147,0.00000),complex(0.00000,-0.324017),complex(-0.276165,0.00000),complex(0.00000,0.199716),complex(0.123360,0.00000),complex(0.00000,-0.0645649),complex(-0.0279541,0.00000),complex(0.00000,0.00949479),complex(0.00220062,-0.00000)],$
            [complex(0.00622447,0.00000),complex(0.00000,-0.0244141),complex(-0.0645649,0.00000),complex(0.00000,0.131836),complex(0.217779,0.00000),complex(0.00000,-0.294721),complex(-0.321634,0.00000),complex(0.00000,0.264064),complex(0.119691,0.00000),complex(0.00000,0.0677075),complex(0.220700,0.00000),complex(0.00000,-0.266174),complex(-0.178555,0.00000),complex(0.00000,3.79514e-08),complex(-0.178555,0.00000),complex(0.00000,0.266174),complex(0.220700,0.00000),complex(0.00000,-0.0677075),complex(0.119691,0.00000),complex(0.00000,-0.264064),complex(-0.321634,0.00000),complex(0.00000,0.294721),complex(0.217779,0.00000),complex(0.00000,-0.131836),complex(-0.0645650,0.00000),complex(0.00000,0.0244140),complex(0.00622438,-0.00000)],$
            [complex(-0.0149255,0.00000),complex(0.00000,0.0526886),complex(0.123360,0.00000),complex(0.00000,-0.217780),complex(-0.299561,0.00000),complex(0.00000,0.314950),complex(0.225148,0.00000),complex(0.00000,-0.0437250),complex(0.152240,0.00000),complex(0.00000,-0.257653),complex(-0.207086,0.00000),complex(0.00000,0.0277505),complex(-0.167541,0.00000),complex(0.00000,0.251139),complex(0.167541,0.00000),complex(0.00000,0.0277507),complex(0.207086,0.00000),complex(0.00000,-0.257653),complex(-0.152240,0.00000),complex(0.00000,-0.0437251),complex(-0.225148,0.00000),complex(0.00000,0.314950),complex(0.299561,0.00000),complex(0.00000,-0.217779),complex(-0.123360,0.00000),complex(0.00000,0.0526886),complex(0.0149256,-0.00000)],$
            [complex(0.0313081,0.00000),complex(0.00000,-0.0982404),complex(-0.199716,0.00000),complex(0.00000,0.294721),complex(0.314950,0.00000),complex(0.00000,-0.210938),complex(-0.00639448,0.00000),complex(0.00000,-0.191466),complex(-0.254616,0.00000),complex(0.00000,0.134610),complex(-0.0833672,0.00000),complex(0.00000,0.232840),complex(0.195242,0.00000),complex(0.00000,-5.58794e-09),complex(0.195242,0.00000),complex(0.00000,-0.232840),complex(-0.0833672,0.00000),complex(0.00000,-0.134610),complex(-0.254616,0.00000),complex(0.00000,0.191466),complex(-0.00639446,0.00000),complex(0.00000,0.210938),complex(0.314950,0.00000),complex(0.00000,-0.294721),complex(-0.199716,0.00000),complex(0.00000,0.0982404),complex(0.0313080,-0.00000)],$
            [complex(-0.0585721,0.00000),complex(0.00000,0.160817),complex(0.276165,0.00000),complex(0.00000,-0.321634),complex(-0.225148,0.00000),complex(0.00000,0.00639440),complex(-0.202393,0.00000),complex(0.00000,0.245541),complex(0.0845842,0.00000),complex(0.00000,0.144804),complex(0.238053,0.00000),complex(0.00000,-0.108901),complex(0.121755,0.00000),complex(0.00000,-0.234652),complex(-0.121755,0.00000),complex(0.00000,-0.108901),complex(-0.238053,0.00000),complex(0.00000,0.144803),complex(-0.0845843,0.00000),complex(0.00000,0.245540),complex(0.202392,0.00000),complex(0.00000,0.00639458),complex(0.225148,0.00000),complex(0.00000,-0.321634),complex(-0.276165,0.00000),complex(0.00000,0.160817),complex(0.0585721,-0.00000)],$
            [complex(0.0990049,0.00000),complex(0.00000,-0.232998),complex(-0.324017,0.00000),complex(0.00000,0.264064),complex(0.0437250,0.00000),complex(0.00000,0.191466),complex(0.245540,0.00000),complex(0.00000,-0.0673827),complex(0.170063,0.00000),complex(0.00000,-0.225607),complex(-0.0416257,0.00000),complex(0.00000,-0.184076),complex(-0.205803,0.00000),complex(0.00000,-4.84288e-08),complex(-0.205803,0.00000),complex(0.00000,0.184076),complex(-0.0416258,0.00000),complex(0.00000,0.225607),complex(0.170063,0.00000),complex(0.00000,0.0673829),complex(0.245540,0.00000),complex(0.00000,-0.191466),complex(0.0437250,0.00000),complex(0.00000,-0.264064),complex(-0.324017,0.00000),complex(0.00000,0.232998),complex(0.0990049,-0.00000)],$
            [complex(-0.152577,0.00000),complex(0.00000,0.299228),complex(0.313148,0.00000),complex(0.00000,-0.119691),complex(0.152240,0.00000),complex(0.00000,-0.254616),complex(-0.0845843,0.00000),complex(0.00000,-0.170063),complex(-0.219116,0.00000),complex(0.00000,0.00742324),complex(-0.208205,0.00000),complex(0.00000,0.164236),complex(-0.0834643,0.00000),complex(0.00000,0.225199),complex(0.0834643,0.00000),complex(0.00000,0.164236),complex(0.208205,0.00000),complex(0.00000,0.00742321),complex(0.219116,0.00000),complex(0.00000,-0.170063),complex(0.0845842,0.00000),complex(0.00000,-0.254616),complex(-0.152240,0.00000),complex(0.00000,-0.119691),complex(-0.313147,0.00000),complex(0.00000,0.299228),complex(0.152577,-0.00000)],$
            [complex(0.215776,0.00000),complex(0.00000,-0.338538),complex(-0.227413,0.00000),complex(0.00000,-0.0677073),complex(-0.257653,0.00000),complex(0.00000,0.134610),complex(-0.144803,0.00000),complex(0.00000,0.225607),complex(0.00742313,0.00000),complex(0.00000,0.213867),complex(0.138469,0.00000),complex(0.00000,0.126690),complex(0.212466,0.00000),complex(0.00000,-8.94070e-08),complex(0.212466,0.00000),complex(0.00000,-0.126690),complex(0.138469,0.00000),complex(0.00000,-0.213867),complex(0.00742330,0.00000),complex(0.00000,-0.225607),complex(-0.144803,0.00000),complex(0.00000,-0.134610),complex(-0.257653,0.00000),complex(0.00000,0.0677076),complex(-0.227413,0.00000),complex(0.00000,0.338538),complex(0.215776,-0.00000)],$
            [complex(-0.281338,0.00000),complex(0.00000,0.331050),complex(0.0780291,0.00000),complex(0.00000,0.220700),complex(0.207086,0.00000),complex(0.00000,0.0833672),complex(0.238053,0.00000),complex(0.00000,-0.0416259),complex(0.208205,0.00000),complex(0.00000,-0.138469),complex(0.139526,0.00000),complex(0.00000,-0.199192),complex(0.0488860,0.00000),complex(0.00000,-0.219836),complex(-0.0488860,0.00000),complex(0.00000,-0.199192),complex(-0.139526,0.00000),complex(0.00000,-0.138469),complex(-0.208205,0.00000),complex(0.00000,-0.0416258),complex(-0.238053,0.00000),complex(0.00000,0.0833672),complex(-0.207086,0.00000),complex(0.00000,0.220700),complex(-0.0780291,0.00000),complex(0.00000,0.331049),complex(0.281338,-0.00000)],$
            [complex(0.339306,0.00000),complex(0.00000,-0.266174),complex(0.0941066,0.00000),complex(0.00000,-0.266174),complex(-0.0277507,0.00000),complex(0.00000,-0.232840),complex(-0.108901,0.00000),complex(0.00000,-0.184076),complex(-0.164236,0.00000),complex(0.00000,-0.126690),complex(-0.199192,0.00000),complex(0.00000,-0.0644530),complex(-0.216182,0.00000),complex(0.00000,4.47035e-08),complex(-0.216182,0.00000),complex(0.00000,0.0644532),complex(-0.199192,0.00000),complex(0.00000,0.126690),complex(-0.164236,0.00000),complex(0.00000,0.184076),complex(-0.108901,0.00000),complex(0.00000,0.232840),complex(-0.0277506,0.00000),complex(0.00000,0.266174),complex(0.0941067,0.00000),complex(0.00000,0.266174),complex(0.339306,-0.00000)],$
            [complex(-0.379356,0.00000),complex(0.00000,0.148796),complex(-0.231472,0.00000),complex(0.00000,0.178555),complex(-0.167541,0.00000),complex(0.00000,0.195242),complex(-0.121755,0.00000),complex(0.00000,0.205803),complex(-0.0834643,0.00000),complex(0.00000,0.212465),complex(-0.0488861,0.00000),complex(0.00000,0.216182),complex(-0.0161133,0.00000),complex(0.00000,0.217380),complex(0.0161132,0.00000),complex(0.00000,0.216182),complex(0.0488861,0.00000),complex(0.00000,0.212466),complex(0.0834643,0.00000),complex(0.00000,0.205803),complex(0.121755,0.00000),complex(0.00000,0.195242),complex(0.167541,0.00000),complex(0.00000,0.178555),complex(0.231472,0.00000),complex(0.00000,0.148796),complex(0.379356,-0.00000)],$
            [complex(0.393676,0.00000),complex(0.00000,-2.98023e-08),complex(0.283884,0.00000),complex(0.00000,2.98023e-08),complex(0.251139,0.00000),complex(0.00000,-5.96046e-08),complex(0.234652,0.00000),0,complex(0.225199,0.00000),complex(0.00000,-4.47035e-08),complex(0.219836,0.00000),complex(0.00000,4.47035e-08),complex(0.217380,0.00000),complex(0.00000,1.63913e-07),complex(0.217380,0.00000),0,complex(0.219836,0.00000),complex(0.00000,7.45058e-08),complex(0.225199,0.00000),complex(0.00000,5.96046e-08),complex(0.234652,0.00000),complex(0.00000,-5.96046e-08),complex(0.251139,0.00000),complex(0.00000,4.47035e-08),complex(0.283884,0.00000),complex(0.00000,1.49012e-08),complex(0.393676,-0.00000)],$
            [complex(-0.379356,0.00000),complex(0.00000,-0.148796),complex(-0.231472,0.00000),complex(0.00000,-0.178555),complex(-0.167541,0.00000),complex(0.00000,-0.195242),complex(-0.121755,0.00000),complex(0.00000,-0.205803),complex(-0.0834643,0.00000),complex(0.00000,-0.212465),complex(-0.0488860,0.00000),complex(0.00000,-0.216182),complex(-0.0161133,0.00000),complex(0.00000,-0.217380),complex(0.0161133,0.00000),complex(0.00000,-0.216182),complex(0.0488860,0.00000),complex(0.00000,-0.212465),complex(0.0834644,0.00000),complex(0.00000,-0.205803),complex(0.121755,0.00000),complex(0.00000,-0.195242),complex(0.167541,0.00000),complex(0.00000,-0.178555),complex(0.231472,0.00000),complex(0.00000,-0.148796),complex(0.379356,-0.00000)],$
            [complex(0.339306,0.00000),complex(0.00000,0.266174),complex(0.0941066,0.00000),complex(0.00000,0.266174),complex(-0.0277505,0.00000),complex(0.00000,0.232840),complex(-0.108901,0.00000),complex(0.00000,0.184076),complex(-0.164236,0.00000),complex(0.00000,0.126690),complex(-0.199192,0.00000),complex(0.00000,0.0644531),complex(-0.216182,0.00000),complex(0.00000,-4.47035e-08),complex(-0.216182,0.00000),complex(0.00000,-0.0644530),complex(-0.199192,0.00000),complex(0.00000,-0.126690),complex(-0.164236,0.00000),complex(0.00000,-0.184076),complex(-0.108901,0.00000),complex(0.00000,-0.232840),complex(-0.0277505,0.00000),complex(0.00000,-0.266174),complex(0.0941066,0.00000),complex(0.00000,-0.266174),complex(0.339306,-0.00000)],$
            [complex(-0.281338,0.00000),complex(0.00000,-0.331050),complex(0.0780291,0.00000),complex(0.00000,-0.220700),complex(0.207086,0.00000),complex(0.00000,-0.0833672),complex(0.238053,0.00000),complex(0.00000,0.0416258),complex(0.208205,0.00000),complex(0.00000,0.138469),complex(0.139526,0.00000),complex(0.00000,0.199192),complex(0.0488860,0.00000),complex(0.00000,0.219836),complex(-0.0488860,0.00000),complex(0.00000,0.199192),complex(-0.139526,0.00000),complex(0.00000,0.138469),complex(-0.208205,0.00000),complex(0.00000,0.0416258),complex(-0.238053,0.00000),complex(0.00000,-0.0833672),complex(-0.207086,0.00000),complex(0.00000,-0.220700),complex(-0.0780291,0.00000),complex(0.00000,-0.331050),complex(0.281338,0.00000)],$
            [complex(0.215776,0.00000),complex(0.00000,0.338538),complex(-0.227413,0.00000),complex(0.00000,0.0677076),complex(-0.257654,0.00000),complex(0.00000,-0.134610),complex(-0.144804,0.00000),complex(0.00000,-0.225607),complex(0.00742321,0.00000),complex(0.00000,-0.213867),complex(0.138469,0.00000),complex(0.00000,-0.126690),complex(0.212465,0.00000),complex(0.00000,1.49012e-08),complex(0.212465,0.00000),complex(0.00000,0.126690),complex(0.138469,0.00000),complex(0.00000,0.213867),complex(0.00742333,0.00000),complex(0.00000,0.225607),complex(-0.144803,0.00000),complex(0.00000,0.134610),complex(-0.257653,0.00000),complex(0.00000,-0.0677074),complex(-0.227413,0.00000),complex(0.00000,-0.338538),complex(0.215776,-0.00000)],$
            [complex(-0.152577,0.00000),complex(0.00000,-0.299228),complex(0.313147,0.00000),complex(0.00000,0.119691),complex(0.152240,0.00000),complex(0.00000,0.254616),complex(-0.0845843,0.00000),complex(0.00000,0.170063),complex(-0.219116,0.00000),complex(0.00000,-0.00742320),complex(-0.208205,0.00000),complex(0.00000,-0.164236),complex(-0.0834643,0.00000),complex(0.00000,-0.225199),complex(0.0834643,0.00000),complex(0.00000,-0.164236),complex(0.208205,0.00000),complex(0.00000,-0.00742321),complex(0.219116,0.00000),complex(0.00000,0.170063),complex(0.0845842,0.00000),complex(0.00000,0.254616),complex(-0.152240,0.00000),complex(0.00000,0.119691),complex(-0.313147,0.00000),complex(0.00000,-0.299228),complex(0.152577,-0.00000)],$
            [complex(0.0990049,0.00000),complex(0.00000,0.232998),complex(-0.324017,0.00000),complex(0.00000,-0.264064),complex(0.0437250,0.00000),complex(0.00000,-0.191466),complex(0.245540,0.00000),complex(0.00000,0.0673828),complex(0.170063,0.00000),complex(0.00000,0.225607),complex(-0.0416258,0.00000),complex(0.00000,0.184076),complex(-0.205803,0.00000),complex(0.00000,3.35276e-08),complex(-0.205803,0.00000),complex(0.00000,-0.184076),complex(-0.0416259,0.00000),complex(0.00000,-0.225607),complex(0.170063,0.00000),complex(0.00000,-0.0673828),complex(0.245540,0.00000),complex(0.00000,0.191466),complex(0.0437251,0.00000),complex(0.00000,0.264064),complex(-0.324017,0.00000),complex(0.00000,-0.232998),complex(0.0990050,-0.00000)],$
            [complex(-0.0585720,0.00000),complex(0.00000,-0.160817),complex(0.276165,0.00000),complex(0.00000,0.321634),complex(-0.225148,0.00000),complex(0.00000,-0.00639440),complex(-0.202393,0.00000),complex(0.00000,-0.245540),complex(0.0845843,0.00000),complex(0.00000,-0.144803),complex(0.238053,0.00000),complex(0.00000,0.108901),complex(0.121755,0.00000),complex(0.00000,0.234652),complex(-0.121755,0.00000),complex(0.00000,0.108901),complex(-0.238053,0.00000),complex(0.00000,-0.144804),complex(-0.0845842,0.00000),complex(0.00000,-0.245540),complex(0.202393,0.00000),complex(0.00000,-0.00639433),complex(0.225148,0.00000),complex(0.00000,0.321634),complex(-0.276165,0.00000),complex(0.00000,-0.160817),complex(0.0585720,-0.00000)],$
            [complex(0.0313081,0.00000),complex(0.00000,0.0982405),complex(-0.199716,0.00000),complex(0.00000,-0.294721),complex(0.314950,0.00000),complex(0.00000,0.210938),complex(-0.00639442,0.00000),complex(0.00000,0.191466),complex(-0.254616,0.00000),complex(0.00000,-0.134610),complex(-0.0833671,0.00000),complex(0.00000,-0.232840),complex(0.195242,0.00000),complex(0.00000,1.49012e-08),complex(0.195242,0.00000),complex(0.00000,0.232840),complex(-0.0833672,0.00000),complex(0.00000,0.134610),complex(-0.254616,0.00000),complex(0.00000,-0.191466),complex(-0.00639446,0.00000),complex(0.00000,-0.210938),complex(0.314950,-0.00000),complex(0.00000,0.294721),complex(-0.199716,0.00000),complex(0.00000,-0.0982405),complex(0.0313080,-0.00000)],$
            [complex(-0.0149256,0.00000),complex(0.00000,-0.0526886),complex(0.123360,0.00000),complex(0.00000,0.217779),complex(-0.299561,0.00000),complex(0.00000,-0.314950),complex(0.225148,0.00000),complex(0.00000,0.0437250),complex(0.152240,0.00000),complex(0.00000,0.257654),complex(-0.207086,0.00000),complex(0.00000,-0.0277506),complex(-0.167541,0.00000),complex(0.00000,-0.251138),complex(0.167540,0.00000),complex(0.00000,-0.0277506),complex(0.207086,0.00000),complex(0.00000,0.257654),complex(-0.152240,0.00000),complex(0.00000,0.0437251),complex(-0.225148,0.00000),complex(0.00000,-0.314950),complex(0.299561,-0.00000),complex(0.00000,0.217779),complex(-0.123360,0.00000),complex(0.00000,-0.0526886),complex(0.0149256,-0.00000)],$
            [complex(0.00622441,0.00000),complex(0.00000,0.0244141),complex(-0.0645649,0.00000),complex(0.00000,-0.131836),complex(0.217780,0.00000),complex(0.00000,0.294721),complex(-0.321634,0.00000),complex(0.00000,-0.264064),complex(0.119691,0.00000),complex(0.00000,-0.0677076),complex(0.220700,0.00000),complex(0.00000,0.266174),complex(-0.178555,0.00000),complex(0.00000,7.40401e-08),complex(-0.178555,0.00000),complex(0.00000,-0.266174),complex(0.220700,0.00000),complex(0.00000,0.0677075),complex(0.119691,0.00000),complex(0.00000,0.264064),complex(-0.321634,0.00000),complex(0.00000,-0.294721),complex(0.217779,-0.00000),complex(0.00000,0.131836),complex(-0.0645649,0.00000),complex(0.00000,-0.0244140),complex(0.00622442,-0.00000)],$
            [complex(-0.00220070,0.00000),complex(0.00000,-0.00949482),complex(0.0279541,0.00000),complex(0.00000,0.0645649),complex(-0.123360,0.00000),complex(0.00000,-0.199716),complex(0.276165,0.00000),complex(0.00000,0.324017),complex(-0.313147,0.00000),complex(0.00000,-0.227413),complex(0.0780291,0.00000),complex(0.00000,-0.0941066),complex(0.231472,0.00000),complex(0.00000,0.283884),complex(-0.231472,0.00000),complex(0.00000,-0.0941066),complex(-0.0780291,0.00000),complex(0.00000,-0.227413),complex(0.313147,0.00000),complex(0.00000,0.324017),complex(-0.276165,0.00000),complex(0.00000,-0.199716),complex(0.123360,-0.00000),complex(0.00000,0.0645649),complex(-0.0279541,0.00000),complex(0.00000,-0.00949482),complex(0.00220067,-0.00000)],$
            [complex(0.000622488,0.00000),complex(0.00000,0.00292976),complex(-0.00949479,0.00000),complex(0.00000,-0.0244141),complex(0.0526886,0.00000),complex(0.00000,0.0982406),complex(-0.160817,0.00000),complex(0.00000,-0.232998),complex(0.299228,0.00000),complex(0.00000,0.338538),complex(-0.331050,0.00000),complex(0.00000,-0.266174),complex(0.148796,0.00000),complex(0.00000,-1.88593e-08),complex(0.148796,0.00000),complex(0.00000,0.266174),complex(-0.331049,0.00000),complex(0.00000,-0.338538),complex(0.299228,0.00000),complex(0.00000,0.232998),complex(-0.160817,0.00000),complex(0.00000,-0.0982405),complex(0.0526886,-0.00000),complex(0.00000,0.0244140),complex(-0.00949479,0.00000),complex(0.00000,-0.00292971),complex(0.000622559,-0.00000)],$
            [complex(-0.000122045,0.00000),complex(0.00000,-0.000622452),complex(0.00220060,0.00000),complex(0.00000,0.00622432),complex(-0.0149256,0.00000),complex(0.00000,-0.0313082),complex(0.0585720,0.00000),complex(0.00000,0.0990049),complex(-0.152577,0.00000),complex(0.00000,-0.215776),complex(0.281338,0.00000),complex(0.00000,0.339306),complex(-0.379356,0.00000),complex(0.00000,-0.393676),complex(0.379356,0.00000),complex(0.00000,0.339306),complex(-0.281338,0.00000),complex(0.00000,-0.215776),complex(0.152577,0.00000),complex(0.00000,0.0990049),complex(-0.0585721,0.00000),complex(0.00000,-0.0313081),complex(0.0149256,-0.00000),complex(0.00000,0.00622434),complex(-0.00220060,0.00000),complex(0.00000,-0.000622538),complex(0.000122067,-0.00000)]$
            ]
          end

      25: begin
            matrix[325:350,325:350]=[$
            [complex(0.00000,0.000172734),complex(0.000863161,0.00000),complex(0.00000,-0.00299002),complex(-0.00827928,0.00000),complex(0.00000,0.0194163),complex(0.0397920,0.00000),complex(0.00000,-0.0726497),complex(-0.119691,0.00000),complex(0.00000,0.179537),complex(0.246750,0.00000),complex(0.00000,-0.312116),complex(-0.364473,0.00000),complex(0.00000,0.393676),complex(0.393676,0.00000),complex(0.00000,-0.364473),complex(-0.312116,0.00000),complex(0.00000,0.246750),complex(0.179537,0.00000),complex(0.00000,-0.119691),complex(-0.0726498,0.00000),complex(0.00000,0.0397919),complex(0.0194165,0.00000),complex(0.00000,-0.00827922),complex(-0.00299007,0.00000),complex(0.00000,0.000863181),complex(0.000172733,-0.00000)],$
            [complex(0.00000,-0.000863197),complex(-0.00397062,0.00000),complex(0.00000,0.0125584),complex(0.0314611,0.00000),complex(0.00000,-0.0660160),complex(-0.119376,0.00000),complex(0.00000,0.188889),complex(0.263321,0.00000),complex(0.00000,-0.323166),complex(-0.345449,0.00000),complex(0.00000,0.312116),complex(0.218684,0.00000),complex(0.00000,-0.0787352),complex(0.0787352,0.00000),complex(0.00000,-0.218684),complex(-0.312116,0.00000),complex(0.00000,0.345450),complex(0.323166,0.00000),complex(0.00000,-0.263321),complex(-0.188889,0.00000),complex(0.00000,0.119376),complex(0.0660159,0.00000),complex(0.00000,-0.0314610),complex(-0.0125584,0.00000),complex(0.00000,0.00397065),complex(0.000863233,-0.00000)],$
            [complex(0.00000,0.00299005),complex(0.0125584,0.00000),complex(0.00000,-0.0359077),complex(-0.0803041,0.00000),complex(0.00000,0.147973),complex(0.229739,0.00000),complex(0.00000,-0.302000),complex(-0.331698,0.00000),complex(0.00000,0.290236),complex(0.170953,0.00000),complex(0.00000,2.54950e-08),complex(0.168343,0.00000),complex(0.00000,-0.272747),complex(-0.272747,0.00000),complex(0.00000,0.168343),complex(-8.53324e-08,0.00000),complex(0.00000,0.170953),complex(0.290236,0.00000),complex(0.00000,-0.331698),complex(-0.301999,0.00000),complex(0.00000,0.229739),complex(0.147973,0.00000),complex(0.00000,-0.0803041),complex(-0.0359076,0.00000),complex(0.00000,0.0125585),complex(0.00299008,-0.00000)],$
            [complex(0.00000,-0.00827920),complex(-0.0314610,0.00000),complex(0.00000,0.0803040),complex(0.157442,0.00000),complex(0.00000,-0.247775),complex(-0.315293,0.00000),complex(0.00000,0.315089),complex(0.219625,0.00000),complex(0.00000,-0.0449232),complex(0.144062,0.00000),complex(0.00000,-0.260323),complex(-0.243194,0.00000),complex(0.00000,0.0985046),complex(-0.0985046,0.00000),complex(0.00000,0.243194),complex(0.260323,0.00000),complex(0.00000,-0.144062),complex(0.0449232,0.00000),complex(0.00000,-0.219625),complex(-0.315089,0.00000),complex(0.00000,0.315293),complex(0.247775,0.00000),complex(0.00000,-0.157442),complex(-0.0803040,0.00000),complex(0.00000,0.0314610),complex(0.00827921,-0.00000)],$
            [complex(0.00000,0.0194164),complex(0.0660160,0.00000),complex(0.00000,-0.147973),complex(-0.247775,0.00000),complex(0.00000,0.317991),complex(0.300724,0.00000),complex(0.00000,-0.169235),complex(0.0361822,0.00000),complex(0.00000,-0.213901),complex(-0.258877,0.00000),complex(0.00000,0.138753),complex(-0.0712924,0.00000),complex(0.00000,0.231014),complex(0.231014,0.00000),complex(0.00000,-0.0712923),complex(0.138753,0.00000),complex(0.00000,-0.258877),complex(-0.213901,0.00000),complex(0.00000,0.0361823),complex(-0.169235,0.00000),complex(0.00000,0.300724),complex(0.317991,0.00000),complex(0.00000,-0.247775),complex(-0.147973,0.00000),complex(0.00000,0.0660160),complex(0.0194164,-0.00000)],$
            [complex(0.00000,-0.0397921),complex(-0.119376,0.00000),complex(0.00000,0.229739),complex(0.315293,0.00000),complex(0.00000,-0.300725),complex(-0.151572,0.00000),complex(0.00000,-0.0737531),complex(-0.239902,0.00000),complex(0.00000,0.228998),complex(0.0449609,0.00000),complex(0.00000,0.170615),complex(0.243510,0.00000),complex(0.00000,-0.112723),complex(0.112723,0.00000),complex(0.00000,-0.243510),complex(-0.170615,0.00000),complex(0.00000,-0.0449610),complex(-0.228998,0.00000),complex(0.00000,0.239902),complex(0.0737530,0.00000),complex(0.00000,0.151572),complex(0.300724,0.00000),complex(0.00000,-0.315293),complex(-0.229739,0.00000),complex(0.00000,0.119376),complex(0.0397920,-0.00000)],$
            [complex(0.00000,0.0726497),complex(0.188889,0.00000),complex(0.00000,-0.301999),complex(-0.315089,0.00000),complex(0.00000,0.169235),complex(-0.0737532,0.00000),complex(0.00000,0.245830),complex(0.207054,0.00000),complex(0.00000,0.0119454),complex(0.213427,0.00000),complex(0.00000,-0.207666),complex(-3.35276e-08,0.00000),complex(0.00000,-0.205803),complex(-0.205803,0.00000),complex(0.00000,3.16650e-08),complex(-0.207666,0.00000),complex(0.00000,0.213427),complex(0.0119454,0.00000),complex(0.00000,0.207054),complex(0.245830,0.00000),complex(0.00000,-0.0737530),complex(0.169235,0.00000),complex(0.00000,-0.315089),complex(-0.301999,0.00000),complex(0.00000,0.188889),complex(0.0726497,-0.00000)],$
            [complex(0.00000,-0.119691),complex(-0.263321,0.00000),complex(0.00000,0.331698),complex(0.219624,0.00000),complex(0.00000,0.0361822),complex(0.239902,0.00000),complex(0.00000,-0.207054),complex(0.0303835,-0.00000),complex(0.00000,-0.226840),complex(-0.172252,-0.00000),complex(0.00000,-0.0720277),complex(-0.231303,0.00000),complex(0.00000,0.124918),complex(-0.124918,0.00000),complex(0.00000,0.231303),complex(0.0720277,0.00000),complex(0.00000,0.172252),complex(0.226840,0.00000),complex(0.00000,-0.0303835),complex(0.207054,0.00000),complex(0.00000,-0.239902),complex(-0.0361824,0.00000),complex(0.00000,-0.219625),complex(-0.331698,0.00000),complex(0.00000,0.263321),complex(0.119691,-0.00000)],$
            [complex(0.00000,0.179537),complex(0.323166,0.00000),complex(0.00000,-0.290235),complex(-0.0449233,0.00000),complex(0.00000,-0.213901),complex(-0.228998,0.00000),complex(0.00000,-0.0119454),complex(-0.226840,0.00000),complex(0.00000,0.158650),complex(-0.104633,0.00000),complex(0.00000,0.229588),complex(0.0578257,0.00000),complex(0.00000,0.187377),complex(0.187377,0.00000),complex(0.00000,0.0578257),complex(0.229588,0.00000),complex(0.00000,-0.104633),complex(0.158650,0.00000),complex(0.00000,-0.226840),complex(-0.0119454,0.00000),complex(0.00000,-0.228998),complex(-0.213901,0.00000),complex(0.00000,-0.0449232),complex(-0.290236,0.00000),complex(0.00000,0.323166),complex(0.179537,-0.00000)],$
            [complex(0.00000,-0.246750),complex(-0.345450,0.00000),complex(0.00000,0.170953),complex(-0.144062,0.00000),complex(0.00000,0.258877),complex(0.0449610,0.00000),complex(0.00000,0.213427),complex(0.172252,0.00000),complex(0.00000,0.104633),complex(0.226322,0.00000),complex(0.00000,-0.0229284),complex(0.210372,0.00000),complex(-0.00000,-0.136337),complex(0.136337,0.00000),complex(-0.00000,-0.210372),complex(0.0229285,0.00000),complex(-0.00000,-0.226323),complex(-0.104633,0.00000),complex(-0.00000,-0.172252),complex(-0.213427,0.00000),complex(0.00000,-0.0449609),complex(-0.258877,0.00000),complex(0.00000,0.144062),complex(-0.170953,0.00000),complex(0.00000,0.345450),complex(0.246750,-0.00000)],$
            [complex(0.00000,0.312116),complex(0.312116,0.00000),complex(0.00000,5.96046e-08),complex(0.260323,0.00000),complex(0.00000,-0.138753),complex(0.170615,0.00000),complex(0.00000,-0.207666),complex(0.0720277,0.00000),complex(0.00000,-0.229588),complex(-0.0229285,0.00000),complex(0.00000,-0.215447),complex(-0.106441,0.00000),complex(0.00000,-0.172454),complex(-0.172454,0.00000),complex(0.00000,-0.106441),complex(-0.215447,0.00000),complex(0.00000,-0.0229284),complex(-0.229588,0.00000),complex(0.00000,0.0720276),complex(-0.207666,0.00000),complex(0.00000,0.170615),complex(-0.138753,0.00000),complex(0.00000,0.260323),complex(-4.47035e-08,0.00000),complex(0.00000,0.312116),complex(0.312116,-0.00000)],$
            [complex(0.00000,-0.364473),complex(-0.218684,0.00000),complex(0.00000,-0.168343),complex(-0.243193,0.00000),complex(0.00000,-0.0712923),complex(-0.243510,0.00000),complex(0.00000,8.94070e-08),complex(-0.231303,0.00000),complex(0.00000,0.0578258),complex(-0.210372,0.00000),complex(0.00000,0.106441),complex(-0.182301,0.00000),complex(0.00000,0.147681),complex(-0.147681,0.00000),complex(0.00000,0.182301),complex(-0.106441,0.00000),complex(0.00000,0.210372),complex(-0.0578258,0.00000),complex(0.00000,0.231303),complex(-4.47035e-08,0.00000),complex(0.00000,0.243510),complex(0.0712924,0.00000),complex(0.00000,0.243194),complex(0.168343,0.00000),complex(0.00000,0.218684),complex(0.364473,-0.00000)],$
            [complex(0.00000,0.393676),complex(0.0787352,0.00000),complex(0.00000,0.272747),complex(0.0985047,0.00000),complex(0.00000,0.231014),complex(0.112723,0.00000),complex(0.00000,0.205804),complex(0.124918,0.00000),complex(0.00000,0.187377),complex(0.136337,0.00000),complex(0.00000,0.172454),complex(0.147681,0.00000),complex(0.00000,0.159513),complex(0.159513,0.00000),complex(0.00000,0.147681),complex(0.172454,0.00000),complex(0.00000,0.136337),complex(0.187377,0.00000),complex(0.00000,0.124918),complex(0.205803,0.00000),complex(0.00000,0.112723),complex(0.231014,0.00000),complex(0.00000,0.0985047),complex(0.272747,0.00000),complex(0.00000,0.0787353),complex(0.393676,-0.00000)],$
            [complex(0.00000,-0.393676),complex(0.0787353,0.00000),complex(0.00000,-0.272747),complex(0.0985046,0.00000),complex(0.00000,-0.231014),complex(0.112723,0.00000),complex(0.00000,-0.205803),complex(0.124918,0.00000),complex(0.00000,-0.187377),complex(0.136337,0.00000),complex(0.00000,-0.172454),complex(0.147681,0.00000),complex(0.00000,-0.159513),complex(0.159513,0.00000),complex(0.00000,-0.147681),complex(0.172454,0.00000),complex(0.00000,-0.136337),complex(0.187377,0.00000),complex(0.00000,-0.124918),complex(0.205803,0.00000),complex(0.00000,-0.112723),complex(0.231014,0.00000),complex(0.00000,-0.0985046),complex(0.272747,0.00000),complex(0.00000,-0.0787353),complex(0.393676,-0.00000)],$
            [complex(0.00000,0.364474),complex(-0.218684,0.00000),complex(0.00000,0.168343),complex(-0.243194,0.00000),complex(0.00000,0.0712925),complex(-0.243510,0.00000),complex(0.00000,-1.04308e-07),complex(-0.231303,0.00000),complex(0.00000,-0.0578259),complex(-0.210372,0.00000),complex(0.00000,-0.106441),complex(-0.182301,0.00000),complex(0.00000,-0.147681),complex(-0.147681,0.00000),complex(0.00000,-0.182301),complex(-0.106441,0.00000),complex(0.00000,-0.210372),complex(-0.0578258,0.00000),complex(0.00000,-0.231303),complex(8.94070e-08,0.00000),complex(0.00000,-0.243510),complex(0.0712925,0.00000),complex(0.00000,-0.243193),complex(0.168343,0.00000),complex(0.00000,-0.218684),complex(0.364473,-0.00000)],$
            [complex(0.00000,-0.312116),complex(0.312117,0.00000),complex(0.00000,-2.98023e-08),complex(0.260323,0.00000),complex(0.00000,0.138752),complex(0.170615,0.00000),complex(0.00000,0.207666),complex(0.0720277,0.00000),complex(0.00000,0.229588),complex(-0.0229284,0.00000),complex(0.00000,0.215447),complex(-0.106441,0.00000),complex(0.00000,0.172454),complex(-0.172454,0.00000),complex(0.00000,0.106441),complex(-0.215447,0.00000),complex(0.00000,0.0229284),complex(-0.229588,0.00000),complex(0.00000,-0.0720276),complex(-0.207666,0.00000),complex(0.00000,-0.170615),complex(-0.138753,0.00000),complex(0.00000,-0.260323),complex(2.98023e-08,-0.00000),complex(0.00000,-0.312116),complex(0.312116,-0.00000)],$
            [complex(0.00000,0.246750),complex(-0.345449,0.00000),complex(0.00000,-0.170953),complex(-0.144062,0.00000),complex(0.00000,-0.258877),complex(0.0449609,0.00000),complex(0.00000,-0.213427),complex(0.172252,0.00000),complex(0.00000,-0.104633),complex(0.226323,0.00000),complex(0.00000,0.0229285),complex(0.210372,0.00000),complex(0.00000,0.136337),complex(0.136337,0.00000),complex(0.00000,0.210372),complex(0.0229284,0.00000),complex(0.00000,0.226323),complex(-0.104633,0.00000),complex(0.00000,0.172252),complex(-0.213427,0.00000),complex(0.00000,0.0449610),complex(-0.258877,0.00000),complex(0.00000,-0.144063),complex(-0.170953,0.00000),complex(0.00000,-0.345450),complex(0.246750,-0.00000)],$
            [complex(0.00000,-0.179537),complex(0.323166,0.00000),complex(0.00000,0.290236),complex(-0.0449232,0.00000),complex(0.00000,0.213901),complex(-0.228997,0.00000),complex(0.00000,0.0119454),complex(-0.226840,0.00000),complex(0.00000,-0.158650),complex(-0.104633,0.00000),complex(0.00000,-0.229588),complex(0.0578257,0.00000),complex(0.00000,-0.187377),complex(0.187377,0.00000),complex(0.00000,-0.0578258),complex(0.229588,0.00000),complex(0.00000,0.104633),complex(0.158650,0.00000),complex(0.00000,0.226840),complex(-0.0119454,0.00000),complex(0.00000,0.228998),complex(-0.213901,0.00000),complex(0.00000,0.0449232),complex(-0.290236,0.00000),complex(0.00000,-0.323166),complex(0.179537,0.00000)],$
            [complex(0.00000,0.119691),complex(-0.263321,0.00000),complex(0.00000,-0.331698),complex(0.219625,0.00000),complex(0.00000,-0.0361825),complex(0.239902,0.00000),complex(0.00000,0.207054),complex(0.0303835,0.00000),complex(0.00000,0.226840),complex(-0.172252,0.00000),complex(0.00000,0.0720276),complex(-0.231303,0.00000),complex(0.00000,-0.124918),complex(-0.124918,0.00000),complex(0.00000,-0.231303),complex(0.0720277,0.00000),complex(0.00000,-0.172252),complex(0.226840,0.00000),complex(0.00000,0.0303835),complex(0.207054,0.00000),complex(0.00000,0.239902),complex(-0.0361823,0.00000),complex(-0.00000,0.219625),complex(-0.331698,0.00000),complex(0.00000,-0.263321),complex(0.119691,-0.00000)],$
            [complex(0.00000,-0.0726497),complex(0.188889,0.00000),complex(0.00000,0.301999),complex(-0.315089,0.00000),complex(0.00000,-0.169235),complex(-0.0737532,0.00000),complex(0.00000,-0.245830),complex(0.207054,0.00000),complex(0.00000,-0.0119453),complex(0.213427,0.00000),complex(0.00000,0.207666),complex(2.79397e-08,0.00000),complex(0.00000,0.205803),complex(-0.205803,0.00000),complex(0.00000,9.31323e-08),complex(-0.207666,0.00000),complex(0.00000,-0.213427),complex(0.0119454,0.00000),complex(0.00000,-0.207054),complex(0.245830,0.00000),complex(0.00000,0.0737531),complex(0.169235,0.00000),complex(-0.00000,0.315089),complex(-0.301999,0.00000),complex(0.00000,-0.188889),complex(0.0726497,-0.00000)],$
            [complex(0.00000,0.0397920),complex(-0.119376,0.00000),complex(0.00000,-0.229739),complex(0.315293,0.00000),complex(0.00000,0.300724),complex(-0.151572,0.00000),complex(0.00000,0.0737531),complex(-0.239902,0.00000),complex(0.00000,-0.228997),complex(0.0449609,0.00000),complex(0.00000,-0.170615),complex(0.243510,0.00000),complex(0.00000,0.112723),complex(0.112723,0.00000),complex(0.00000,0.243510),complex(-0.170615,0.00000),complex(0.00000,0.0449609),complex(-0.228998,0.00000),complex(0.00000,-0.239902),complex(0.0737531,0.00000),complex(0.00000,-0.151572),complex(0.300724,0.00000),complex(-0.00000,0.315293),complex(-0.229739,0.00000),complex(0.00000,-0.119376),complex(0.0397920,-0.00000)],$
            [complex(0.00000,-0.0194164),complex(0.0660161,0.00000),complex(0.00000,0.147973),complex(-0.247775,0.00000),complex(0.00000,-0.317991),complex(0.300724,0.00000),complex(0.00000,0.169235),complex(0.0361822,0.00000),complex(0.00000,0.213901),complex(-0.258877,0.00000),complex(0.00000,-0.138753),complex(-0.0712925,0.00000),complex(0.00000,-0.231014),complex(0.231014,0.00000),complex(0.00000,0.0712924),complex(0.138753,0.00000),complex(0.00000,0.258877),complex(-0.213901,0.00000),complex(0.00000,-0.0361822),complex(-0.169235,0.00000),complex(0.00000,-0.300724),complex(0.317991,0.00000),complex(-0.00000,0.247775),complex(-0.147973,0.00000),complex(0.00000,-0.0660160),complex(0.0194164,-0.00000)],$
            [complex(0.00000,0.00827919),complex(-0.0314610,0.00000),complex(0.00000,-0.0803040),complex(0.157442,0.00000),complex(0.00000,0.247775),complex(-0.315293,0.00000),complex(0.00000,-0.315089),complex(0.219625,0.00000),complex(0.00000,0.0449231),complex(0.144062,0.00000),complex(0.00000,0.260323),complex(-0.243193,0.00000),complex(0.00000,-0.0985047),complex(-0.0985046,0.00000),complex(0.00000,-0.243194),complex(0.260323,0.00000),complex(0.00000,0.144062),complex(0.0449231,0.00000),complex(0.00000,0.219625),complex(-0.315089,0.00000),complex(0.00000,-0.315293),complex(0.247775,0.00000),complex(-0.00000,0.157442),complex(-0.0803040,0.00000),complex(0.00000,-0.0314610),complex(0.00827919,-0.00000)],$
            [complex(0.00000,-0.00299009),complex(0.0125585,0.00000),complex(0.00000,0.0359076),complex(-0.0803041,0.00000),complex(-0.00000,-0.147973),complex(0.229739,0.00000),complex(-0.00000,0.301999),complex(-0.331698,0.00000),complex(-0.00000,-0.290236),complex(0.170953,0.00000),complex(-0.00000,1.02445e-08),complex(0.168343,0.00000),complex(0.00000,0.272747),complex(-0.272747,0.00000),complex(0.00000,-0.168343),complex(2.37487e-08,0.00000),complex(0.00000,-0.170953),complex(0.290236,0.00000),complex(0.00000,0.331698),complex(-0.301999,0.00000),complex(0.00000,-0.229739),complex(0.147973,0.00000),complex(-0.00000,0.0803041),complex(-0.0359077,0.00000),complex(0.00000,-0.0125584),complex(0.00299008,-0.00000)],$
            [complex(0.00000,0.000863208),complex(-0.00397063,0.00000),complex(0.00000,-0.0125585),complex(0.0314610,0.00000),complex(0.00000,0.0660160),complex(-0.119376,0.00000),complex(0.00000,-0.188889),complex(0.263321,0.00000),complex(0.00000,0.323166),complex(-0.345450,0.00000),complex(0.00000,-0.312116),complex(0.218684,0.00000),complex(0.00000,0.0787353),complex(0.0787353,0.00000),complex(0.00000,0.218684),complex(-0.312116,0.00000),complex(0.00000,-0.345450),complex(0.323166,0.00000),complex(0.00000,0.263321),complex(-0.188889,0.00000),complex(0.00000,-0.119376),complex(0.0660160,0.00000),complex(-0.00000,0.0314610),complex(-0.0125584,0.00000),complex(0.00000,-0.00397064),complex(0.000863175,-0.00000)],$
            [complex(0.00000,-0.000172706),complex(0.000863226,0.00000),complex(0.00000,0.00299006),complex(-0.00827922,0.00000),complex(0.00000,-0.0194164),complex(0.0397920,0.00000),complex(0.00000,0.0726498),complex(-0.119691,0.00000),complex(0.00000,-0.179537),complex(0.246750,0.00000),complex(0.00000,0.312116),complex(-0.364473,0.00000),complex(0.00000,-0.393676),complex(0.393676,0.00000),complex(0.00000,0.364473),complex(-0.312116,0.00000),complex(0.00000,-0.246750),complex(0.179537,0.00000),complex(0.00000,0.119691),complex(-0.0726497,0.00000),complex(0.00000,-0.0397920),complex(0.0194164,0.00000),complex(-0.00000,0.00827920),complex(-0.00299009,0.00000),complex(0.00000,-0.000863198),complex(0.000172728,-0.00000)]$
            ]
          end

      24: begin
            matrix[300:324,300:324]=[$
            [complex(0.000244082,0.00000),complex(0.00000,-0.00119606),complex(-0.00405596,0.00000),complex(0.00000,0.0109836),complex(0.0251667,0.00000),complex(0.00000,-0.0503332),complex(-0.0895687,0.00000),complex(0.00000,0.143629),complex(0.209374,0.00000),complex(0.00000,-0.279165),complex(-0.341906,0.00000),complex(0.00000,0.385722),complex(0.401473,0.00000),complex(0.00000,-0.385722),complex(-0.341906,0.00000),complex(0.00000,0.279165),complex(0.209374,0.00000),complex(0.00000,-0.143629),complex(-0.0895687,0.00000),complex(0.00000,0.0503332),complex(0.0251667,0.00000),complex(0.00000,-0.0109836),complex(-0.00405603,0.00000),complex(0.00000,0.00119606),complex(0.000244150,-0.00000)],$
            [complex(-0.00119608,0.00000),complex(0.00000,0.00537116),complex(0.0165584,0.00000),complex(0.00000,-0.0403564),complex(-0.0821938,0.00000),complex(0.00000,0.143839),complex(0.219397,0.00000),complex(0.00000,-0.293182),complex(-0.341906,0.00000),complex(0.00000,0.341906),complex(0.279165,0.00000),complex(0.00000,-0.157471),complex(-5.06069e-08,0.00000),complex(0.00000,-0.157471),complex(-0.279165,0.00000),complex(0.00000,0.341906),complex(0.341906,0.00000),complex(0.00000,-0.293182),complex(-0.219398,0.00000),complex(0.00000,0.143839),complex(0.0821939,0.00000),complex(0.00000,-0.0403564),complex(-0.0165585,0.00000),complex(0.00000,0.00537110),complex(0.00119609,-0.00000)],$
            [complex(0.00405596,0.00000),complex(0.00000,-0.0165584),complex(-0.0458984,0.00000),complex(0.00000,0.0991703),complex(0.175723,0.00000),complex(0.00000,-0.260555),complex(-0.323484,0.00000),complex(0.00000,0.328528),complex(0.252057,0.00000),complex(0.00000,-0.100823),complex(0.0823213,0.00000),complex(0.00000,-0.232178),complex(-0.289990,0.00000),complex(0.00000,0.232178),complex(0.0823214,0.00000),complex(0.00000,0.100823),complex(0.252057,0.00000),complex(0.00000,-0.328528),complex(-0.323484,0.00000),complex(0.00000,0.260555),complex(0.175723,0.00000),complex(0.00000,-0.0991704),complex(-0.0458984,0.00000),complex(0.00000,0.0165584),complex(0.00405595,-0.00000)],$
            [complex(-0.0109836,0.00000),complex(0.00000,0.0403565),complex(0.0991706,0.00000),complex(0.00000,-0.186035),complex(-0.277461,0.00000),complex(0.00000,0.328925),complex(0.294654,0.00000),complex(0.00000,-0.159628),complex(0.0372312,0.00000),complex(0.00000,-0.210977),complex(-0.273593,0.00000),complex(0.00000,0.188622),complex(1.61555e-08,0.00000),complex(0.00000,0.188622),complex(0.273593,0.00000),complex(0.00000,-0.210977),complex(-0.0372313,0.00000),complex(0.00000,-0.159628),complex(-0.294654,0.00000),complex(0.00000,0.328925),complex(0.277461,0.00000),complex(0.00000,-0.186035),complex(-0.0991703,0.00000),complex(0.00000,0.0403565),complex(0.0109836,-0.00000)],$
            [complex(0.0251666,0.00000),complex(0.00000,-0.0821938),complex(-0.175723,0.00000),complex(0.00000,0.277461),complex(0.328613,0.00000),complex(0.00000,-0.271484),complex(-0.0990550,0.00000),complex(0.00000,-0.117041),complex(-0.255922,0.00000),complex(0.00000,0.227487),complex(0.0464356,0.00000),complex(0.00000,0.164643),complex(0.257048,0.00000),complex(0.00000,-0.164643),complex(0.0464356,0.00000),complex(0.00000,-0.227487),complex(-0.255923,0.00000),complex(0.00000,0.117041),complex(-0.0990550,0.00000),complex(0.00000,0.271484),complex(0.328613,0.00000),complex(0.00000,-0.277461),complex(-0.175723,0.00000),complex(0.00000,0.0821939),complex(0.0251666,-0.00000)],$
            [complex(-0.0503333,0.00000),complex(0.00000,0.143839),complex(0.260555,0.00000),complex(0.00000,-0.328925),complex(-0.271484,0.00000),complex(0.00000,0.0786133),complex(-0.151189,0.00000),complex(0.00000,0.263342),complex(0.170615,0.00000),complex(0.00000,0.0568716),complex(0.232178,0.00000),complex(0.00000,-0.205804),complex(3.53515e-08,0.00000),complex(0.00000,-0.205803),complex(-0.232178,0.00000),complex(0.00000,0.0568717),complex(-0.170615,0.00000),complex(0.00000,0.263342),complex(0.151189,0.00000),complex(0.00000,0.0786132),complex(0.271484,0.00000),complex(0.00000,-0.328925),complex(-0.260555,0.00000),complex(0.00000,0.143839),complex(0.0503333,-0.00000)],$
            [complex(0.0895687,0.00000),complex(0.00000,-0.219398),complex(-0.323484,0.00000),complex(0.00000,0.294654),complex(0.0990549,0.00000),complex(0.00000,0.151189),complex(0.262695,0.00000),complex(0.00000,-0.137024),complex(0.111857,0.00000),complex(0.00000,-0.245020),complex(-0.130473,0.00000),complex(0.00000,-0.115652),complex(-0.240748,0.00000),complex(0.00000,0.115652),complex(-0.130473,0.00000),complex(0.00000,0.245020),complex(0.111857,0.00000),complex(0.00000,0.137024),complex(0.262695,0.00000),complex(0.00000,-0.151189),complex(0.0990549,0.00000),complex(0.00000,-0.294654),complex(-0.323484,0.00000),complex(0.00000,0.219398),complex(0.0895687,-0.00000)],$
            [complex(-0.143629,0.00000),complex(0.00000,0.293182),complex(0.328528,0.00000),complex(0.00000,-0.159628),complex(0.117041,0.00000),complex(0.00000,-0.263342),complex(-0.137023,0.00000),complex(0.00000,-0.128418),complex(-0.242007,0.00000),complex(0.00000,0.0768730),complex(-0.174351,0.00000),complex(0.00000,0.216364),complex(3.42729e-08,0.00000),complex(0.00000,0.216364),complex(0.174351,0.00000),complex(0.00000,0.0768729),complex(0.242007,0.00000),complex(0.00000,-0.128418),complex(0.137023,0.00000),complex(0.00000,-0.263342),complex(-0.117041,0.00000),complex(0.00000,-0.159628),complex(-0.328528,0.00000),complex(0.00000,0.293182),complex(0.143629,-0.00000)],$
            [complex(0.209374,0.00000),complex(0.00000,-0.341906),complex(-0.252057,0.00000),complex(0.00000,-0.0372312),complex(-0.255922,0.00000),complex(0.00000,0.170615),complex(-0.111857,0.00000),complex(0.00000,0.242007),complex(0.0583497,0.00000),complex(0.00000,0.196289),complex(0.185386,0.00000),complex(0.00000,0.0742123),complex(0.231728,0.00000),complex(0.00000,-0.0742123),complex(0.185386,0.00000),complex(0.00000,-0.196289),complex(0.0583496,0.00000),complex(0.00000,-0.242007),complex(-0.111857,0.00000),complex(0.00000,-0.170615),complex(-0.255922,0.00000),complex(0.00000,0.0372313),complex(-0.252057,0.00000),complex(0.00000,0.341906),complex(0.209374,-0.00000)],$
            [complex(-0.279165,0.00000),complex(0.00000,0.341906),complex(0.100823,0.00000),complex(0.00000,0.210977),complex(0.227487,0.00000),complex(0.00000,0.0568717),complex(0.245020,0.00000),complex(0.00000,-0.0768729),complex(0.196289,0.00000),complex(0.00000,-0.172852),complex(0.107644,0.00000),complex(0.00000,-0.222637),complex(-6.07014e-09,0.00000),complex(0.00000,-0.222637),complex(-0.107644,0.00000),complex(0.00000,-0.172852),complex(-0.196289,0.00000),complex(0.00000,-0.0768729),complex(-0.245020,0.00000),complex(0.00000,0.0568717),complex(-0.227487,0.00000),complex(0.00000,0.210977),complex(-0.100823,0.00000),complex(0.00000,0.341906),complex(0.279165,-0.00000)],$
            [complex(0.341906,0.00000),complex(0.00000,-0.279165),complex(0.0823214,0.00000),complex(0.00000,-0.273593),complex(-0.0464356,0.00000),complex(0.00000,-0.232178),complex(-0.130473,0.00000),complex(0.00000,-0.174351),complex(-0.185386,0.00000),complex(0.00000,-0.107644),complex(-0.216797,0.00000),complex(0.00000,-0.0363565),complex(-0.227046,0.00000),complex(0.00000,0.0363564),complex(-0.216797,0.00000),complex(0.00000,0.107644),complex(-0.185386,0.00000),complex(0.00000,0.174351),complex(-0.130473,0.00000),complex(0.00000,0.232178),complex(-0.0464355,0.00000),complex(0.00000,0.273593),complex(0.0823213,0.00000),complex(0.00000,0.279165),complex(0.341906,-0.00000)],$
            [complex(-0.385722,0.00000),complex(0.00000,0.157470),complex(-0.232178,0.00000),complex(0.00000,0.188622),complex(-0.164643,0.00000),complex(0.00000,0.205803),complex(-0.115652,0.00000),complex(0.00000,0.216364),complex(-0.0742123,0.00000),complex(0.00000,0.222637),complex(-0.0363565,0.00000),complex(0.00000,0.225586),complex(-2.45136e-08,0.00000),complex(0.00000,0.225586),complex(0.0363564,0.00000),complex(0.00000,0.222637),complex(0.0742123,0.00000),complex(0.00000,0.216364),complex(0.115652,0.00000),complex(0.00000,0.205803),complex(0.164643,0.00000),complex(0.00000,0.188622),complex(0.232178,0.00000),complex(0.00000,0.157471),complex(0.385722,-0.00000)],$
            [complex(0.401473,0.00000),complex(0.00000,-8.94070e-08),complex(0.289990,0.00000),complex(0.00000,1.19209e-07),complex(0.257048,0.00000),complex(0.00000,1.56462e-07),complex(0.240748,0.00000),complex(0.00000,5.96046e-08),complex(0.231728,0.00000),complex(0.00000,4.47035e-08),complex(0.227046,0.00000),complex(0.00000,8.94070e-08),complex(0.225586,0.00000),0,complex(0.227046,0.00000),complex(0.00000,1.49012e-08),complex(0.231728,0.00000),complex(0.00000,5.96046e-08),complex(0.240748,0.00000),complex(0.00000,1.11759e-07),complex(0.257048,0.00000),complex(0.00000,2.98023e-08),complex(0.289990,0.00000),0,complex(0.401473,-0.00000)],$
            [complex(-0.385722,0.00000),complex(0.00000,-0.157470),complex(-0.232178,0.00000),complex(0.00000,-0.188622),complex(-0.164643,0.00000),complex(0.00000,-0.205803),complex(-0.115652,0.00000),complex(0.00000,-0.216364),complex(-0.0742123,0.00000),complex(0.00000,-0.222637),complex(-0.0363565,0.00000),complex(0.00000,-0.225586),complex(-1.00664e-07,0.00000),complex(0.00000,-0.225586),complex(0.0363564,0.00000),complex(0.00000,-0.222637),complex(0.0742123,0.00000),complex(0.00000,-0.216364),complex(0.115652,0.00000),complex(0.00000,-0.205803),complex(0.164643,0.00000),complex(0.00000,-0.188622),complex(0.232178,0.00000),complex(0.00000,-0.157471),complex(0.385722,-0.00000)],$
            [complex(0.341906,0.00000),complex(0.00000,0.279165),complex(0.0823214,0.00000),complex(0.00000,0.273593),complex(-0.0464355,0.00000),complex(0.00000,0.232178),complex(-0.130473,0.00000),complex(0.00000,0.174351),complex(-0.185386,0.00000),complex(0.00000,0.107644),complex(-0.216797,0.00000),complex(0.00000,0.0363565),complex(-0.227046,0.00000),complex(0.00000,-0.0363564),complex(-0.216797,0.00000),complex(0.00000,-0.107643),complex(-0.185386,0.00000),complex(0.00000,-0.174351),complex(-0.130472,0.00000),complex(0.00000,-0.232178),complex(-0.0464355,0.00000),complex(0.00000,-0.273593),complex(0.0823213,0.00000),complex(0.00000,-0.279165),complex(0.341906,-0.00000)],$
            [complex(-0.279165,0.00000),complex(0.00000,-0.341906),complex(0.100823,0.00000),complex(0.00000,-0.210977),complex(0.227487,0.00000),complex(0.00000,-0.0568717),complex(0.245020,0.00000),complex(0.00000,0.0768729),complex(0.196289,0.00000),complex(0.00000,0.172852),complex(0.107644,0.00000),complex(0.00000,0.222637),complex(5.50675e-08,0.00000),complex(0.00000,0.222637),complex(-0.107644,0.00000),complex(0.00000,0.172852),complex(-0.196289,0.00000),complex(0.00000,0.0768729),complex(-0.245020,0.00000),complex(0.00000,-0.0568717),complex(-0.227487,0.00000),complex(0.00000,-0.210977),complex(-0.100823,0.00000),complex(0.00000,-0.341906),complex(0.279165,-0.00000)],$
            [complex(0.209374,0.00000),complex(0.00000,0.341907),complex(-0.252057,0.00000),complex(0.00000,0.0372313),complex(-0.255922,0.00000),complex(0.00000,-0.170615),complex(-0.111857,0.00000),complex(0.00000,-0.242007),complex(0.0583496,0.00000),complex(0.00000,-0.196289),complex(0.185386,0.00000),complex(0.00000,-0.0742123),complex(0.231728,0.00000),complex(0.00000,0.0742123),complex(0.185386,0.00000),complex(0.00000,0.196289),complex(0.0583495,0.00000),complex(0.00000,0.242007),complex(-0.111857,0.00000),complex(0.00000,0.170615),complex(-0.255923,0.00000),complex(0.00000,-0.0372312),complex(-0.252057,0.00000),complex(0.00000,-0.341906),complex(0.209374,-0.00000)],$
            [complex(-0.143629,0.00000),complex(0.00000,-0.293182),complex(0.328528,0.00000),complex(0.00000,0.159628),complex(0.117041,0.00000),complex(0.00000,0.263342),complex(-0.137024,0.00000),complex(0.00000,0.128418),complex(-0.242007,0.00000),complex(0.00000,-0.0768728),complex(-0.174351,0.00000),complex(0.00000,-0.216364),complex(2.70934e-08,0.00000),complex(0.00000,-0.216364),complex(0.174351,0.00000),complex(0.00000,-0.0768729),complex(0.242007,0.00000),complex(0.00000,0.128418),complex(0.137023,0.00000),complex(0.00000,0.263342),complex(-0.117041,0.00000),complex(0.00000,0.159628),complex(-0.328528,0.00000),complex(0.00000,-0.293182),complex(0.143629,-0.00000)],$
            [complex(0.0895686,0.00000),complex(0.00000,0.219398),complex(-0.323484,0.00000),complex(-0.00000,-0.294654),complex(0.0990550,0.00000),complex(0.00000,-0.151189),complex(0.262695,0.00000),complex(0.00000,0.137024),complex(0.111857,0.00000),complex(0.00000,0.245020),complex(-0.130473,0.00000),complex(0.00000,0.115652),complex(-0.240748,0.00000),complex(0.00000,-0.115651),complex(-0.130473,0.00000),complex(0.00000,-0.245020),complex(0.111857,0.00000),complex(0.00000,-0.137024),complex(0.262695,0.00000),complex(0.00000,0.151189),complex(0.0990550,0.00000),complex(0.00000,0.294654),complex(-0.323484,0.00000),complex(0.00000,-0.219398),complex(0.0895687,-0.00000)],$
            [complex(-0.0503332,0.00000),complex(0.00000,-0.143839),complex(0.260555,0.00000),complex(0.00000,0.328925),complex(-0.271484,0.00000),complex(0.00000,-0.0786133),complex(-0.151189,0.00000),complex(0.00000,-0.263342),complex(0.170615,0.00000),complex(0.00000,-0.0568717),complex(0.232178,0.00000),complex(0.00000,0.205803),complex(4.61709e-08,0.00000),complex(0.00000,0.205803),complex(-0.232178,0.00000),complex(0.00000,-0.0568716),complex(-0.170615,0.00000),complex(0.00000,-0.263342),complex(0.151189,0.00000),complex(0.00000,-0.0786132),complex(0.271484,0.00000),complex(0.00000,0.328925),complex(-0.260555,0.00000),complex(0.00000,-0.143839),complex(0.0503332,-0.00000)],$
            [complex(0.0251667,0.00000),complex(0.00000,0.0821938),complex(-0.175723,0.00000),complex(0.00000,-0.277461),complex(0.328613,0.00000),complex(0.00000,0.271484),complex(-0.0990550,0.00000),complex(0.00000,0.117041),complex(-0.255922,0.00000),complex(0.00000,-0.227487),complex(0.0464356,0.00000),complex(0.00000,-0.164643),complex(0.257048,0.00000),complex(0.00000,0.164643),complex(0.0464355,0.00000),complex(0.00000,0.227487),complex(-0.255923,0.00000),complex(0.00000,-0.117041),complex(-0.0990549,0.00000),complex(0.00000,-0.271484),complex(0.328613,0.00000),complex(0.00000,0.277461),complex(-0.175723,0.00000),complex(0.00000,-0.0821939),complex(0.0251666,-0.00000)],$
            [complex(-0.0109836,0.00000),complex(0.00000,-0.0403565),complex(0.0991704,0.00000),complex(0.00000,0.186035),complex(-0.277461,0.00000),complex(0.00000,-0.328925),complex(0.294654,0.00000),complex(0.00000,0.159628),complex(0.0372312,0.00000),complex(0.00000,0.210977),complex(-0.273593,0.00000),complex(0.00000,-0.188622),complex(-6.43057e-09,0.00000),complex(0.00000,-0.188622),complex(0.273593,0.00000),complex(0.00000,0.210977),complex(-0.0372312,0.00000),complex(0.00000,0.159628),complex(-0.294654,0.00000),complex(0.00000,-0.328925),complex(0.277461,0.00000),complex(0.00000,0.186035),complex(-0.0991703,0.00000),complex(0.00000,-0.0403564),complex(0.0109836,-0.00000)],$
            [complex(0.00405596,0.00000),complex(0.00000,0.0165585),complex(-0.0458984,0.00000),complex(0.00000,-0.0991704),complex(0.175723,0.00000),complex(0.00000,0.260555),complex(-0.323484,0.00000),complex(0.00000,-0.328528),complex(0.252057,0.00000),complex(0.00000,0.100823),complex(0.0823213,0.00000),complex(0.00000,0.232178),complex(-0.289990,0.00000),complex(0.00000,-0.232178),complex(0.0823214,0.00000),complex(0.00000,-0.100823),complex(0.252057,0.00000),complex(0.00000,0.328528),complex(-0.323484,0.00000),complex(0.00000,-0.260555),complex(0.175723,0.00000),complex(0.00000,0.0991704),complex(-0.0458985,0.00000),complex(0.00000,-0.0165585),complex(0.00405592,-0.00000)],$
            [complex(-0.00119608,0.00000),complex(0.00000,-0.00537109),complex(0.0165584,0.00000),complex(0.00000,0.0403563),complex(-0.0821939,0.00000),complex(0.00000,-0.143839),complex(0.219397,0.00000),complex(0.00000,0.293182),complex(-0.341906,0.00000),complex(0.00000,-0.341906),complex(0.279165,0.00000),complex(0.00000,0.157471),complex(-7.38608e-08,0.00000),complex(0.00000,0.157471),complex(-0.279165,0.00000),complex(0.00000,-0.341906),complex(0.341906,0.00000),complex(0.00000,0.293182),complex(-0.219398,0.00000),complex(0.00000,-0.143839),complex(0.0821938,0.00000),complex(0.00000,0.0403565),complex(-0.0165584,0.00000),complex(0.00000,-0.00537111),complex(0.00119606,-0.00000)],$
            [complex(0.000244071,0.00000),complex(0.00000,0.00119611),complex(-0.00405598,0.00000),complex(0.00000,-0.0109837),complex(0.0251667,0.00000),complex(0.00000,0.0503332),complex(-0.0895687,0.00000),complex(0.00000,-0.143629),complex(0.209374,0.00000),complex(0.00000,0.279165),complex(-0.341906,0.00000),complex(0.00000,-0.385722),complex(0.401473,0.00000),complex(0.00000,0.385722),complex(-0.341906,0.00000),complex(0.00000,-0.279165),complex(0.209374,0.00000),complex(0.00000,0.143629),complex(-0.0895686,0.00000),complex(0.00000,-0.0503332),complex(0.0251666,0.00000),complex(0.00000,0.0109836),complex(-0.00405599,0.00000),complex(0.00000,-0.00119606),complex(0.000244159,-0.00000)]$
            ]
          end

      23: begin
            matrix[276:299,276:299]=[$
            [complex(0.00000,-0.000345311),complex(-0.00165588,0.00000),complex(0.00000,0.00549179),complex(0.0145300,0.00000),complex(0.00000,-0.0324900),complex(-0.0633347,0.00000),complex(0.00000,0.109699),complex(0.170953,0.00000),complex(0.00000,-0.241764),complex(-0.312116,0.00000),complex(0.00000,0.369301),complex(0.401473,0.00000),complex(0.00000,-0.401473),complex(-0.369301,0.00000),complex(0.00000,0.312116),complex(0.241764,0.00000),complex(0.00000,-0.170953),complex(-0.109699,0.00000),complex(0.00000,0.0633346),complex(0.0324900,0.00000),complex(0.00000,-0.0145300),complex(-0.00549178,0.00000),complex(0.00000,0.00165587),complex(0.000345300,-0.00000)],$
            [complex(0.00000,0.00165590),complex(0.00725073,0.00000),complex(0.00000,-0.0217573),complex(-0.0515050,0.00000),complex(0.00000,0.101619),complex(0.171680,0.00000),complex(0.00000,-0.251612),complex(-0.320816,0.00000),complex(0.00000,0.352879),complex(0.325404,0.00000),complex(0.00000,-0.231014),complex(-0.0837128,0.00000),complex(0.00000,-0.0837128),complex(-0.231014,0.00000),complex(0.00000,0.325404),complex(0.352879,0.00000),complex(0.00000,-0.320816),complex(-0.251612,0.00000),complex(0.00000,0.171680),complex(0.101619,0.00000),complex(0.00000,-0.0515050),complex(-0.0217573,0.00000),complex(0.00000,0.00725067),complex(0.00165583,-0.00000)],$
            [complex(0.00000,-0.00549176),complex(-0.0217573,0.00000),complex(0.00000,0.0583501),complex(0.121494,0.00000),complex(0.00000,-0.206305),complex(-0.290672,0.00000),complex(0.00000,0.337939),complex(0.311684,0.00000),complex(0.00000,-0.197595),complex(-0.0196226,0.00000),complex(0.00000,-0.162524),complex(-0.277644,0.00000),complex(0.00000,0.277644),complex(0.162524,0.00000),complex(0.00000,0.0196226),complex(0.197595,0.00000),complex(0.00000,-0.311684),complex(-0.337939,0.00000),complex(0.00000,0.290672),complex(0.206305,0.00000),complex(0.00000,-0.121494),complex(-0.0583500,0.00000),complex(0.00000,0.0217573),complex(0.00549176,-0.00000)],$
            [complex(0.00000,0.0145299),complex(0.0515051,0.00000),complex(0.00000,-0.121494),complex(-0.217173,0.00000),complex(0.00000,0.304956),complex(0.332601,0.00000),complex(0.00000,-0.258064),complex(-0.0853076,0.00000),complex(0.00000,-0.120643),complex(-0.259582,0.00000),complex(0.00000,0.254489),complex(0.104940,0.00000),complex(0.00000,0.104940),complex(0.254489,0.00000),complex(0.00000,-0.259582),complex(-0.120643,0.00000),complex(0.00000,-0.0853075),complex(-0.258064,0.00000),complex(0.00000,0.332601),complex(0.304956,0.00000),complex(0.00000,-0.217173),complex(-0.121494,0.00000),complex(0.00000,0.0515050),complex(0.0145300,-0.00000)],$
            [complex(0.00000,-0.0324899),complex(-0.101619,0.00000),complex(0.00000,0.206305),complex(0.304956,0.00000),complex(0.00000,-0.328349),complex(-0.225472,0.00000),complex(0.00000,0.0174863),complex(-0.190753,0.00000),complex(0.00000,0.269766),complex(0.162524,0.00000),complex(0.00000,0.0588677),complex(0.234652,0.00000),complex(0.00000,-0.234652),complex(-0.0588677,0.00000),complex(0.00000,-0.162524),complex(-0.269766,0.00000),complex(0.00000,0.190753),complex(-0.0174863,0.00000),complex(0.00000,0.225472),complex(0.328349,0.00000),complex(0.00000,-0.304956),complex(-0.206305,0.00000),complex(0.00000,0.101619),complex(0.0324899,-0.00000)],$
            [complex(-0.00000,0.0633346),complex(0.171680,0.00000),complex(0.00000,-0.290672),complex(-0.332602,0.00000),complex(-0.00000,0.225472),complex(-0.00448846,0.00000),complex(0.00000,0.217081),complex(0.254421,0.00000),complex(0.00000,-0.0830323),complex(0.154836,0.00000),complex(0.00000,-0.251654),complex(-0.120374,0.00000),complex(0.00000,-0.120374),complex(-0.251654,0.00000),complex(0.00000,0.154836),complex(-0.0830323,0.00000),complex(0.00000,0.254421),complex(0.217081,0.00000),complex(0.00000,-0.00448842),complex(0.225472,0.00000),complex(0.00000,-0.332601),complex(-0.290672,0.00000),complex(0.00000,0.171680),complex(0.0633346,-0.00000)],$
            [complex(0.00000,-0.109699),complex(-0.251612,0.00000),complex(0.00000,0.337938),complex(0.258064,0.00000),complex(0.00000,-0.0174863),complex(0.217081,0.00000),complex(0.00000,-0.246175),complex(-0.0414307,0.00000),complex(0.00000,-0.197081),complex(-0.226925,0.00000),complex(0.00000,0.0174351),complex(-0.208494,0.00000),complex(0.00000,0.208494),complex(-0.0174351,0.00000),complex(0.00000,0.226925),complex(0.197081,0.00000),complex(0.00000,0.0414306),complex(0.246175,0.00000),complex(0.00000,-0.217081),complex(0.0174864,0.00000),complex(0.00000,-0.258064),complex(-0.337938,0.00000),complex(0.00000,0.251612),complex(0.109699,-0.00000)],$
            [complex(0.00000,0.170953),complex(0.320816,0.00000),complex(0.00000,-0.311684),complex(-0.0853076,0.00000),complex(0.00000,-0.190753),complex(-0.254422,0.00000),complex(0.00000,0.0414305),complex(-0.208196,0.00000),complex(0.00000,0.205566),complex(-0.0434955,0.00000),complex(0.00000,0.234946),complex(0.133788,0.00000),complex(0.00000,0.133788),complex(0.234946,0.00000),complex(0.00000,-0.0434954),complex(0.205566,0.00000),complex(0.00000,-0.208196),complex(0.0414307,0.00000),complex(0.00000,-0.254422),complex(-0.190753,0.00000),complex(0.00000,-0.0853075),complex(-0.311684,0.00000),complex(0.00000,0.320816),complex(0.170953,-0.00000)],$
            [complex(0.00000,-0.241764),complex(-0.352880,0.00000),complex(0.00000,0.197595),complex(-0.120643,0.00000),complex(-0.00000,0.269766),complex(0.0830322,0.00000),complex(0.00000,0.197081),complex(0.205566,0.00000),complex(0.00000,0.0628385),complex(0.238024,0.00000),complex(0.00000,-0.0791107),complex(0.189205,0.00000),complex(0.00000,-0.189205),complex(0.0791106,0.00000),complex(0.00000,-0.238024),complex(-0.0628386,0.00000),complex(0.00000,-0.205566),complex(-0.197081,0.00000),complex(0.00000,-0.0830322),complex(-0.269766,0.00000),complex(0.00000,0.120643),complex(-0.197595,0.00000),complex(0.00000,0.352879),complex(0.241764,-0.00000)],$
            [complex(0.00000,0.312116),complex(0.325404,0.00000),complex(0.00000,-0.0196226),complex(0.259582,0.00000),complex(0.00000,-0.162524),complex(0.154836,0.00000),complex(0.00000,-0.226925),complex(0.0434954,0.00000),complex(0.00000,-0.238024),complex(-0.0600765,0.00000),complex(0.00000,-0.208348),complex(-0.146558,0.00000),complex(0.00000,-0.146558),complex(-0.208348,0.00000),complex(0.00000,-0.0600765),complex(-0.238024,0.00000),complex(0.00000,0.0434954),complex(-0.226925,0.00000),complex(0.00000,0.154836),complex(-0.162524,0.00000),complex(0.00000,0.259582),complex(-0.0196226,0.00000),complex(0.00000,0.325404),complex(0.312116,-0.00000)],$
            [complex(0.00000,-0.369301),complex(-0.231014,0.00000),complex(0.00000,-0.162524),complex(-0.254489,0.00000),complex(0.00000,-0.0588678),complex(-0.251654,0.00000),complex(0.00000,0.0174351),complex(-0.234947,0.00000),complex(0.00000,0.0791106),complex(-0.208348,0.00000),complex(0.00000,0.130511),complex(-0.173409,0.00000),complex(0.00000,0.173409),complex(-0.130511,0.00000),complex(0.00000,0.208348),complex(-0.0791106,0.00000),complex(0.00000,0.234947),complex(-0.0174351,0.00000),complex(0.00000,0.251654),complex(0.0588678,0.00000),complex(0.00000,0.254489),complex(0.162524,0.00000),complex(0.00000,0.231014),complex(0.369301,-0.00000)],$
            [complex(0.00000,0.401473),complex(0.0837128,0.00000),complex(0.00000,0.277644),complex(0.104939,0.00000),complex(0.00000,0.234652),complex(0.120374,0.00000),complex(0.00000,0.208494),complex(0.133788,0.00000),complex(0.00000,0.189205),complex(0.146558,0.00000),complex(0.00000,0.173409),complex(0.159513,0.00000),complex(0.00000,0.159513),complex(0.173409,0.00000),complex(0.00000,0.146558),complex(0.189205,0.00000),complex(0.00000,0.133788),complex(0.208494,0.00000),complex(0.00000,0.120374),complex(0.234652,0.00000),complex(0.00000,0.104940),complex(0.277644,0.00000),complex(0.00000,0.0837130),complex(0.401473,-0.00000)],$
            [complex(0.00000,-0.401473),complex(0.0837129,0.00000),complex(0.00000,-0.277644),complex(0.104940,0.00000),complex(0.00000,-0.234652),complex(0.120374,0.00000),complex(0.00000,-0.208494),complex(0.133788,0.00000),complex(0.00000,-0.189205),complex(0.146558,0.00000),complex(0.00000,-0.173409),complex(0.159513,0.00000),complex(0.00000,-0.159513),complex(0.173409,0.00000),complex(0.00000,-0.146558),complex(0.189205,0.00000),complex(0.00000,-0.133788),complex(0.208494,0.00000),complex(0.00000,-0.120374),complex(0.234652,0.00000),complex(0.00000,-0.104940),complex(0.277644,0.00000),complex(0.00000,-0.0837128),complex(0.401473,-0.00000)],$
            [complex(0.00000,0.369301),complex(-0.231014,0.00000),complex(0.00000,0.162524),complex(-0.254489,0.00000),complex(0.00000,0.0588678),complex(-0.251654,0.00000),complex(0.00000,-0.0174351),complex(-0.234947,0.00000),complex(0.00000,-0.0791105),complex(-0.208348,0.00000),complex(0.00000,-0.130511),complex(-0.173409,0.00000),complex(0.00000,-0.173409),complex(-0.130511,0.00000),complex(0.00000,-0.208348),complex(-0.0791107,0.00000),complex(0.00000,-0.234947),complex(-0.0174352,0.00000),complex(0.00000,-0.251654),complex(0.0588676,0.00000),complex(0.00000,-0.254489),complex(0.162524,0.00000),complex(0.00000,-0.231014),complex(0.369301,-0.00000)],$
            [complex(0.00000,-0.312116),complex(0.325404,0.00000),complex(0.00000,0.0196225),complex(0.259582,0.00000),complex(0.00000,0.162524),complex(0.154836,0.00000),complex(0.00000,0.226925),complex(0.0434954,0.00000),complex(0.00000,0.238024),complex(-0.0600764,0.00000),complex(0.00000,0.208348),complex(-0.146558,0.00000),complex(0.00000,0.146558),complex(-0.208348,-0.00000),complex(0.00000,0.0600764),complex(-0.238024,-0.00000),complex(0.00000,-0.0434954),complex(-0.226925,-0.00000),complex(0.00000,-0.154836),complex(-0.162524,0.00000),complex(0.00000,-0.259582),complex(-0.0196226,0.00000),complex(0.00000,-0.325404),complex(0.312116,0.00000)],$
            [complex(0.00000,0.241764),complex(-0.352880,0.00000),complex(0.00000,-0.197595),complex(-0.120643,0.00000),complex(0.00000,-0.269766),complex(0.0830322,0.00000),complex(0.00000,-0.197081),complex(0.205566,0.00000),complex(0.00000,-0.0628386),complex(0.238024,0.00000),complex(0.00000,0.0791107),complex(0.189205,0.00000),complex(0.00000,0.189205),complex(0.0791105,0.00000),complex(0.00000,0.238024),complex(-0.0628387,0.00000),complex(0.00000,0.205566),complex(-0.197081,0.00000),complex(0.00000,0.0830322),complex(-0.269766,0.00000),complex(0.00000,-0.120643),complex(-0.197595,0.00000),complex(0.00000,-0.352879),complex(0.241764,-0.00000)],$
            [complex(0.00000,-0.170953),complex(0.320816,0.00000),complex(0.00000,0.311684),complex(-0.0853075,0.00000),complex(0.00000,0.190753),complex(-0.254422,0.00000),complex(0.00000,-0.0414307),complex(-0.208196,0.00000),complex(0.00000,-0.205566),complex(-0.0434955,0.00000),complex(0.00000,-0.234946),complex(0.133788,0.00000),complex(0.00000,-0.133788),complex(0.234946,0.00000),complex(0.00000,0.0434954),complex(0.205566,0.00000),complex(0.00000,0.208196),complex(0.0414306,0.00000),complex(0.00000,0.254421),complex(-0.190753,0.00000),complex(0.00000,0.0853075),complex(-0.311684,0.00000),complex(0.00000,-0.320816),complex(0.170953,-0.00000)],$
            [complex(0.00000,0.109699),complex(-0.251612,0.00000),complex(0.00000,-0.337938),complex(0.258064,0.00000),complex(0.00000,0.0174864),complex(0.217081,0.00000),complex(0.00000,0.246175),complex(-0.0414306,0.00000),complex(0.00000,0.197081),complex(-0.226925,0.00000),complex(0.00000,-0.0174351),complex(-0.208494,0.00000),complex(0.00000,-0.208494),complex(-0.0174351,0.00000),complex(0.00000,-0.226925),complex(0.197081,0.00000),complex(0.00000,-0.0414306),complex(0.246175,0.00000),complex(0.00000,0.217081),complex(0.0174864,0.00000),complex(0.00000,0.258064),complex(-0.337938,0.00000),complex(0.00000,-0.251612),complex(0.109699,-0.00000)],$
            [complex(0.00000,-0.0633346),complex(0.171680,0.00000),complex(0.00000,0.290672),complex(-0.332601,0.00000),complex(0.00000,-0.225471),complex(-0.00448831,0.00000),complex(0.00000,-0.217081),complex(0.254422,0.00000),complex(0.00000,0.0830324),complex(0.154836,0.00000),complex(0.00000,0.251654),complex(-0.120374,0.00000),complex(0.00000,0.120374),complex(-0.251654,0.00000),complex(0.00000,-0.154836),complex(-0.0830322,0.00000),complex(0.00000,-0.254421),complex(0.217081,0.00000),complex(0.00000,0.00448846),complex(0.225471,0.00000),complex(0.00000,0.332601),complex(-0.290672,0.00000),complex(0.00000,-0.171680),complex(0.0633347,-0.00000)],$
            [complex(0.00000,0.0324900),complex(-0.101619,0.00000),complex(0.00000,-0.206305),complex(0.304956,0.00000),complex(0.00000,0.328349),complex(-0.225472,0.00000),complex(0.00000,-0.0174863),complex(-0.190753,0.00000),complex(0.00000,-0.269766),complex(0.162524,0.00000),complex(0.00000,-0.0588677),complex(0.234652,0.00000),complex(0.00000,0.234652),complex(-0.0588678,0.00000),complex(0.00000,0.162524),complex(-0.269766,0.00000),complex(0.00000,-0.190753),complex(-0.0174864,0.00000),complex(0.00000,-0.225472),complex(0.328349,0.00000),complex(0.00000,0.304956),complex(-0.206305,0.00000),complex(0.00000,-0.101619),complex(0.0324900,-0.00000)],$
            [complex(0.00000,-0.0145300),complex(0.0515050,0.00000),complex(0.00000,0.121494),complex(-0.217173,0.00000),complex(0.00000,-0.304956),complex(0.332601,0.00000),complex(0.00000,0.258064),complex(-0.0853075,0.00000),complex(0.00000,0.120643),complex(-0.259582,0.00000),complex(0.00000,-0.254489),complex(0.104940,0.00000),complex(0.00000,-0.104940),complex(0.254489,0.00000),complex(0.00000,0.259582),complex(-0.120643,0.00000),complex(0.00000,0.0853075),complex(-0.258064,0.00000),complex(0.00000,-0.332601),complex(0.304956,0.00000),complex(0.00000,0.217173),complex(-0.121494,0.00000),complex(0.00000,-0.0515050),complex(0.0145300,-0.00000)],$
            [complex(0.00000,0.00549177),complex(-0.0217574,0.00000),complex(0.00000,-0.0583501),complex(0.121494,0.00000),complex(0.00000,0.206305),complex(-0.290672,0.00000),complex(0.00000,-0.337938),complex(0.311684,0.00000),complex(0.00000,0.197595),complex(-0.0196226,0.00000),complex(0.00000,0.162524),complex(-0.277644,0.00000),complex(0.00000,-0.277644),complex(0.162524,0.00000),complex(0.00000,-0.0196226),complex(0.197595,0.00000),complex(0.00000,0.311684),complex(-0.337939,0.00000),complex(0.00000,-0.290672),complex(0.206305,0.00000),complex(0.00000,0.121494),complex(-0.0583501,0.00000),complex(0.00000,-0.0217573),complex(0.00549184,-0.00000)],$
            [complex(0.00000,-0.00165585),complex(0.00725062,0.00000),complex(0.00000,0.0217573),complex(-0.0515051,0.00000),complex(0.00000,-0.101619),complex(0.171680,0.00000),complex(0.00000,0.251612),complex(-0.320816,0.00000),complex(0.00000,-0.352879),complex(0.325404,0.00000),complex(0.00000,0.231014),complex(-0.0837129,0.00000),complex(0.00000,0.0837127),complex(-0.231014,0.00000),complex(0.00000,-0.325404),complex(0.352879,0.00000),complex(0.00000,0.320816),complex(-0.251612,0.00000),complex(0.00000,-0.171680),complex(0.101619,0.00000),complex(0.00000,0.0515050),complex(-0.0217573,0.00000),complex(0.00000,-0.00725060),complex(0.00165585,-0.00000)],$
            [complex(0.00000,0.000345349),complex(-0.00165587,0.00000),complex(0.00000,-0.00549182),complex(0.0145299,0.00000),complex(0.00000,0.0324900),complex(-0.0633346,0.00000),complex(0.00000,-0.109699),complex(0.170953,0.00000),complex(0.00000,0.241764),complex(-0.312116,0.00000),complex(0.00000,-0.369301),complex(0.401473,0.00000),complex(0.00000,0.401473),complex(-0.369301,0.00000),complex(0.00000,-0.312116),complex(0.241764,0.00000),complex(0.00000,0.170953),complex(-0.109699,0.00000),complex(0.00000,-0.0633346),complex(0.0324900,0.00000),complex(0.00000,0.0145300),complex(-0.00549181,0.00000),complex(0.00000,-0.00165575),complex(0.000345321,-0.00000)]$
            ]
          end

      22: begin
            matrix[253:275,253:275]=[$
            [complex(-0.000488214,0.00000),complex(0.00000,0.00229034),complex(0.00742119,0.00000),complex(0.00000,-0.0191616),complex(-0.0417616,0.00000),complex(0.00000,0.0792371),complex(0.133376,0.00000),complex(0.00000,-0.201645),complex(-0.276114,0.00000),complex(0.00000,0.344375),complex(0.392648,0.00000),complex(0.00000,-0.410107),complex(-0.392648,0.00000),complex(0.00000,0.344375),complex(0.276114,0.00000),complex(0.00000,-0.201645),complex(-0.133376,0.00000),complex(0.00000,0.0792371),complex(0.0417615,0.00000),complex(0.00000,-0.0191616),complex(-0.00742117,0.00000),complex(0.00000,0.00229033),complex(0.000488286,-0.00000)],$
            [complex(0.00229043,0.00000),complex(0.00000,-0.00976570),complex(-0.0284798,0.00000),complex(0.00000,0.0653641),complex(0.124651,0.00000),complex(0.00000,-0.202721),complex(-0.284358,0.00000),complex(0.00000,0.343928),complex(0.353207,0.00000),complex(0.00000,-0.293684),complex(-0.167426,0.00000),complex(0.00000,3.37604e-09),complex(-0.167426,0.00000),complex(0.00000,0.293684),complex(0.353207,0.00000),complex(0.00000,-0.343928),complex(-0.284358,0.00000),complex(0.00000,0.202721),complex(0.124650,0.00000),complex(0.00000,-0.0653640),complex(-0.0284798,0.00000),complex(0.00000,0.00976562),complex(0.00229025,-0.00000)],$
            [complex(-0.00742120,0.00000),complex(0.00000,0.0284798),complex(0.0737304,0.00000),complex(0.00000,-0.147506),complex(-0.239051,0.00000),complex(0.00000,0.318018),complex(0.342244,0.00000),complex(0.00000,-0.278613),complex(-0.127169,0.00000),complex(0.00000,-0.0679746),complex(-0.232509,0.00000),complex(0.00000,0.296814),complex(0.232509,0.00000),complex(0.00000,-0.0679747),complex(0.127169,0.00000),complex(0.00000,-0.278613),complex(-0.342244,0.00000),complex(0.00000,0.318018),complex(0.239051,0.00000),complex(0.00000,-0.147506),complex(-0.0737305,0.00000),complex(0.00000,0.0284797),complex(0.00742119,-0.00000)],$
            [complex(0.0191616,0.00000),complex(0.00000,-0.0653641),complex(-0.147506,0.00000),complex(0.00000,0.250000),complex(0.327769,0.00000),complex(0.00000,-0.323064),complex(-0.203924,0.00000),complex(0.00000,6.79865e-08),complex(-0.197009,0.00000),complex(0.00000,0.280816),complex(0.200112,0.00000),complex(0.00000,-3.72529e-08),complex(0.200112,0.00000),complex(0.00000,-0.280816),complex(-0.197009,0.00000),complex(0.00000,-9.22009e-08),complex(-0.203924,0.00000),complex(0.00000,0.323064),complex(0.327769,0.00000),complex(0.00000,-0.250000),complex(-0.147506,0.00000),complex(0.00000,0.0653640),complex(0.0191615,-0.00000)],$
            [complex(-0.0417615,0.00000),complex(0.00000,0.124650),complex(0.239051,0.00000),complex(0.00000,-0.327769),complex(-0.313965,0.00000),complex(0.00000,0.162128),complex(-0.0701752,0.00000),complex(0.00000,0.247554),complex(0.248584,0.00000),complex(0.00000,-0.0684500),complex(0.160681,0.00000),complex(0.00000,-0.263726),complex(-0.160681,0.00000),complex(0.00000,-0.0684499),complex(-0.248584,0.00000),complex(0.00000,0.247554),complex(0.0701750,0.00000),complex(0.00000,0.162128),complex(0.313965,0.00000),complex(0.00000,-0.327769),complex(-0.239051,0.00000),complex(0.00000,0.124650),complex(0.0417615,-0.00000)],$
            [complex(0.0792371,0.00000),complex(0.00000,-0.202721),complex(-0.318018,0.00000),complex(0.00000,0.323064),complex(0.162128,0.00000),complex(0.00000,0.0917966),complex(0.261364,0.00000),complex(0.00000,-0.208756),complex(0.0238209,0.00000),complex(0.00000,-0.229191),complex(-0.217765,0.00000),complex(0.00000,-1.11759e-08),complex(-0.217765,0.00000),complex(0.00000,0.229191),complex(0.0238210,0.00000),complex(0.00000,0.208756),complex(0.261364,0.00000),complex(0.00000,-0.0917970),complex(0.162128,0.00000),complex(0.00000,-0.323064),complex(-0.318019,0.00000),complex(0.00000,0.202721),complex(0.0792371,-0.00000)],$
            [complex(-0.133376,0.00000),complex(0.00000,0.284358),complex(0.342244,0.00000),complex(0.00000,-0.203924),complex(0.0701750,0.00000),complex(0.00000,-0.261364),complex(-0.192871,0.00000),complex(0.00000,-0.0671773),complex(-0.247655,0.00000),complex(0.00000,0.155071),complex(-0.107810,0.00000),complex(0.00000,0.247728),complex(0.107810,0.00000),complex(0.00000,0.155071),complex(0.247655,0.00000),complex(0.00000,-0.0671773),complex(0.192871,0.00000),complex(0.00000,-0.261364),complex(-0.0701751,0.00000),complex(0.00000,-0.203924),complex(-0.342244,0.00000),complex(0.00000,0.284358),complex(0.133376,-0.00000)],$
            [complex(0.201645,0.00000),complex(0.00000,-0.343928),complex(-0.278613,0.00000),complex(0.00000,1.49012e-08),complex(-0.247554,0.00000),complex(0.00000,0.208756),complex(-0.0671773,0.00000),complex(0.00000,0.250000),complex(0.117675,0.00000),complex(0.00000,0.160109),complex(0.228190,0.00000),complex(0.00000,2.98023e-08),complex(0.228190,0.00000),complex(0.00000,-0.160109),complex(0.117675,0.00000),complex(0.00000,-0.250000),complex(-0.0671772,0.00000),complex(0.00000,-0.208756),complex(-0.247554,0.00000),complex(0.00000,-7.45058e-09),complex(-0.278613,0.00000),complex(0.00000,0.343928),complex(0.201645,-0.00000)],$
            [complex(-0.276114,0.00000),complex(0.00000,0.353206),complex(0.127169,0.00000),complex(0.00000,0.197009),complex(0.248583,0.00000),complex(0.00000,0.0238209),complex(0.247655,0.00000),complex(0.00000,-0.117675),complex(0.174805,0.00000),complex(0.00000,-0.208276),complex(0.0624923,0.00000),complex(0.00000,-0.239328),complex(-0.0624924,0.00000),complex(0.00000,-0.208276),complex(-0.174805,0.00000),complex(0.00000,-0.117675),complex(-0.247655,0.00000),complex(0.00000,0.0238210),complex(-0.248584,0.00000),complex(0.00000,0.197009),complex(-0.127169,0.00000),complex(0.00000,0.353207),complex(0.276114,-0.00000)],$
            [complex(0.344375,0.00000),complex(0.00000,-0.293684),complex(0.0679747,0.00000),complex(0.00000,-0.280816),complex(-0.0684500,0.00000),complex(0.00000,-0.229191),complex(-0.155071,0.00000),complex(0.00000,-0.160109),complex(-0.208276,0.00000),complex(0.00000,-0.0820312),complex(-0.233825,0.00000),complex(0.00000,-1.19209e-07),complex(-0.233825,0.00000),complex(0.00000,0.0820312),complex(-0.208276,0.00000),complex(0.00000,0.160109),complex(-0.155071,0.00000),complex(0.00000,0.229191),complex(-0.0684500,0.00000),complex(0.00000,0.280816),complex(0.0679746,0.00000),complex(0.00000,0.293684),complex(0.344375,-0.00000)],$
            [complex(-0.392648,0.00000),complex(0.00000,0.167426),complex(-0.232509,0.00000),complex(0.00000,0.200112),complex(-0.160681,0.00000),complex(0.00000,0.217765),complex(-0.107810,0.00000),complex(0.00000,0.228190),complex(-0.0624924,0.00000),complex(0.00000,0.233825),complex(-0.0205078,0.00000),complex(0.00000,0.235617),complex(0.0205078,0.00000),complex(0.00000,0.233825),complex(0.0624924,0.00000),complex(0.00000,0.228190),complex(0.107810,0.00000),complex(0.00000,0.217765),complex(0.160681,0.00000),complex(0.00000,0.200112),complex(0.232509,0.00000),complex(0.00000,0.167426),complex(0.392648,-0.00000)],$
            [complex(0.410107,0.00000),complex(0.00000,1.49012e-08),complex(0.296814,0.00000),complex(0.00000,1.04308e-07),complex(0.263726,0.00000),complex(0.00000,2.98023e-08),complex(0.247728,0.00000),complex(0.00000,1.49012e-08),complex(0.239327,0.00000),complex(0.00000,-8.94070e-08),complex(0.235617,0.00000),complex(0.00000,5.96046e-08),complex(0.235617,0.00000),complex(0.00000,1.19209e-07),complex(0.239328,0.00000),complex(0.00000,4.47035e-08),complex(0.247728,0.00000),complex(0.00000,-4.47035e-08),complex(0.263726,-0.00000),complex(0.00000,4.47035e-08),complex(0.296814,0.00000),complex(0.00000,-1.49012e-08),complex(0.410107,-0.00000)],$
            [complex(-0.392648,0.00000),complex(0.00000,-0.167426),complex(-0.232509,0.00000),complex(0.00000,-0.200112),complex(-0.160681,0.00000),complex(0.00000,-0.217765),complex(-0.107810,0.00000),complex(0.00000,-0.228190),complex(-0.0624923,0.00000),complex(0.00000,-0.233825),complex(-0.0205078,0.00000),complex(0.00000,-0.235617),complex(0.0205077,0.00000),complex(0.00000,-0.233825),complex(0.0624923,0.00000),complex(0.00000,-0.228190),complex(0.107810,0.00000),complex(0.00000,-0.217765),complex(0.160681,0.00000),complex(0.00000,-0.200112),complex(0.232509,0.00000),complex(0.00000,-0.167426),complex(0.392648,-0.00000)],$
            [complex(0.344375,0.00000),complex(0.00000,0.293684),complex(0.0679747,0.00000),complex(0.00000,0.280816),complex(-0.0684500,0.00000),complex(0.00000,0.229191),complex(-0.155071,0.00000),complex(0.00000,0.160109),complex(-0.208276,0.00000),complex(0.00000,0.0820313),complex(-0.233825,0.00000),complex(0.00000,-7.45058e-08),complex(-0.233825,0.00000),complex(0.00000,-0.0820311),complex(-0.208276,0.00000),complex(0.00000,-0.160109),complex(-0.155071,0.00000),complex(0.00000,-0.229191),complex(-0.0684499,-0.00000),complex(0.00000,-0.280816),complex(0.0679746,0.00000),complex(0.00000,-0.293684),complex(0.344375,-0.00000)],$
            [complex(-0.276114,0.00000),complex(0.00000,-0.353207),complex(0.127169,0.00000),complex(0.00000,-0.197009),complex(0.248584,0.00000),complex(0.00000,-0.0238209),complex(0.247655,0.00000),complex(0.00000,0.117675),complex(0.174805,0.00000),complex(0.00000,0.208276),complex(0.0624923,0.00000),complex(0.00000,0.239327),complex(-0.0624924,0.00000),complex(0.00000,0.208276),complex(-0.174805,0.00000),complex(0.00000,0.117675),complex(-0.247655,0.00000),complex(0.00000,-0.0238209),complex(-0.248584,0.00000),complex(0.00000,-0.197009),complex(-0.127169,0.00000),complex(0.00000,-0.353207),complex(0.276114,0.00000)],$
            [complex(0.201645,0.00000),complex(0.00000,0.343928),complex(-0.278613,0.00000),complex(0.00000,-7.45058e-08),complex(-0.247554,0.00000),complex(0.00000,-0.208756),complex(-0.0671773,0.00000),complex(0.00000,-0.250000),complex(0.117675,0.00000),complex(0.00000,-0.160108),complex(0.228190,0.00000),complex(0.00000,4.47035e-08),complex(0.228190,0.00000),complex(0.00000,0.160109),complex(0.117675,0.00000),complex(0.00000,0.250000),complex(-0.0671772,0.00000),complex(0.00000,0.208756),complex(-0.247554,0.00000),complex(0.00000,2.23517e-08),complex(-0.278613,0.00000),complex(0.00000,-0.343927),complex(0.201645,-0.00000)],$
            [complex(-0.133376,0.00000),complex(0.00000,-0.284358),complex(0.342244,0.00000),complex(0.00000,0.203924),complex(0.0701751,0.00000),complex(0.00000,0.261364),complex(-0.192871,0.00000),complex(0.00000,0.0671773),complex(-0.247655,0.00000),complex(0.00000,-0.155071),complex(-0.107809,0.00000),complex(0.00000,-0.247728),complex(0.107810,0.00000),complex(0.00000,-0.155071),complex(0.247655,0.00000),complex(0.00000,0.0671772),complex(0.192871,0.00000),complex(0.00000,0.261364),complex(-0.0701751,0.00000),complex(0.00000,0.203924),complex(-0.342244,0.00000),complex(0.00000,-0.284358),complex(0.133376,-0.00000)],$
            [complex(0.0792371,0.00000),complex(0.00000,0.202721),complex(-0.318018,0.00000),complex(0.00000,-0.323064),complex(0.162128,0.00000),complex(0.00000,-0.0917969),complex(0.261364,0.00000),complex(0.00000,0.208756),complex(0.0238209,0.00000),complex(0.00000,0.229191),complex(-0.217765,0.00000),complex(0.00000,-3.72529e-08),complex(-0.217765,0.00000),complex(0.00000,-0.229191),complex(0.0238210,0.00000),complex(0.00000,-0.208756),complex(0.261364,0.00000),complex(0.00000,0.0917969),complex(0.162128,0.00000),complex(0.00000,0.323064),complex(-0.318018,0.00000),complex(0.00000,-0.202721),complex(0.0792371,-0.00000)],$
            [complex(-0.0417616,0.00000),complex(0.00000,-0.124650),complex(0.239051,0.00000),complex(0.00000,0.327769),complex(-0.313965,0.00000),complex(0.00000,-0.162128),complex(-0.0701751,0.00000),complex(0.00000,-0.247554),complex(0.248584,0.00000),complex(0.00000,0.0684500),complex(0.160681,0.00000),complex(0.00000,0.263726),complex(-0.160681,0.00000),complex(0.00000,0.0684500),complex(-0.248584,0.00000),complex(0.00000,-0.247554),complex(0.0701751,0.00000),complex(0.00000,-0.162129),complex(0.313965,0.00000),complex(0.00000,0.327769),complex(-0.239051,0.00000),complex(0.00000,-0.124650),complex(0.0417616,-0.00000)],$
            [complex(0.0191615,0.00000),complex(0.00000,0.0653641),complex(-0.147506,0.00000),complex(0.00000,-0.250000),complex(0.327769,0.00000),complex(0.00000,0.323064),complex(-0.203924,0.00000),complex(0.00000,-5.77420e-08),complex(-0.197009,0.00000),complex(0.00000,-0.280816),complex(0.200112,0.00000),complex(0.00000,3.53903e-08),complex(0.200112,0.00000),complex(0.00000,0.280816),complex(-0.197009,0.00000),complex(0.00000,-1.95578e-08),complex(-0.203924,0.00000),complex(0.00000,-0.323064),complex(0.327769,0.00000),complex(0.00000,0.250000),complex(-0.147506,0.00000),complex(0.00000,-0.0653641),complex(0.0191616,-0.00000)],$
            [complex(-0.00742120,0.00000),complex(0.00000,-0.0284798),complex(0.0737304,0.00000),complex(0.00000,0.147506),complex(-0.239051,0.00000),complex(0.00000,-0.318019),complex(0.342244,0.00000),complex(0.00000,0.278613),complex(-0.127169,0.00000),complex(0.00000,0.0679746),complex(-0.232509,0.00000),complex(0.00000,-0.296814),complex(0.232509,0.00000),complex(0.00000,0.0679747),complex(0.127169,0.00000),complex(0.00000,0.278613),complex(-0.342244,0.00000),complex(0.00000,-0.318019),complex(0.239051,0.00000),complex(0.00000,0.147506),complex(-0.0737305,0.00000),complex(0.00000,-0.0284798),complex(0.00742123,-0.00000)],$
            [complex(0.00229036,0.00000),complex(0.00000,0.00976548),complex(-0.0284798,0.00000),complex(0.00000,-0.0653641),complex(0.124650,0.00000),complex(0.00000,0.202721),complex(-0.284358,0.00000),complex(0.00000,-0.343928),complex(0.353207,0.00000),complex(0.00000,0.293684),complex(-0.167426,0.00000),complex(0.00000,4.06289e-08),complex(-0.167426,0.00000),complex(0.00000,-0.293684),complex(0.353207,0.00000),complex(0.00000,0.343928),complex(-0.284358,0.00000),complex(0.00000,-0.202721),complex(0.124650,0.00000),complex(0.00000,0.0653641),complex(-0.0284798,0.00000),complex(0.00000,-0.00976562),complex(0.00229030,-0.00000)],$
            [complex(-0.000488264,0.00000),complex(0.00000,-0.00229027),complex(0.00742117,0.00000),complex(0.00000,0.0191615),complex(-0.0417616,0.00000),complex(0.00000,-0.0792370),complex(0.133376,0.00000),complex(0.00000,0.201645),complex(-0.276114,0.00000),complex(0.00000,-0.344375),complex(0.392648,0.00000),complex(0.00000,0.410107),complex(-0.392648,0.00000),complex(0.00000,-0.344375),complex(0.276114,0.00000),complex(0.00000,0.201645),complex(-0.133376,0.00000),complex(0.00000,-0.0792371),complex(0.0417616,0.00000),complex(0.00000,0.0191616),complex(-0.00742119,0.00000),complex(0.00000,-0.00229029),complex(0.000488267,-0.00000)]$
            ]
          end

      21: begin
            matrix[231:252,231:252]=[$
            [complex(0.00000,0.000690663),complex(0.00316445,0.00000),complex(0.00000,-0.0100067),complex(-0.0251831,0.00000),complex(0.00000,0.0534217),complex(0.0985046,0.00000),complex(0.00000,-0.160857),complex(-0.235471,0.00000),complex(0.00000,0.311499),complex(0.374375,0.00000),complex(0.00000,-0.410107),complex(-0.410107,0.00000),complex(0.00000,0.374375),complex(0.311499,0.00000),complex(0.00000,-0.235471),complex(-0.160857,0.00000),complex(-0.00000,0.0985046),complex(0.0534216,0.00000),complex(0.00000,-0.0251832),complex(-0.0100068,0.00000),complex(0.00000,0.00316446),complex(0.000690611,-0.00000)],$
            [complex(0.00000,-0.00316447),complex(-0.0131200,0.00000),complex(0.00000,0.0371222),complex(0.0824315,0.00000),complex(0.00000,-0.151548),complex(-0.236450,0.00000),complex(0.00000,0.315918),complex(0.359688,0.00000),complex(0.00000,-0.339873),complex(-0.245086,0.00000),complex(0.00000,0.0894927),complex(-0.0894928,0.00000),complex(0.00000,0.245086),complex(0.339873,0.00000),complex(0.00000,-0.359688),complex(-0.315918,0.00000),complex(-0.00000,0.236450),complex(0.151548,0.00000),complex(0.00000,-0.0824313),complex(-0.0371223,0.00000),complex(0.00000,0.0131200),complex(0.00316445,-0.00000)],$
            [complex(0.00000,0.0100068),complex(0.0371223,0.00000),complex(0.00000,-0.0925315),complex(-0.177256,0.00000),complex(0.00000,0.272797),complex(0.339873,0.00000),complex(0.00000,-0.333006),complex(-0.227487,0.00000),complex(0.00000,0.0429909),complex(-0.155006,0.00000),complex(0.00000,0.283001),complex(0.283001,0.00000),complex(0.00000,-0.155006),complex(0.0429909,0.00000),complex(0.00000,-0.227487),complex(-0.333006,0.00000),complex(-0.00000,0.339873),complex(0.272797,0.00000),complex(0.00000,-0.177256),complex(-0.0925315,0.00000),complex(0.00000,0.0371221),complex(0.0100068,-0.00000)],$
            [complex(0.00000,-0.0251832),complex(-0.0824313,0.00000),complex(0.00000,0.177256),complex(0.283119,0.00000),complex(0.00000,-0.342773),complex(-0.297114,0.00000),complex(0.00000,0.132323),complex(-0.0903941,0.00000),complex(0.00000,0.256243),complex(0.266904,0.00000),complex(0.00000,-0.112453),complex(0.112453,0.00000),complex(0.00000,-0.266904),complex(-0.256243,0.00000),complex(0.00000,0.0903940),complex(-0.132323,0.00000),complex(-0.00000,0.297114),complex(0.342773,0.00000),complex(0.00000,-0.283119),complex(-0.177256,0.00000),complex(0.00000,0.0824313),complex(0.0251831,-0.00000)],$
            [complex(0.00000,0.0534216),complex(0.151548,0.00000),complex(0.00000,-0.272797),complex(-0.342773,0.00000),complex(0.00000,0.282428),complex(0.0827633,0.00000),complex(0.00000,0.155945),complex(0.276979,0.00000),complex(0.00000,-0.189244),complex(0.0435529,0.00000),complex(0.00000,-0.238549),complex(-0.238549,0.00000),complex(0.00000,0.0435529),complex(-0.189244,0.00000),complex(0.00000,0.276979),complex(0.155945,0.00000),complex(-0.00000,0.0827633),complex(0.282428,0.00000),complex(0.00000,-0.342773),complex(-0.272797,0.00000),complex(0.00000,0.151548),complex(0.0534216,-0.00000)],$
            [complex(0.00000,-0.0985045),complex(-0.236450,0.00000),complex(0.00000,0.339873),complex(0.297114,0.00000),complex(0.00000,-0.0827633),complex(0.174705,0.00000),complex(0.00000,-0.274016),complex(-0.127103,-0.00000),complex(0.00000,-0.133203),complex(-0.259819,-0.00000),complex(0.00000,0.129371),complex(-0.129372,-0.00000),complex(0.00000,0.259819),complex(0.133203,0.00000),complex(0.00000,0.127103),complex(0.274016,0.00000),complex(0.00000,-0.174705),complex(0.0827634,0.00000),complex(0.00000,-0.297114),complex(-0.339873,0.00000),complex(0.00000,0.236450),complex(0.0985046,-0.00000)],$
            [complex(0.00000,0.160857),complex(0.315918,0.00000),complex(0.00000,-0.333006),complex(-0.132323,0.00000),complex(0.00000,-0.155944),complex(-0.274016,0.00000),complex(0.00000,0.104961),complex(-0.169821,0.00000),complex(0.00000,0.246047),complex(0.0385711,0.00000),complex(0.00000,0.211263),complex(0.211263,0.00000),complex(0.00000,0.0385710),complex(0.246047,0.00000),complex(0.00000,-0.169821),complex(0.104961,0.00000),complex(0.00000,-0.274016),complex(-0.155945,0.00000),complex(0.00000,-0.132323),complex(-0.333006,0.00000),complex(0.00000,0.315918),complex(0.160857,-0.00000)],$
            [complex(-0.00000,-0.235471),complex(-0.359688,0.00000),complex(0.00000,0.227487),complex(-0.0903941,0.00000),complex(0.00000,0.276979),complex(0.127103,0.00000),complex(0.00000,0.169821),complex(0.237544,0.00000),complex(0.00000,0.00730794),complex(0.237142,0.00000),complex(-0.00000,-0.144320),complex(0.144320,0.00000),complex(-0.00000,-0.237142),complex(-0.00730792,0.00000),complex(-0.00000,-0.237544),complex(-0.169821,0.00000),complex(-0.00000,-0.127103),complex(-0.276979,0.00000),complex(0.00000,0.0903941),complex(-0.227487,0.00000),complex(0.00000,0.359688),complex(0.235471,-0.00000)],$
            [complex(0.00000,0.311499),complex(0.339873,0.00000),complex(0.00000,-0.0429910),complex(0.256243,0.00000),complex(0.00000,-0.189244),complex(0.133203,0.00000),complex(0.00000,-0.246047),complex(0.00730783,0.00000),complex(0.00000,-0.241687),complex(-0.104570,0.00000),complex(0.00000,-0.190917),complex(-0.190917,0.00000),complex(0.00000,-0.104570),complex(-0.241687,0.00000),complex(0.00000,0.00730795),complex(-0.246047,0.00000),complex(0.00000,0.133203),complex(-0.189244,0.00000),complex(0.00000,0.256243),complex(-0.0429909,0.00000),complex(0.00000,0.339873),complex(0.311499,-0.00000)],$
            [complex(0.00000,-0.374375),complex(-0.245086,0.00000),complex(0.00000,-0.155006),complex(-0.266904,0.00000),complex(0.00000,-0.0435530),complex(-0.259819,0.00000),complex(0.00000,0.0385711),complex(-0.237142,0.00000),complex(0.00000,0.104570),complex(-0.203017,0.00000),complex(0.00000,0.158853),complex(-0.158853,0.00000),complex(0.00000,0.203017),complex(-0.104570,0.00000),complex(0.00000,0.237142),complex(-0.0385711,0.00000),complex(0.00000,0.259819),complex(0.0435529,0.00000),complex(0.00000,0.266904),complex(0.155006,0.00000),complex(0.00000,0.245086),complex(0.374375,-0.00000)],$
            [complex(0.00000,0.410107),complex(0.0894928,0.00000),complex(0.00000,0.283001),complex(0.112453,0.00000),complex(0.00000,0.238549),complex(0.129372,0.00000),complex(0.00000,0.211263),complex(0.144320,0.00000),complex(0.00000,0.190917),complex(0.158853,0.00000),complex(0.00000,0.174015),complex(0.174014,0.00000),complex(0.00000,0.158853),complex(0.190917,0.00000),complex(0.00000,0.144320),complex(0.211263,0.00000),complex(0.00000,0.129371),complex(0.238549,0.00000),complex(0.00000,0.112453),complex(0.283001,0.00000),complex(0.00000,0.0894927),complex(0.410107,-0.00000)],$
            [complex(0.00000,-0.410107),complex(0.0894928,0.00000),complex(0.00000,-0.283001),complex(0.112453,0.00000),complex(0.00000,-0.238549),complex(0.129371,0.00000),complex(0.00000,-0.211263),complex(0.144320,0.00000),complex(0.00000,-0.190917),complex(0.158853,0.00000),complex(0.00000,-0.174014),complex(0.174015,0.00000),complex(0.00000,-0.158853),complex(0.190917,0.00000),complex(0.00000,-0.144320),complex(0.211263,0.00000),complex(0.00000,-0.129371),complex(0.238549,0.00000),complex(0.00000,-0.112453),complex(0.283001,0.00000),complex(0.00000,-0.0894928),complex(0.410107,-0.00000)],$
            [complex(0.00000,0.374375),complex(-0.245086,0.00000),complex(0.00000,0.155006),complex(-0.266904,0.00000),complex(0.00000,0.0435529),complex(-0.259819,0.00000),complex(0.00000,-0.0385712),complex(-0.237142,0.00000),complex(0.00000,-0.104570),complex(-0.203017,0.00000),complex(0.00000,-0.158853),complex(-0.158853,0.00000),complex(0.00000,-0.203017),complex(-0.104570,0.00000),complex(0.00000,-0.237142),complex(-0.0385711,0.00000),complex(0.00000,-0.259819),complex(0.0435530,-0.00000),complex(0.00000,-0.266904),complex(0.155006,0.00000),complex(0.00000,-0.245086),complex(0.374375,-0.00000)],$
            [complex(0.00000,-0.311499),complex(0.339873,0.00000),complex(0.00000,0.0429910),complex(0.256243,0.00000),complex(0.00000,0.189244),complex(0.133203,0.00000),complex(0.00000,0.246047),complex(0.00730795,0.00000),complex(0.00000,0.241687),complex(-0.104570,0.00000),complex(0.00000,0.190917),complex(-0.190917,0.00000),complex(0.00000,0.104570),complex(-0.241687,0.00000),complex(0.00000,-0.00730792),complex(-0.246047,0.00000),complex(0.00000,-0.133203),complex(-0.189244,0.00000),complex(0.00000,-0.256243),complex(-0.0429909,-0.00000),complex(0.00000,-0.339873),complex(0.311499,-0.00000)],$
            [complex(0.00000,0.235471),complex(-0.359688,0.00000),complex(0.00000,-0.227487),complex(-0.0903940,0.00000),complex(0.00000,-0.276979),complex(0.127103,0.00000),complex(0.00000,-0.169821),complex(0.237544,0.00000),complex(0.00000,-0.00730801),complex(0.237142,0.00000),complex(0.00000,0.144320),complex(0.144320,0.00000),complex(0.00000,0.237142),complex(-0.00730789,0.00000),complex(0.00000,0.237544),complex(-0.169821,0.00000),complex(-0.00000,0.127103),complex(-0.276979,0.00000),complex(0.00000,-0.0903940),complex(-0.227487,0.00000),complex(0.00000,-0.359688),complex(0.235471,-0.00000)],$
            [complex(0.00000,-0.160857),complex(0.315918,0.00000),complex(0.00000,0.333006),complex(-0.132323,0.00000),complex(0.00000,0.155944),complex(-0.274016,0.00000),complex(0.00000,-0.104961),complex(-0.169821,0.00000),complex(0.00000,-0.246047),complex(0.0385711,0.00000),complex(0.00000,-0.211263),complex(0.211263,0.00000),complex(0.00000,-0.0385711),complex(0.246047,0.00000),complex(0.00000,0.169821),complex(0.104961,0.00000),complex(-0.00000,0.274016),complex(-0.155945,0.00000),complex(0.00000,0.132323),complex(-0.333006,0.00000),complex(0.00000,-0.315918),complex(0.160857,-0.00000)],$
            [complex(0.00000,0.0985046),complex(-0.236450,0.00000),complex(0.00000,-0.339873),complex(0.297114,0.00000),complex(0.00000,0.0827633),complex(0.174705,0.00000),complex(0.00000,0.274016),complex(-0.127103,0.00000),complex(0.00000,0.133203),complex(-0.259819,0.00000),complex(0.00000,-0.129372),complex(-0.129372,0.00000),complex(0.00000,-0.259819),complex(0.133203,0.00000),complex(0.00000,-0.127103),complex(0.274016,0.00000),complex(0.00000,0.174705),complex(0.0827633,0.00000),complex(0.00000,0.297114),complex(-0.339873,0.00000),complex(0.00000,-0.236450),complex(0.0985046,0.00000)],$
            [complex(0.00000,-0.0534216),complex(0.151548,0.00000),complex(0.00000,0.272797),complex(-0.342774,0.00000),complex(0.00000,-0.282428),complex(0.0827635,0.00000),complex(0.00000,-0.155945),complex(0.276979,0.00000),complex(0.00000,0.189244),complex(0.0435529,0.00000),complex(0.00000,0.238549),complex(-0.238549,0.00000),complex(0.00000,-0.0435529),complex(-0.189244,0.00000),complex(0.00000,-0.276979),complex(0.155945,0.00000),complex(0.00000,-0.0827633),complex(0.282428,0.00000),complex(-0.00000,0.342773),complex(-0.272797,0.00000),complex(0.00000,-0.151548),complex(0.0534216,-0.00000)],$
            [complex(0.00000,0.0251832),complex(-0.0824314,0.00000),complex(0.00000,-0.177256),complex(0.283119,0.00000),complex(0.00000,0.342773),complex(-0.297114,0.00000),complex(0.00000,-0.132323),complex(-0.0903941,0.00000),complex(0.00000,-0.256243),complex(0.266904,0.00000),complex(0.00000,0.112453),complex(0.112453,0.00000),complex(0.00000,0.266904),complex(-0.256243,0.00000),complex(0.00000,-0.0903940),complex(-0.132323,0.00000),complex(0.00000,-0.297114),complex(0.342773,0.00000),complex(-0.00000,0.283119),complex(-0.177256,0.00000),complex(0.00000,-0.0824314),complex(0.0251832,-0.00000)],$
            [complex(0.00000,-0.0100068),complex(0.0371222,0.00000),complex(0.00000,0.0925315),complex(-0.177256,0.00000),complex(0.00000,-0.272797),complex(0.339873,0.00000),complex(0.00000,0.333006),complex(-0.227487,0.00000),complex(0.00000,-0.0429909),complex(-0.155006,0.00000),complex(0.00000,-0.283001),complex(0.283001,0.00000),complex(0.00000,0.155006),complex(0.0429909,0.00000),complex(0.00000,0.227487),complex(-0.333006,0.00000),complex(0.00000,-0.339873),complex(0.272797,0.00000),complex(-0.00000,0.177256),complex(-0.0925314,0.00000),complex(0.00000,-0.0371222),complex(0.0100068,-0.00000)],$
            [complex(0.00000,0.00316442),complex(-0.0131201,0.00000),complex(0.00000,-0.0371222),complex(0.0824315,0.00000),complex(0.00000,0.151548),complex(-0.236450,0.00000),complex(0.00000,-0.315918),complex(0.359688,0.00000),complex(0.00000,0.339873),complex(-0.245086,0.00000),complex(0.00000,-0.0894927),complex(-0.0894928,0.00000),complex(0.00000,-0.245086),complex(0.339873,0.00000),complex(0.00000,0.359688),complex(-0.315918,0.00000),complex(0.00000,-0.236450),complex(0.151548,0.00000),complex(-0.00000,0.0824313),complex(-0.0371222,0.00000),complex(0.00000,-0.0131201),complex(0.00316449,-0.00000)],$
            [complex(0.00000,-0.000690665),complex(0.00316445,0.00000),complex(0.00000,0.0100068),complex(-0.0251833,0.00000),complex(0.00000,-0.0534216),complex(0.0985046,0.00000),complex(0.00000,0.160857),complex(-0.235471,0.00000),complex(0.00000,-0.311499),complex(0.374375,0.00000),complex(0.00000,0.410107),complex(-0.410107,0.00000),complex(0.00000,-0.374375),complex(0.311499,0.00000),complex(0.00000,0.235471),complex(-0.160857,0.00000),complex(0.00000,-0.0985046),complex(0.0534216,0.00000),complex(-0.00000,0.0251832),complex(-0.0100068,0.00000),complex(0.00000,-0.00316444),complex(0.000690602,-0.00000)]$
            ]
          end

      20: begin
            matrix[210:230,210:230]=[$
            [complex(+9.766079E-04,+0.000000E+00),complex(+0.000000E+00,-4.367333E-03),complex(-1.346101E-02,+0.000000E+00),complex(+0.000000E+00,+3.297254E-02),complex(+6.797464E-02,+0.000000E+00),complex(+0.000000E+00,-1.215968E-01),complex(-1.922613E-01,+0.000000E+00),complex(+0.000000E+00,+2.718986E-01),complex(+3.466041E-01,+0.000000E+00),complex(+0.000000E+00,-4.002238E-01),complex(-4.197583E-01,+0.000000E+00),complex(+0.000000E+00,+4.002239E-01),complex(+3.466040E-01,+0.000000E+00),complex(+0.000000E+00,-2.718985E-01),complex(-1.922614E-01,+0.000000E+00),complex(+0.000000E+00,+1.215968E-01),complex(+6.797463E-02,+0.000000E+00),complex(+0.000000E+00,-3.297250E-02),complex(-1.346102E-02,+0.000000E+00),complex(+0.000000E+00,+4.367323E-03),complex(+9.766098E-04,-0.000000E+00)],$
            [complex(-4.367294E-03,+0.000000E+00),complex(+0.000000E+00,+1.757806E-02),complex(+4.815942E-02,+0.000000E+00),complex(+0.000000E+00,-1.032204E-01),complex(-1.823951E-01,+0.000000E+00),complex(+0.000000E+00,+2.718986E-01),complex(+3.439275E-01,+0.000000E+00),complex(+0.000000E+00,-3.647903E-01),complex(-3.100121E-01,+0.000000E+00),complex(+0.000000E+00,+1.789856E-01),complex(+7.135282E-08,+0.000000E+00),complex(+0.000000E+00,+1.789856E-01),complex(+3.100120E-01,+0.000000E+00),complex(+0.000000E+00,-3.647902E-01),complex(-3.439276E-01,+0.000000E+00),complex(+0.000000E+00,+2.718985E-01),complex(+1.823951E-01,+0.000000E+00),complex(+0.000000E+00,-1.032203E-01),complex(-4.815943E-02,+0.000000E+00),complex(+0.000000E+00,+1.757816E-02),complex(+4.367352E-03,-0.000000E+00)],$
            [complex(+1.346106E-02,+0.000000E+00),complex(+0.000000E+00,-4.815952E-02),complex(-1.152344E-01,+0.000000E+00),complex(+0.000000E+00,+2.105031E-01),complex(+3.057467E-01,+0.000000E+00),complex(+0.000000E+00,-3.528621E-01),complex(-3.068583E-01,+0.000000E+00),complex(+0.000000E+00,+1.578048E-01),complex(-5.029058E-02,+0.000000E+00),complex(+0.000000E+00,+2.322823E-01),complex(+3.045247E-01,+0.000000E+00),complex(+0.000000E+00,-2.322823E-01),complex(-5.029059E-02,+0.000000E+00),complex(+0.000000E+00,-1.578048E-01),complex(-3.068583E-01,+0.000000E+00),complex(+0.000000E+00,+3.528622E-01),complex(+3.057467E-01,+0.000000E+00),complex(+0.000000E+00,-2.105030E-01),complex(-1.152343E-01,+0.000000E+00),complex(+0.000000E+00,+4.815954E-02),complex(+1.346104E-02,-0.000000E+00)],$
            [complex(-3.297254E-02,+0.000000E+00),complex(+0.000000E+00,+1.032205E-01),complex(+2.105030E-01,+0.000000E+00),complex(+0.000000E+00,-3.144532E-01),complex(-3.462765E-01,+0.000000E+00),complex(+0.000000E+00,+2.520970E-01),complex(+4.555434E-02,+0.000000E+00),complex(+0.000000E+00,+1.771647E-01),complex(+2.874347E-01,+0.000000E+00),complex(+0.000000E+00,-2.133649E-01),complex(-1.065832E-08,+0.000000E+00),complex(+0.000000E+00,-2.133649E-01),complex(-2.874347E-01,+0.000000E+00),complex(+0.000000E+00,+1.771647E-01),complex(-4.555434E-02,+0.000000E+00),complex(+0.000000E+00,+2.520969E-01),complex(+3.462765E-01,+0.000000E+00),complex(+0.000000E+00,-3.144531E-01),complex(-2.105030E-01,+0.000000E+00),complex(+0.000000E+00,+1.032204E-01),complex(+3.297254E-02,-0.000000E+00)],$
            [complex(+6.797474E-02,+0.000000E+00),complex(+0.000000E+00,-1.823952E-01),complex(-3.057468E-01,+0.000000E+00),complex(+0.000000E+00,+3.462763E-01),complex(+2.314453E-01,+0.000000E+00),complex(+0.000000E+00,+8.734811E-03),complex(+2.292573E-01,+0.000000E+00),complex(+0.000000E+00,-2.695312E-01),complex(-9.461068E-02,+0.000000E+00),complex(+0.000000E+00,-1.552459E-01),complex(-2.713720E-01,+0.000000E+00),complex(+0.000000E+00,+1.552458E-01),complex(-9.461080E-02,+0.000000E+00),complex(+0.000000E+00,+2.695312E-01),complex(+2.292573E-01,+0.000000E+00),complex(+0.000000E+00,-8.734662E-03),complex(+2.314453E-01,+0.000000E+00),complex(+0.000000E+00,-3.462764E-01),complex(-3.057467E-01,+0.000000E+00),complex(+0.000000E+00,+1.823951E-01),complex(+6.797466E-02,-0.000000E+00)],$
            [complex(-1.215968E-01,+0.000000E+00),complex(+0.000000E+00,+2.718985E-01),complex(+3.528622E-01,+0.000000E+00),complex(-0.000000E+00,-2.520969E-01),complex(+8.734666E-03,+0.000000E+00),complex(-0.000000E+00,-2.421875E-01),complex(-2.470529E-01,+0.000000E+00),complex(+0.000000E+00,+1.746923E-02),complex(-2.226905E-01,+0.000000E+00),complex(+0.000000E+00,+2.314267E-01),complex(+4.461672E-08,+0.000000E+00),complex(+0.000000E+00,+2.314268E-01),complex(+2.226905E-01,+0.000000E+00),complex(+0.000000E+00,+1.746927E-02),complex(+2.470529E-01,+0.000000E+00),complex(-0.000000E+00,-2.421875E-01),complex(-8.734640E-03,+0.000000E+00),complex(-0.000000E+00,-2.520969E-01),complex(-3.528622E-01,+0.000000E+00),complex(+0.000000E+00,+2.718985E-01),complex(+1.215967E-01,-0.000000E+00)],$
            [complex(+1.922613E-01,+0.000000E+00),complex(+0.000000E+00,-3.439276E-01),complex(-3.068582E-01,+0.000000E+00),complex(+0.000000E+00,+4.555433E-02),complex(-2.292573E-01,+0.000000E+00),complex(+0.000000E+00,+2.470530E-01),complex(-7.812463E-03,+0.000000E+00),complex(+0.000000E+00,+2.430679E-01),complex(+1.830945E-01,+0.000000E+00),complex(+0.000000E+00,+9.757794E-02),complex(+2.558520E-01,+0.000000E+00),complex(+0.000000E+00,-9.757799E-02),complex(+1.830943E-01,+0.000000E+00),complex(+0.000000E+00,-2.430679E-01),complex(-7.812504E-03,+0.000000E+00),complex(+0.000000E+00,-2.470529E-01),complex(-2.292573E-01,+0.000000E+00),complex(+0.000000E+00,-4.555430E-02),complex(-3.068581E-01,+0.000000E+00),complex(+0.000000E+00,+3.439276E-01),complex(+1.922613E-01,-0.000000E+00)],$
            [complex(-2.718983E-01,+0.000000E+00),complex(+0.000000E+00,+3.647900E-01),complex(+1.578048E-01,+0.000000E+00),complex(+0.000000E+00,+1.771647E-01),complex(+2.695312E-01,+0.000000E+00),complex(+0.000000E+00,-1.746941E-02),complex(+2.430680E-01,+0.000000E+00),complex(+0.000000E+00,-1.640625E-01),complex(+1.394264E-01,+0.000000E+00),complex(+0.000000E+00,-2.414935E-01),complex(-4.896881E-09,+0.000000E+00),complex(+0.000000E+00,-2.414934E-01),complex(-1.394263E-01,+0.000000E+00),complex(+0.000000E+00,-1.640625E-01),complex(-2.430680E-01,+0.000000E+00),complex(+0.000000E+00,-1.746932E-02),complex(-2.695313E-01,+0.000000E+00),complex(+0.000000E+00,+1.771648E-01),complex(-1.578048E-01,+0.000000E+00),complex(+0.000000E+00,+3.647903E-01),complex(+2.718985E-01,-0.000000E+00)],$
            [complex(+3.466041E-01,+0.000000E+00),complex(+0.000000E+00,-3.100121E-01),complex(+5.029054E-02,+0.000000E+00),complex(+0.000000E+00,-2.874346E-01),complex(-9.461074E-02,+0.000000E+00),complex(+0.000000E+00,-2.226904E-01),complex(-1.830944E-01,+0.000000E+00),complex(+0.000000E+00,-1.394264E-01),complex(-2.324218E-01,+0.000000E+00),complex(+0.000000E+00,-4.736078E-02),complex(-2.483620E-01,+0.000000E+00),complex(+0.000000E+00,+4.736075E-02),complex(-2.324219E-01,+0.000000E+00),complex(+0.000000E+00,+1.394264E-01),complex(-1.830944E-01,+0.000000E+00),complex(+0.000000E+00,+2.226905E-01),complex(-9.461071E-02,+0.000000E+00),complex(+0.000000E+00,+2.874348E-01),complex(+5.029058E-02,+0.000000E+00),complex(+0.000000E+00,+3.100121E-01),complex(+3.466040E-01,-0.000000E+00)],$
            [complex(-4.002237E-01,+0.000000E+00),complex(+0.000000E+00,+1.789856E-01),complex(-2.322824E-01,+0.000000E+00),complex(+0.000000E+00,+2.133650E-01),complex(-1.552458E-01,+0.000000E+00),complex(+0.000000E+00,+2.314267E-01),complex(-9.757808E-02,+0.000000E+00),complex(+0.000000E+00,+2.414934E-01),complex(-4.736071E-02,+0.000000E+00),complex(+0.000000E+00,+2.460937E-01),complex(+2.203462E-08,+0.000000E+00),complex(+0.000000E+00,+2.460937E-01),complex(+4.736078E-02,+0.000000E+00),complex(+0.000000E+00,+2.414934E-01),complex(+9.757807E-02,+0.000000E+00),complex(+0.000000E+00,+2.314268E-01),complex(+1.552458E-01,+0.000000E+00),complex(+0.000000E+00,+2.133649E-01),complex(+2.322824E-01,+0.000000E+00),complex(+0.000000E+00,+1.789856E-01),complex(+4.002238E-01,-0.000000E+00)],$
            [complex(+4.197584E-01,+0.000000E+00),complex(+0.000000E+00,+1.490116E-07),complex(+3.045246E-01,+0.000000E+00),complex(+0.000000E+00,+4.470348E-08),complex(+2.713720E-01,+0.000000E+00),complex(+0.000000E+00,+4.470348E-08),complex(+2.558520E-01,+0.000000E+00),0,complex(+2.483619E-01,+0.000000E+00),complex(+0.000000E+00,+2.980232E-08),complex(+2.460937E-01,+0.000000E+00),complex(+0.000000E+00,-2.980232E-08),complex(+2.483619E-01,+0.000000E+00),complex(+0.000000E+00,+2.980232E-08),complex(+2.558519E-01,+0.000000E+00),0,complex(+2.713719E-01,+0.000000E+00),complex(+0.000000E+00,-2.980232E-08),complex(+3.045248E-01,+0.000000E+00),complex(+0.000000E+00,+8.940697E-08),complex(+4.197583E-01,-0.000000E+00)],$
            [complex(-4.002239E-01,+0.000000E+00),complex(+0.000000E+00,-1.789855E-01),complex(-2.322823E-01,+0.000000E+00),complex(+0.000000E+00,-2.133649E-01),complex(-1.552458E-01,+0.000000E+00),complex(+0.000000E+00,-2.314267E-01),complex(-9.757812E-02,+0.000000E+00),complex(+0.000000E+00,-2.414934E-01),complex(-4.736073E-02,+0.000000E+00),complex(+0.000000E+00,-2.460937E-01),complex(+1.356904E-08,+0.000000E+00),complex(+0.000000E+00,-2.460938E-01),complex(+4.736075E-02,+0.000000E+00),complex(+0.000000E+00,-2.414934E-01),complex(+9.757813E-02,+0.000000E+00),complex(+0.000000E+00,-2.314267E-01),complex(+1.552458E-01,+0.000000E+00),complex(+0.000000E+00,-2.133650E-01),complex(+2.322823E-01,+0.000000E+00),complex(+0.000000E+00,-1.789856E-01),complex(+4.002238E-01,-0.000000E+00)],$
            [complex(+3.466041E-01,+0.000000E+00),complex(+0.000000E+00,+3.100122E-01),complex(+5.029061E-02,+0.000000E+00),complex(+0.000000E+00,+2.874346E-01),complex(-9.461063E-02,+0.000000E+00),complex(+0.000000E+00,+2.226905E-01),complex(-1.830944E-01,+0.000000E+00),complex(+0.000000E+00,+1.394264E-01),complex(-2.324219E-01,+0.000000E+00),complex(+0.000000E+00,+4.736082E-02),complex(-2.483619E-01,+0.000000E+00),complex(+0.000000E+00,-4.736091E-02),complex(-2.324219E-01,+0.000000E+00),complex(+0.000000E+00,-1.394263E-01),complex(-1.830944E-01,+0.000000E+00),complex(+0.000000E+00,-2.226905E-01),complex(-9.461068E-02,+0.000000E+00),complex(+0.000000E+00,-2.874347E-01),complex(+5.029055E-02,+0.000000E+00),complex(+0.000000E+00,-3.100121E-01),complex(+3.466040E-01,-0.000000E+00)],$
            [complex(-2.718986E-01,+0.000000E+00),complex(+0.000000E+00,-3.647902E-01),complex(+1.578047E-01,+0.000000E+00),complex(+0.000000E+00,-1.771647E-01),complex(+2.695314E-01,+0.000000E+00),complex(+0.000000E+00,+1.746935E-02),complex(+2.430680E-01,+0.000000E+00),complex(+0.000000E+00,+1.640625E-01),complex(+1.394263E-01,+0.000000E+00),complex(+0.000000E+00,+2.414935E-01),complex(+1.623776E-08,+0.000000E+00),complex(+0.000000E+00,+2.414935E-01),complex(-1.394263E-01,+0.000000E+00),complex(+0.000000E+00,+1.640624E-01),complex(-2.430680E-01,+0.000000E+00),complex(+0.000000E+00,+1.746926E-02),complex(-2.695312E-01,+0.000000E+00),complex(+0.000000E+00,-1.771647E-01),complex(-1.578048E-01,+0.000000E+00),complex(+0.000000E+00,-3.647903E-01),complex(+2.718985E-01,+0.000000E+00)],$
            [complex(+1.922613E-01,+0.000000E+00),complex(+0.000000E+00,+3.439275E-01),complex(-3.068582E-01,+0.000000E+00),complex(+0.000000E+00,-4.555446E-02),complex(-2.292572E-01,+0.000000E+00),complex(+0.000000E+00,-2.470529E-01),complex(-7.812504E-03,+0.000000E+00),complex(+0.000000E+00,-2.430679E-01),complex(+1.830943E-01,+0.000000E+00),complex(+0.000000E+00,-9.757810E-02),complex(+2.558520E-01,+0.000000E+00),complex(+0.000000E+00,+9.757826E-02),complex(+1.830944E-01,+0.000000E+00),complex(-0.000000E+00,+2.430680E-01),complex(-7.812522E-03,+0.000000E+00),complex(+0.000000E+00,+2.470530E-01),complex(-2.292572E-01,+0.000000E+00),complex(+0.000000E+00,+4.555430E-02),complex(-3.068583E-01,+0.000000E+00),complex(+0.000000E+00,-3.439276E-01),complex(+1.922614E-01,-0.000000E+00)],$
            [complex(-1.215968E-01,-0.000000E+00),complex(+0.000000E+00,-2.718986E-01),complex(+3.528623E-01,+0.000000E+00),complex(+0.000000E+00,+2.520969E-01),complex(+8.734636E-03,+0.000000E+00),complex(+0.000000E+00,+2.421875E-01),complex(-2.470530E-01,-0.000000E+00),complex(+0.000000E+00,-1.746923E-02),complex(-2.226906E-01,-0.000000E+00),complex(+0.000000E+00,-2.314267E-01),complex(+7.702955E-08,+0.000000E+00),complex(+0.000000E+00,-2.314268E-01),complex(+2.226906E-01,+0.000000E+00),complex(+0.000000E+00,-1.746928E-02),complex(+2.470530E-01,+0.000000E+00),complex(+0.000000E+00,+2.421875E-01),complex(-8.734655E-03,+0.000000E+00),complex(+0.000000E+00,+2.520969E-01),complex(-3.528623E-01,+0.000000E+00),complex(+0.000000E+00,-2.718986E-01),complex(+1.215967E-01,+0.000000E+00)],$
            [complex(+6.797461E-02,+0.000000E+00),complex(+0.000000E+00,+1.823952E-01),complex(-3.057467E-01,+0.000000E+00),complex(+0.000000E+00,-3.462765E-01),complex(+2.314453E-01,+0.000000E+00),complex(+0.000000E+00,-8.734640E-03),complex(+2.292572E-01,+0.000000E+00),complex(+0.000000E+00,+2.695312E-01),complex(-9.461067E-02,+0.000000E+00),complex(+0.000000E+00,+1.552459E-01),complex(-2.713719E-01,+0.000000E+00),complex(+0.000000E+00,-1.552459E-01),complex(-9.461068E-02,+0.000000E+00),complex(+0.000000E+00,-2.695312E-01),complex(+2.292573E-01,+0.000000E+00),complex(+0.000000E+00,+8.734610E-03),complex(+2.314453E-01,+0.000000E+00),complex(+0.000000E+00,+3.462765E-01),complex(-3.057468E-01,+0.000000E+00),complex(+0.000000E+00,-1.823951E-01),complex(+6.797467E-02,-0.000000E+00)],$
            [complex(-3.297258E-02,+0.000000E+00),complex(+0.000000E+00,-1.032203E-01),complex(+2.105031E-01,+0.000000E+00),complex(+0.000000E+00,+3.144532E-01),complex(-3.462766E-01,+0.000000E+00),complex(+0.000000E+00,-2.520970E-01),complex(+4.555432E-02,+0.000000E+00),complex(+0.000000E+00,-1.771646E-01),complex(+2.874348E-01,+0.000000E+00),complex(+0.000000E+00,+2.133650E-01),complex(-1.063869E-07,+0.000000E+00),complex(+0.000000E+00,+2.133648E-01),complex(-2.874346E-01,+0.000000E+00),complex(+0.000000E+00,-1.771647E-01),complex(-4.555424E-02,+0.000000E+00),complex(+0.000000E+00,-2.520968E-01),complex(+3.462765E-01,+0.000000E+00),complex(+0.000000E+00,+3.144531E-01),complex(-2.105030E-01,+0.000000E+00),complex(+0.000000E+00,-1.032204E-01),complex(+3.297259E-02,-0.000000E+00)],$
            [complex(+1.346107E-02,+0.000000E+00),complex(+0.000000E+00,+4.815948E-02),complex(-1.152343E-01,+0.000000E+00),complex(+0.000000E+00,-2.105030E-01),complex(+3.057466E-01,+0.000000E+00),complex(+0.000000E+00,+3.528622E-01),complex(-3.068583E-01,+0.000000E+00),complex(+0.000000E+00,-1.578048E-01),complex(-5.029061E-02,+0.000000E+00),complex(+0.000000E+00,-2.322824E-01),complex(+3.045247E-01,+0.000000E+00),complex(+0.000000E+00,+2.322823E-01),complex(-5.029062E-02,+0.000000E+00),complex(+0.000000E+00,+1.578047E-01),complex(-3.068583E-01,+0.000000E+00),complex(+0.000000E+00,-3.528621E-01),complex(+3.057467E-01,+0.000000E+00),complex(+0.000000E+00,+2.105030E-01),complex(-1.152343E-01,+0.000000E+00),complex(+0.000000E+00,-4.815945E-02),complex(+1.346103E-02,-0.000000E+00)],$
            [complex(-4.367277E-03,+0.000000E+00),complex(+0.000000E+00,-1.757822E-02),complex(+4.815949E-02,+0.000000E+00),complex(+0.000000E+00,+1.032204E-01),complex(-1.823952E-01,+0.000000E+00),complex(+0.000000E+00,-2.718987E-01),complex(+3.439276E-01,+0.000000E+00),complex(+0.000000E+00,+3.647901E-01),complex(-3.100121E-01,+0.000000E+00),complex(+0.000000E+00,-1.789855E-01),complex(+3.433274E-08,+0.000000E+00),complex(+0.000000E+00,-1.789856E-01),complex(+3.100120E-01,+0.000000E+00),complex(+0.000000E+00,+3.647901E-01),complex(-3.439274E-01,+0.000000E+00),complex(+0.000000E+00,-2.718985E-01),complex(+1.823951E-01,+0.000000E+00),complex(+0.000000E+00,+1.032204E-01),complex(-4.815942E-02,+0.000000E+00),complex(+0.000000E+00,-1.757808E-02),complex(+4.367341E-03,-0.000000E+00)],$
            [complex(+9.766088E-04,+0.000000E+00),complex(+0.000000E+00,+4.367339E-03),complex(-1.346097E-02,+0.000000E+00),complex(+0.000000E+00,-3.297256E-02),complex(+6.797460E-02,+0.000000E+00),complex(+0.000000E+00,+1.215967E-01),complex(-1.922613E-01,+0.000000E+00),complex(+0.000000E+00,-2.718985E-01),complex(+3.466040E-01,+0.000000E+00),complex(+0.000000E+00,+4.002239E-01),complex(-4.197584E-01,+0.000000E+00),complex(+0.000000E+00,-4.002238E-01),complex(+3.466040E-01,+0.000000E+00),complex(+0.000000E+00,+2.718985E-01),complex(-1.922613E-01,+0.000000E+00),complex(+0.000000E+00,-1.215967E-01),complex(+6.797465E-02,+0.000000E+00),complex(+0.000000E+00,+3.297253E-02),complex(-1.346099E-02,+0.000000E+00),complex(+0.000000E+00,-4.367317E-03),complex(+9.766067E-04,-0.000000E+00)]$
            ]
          end

      19: begin
            matrix[190:209,190:209]=[$
            [complex(+0.000000E+00,-1.381070E-03),complex(-6.019983E-03,+0.000000E+00),complex(+0.000000E+00,+1.805984E-02),complex(+4.299090E-02,+0.000000E+00),complex(+0.000000E+00,-8.598191E-02),complex(-1.489250E-01,+0.000000E+00),complex(+0.000000E+00,+2.274867E-01),complex(+3.100120E-01,+0.000000E+00),complex(+0.000000E+00,-3.796857E-01),complex(-4.197583E-01,+0.000000E+00),complex(+0.000000E+00,+4.197584E-01),complex(+3.796858E-01,+0.000000E+00),complex(+0.000000E+00,-3.100120E-01),complex(-2.274867E-01,+0.000000E+00),complex(+0.000000E+00,+1.489250E-01),complex(+8.598192E-02,+0.000000E+00),complex(+0.000000E+00,-4.299096E-02),complex(-1.805978E-02,+0.000000E+00),complex(+0.000000E+00,+6.019946E-03),complex(+1.381068E-03,-0.000000E+00)],$
            [complex(+0.000000E+00,+6.019848E-03),complex(+2.347821E-02,+0.000000E+00),complex(+0.000000E+00,-6.214806E-02),complex(-1.282164E-01,+0.000000E+00),complex(+0.000000E+00,+2.169816E-01),complex(+3.074916E-01,+0.000000E+00),complex(+0.000000E+00,-3.653231E-01),complex(-3.556083E-01,+0.000000E+00),complex(+0.000000E+00,+2.613176E-01),complex(+9.629915E-02,+0.000000E+00),complex(+0.000000E+00,+9.629919E-02),complex(+2.613176E-01,+0.000000E+00),complex(+0.000000E+00,-3.556082E-01),complex(-3.653231E-01,+0.000000E+00),complex(+0.000000E+00,+3.074917E-01),complex(+2.169815E-01,+0.000000E+00),complex(+0.000000E+00,-1.282163E-01),complex(-6.214804E-02,+0.000000E+00),complex(+0.000000E+00,+2.347819E-02),complex(+6.019956E-03,-0.000000E+00)],$
            [complex(+0.000000E+00,-1.805978E-02),complex(-6.214804E-02,+0.000000E+00),complex(+0.000000E+00,+1.422499E-01),complex(+2.465700E-01,+0.000000E+00),complex(+0.000000E+00,-3.353351E-01),complex(-3.530460E-01,+0.000000E+00),complex(+0.000000E+00,+2.609451E-01),complex(+7.112164E-02,+0.000000E+00),complex(+0.000000E+00,+1.451764E-01),complex(+2.888975E-01,+0.000000E+00),complex(+0.000000E+00,-2.888975E-01),complex(-1.451764E-01,+0.000000E+00),complex(+0.000000E+00,-7.112169E-02),complex(-2.609451E-01,+0.000000E+00),complex(+0.000000E+00,+3.530460E-01),complex(+3.353351E-01,+0.000000E+00),complex(+0.000000E+00,-2.465699E-01),complex(-1.422499E-01,+0.000000E+00),complex(+0.000000E+00,+6.214809E-02),complex(+1.805982E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,+4.299101E-02),complex(+1.282164E-01,+0.000000E+00),complex(+0.000000E+00,-2.465700E-01),complex(-3.411238E-01,+0.000000E+00),complex(+0.000000E+00,+3.342186E-01),complex(+1.865822E-01,+0.000000E+00),complex(+0.000000E+00,+5.115540E-02),complex(+2.489754E-01,+0.000000E+00),complex(+0.000000E+00,-2.805371E-01),complex(-1.213613E-01,+0.000000E+00),complex(+0.000000E+00,-1.213611E-01),complex(-2.805370E-01,+0.000000E+00),complex(+0.000000E+00,+2.489755E-01),complex(+5.115554E-02,+0.000000E+00),complex(+0.000000E+00,+1.865822E-01),complex(+3.342184E-01,+0.000000E+00),complex(+0.000000E+00,-3.411238E-01),complex(-2.465699E-01,+0.000000E+00),complex(+0.000000E+00,+1.282163E-01),complex(+4.299093E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-8.598181E-02),complex(-2.169816E-01,+0.000000E+00),complex(+0.000000E+00,+3.353350E-01),complex(+3.342184E-01,+0.000000E+00),complex(+0.000000E+00,-1.602038E-01),complex(+1.052515E-01,+0.000000E+00),complex(+0.000000E+00,-2.777012E-01),complex(-2.190986E-01,+0.000000E+00),complex(+0.000000E+00,-2.439453E-02),complex(-2.427224E-01,+0.000000E+00),complex(+0.000000E+00,+2.427225E-01),complex(+2.439452E-02,+0.000000E+00),complex(+0.000000E+00,+2.190985E-01),complex(+2.777011E-01,+0.000000E+00),complex(+0.000000E+00,-1.052515E-01),complex(+1.602039E-01,+0.000000E+00),complex(+0.000000E+00,-3.342184E-01),complex(-3.353351E-01,+0.000000E+00),complex(+0.000000E+00,+2.169815E-01),complex(+8.598185E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,+1.489248E-01),complex(+3.074917E-01,+0.000000E+00),complex(+0.000000E+00,-3.530459E-01),complex(-1.865823E-01,+0.000000E+00),complex(+0.000000E+00,-1.052515E-01),complex(-2.817379E-01,+0.000000E+00),complex(+0.000000E+00,+1.772078E-01),complex(-1.034972E-01,+0.000000E+00),complex(+0.000000E+00,+2.675996E-01),complex(+1.401360E-01,+0.000000E+00),complex(+0.000000E+00,+1.401360E-01),complex(+2.675995E-01,+0.000000E+00),complex(+0.000000E+00,-1.034971E-01),complex(+1.772077E-01,+0.000000E+00),complex(+0.000000E+00,-2.817379E-01),complex(-1.052515E-01,+0.000000E+00),complex(+0.000000E+00,-1.865822E-01),complex(-3.530459E-01,+0.000000E+00),complex(+0.000000E+00,+3.074916E-01),complex(+1.489250E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-2.274867E-01),complex(-3.653231E-01,+0.000000E+00),complex(+0.000000E+00,+2.609451E-01),complex(-5.115554E-02,+0.000000E+00),complex(+0.000000E+00,+2.777012E-01),complex(+1.772077E-01,+0.000000E+00),complex(+0.000000E+00,+1.270582E-01),complex(+2.634910E-01,+0.000000E+00),complex(+0.000000E+00,-6.454185E-02),complex(+2.140611E-01,+0.000000E+00),complex(-0.000000E+00,-2.140611E-01),complex(+6.454188E-02,+0.000000E+00),complex(+0.000000E+00,-2.634909E-01),complex(-1.270582E-01,+0.000000E+00),complex(-0.000000E+00,-1.772078E-01),complex(-2.777011E-01,+0.000000E+00),complex(+0.000000E+00,+5.115551E-02),complex(-2.609451E-01,+0.000000E+00),complex(+0.000000E+00,+3.653231E-01),complex(+2.274867E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.100121E-01),complex(+3.556082E-01,+0.000000E+00),complex(+0.000000E+00,-7.112168E-02),complex(+2.489755E-01,+0.000000E+00),complex(+0.000000E+00,-2.190986E-01),complex(+1.034971E-01,+0.000000E+00),complex(+0.000000E+00,-2.634910E-01),complex(-3.866996E-02,+0.000000E+00),complex(+0.000000E+00,-2.368038E-01),complex(-1.570778E-01,+0.000000E+00),complex(+0.000000E+00,-1.570780E-01),complex(-2.368038E-01,+0.000000E+00),complex(+0.000000E+00,-3.866991E-02),complex(-2.634909E-01,+0.000000E+00),complex(+0.000000E+00,+1.034972E-01),complex(-2.190985E-01,+0.000000E+00),complex(+0.000000E+00,+2.489756E-01),complex(-7.112166E-02,+0.000000E+00),complex(+0.000000E+00,+3.556082E-01),complex(+3.100122E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-3.796856E-01),complex(-2.613176E-01,+0.000000E+00),complex(+0.000000E+00,-1.451765E-01),complex(-2.805370E-01,+0.000000E+00),complex(+0.000000E+00,-2.439454E-02),complex(-2.675995E-01,+0.000000E+00),complex(+0.000000E+00,+6.454179E-02),complex(-2.368038E-01,+0.000000E+00),complex(+0.000000E+00,+1.353447E-01),complex(-1.923804E-01,+0.000000E+00),complex(+0.000000E+00,+1.923804E-01),complex(-1.353447E-01,+0.000000E+00),complex(+0.000000E+00,+2.368039E-01),complex(-6.454182E-02,+0.000000E+00),complex(+0.000000E+00,+2.675995E-01),complex(+2.439452E-02,+0.000000E+00),complex(+0.000000E+00,+2.805370E-01),complex(+1.451765E-01,+0.000000E+00),complex(+0.000000E+00,+2.613176E-01),complex(+3.796857E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+4.197583E-01),complex(+9.629914E-02,+0.000000E+00),complex(+0.000000E+00,+2.888974E-01),complex(+1.213612E-01,+0.000000E+00),complex(+0.000000E+00,+2.427225E-01),complex(+1.401358E-01,+0.000000E+00),complex(+0.000000E+00,+2.140611E-01),complex(+1.570779E-01,+0.000000E+00),complex(+0.000000E+00,+1.923804E-01),complex(+1.740145E-01,+0.000000E+00),complex(+0.000000E+00,+1.740146E-01),complex(+1.923802E-01,+0.000000E+00),complex(+0.000000E+00,+1.570780E-01),complex(+2.140610E-01,+0.000000E+00),complex(+0.000000E+00,+1.401359E-01),complex(+2.427225E-01,+0.000000E+00),complex(+0.000000E+00,+1.213612E-01),complex(+2.888976E-01,+0.000000E+00),complex(+0.000000E+00,+9.629908E-02),complex(+4.197583E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.197583E-01),complex(+9.629914E-02,+0.000000E+00),complex(+0.000000E+00,-2.888975E-01),complex(+1.213612E-01,+0.000000E+00),complex(+0.000000E+00,-2.427225E-01),complex(+1.401358E-01,+0.000000E+00),complex(+0.000000E+00,-2.140611E-01),complex(+1.570778E-01,+0.000000E+00),complex(+0.000000E+00,-1.923803E-01),complex(+1.740146E-01,+0.000000E+00),complex(+0.000000E+00,-1.740146E-01),complex(+1.923803E-01,+0.000000E+00),complex(+0.000000E+00,-1.570778E-01),complex(+2.140611E-01,+0.000000E+00),complex(+0.000000E+00,-1.401359E-01),complex(+2.427225E-01,+0.000000E+00),complex(+0.000000E+00,-1.213613E-01),complex(+2.888976E-01,+0.000000E+00),complex(+0.000000E+00,-9.629925E-02),complex(+4.197583E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.796858E-01),complex(-2.613177E-01,+0.000000E+00),complex(+0.000000E+00,+1.451765E-01),complex(-2.805370E-01,+0.000000E+00),complex(+0.000000E+00,+2.439452E-02),complex(-2.675996E-01,+0.000000E+00),complex(+0.000000E+00,-6.454179E-02),complex(-2.368038E-01,+0.000000E+00),complex(+0.000000E+00,-1.353446E-01),complex(-1.923803E-01,+0.000000E+00),complex(+0.000000E+00,-1.923804E-01),complex(-1.353447E-01,+0.000000E+00),complex(+0.000000E+00,-2.368039E-01),complex(-6.454185E-02,+0.000000E+00),complex(-0.000000E+00,-2.675995E-01),complex(+2.439458E-02,+0.000000E+00),complex(+0.000000E+00,-2.805371E-01),complex(+1.451765E-01,+0.000000E+00),complex(+0.000000E+00,-2.613177E-01),complex(+3.796857E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-3.100119E-01),complex(+3.556081E-01,+0.000000E+00),complex(+0.000000E+00,+7.112166E-02),complex(+2.489757E-01,+0.000000E+00),complex(+0.000000E+00,+2.190985E-01),complex(+1.034972E-01,+0.000000E+00),complex(+0.000000E+00,+2.634909E-01),complex(-3.866990E-02,+0.000000E+00),complex(+0.000000E+00,+2.368038E-01),complex(-1.570779E-01,+0.000000E+00),complex(+0.000000E+00,+1.570778E-01),complex(-2.368039E-01,+0.000000E+00),complex(+0.000000E+00,+3.866974E-02),complex(-2.634909E-01,+0.000000E+00),complex(-0.000000E+00,-1.034973E-01),complex(-2.190985E-01,+0.000000E+00),complex(+0.000000E+00,-2.489756E-01),complex(-7.112163E-02,+0.000000E+00),complex(+0.000000E+00,-3.556082E-01),complex(+3.100121E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+2.274867E-01),complex(-3.653232E-01,+0.000000E+00),complex(+0.000000E+00,-2.609451E-01),complex(-5.115546E-02,+0.000000E+00),complex(+0.000000E+00,-2.777012E-01),complex(+1.772079E-01,+0.000000E+00),complex(+0.000000E+00,-1.270582E-01),complex(+2.634910E-01,+0.000000E+00),complex(+0.000000E+00,+6.454185E-02),complex(+2.140611E-01,+0.000000E+00),complex(+0.000000E+00,+2.140610E-01),complex(+6.454182E-02,+0.000000E+00),complex(+0.000000E+00,+2.634909E-01),complex(-1.270582E-01,+0.000000E+00),complex(-0.000000E+00,+1.772078E-01),complex(-2.777011E-01,+0.000000E+00),complex(+0.000000E+00,-5.115544E-02),complex(-2.609451E-01,+0.000000E+00),complex(+0.000000E+00,-3.653231E-01),complex(+2.274867E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-1.489251E-01),complex(+3.074915E-01,+0.000000E+00),complex(-0.000000E+00,+3.530459E-01),complex(-1.865822E-01,-0.000000E+00),complex(-0.000000E+00,+1.052515E-01),complex(-2.817379E-01,-0.000000E+00),complex(+0.000000E+00,-1.772077E-01),complex(-1.034973E-01,+0.000000E+00),complex(+0.000000E+00,-2.675995E-01),complex(+1.401359E-01,+0.000000E+00),complex(+0.000000E+00,-1.401359E-01),complex(+2.675996E-01,+0.000000E+00),complex(+0.000000E+00,+1.034972E-01),complex(+1.772078E-01,+0.000000E+00),complex(-0.000000E+00,+2.817378E-01),complex(-1.052515E-01,+0.000000E+00),complex(+0.000000E+00,+1.865822E-01),complex(-3.530459E-01,+0.000000E+00),complex(+0.000000E+00,-3.074916E-01),complex(+1.489250E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+8.598188E-02),complex(-2.169817E-01,+0.000000E+00),complex(+0.000000E+00,-3.353353E-01),complex(+3.342184E-01,+0.000000E+00),complex(+0.000000E+00,+1.602040E-01),complex(+1.052515E-01,+0.000000E+00),complex(+0.000000E+00,+2.777011E-01),complex(-2.190985E-01,+0.000000E+00),complex(+0.000000E+00,+2.439453E-02),complex(-2.427224E-01,+0.000000E+00),complex(+0.000000E+00,-2.427224E-01),complex(+2.439458E-02,+0.000000E+00),complex(+0.000000E+00,-2.190984E-01),complex(+2.777011E-01,+0.000000E+00),complex(+0.000000E+00,+1.052515E-01),complex(+1.602038E-01,+0.000000E+00),complex(+0.000000E+00,+3.342185E-01),complex(-3.353351E-01,+0.000000E+00),complex(+0.000000E+00,-2.169815E-01),complex(+8.598190E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.299088E-02),complex(+1.282164E-01,+0.000000E+00),complex(+0.000000E+00,+2.465699E-01),complex(-3.411239E-01,+0.000000E+00),complex(+0.000000E+00,-3.342184E-01),complex(+1.865823E-01,+0.000000E+00),complex(+0.000000E+00,-5.115548E-02),complex(+2.489755E-01,+0.000000E+00),complex(+0.000000E+00,+2.805370E-01),complex(-1.213612E-01,+0.000000E+00),complex(+0.000000E+00,+1.213612E-01),complex(-2.805370E-01,+0.000000E+00),complex(+0.000000E+00,-2.489754E-01),complex(+5.115550E-02,+0.000000E+00),complex(+0.000000E+00,-1.865823E-01),complex(+3.342184E-01,+0.000000E+00),complex(+0.000000E+00,+3.411238E-01),complex(-2.465699E-01,+0.000000E+00),complex(+0.000000E+00,-1.282164E-01),complex(+4.299095E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,+1.805981E-02),complex(-6.214803E-02,+0.000000E+00),complex(+0.000000E+00,-1.422500E-01),complex(+2.465699E-01,+0.000000E+00),complex(+0.000000E+00,+3.353352E-01),complex(-3.530459E-01,+0.000000E+00),complex(+0.000000E+00,-2.609452E-01),complex(+7.112165E-02,+0.000000E+00),complex(+0.000000E+00,-1.451764E-01),complex(+2.888975E-01,+0.000000E+00),complex(+0.000000E+00,+2.888975E-01),complex(-1.451765E-01,+0.000000E+00),complex(+0.000000E+00,+7.112169E-02),complex(-2.609450E-01,+0.000000E+00),complex(+0.000000E+00,-3.530459E-01),complex(+3.353351E-01,+0.000000E+00),complex(+0.000000E+00,+2.465699E-01),complex(-1.422500E-01,+0.000000E+00),complex(+0.000000E+00,-6.214809E-02),complex(+1.805980E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-6.019942E-03),complex(+2.347817E-02,+0.000000E+00),complex(+0.000000E+00,+6.214799E-02),complex(-1.282164E-01,+0.000000E+00),complex(+0.000000E+00,-2.169815E-01),complex(+3.074915E-01,+0.000000E+00),complex(+0.000000E+00,+3.653231E-01),complex(-3.556083E-01,+0.000000E+00),complex(+0.000000E+00,-2.613176E-01),complex(+9.629907E-02,+0.000000E+00),complex(+0.000000E+00,-9.629915E-02),complex(+2.613176E-01,+0.000000E+00),complex(+0.000000E+00,+3.556083E-01),complex(-3.653231E-01,+0.000000E+00),complex(+0.000000E+00,-3.074916E-01),complex(+2.169815E-01,+0.000000E+00),complex(+0.000000E+00,+1.282164E-01),complex(-6.214802E-02,+0.000000E+00),complex(+0.000000E+00,-2.347822E-02),complex(+6.019992E-03,-0.000000E+00)],$
            [complex(+0.000000E+00,+1.381037E-03),complex(-6.019881E-03,+0.000000E+00),complex(+0.000000E+00,-1.805978E-02),complex(+4.299092E-02,+0.000000E+00),complex(+0.000000E+00,+8.598189E-02),complex(-1.489249E-01,+0.000000E+00),complex(+0.000000E+00,-2.274867E-01),complex(+3.100120E-01,+0.000000E+00),complex(+0.000000E+00,+3.796857E-01),complex(-4.197584E-01,+0.000000E+00),complex(+0.000000E+00,-4.197583E-01),complex(+3.796857E-01,+0.000000E+00),complex(+0.000000E+00,+3.100120E-01),complex(-2.274866E-01,+0.000000E+00),complex(+0.000000E+00,-1.489250E-01),complex(+8.598188E-02,+0.000000E+00),complex(+0.000000E+00,+4.299094E-02),complex(-1.805981E-02,+0.000000E+00),complex(+0.000000E+00,-6.019914E-03),complex(+1.381090E-03,-0.000000E+00)]$
            ]
          end

      18: begin
            matrix[171:189,171:189]=[$
            [complex(-1.953075E-03,+0.000000E+00),complex(+0.000000E+00,+8.286358E-03),complex(+2.415878E-02,+0.000000E+00),complex(+0.000000E+00,-5.579232E-02),complex(-1.080415E-01,+0.000000E+00),complex(+0.000000E+00,+1.807880E-01),complex(+2.661128E-01,+0.000000E+00),complex(+0.000000E+00,-3.484234E-01),complex(-4.085627E-01,+0.000000E+00),complex(+0.000000E+00,+4.306629E-01),complex(+4.085628E-01,+0.000000E+00),complex(+0.000000E+00,-3.484235E-01),complex(-2.661128E-01,+0.000000E+00),complex(+0.000000E+00,+1.807880E-01),complex(+1.080415E-01,+0.000000E+00),complex(+0.000000E+00,-5.579240E-02),complex(-2.415880E-02,+0.000000E+00),complex(+0.000000E+00,+8.286353E-03),complex(+1.953126E-03,-0.000000E+00)],$
            [complex(+8.286435E-03,+0.000000E+00),complex(+0.000000E+00,-3.124997E-02),complex(-7.972010E-02,+0.000000E+00),complex(+0.000000E+00,+1.578049E-01),complex(+2.546564E-01,+0.000000E+00),complex(+0.000000E+00,-3.408973E-01),complex(-3.763403E-01,+0.000000E+00),complex(+0.000000E+00,+3.284968E-01),complex(+1.925983E-01,+0.000000E+00),complex(+0.000000E+00,+3.196842E-08),complex(+1.925983E-01,+0.000000E+00),complex(+0.000000E+00,-3.284968E-01),complex(-3.763405E-01,+0.000000E+00),complex(+0.000000E+00,+3.408972E-01),complex(+2.546563E-01,+0.000000E+00),complex(+0.000000E+00,-1.578047E-01),complex(-7.972004E-02,+0.000000E+00),complex(+0.000000E+00,+3.125000E-02),complex(+8.286423E-03,+0.000000E+00)],$
            [complex(-2.415886E-02,+0.000000E+00),complex(+0.000000E+00,+7.972001E-02),complex(+1.738281E-01,+0.000000E+00),complex(+0.000000E+00,-2.841645E-01),complex(-3.581202E-01,+0.000000E+00),complex(+0.000000E+00,+3.361645E-01),complex(+1.936255E-01,+0.000000E+00),complex(+0.000000E+00,+2.816837E-02),complex(+2.312124E-01,+0.000000E+00),complex(+0.000000E+00,-3.133533E-01),complex(-2.312124E-01,+0.000000E+00),complex(+0.000000E+00,+2.816836E-02),complex(-1.936255E-01,+0.000000E+00),complex(+0.000000E+00,+3.361645E-01),complex(+3.581203E-01,+0.000000E+00),complex(+0.000000E+00,-2.841646E-01),complex(-1.738281E-01,+0.000000E+00),complex(+0.000000E+00,+7.972005E-02),complex(+2.415883E-02,-0.000000E+00)],$
            [complex(+5.579227E-02,+0.000000E+00),complex(+0.000000E+00,-1.578047E-01),complex(-2.841644E-01,+0.000000E+00),complex(+0.000000E+00,+3.593751E-01),complex(+3.025768E-01,+0.000000E+00),complex(+0.000000E+00,-1.012616E-01),complex(+1.490529E-01,+0.000000E+00),complex(+0.000000E+00,-2.927343E-01),complex(-2.288409E-01,+0.000000E+00),complex(+0.000000E+00,+4.556662E-08),complex(-2.288409E-01,+0.000000E+00),complex(+0.000000E+00,+2.927343E-01),complex(+1.490531E-01,+0.000000E+00),complex(+0.000000E+00,+1.012615E-01),complex(+3.025768E-01,+0.000000E+00),complex(+0.000000E+00,-3.593750E-01),complex(-2.841647E-01,+0.000000E+00),complex(+0.000000E+00,+1.578047E-01),complex(+5.579240E-02,-0.000000E+00)],$
            [complex(-1.080416E-01,+0.000000E+00),complex(-0.000000E+00,+2.546563E-01),complex(+3.581203E-01,+0.000000E+00),complex(+0.000000E+00,-3.025768E-01),complex(-7.031251E-02,+0.000000E+00),complex(+0.000000E+00,-1.960921E-01),complex(-2.886399E-01,+0.000000E+00),complex(-0.000000E+00,+1.259728E-01),complex(-1.477161E-01,+0.000000E+00),complex(+0.000000E+00,+2.802717E-01),complex(+1.477161E-01,+0.000000E+00),complex(+0.000000E+00,+1.259729E-01),complex(+2.886400E-01,+0.000000E+00),complex(+0.000000E+00,-1.960922E-01),complex(+7.031247E-02,+0.000000E+00),complex(+0.000000E+00,-3.025768E-01),complex(-3.581203E-01,+0.000000E+00),complex(+0.000000E+00,+2.546563E-01),complex(+1.080416E-01,+0.000000E+00)],$
            [complex(+1.807881E-01,+0.000000E+00),complex(+0.000000E+00,-3.408973E-01),complex(-3.361646E-01,+0.000000E+00),complex(+0.000000E+00,+1.012616E-01),complex(-1.960921E-01,+0.000000E+00),complex(+0.000000E+00,+2.812501E-01),complex(+6.899814E-02,+0.000000E+00),complex(+0.000000E+00,+2.107927E-01),complex(+2.471766E-01,+0.000000E+00),complex(+0.000000E+00,-2.862802E-09),complex(+2.471765E-01,+0.000000E+00),complex(+0.000000E+00,-2.107928E-01),complex(+6.899820E-02,+0.000000E+00),complex(+0.000000E+00,-2.812500E-01),complex(-1.960922E-01,+0.000000E+00),complex(+0.000000E+00,-1.012616E-01),complex(-3.361646E-01,+0.000000E+00),complex(+0.000000E+00,+3.408972E-01),complex(+1.807881E-01,-0.000000E+00)],$
            [complex(-2.661129E-01,+0.000000E+00),complex(+0.000000E+00,+3.763404E-01),complex(+1.936255E-01,+0.000000E+00),complex(+0.000000E+00,+1.490531E-01),complex(+2.886400E-01,+0.000000E+00),complex(+0.000000E+00,-6.899807E-02),complex(+2.265625E-01,+0.000000E+00),complex(+0.000000E+00,-2.148082E-01),complex(+8.396158E-02,+0.000000E+00),complex(+0.000000E+00,-2.655101E-01),complex(-8.396161E-02,+0.000000E+00),complex(+0.000000E+00,-2.148082E-01),complex(-2.265625E-01,+0.000000E+00),complex(+0.000000E+00,-6.899810E-02),complex(-2.886399E-01,+0.000000E+00),complex(+0.000000E+00,+1.490530E-01),complex(-1.936255E-01,+0.000000E+00),complex(+0.000000E+00,+3.763403E-01),complex(+2.661128E-01,-0.000000E+00)],$
            [complex(+3.484236E-01,+0.000000E+00),complex(+0.000000E+00,-3.284968E-01),complex(+2.816839E-02,+0.000000E+00),complex(+0.000000E+00,-2.927343E-01),complex(-1.259728E-01,+0.000000E+00),complex(+0.000000E+00,-2.107927E-01),complex(-2.148082E-01,+0.000000E+00),complex(-0.000000E+00,-1.093751E-01),complex(-2.565072E-01,+0.000000E+00),complex(+0.000000E+00,+5.191914E-08),complex(-2.565071E-01,+0.000000E+00),complex(+0.000000E+00,+1.093750E-01),complex(-2.148083E-01,+0.000000E+00),complex(+0.000000E+00,+2.107928E-01),complex(-1.259728E-01,+0.000000E+00),complex(+0.000000E+00,+2.927343E-01),complex(+2.816839E-02,+0.000000E+00),complex(+0.000000E+00,+3.284968E-01),complex(+3.484235E-01,-0.000000E+00)],$
            [complex(-4.085627E-01,+0.000000E+00),complex(+0.000000E+00,+1.925983E-01),complex(-2.312124E-01,+0.000000E+00),complex(+0.000000E+00,+2.288410E-01),complex(-1.477162E-01,+0.000000E+00),complex(+0.000000E+00,+2.471765E-01),complex(-8.396155E-02,+0.000000E+00),complex(+0.000000E+00,+2.565070E-01),complex(-2.734384E-02,+0.000000E+00),complex(+0.000000E+00,+2.594055E-01),complex(+2.734369E-02,+0.000000E+00),complex(+0.000000E+00,+2.565071E-01),complex(+8.396161E-02,+0.000000E+00),complex(+0.000000E+00,+2.471765E-01),complex(+1.477161E-01,+0.000000E+00),complex(+0.000000E+00,+2.288410E-01),complex(+2.312124E-01,+0.000000E+00),complex(+0.000000E+00,+1.925983E-01),complex(+4.085627E-01,-0.000000E+00)],$
            [complex(+4.306633E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+3.133534E-01,+0.000000E+00),0,complex(+2.802718E-01,+0.000000E+00),complex(+0.000000E+00,-2.980232E-08),complex(+2.655101E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+2.594054E-01,+0.000000E+00),complex(+0.000000E+00,-1.444425E-07),complex(+2.594055E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+2.655100E-01,+0.000000E+00),complex(+0.000000E+00,+1.490116E-07),complex(+2.802718E-01,+0.000000E+00),0,complex(+3.133533E-01,+0.000000E+00),complex(+0.000000E+00,+8.940697E-08),complex(+4.306629E-01,-0.000000E+00)],$
            [complex(-4.085626E-01,+0.000000E+00),complex(+0.000000E+00,-1.925983E-01),complex(-2.312124E-01,+0.000000E+00),complex(+0.000000E+00,-2.288409E-01),complex(-1.477162E-01,+0.000000E+00),complex(+0.000000E+00,-2.471764E-01),complex(-8.396168E-02,+0.000000E+00),complex(+0.000000E+00,-2.565071E-01),complex(-2.734378E-02,+0.000000E+00),complex(+0.000000E+00,-2.594056E-01),complex(+2.734366E-02,+0.000000E+00),complex(+0.000000E+00,-2.565070E-01),complex(+8.396164E-02,+0.000000E+00),complex(+0.000000E+00,-2.471765E-01),complex(+1.477161E-01,+0.000000E+00),complex(+0.000000E+00,-2.288409E-01),complex(+2.312124E-01,+0.000000E+00),complex(+0.000000E+00,-1.925983E-01),complex(+4.085627E-01,-0.000000E+00)],$
            [complex(+3.484235E-01,+0.000000E+00),complex(+0.000000E+00,+3.284968E-01),complex(+2.816842E-02,+0.000000E+00),complex(+0.000000E+00,+2.927343E-01),complex(-1.259728E-01,+0.000000E+00),complex(+0.000000E+00,+2.107928E-01),complex(-2.148082E-01,+0.000000E+00),complex(+0.000000E+00,+1.093750E-01),complex(-2.565071E-01,+0.000000E+00),complex(+0.000000E+00,-7.685507E-09),complex(-2.565072E-01,+0.000000E+00),complex(+0.000000E+00,-1.093750E-01),complex(-2.148083E-01,+0.000000E+00),complex(+0.000000E+00,-2.107928E-01),complex(-1.259728E-01,+0.000000E+00),complex(+0.000000E+00,-2.927342E-01),complex(+2.816834E-02,+0.000000E+00),complex(+0.000000E+00,-3.284968E-01),complex(+3.484235E-01,-0.000000E+00)],$
            [complex(-2.661129E-01,+0.000000E+00),complex(+0.000000E+00,-3.763404E-01),complex(+1.936256E-01,+0.000000E+00),complex(+0.000000E+00,-1.490531E-01),complex(+2.886399E-01,+0.000000E+00),complex(+0.000000E+00,+6.899808E-02),complex(+2.265625E-01,+0.000000E+00),complex(+0.000000E+00,+2.148082E-01),complex(+8.396161E-02,+0.000000E+00),complex(+0.000000E+00,+2.655101E-01),complex(-8.396169E-02,+0.000000E+00),complex(+0.000000E+00,+2.148082E-01),complex(-2.265625E-01,-0.000000E+00),complex(+0.000000E+00,+6.899808E-02),complex(-2.886399E-01,+0.000000E+00),complex(+0.000000E+00,-1.490530E-01),complex(-1.936255E-01,+0.000000E+00),complex(+0.000000E+00,-3.763404E-01),complex(+2.661129E-01,+0.000000E+00)],$
            [complex(+1.807882E-01,+0.000000E+00),complex(+0.000000E+00,+3.408972E-01),complex(-3.361645E-01,+0.000000E+00),complex(+0.000000E+00,-1.012617E-01),complex(-1.960922E-01,+0.000000E+00),complex(+0.000000E+00,-2.812501E-01),complex(+6.899805E-02,+0.000000E+00),complex(+0.000000E+00,-2.107929E-01),complex(+2.471763E-01,+0.000000E+00),complex(+0.000000E+00,-2.862782E-09),complex(+2.471764E-01,+0.000000E+00),complex(+0.000000E+00,+2.107927E-01),complex(+6.899816E-02,+0.000000E+00),complex(+0.000000E+00,+2.812500E-01),complex(-1.960922E-01,+0.000000E+00),complex(+0.000000E+00,+1.012616E-01),complex(-3.361645E-01,+0.000000E+00),complex(+0.000000E+00,-3.408972E-01),complex(+1.807882E-01,+0.000000E+00)],$
            [complex(-1.080416E-01,+0.000000E+00),complex(+0.000000E+00,-2.546565E-01),complex(+3.581203E-01,+0.000000E+00),complex(+0.000000E+00,+3.025768E-01),complex(-7.031244E-02,+0.000000E+00),complex(+0.000000E+00,+1.960922E-01),complex(-2.886399E-01,+0.000000E+00),complex(+0.000000E+00,-1.259727E-01),complex(-1.477161E-01,+0.000000E+00),complex(+0.000000E+00,-2.802717E-01),complex(+1.477161E-01,+0.000000E+00),complex(+0.000000E+00,-1.259728E-01),complex(+2.886399E-01,+0.000000E+00),complex(+0.000000E+00,+1.960922E-01),complex(+7.031244E-02,+0.000000E+00),complex(+0.000000E+00,+3.025769E-01),complex(-3.581203E-01,+0.000000E+00),complex(+0.000000E+00,-2.546564E-01),complex(+1.080416E-01,-0.000000E+00)],$
            [complex(+5.579235E-02,+0.000000E+00),complex(+0.000000E+00,+1.578048E-01),complex(-2.841645E-01,+0.000000E+00),complex(+0.000000E+00,-3.593749E-01),complex(+3.025769E-01,+0.000000E+00),complex(+0.000000E+00,+1.012617E-01),complex(+1.490530E-01,+0.000000E+00),complex(+0.000000E+00,+2.927344E-01),complex(-2.288408E-01,+0.000000E+00),complex(+0.000000E+00,+8.313713E-09),complex(-2.288409E-01,+0.000000E+00),complex(+0.000000E+00,-2.927343E-01),complex(+1.490531E-01,+0.000000E+00),complex(+0.000000E+00,-1.012615E-01),complex(+3.025768E-01,+0.000000E+00),complex(+0.000000E+00,+3.593750E-01),complex(-2.841646E-01,+0.000000E+00),complex(+0.000000E+00,-1.578048E-01),complex(+5.579238E-02,-0.000000E+00)],$
            [complex(-2.415886E-02,+0.000000E+00),complex(+0.000000E+00,-7.972009E-02),complex(+1.738282E-01,+0.000000E+00),complex(+0.000000E+00,+2.841645E-01),complex(-3.581203E-01,+0.000000E+00),complex(+0.000000E+00,-3.361645E-01),complex(+1.936255E-01,+0.000000E+00),complex(+0.000000E+00,-2.816837E-02),complex(+2.312124E-01,+0.000000E+00),complex(+0.000000E+00,+3.133533E-01),complex(-2.312124E-01,+0.000000E+00),complex(+0.000000E+00,-2.816834E-02),complex(-1.936256E-01,+0.000000E+00),complex(+0.000000E+00,-3.361645E-01),complex(+3.581203E-01,+0.000000E+00),complex(+0.000000E+00,+2.841646E-01),complex(-1.738282E-01,+0.000000E+00),complex(+0.000000E+00,-7.972000E-02),complex(+2.415885E-02,-0.000000E+00)],$
            [complex(+8.286430E-03,+0.000000E+00),complex(+0.000000E+00,+3.124999E-02),complex(-7.972006E-02,+0.000000E+00),complex(+0.000000E+00,-1.578049E-01),complex(+2.546563E-01,+0.000000E+00),complex(+0.000000E+00,+3.408973E-01),complex(-3.763403E-01,+0.000000E+00),complex(+0.000000E+00,-3.284969E-01),complex(+1.925982E-01,+0.000000E+00),complex(+0.000000E+00,-2.297961E-08),complex(+1.925984E-01,+0.000000E+00),complex(+0.000000E+00,+3.284968E-01),complex(-3.763404E-01,+0.000000E+00),complex(+0.000000E+00,-3.408972E-01),complex(+2.546564E-01,+0.000000E+00),complex(+0.000000E+00,+1.578047E-01),complex(-7.972008E-02,+0.000000E+00),complex(+0.000000E+00,-3.124996E-02),complex(+8.286454E-03,-0.000000E+00)],$
            [complex(-1.953087E-03,+0.000000E+00),complex(+0.000000E+00,-8.286375E-03),complex(+2.415880E-02,+0.000000E+00),complex(+0.000000E+00,+5.579235E-02),complex(-1.080416E-01,+0.000000E+00),complex(+0.000000E+00,-1.807881E-01),complex(+2.661129E-01,+0.000000E+00),complex(+0.000000E+00,+3.484235E-01),complex(-4.085629E-01,+0.000000E+00),complex(+0.000000E+00,-4.306630E-01),complex(+4.085628E-01,+0.000000E+00),complex(+0.000000E+00,+3.484235E-01),complex(-2.661128E-01,+0.000000E+00),complex(+0.000000E+00,-1.807880E-01),complex(+1.080416E-01,+0.000000E+00),complex(+0.000000E+00,+5.579237E-02),complex(-2.415884E-02,+0.000000E+00),complex(+0.000000E+00,-8.286396E-03),complex(+1.953162E-03,-0.000000E+00)]$
            ]
          end

      17: begin
            matrix[153:170,153:170]=[$
            [complex(+0.000000E+00,+2.762188E-03),complex(+1.138857E-02,+0.000000E+00),complex(+0.000000E+00,-3.221178E-02),complex(-7.202771E-02,+0.000000E+00),complex(+0.000000E+00,+1.347514E-01),complex(+2.172802E-01,+0.000000E+00),complex(+0.000000E+00,-3.072807E-01),complex(-3.851967E-01,+0.000000E+00),complex(+0.000000E+00,+4.306630E-01),complex(+4.306630E-01,+0.000000E+00),complex(+0.000000E+00,-3.851967E-01),complex(-3.072806E-01,+0.000000E+00),complex(+0.000000E+00,+2.172803E-01),complex(+1.347515E-01,+0.000000E+00),complex(+0.000000E+00,-7.202768E-02),complex(-3.221180E-02,+0.000000E+00),complex(+0.000000E+00,+1.138854E-02),complex(+2.762143E-03,-0.000000E+00)],$
            [complex(+0.000000E+00,-1.138860E-02),complex(-4.143199E-02,+0.000000E+00),complex(+0.000000E+00,+1.015625E-01),complex(+1.921622E-01,+0.000000E+00),complex(+0.000000E+00,-2.941383E-01),complex(-3.688874E-01,+0.000000E+00),complex(+0.000000E+00,+3.726324E-01),complex(+2.802717E-01,+0.000000E+00),complex(+0.000000E+00,-1.044512E-01),complex(+1.044513E-01,+0.000000E+00),complex(+0.000000E+00,-2.802717E-01),complex(-3.726325E-01,+0.000000E+00),complex(+0.000000E+00,+3.688873E-01),complex(+2.941383E-01,+0.000000E+00),complex(+0.000000E+00,-1.921621E-01),complex(-1.015625E-01,+0.000000E+00),complex(+0.000000E+00,+4.143206E-02),complex(+1.138858E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.221174E-02),complex(+1.015624E-01,+0.000000E+00),complex(+0.000000E+00,-2.099223E-01),complex(-3.211688E-01,+0.000000E+00),complex(+0.000000E+00,+3.697549E-01),complex(+2.981060E-01,+0.000000E+00),complex(+0.000000E+00,-1.053964E-01),complex(+1.321214E-01,+0.000000E+00),complex(+0.000000E+00,-2.954324E-01),complex(-2.954323E-01,+0.000000E+00),complex(+0.000000E+00,+1.321214E-01),complex(-1.053964E-01,+0.000000E+00),complex(+0.000000E+00,+2.981060E-01),complex(+3.697550E-01,+0.000000E+00),complex(+0.000000E+00,-3.211688E-01),complex(-2.099223E-01,+0.000000E+00),complex(+0.000000E+00,+1.015625E-01),complex(+3.221176E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-7.202780E-02),complex(-1.921622E-01,+0.000000E+00),complex(+0.000000E+00,+3.211688E-01),complex(+3.646021E-01,+0.000000E+00),complex(+0.000000E+00,-2.480392E-01),complex(-2.980232E-08,+0.000000E+00),complex(+0.000000E+00,-2.356734E-01),complex(-2.954323E-01,+0.000000E+00),complex(+0.000000E+00,+1.321214E-01),complex(-1.321214E-01,+0.000000E+00),complex(+0.000000E+00,+2.954324E-01),complex(+2.356734E-01,+0.000000E+00),complex(+0.000000E+00,+1.024455E-07),complex(+2.480392E-01,+0.000000E+00),complex(+0.000000E+00,-3.646019E-01),complex(-3.211688E-01,+0.000000E+00),complex(+0.000000E+00,+1.921620E-01),complex(+7.202768E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,+1.347514E-01),complex(+2.941382E-01,+0.000000E+00),complex(-0.000000E+00,-3.697551E-01),complex(-2.480392E-01,+0.000000E+00),complex(+0.000000E+00,-3.314560E-02),complex(-2.672286E-01,+0.000000E+00),complex(+0.000000E+00,+2.519456E-01),complex(-3.725290E-09,+0.000000E+00),complex(+0.000000E+00,+2.471764E-01),complex(+2.471764E-01,+0.000000E+00),complex(+0.000000E+00,+7.078052E-08),complex(+2.519455E-01,-0.000000E+00),complex(-0.000000E+00,-2.672286E-01),complex(-3.314559E-02,+0.000000E+00),complex(-0.000000E+00,-2.480392E-01),complex(-3.697549E-01,+0.000000E+00),complex(+0.000000E+00,+2.941383E-01),complex(+1.347515E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-2.172803E-01),complex(-3.688873E-01,+0.000000E+00),complex(+0.000000E+00,+2.981058E-01),complex(-2.980232E-08,+0.000000E+00),complex(+0.000000E+00,+2.672287E-01),complex(+2.320194E-01,+0.000000E+00),complex(+0.000000E+00,+6.250002E-02),complex(+2.742175E-01,+0.000000E+00),complex(+0.000000E+00,-1.532922E-01),complex(+1.532923E-01,+0.000000E+00),complex(+0.000000E+00,-2.742176E-01),complex(-6.249999E-02,+0.000000E+00),complex(+0.000000E+00,-2.320195E-01),complex(-2.672286E-01,+0.000000E+00),complex(+0.000000E+00,+3.725290E-08),complex(-2.981060E-01,+0.000000E+00),complex(+0.000000E+00,+3.688874E-01),complex(+2.172803E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.072806E-01),complex(+3.726325E-01,+0.000000E+00),complex(+0.000000E+00,-1.053964E-01),complex(+2.356735E-01,+0.000000E+00),complex(+0.000000E+00,-2.519457E-01),complex(+6.250007E-02,+0.000000E+00),complex(+0.000000E+00,-2.762137E-01),complex(-9.695055E-02,+0.000000E+00),complex(+0.000000E+00,-2.167879E-01),complex(-2.167881E-01,+0.000000E+00),complex(+0.000000E+00,-9.695060E-02),complex(-2.762135E-01,+0.000000E+00),complex(+0.000000E+00,+6.250004E-02),complex(-2.519455E-01,+0.000000E+00),complex(+0.000000E+00,+2.356735E-01),complex(-1.053963E-01,+0.000000E+00),complex(+0.000000E+00,+3.726325E-01),complex(+3.072807E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-3.851967E-01),complex(-2.802718E-01,+0.000000E+00),complex(+0.000000E+00,-1.321213E-01),complex(-2.954323E-01,+0.000000E+00),complex(+0.000000E+00,-7.450581E-08),complex(-2.742176E-01,+0.000000E+00),complex(+0.000000E+00,+9.695053E-02),complex(-2.320194E-01,+0.000000E+00),complex(+0.000000E+00,+1.729370E-01),complex(-1.729370E-01,+0.000000E+00),complex(+0.000000E+00,+2.320193E-01),complex(-9.695056E-02,+0.000000E+00),complex(+0.000000E+00,+2.742176E-01),complex(+7.450581E-08,+0.000000E+00),complex(+0.000000E+00,+2.954324E-01),complex(+1.321214E-01,+0.000000E+00),complex(+0.000000E+00,+2.802717E-01),complex(+3.851967E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+4.306630E-01),complex(+1.044511E-01,+0.000000E+00),complex(+0.000000E+00,+2.954324E-01),complex(+1.321215E-01,+0.000000E+00),complex(+0.000000E+00,+2.471763E-01),complex(+1.532923E-01,+0.000000E+00),complex(+0.000000E+00,+2.167881E-01),complex(+1.729370E-01,+0.000000E+00),complex(+0.000000E+00,+1.933496E-01),complex(+1.933496E-01,+0.000000E+00),complex(+0.000000E+00,+1.729371E-01),complex(+2.167881E-01,+0.000000E+00),complex(+0.000000E+00,+1.532923E-01),complex(+2.471765E-01,+0.000000E+00),complex(+0.000000E+00,+1.321214E-01),complex(+2.954324E-01,+0.000000E+00),complex(+0.000000E+00,+1.044512E-01),complex(+4.306630E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.306629E-01),complex(+1.044511E-01,+0.000000E+00),complex(+0.000000E+00,-2.954324E-01),complex(+1.321214E-01,+0.000000E+00),complex(+0.000000E+00,-2.471764E-01),complex(+1.532923E-01,+0.000000E+00),complex(+0.000000E+00,-2.167880E-01),complex(+1.729371E-01,+0.000000E+00),complex(+0.000000E+00,-1.933495E-01),complex(+1.933495E-01,+0.000000E+00),complex(+0.000000E+00,-1.729371E-01),complex(+2.167881E-01,+0.000000E+00),complex(+0.000000E+00,-1.532923E-01),complex(+2.471765E-01,+0.000000E+00),complex(+0.000000E+00,-1.321213E-01),complex(+2.954324E-01,+0.000000E+00),complex(+0.000000E+00,-1.044511E-01),complex(+4.306630E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+3.851968E-01),complex(-2.802716E-01,+0.000000E+00),complex(+0.000000E+00,+1.321214E-01),complex(-2.954323E-01,+0.000000E+00),complex(+0.000000E+00,+1.043081E-07),complex(-2.742176E-01,+0.000000E+00),complex(+0.000000E+00,-9.695059E-02),complex(-2.320194E-01,+0.000000E+00),complex(+0.000000E+00,-1.729371E-01),complex(-1.729370E-01,+0.000000E+00),complex(+0.000000E+00,-2.320194E-01),complex(-9.695050E-02,+0.000000E+00),complex(+0.000000E+00,-2.742176E-01),complex(-4.470348E-08,+0.000000E+00),complex(+0.000000E+00,-2.954323E-01),complex(+1.321212E-01,+0.000000E+00),complex(+0.000000E+00,-2.802717E-01),complex(+3.851967E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-3.072806E-01),complex(+3.726327E-01,+0.000000E+00),complex(+0.000000E+00,+1.053963E-01),complex(+2.356735E-01,+0.000000E+00),complex(+0.000000E+00,+2.519455E-01),complex(+6.250004E-02,+0.000000E+00),complex(+0.000000E+00,+2.762136E-01),complex(-9.695055E-02,+0.000000E+00),complex(+0.000000E+00,+2.167881E-01),complex(-2.167880E-01,+0.000000E+00),complex(+0.000000E+00,+9.695064E-02),complex(-2.762135E-01,+0.000000E+00),complex(+0.000000E+00,-6.250002E-02),complex(-2.519456E-01,+0.000000E+00),complex(+0.000000E+00,-2.356735E-01),complex(-1.053965E-01,+0.000000E+00),complex(+0.000000E+00,-3.726325E-01),complex(+3.072807E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+2.172803E-01),complex(-3.688874E-01,+0.000000E+00),complex(+0.000000E+00,-2.981060E-01),0,complex(+0.000000E+00,-2.672286E-01),complex(+2.320194E-01,+0.000000E+00),complex(+0.000000E+00,-6.249993E-02),complex(+2.742177E-01,+0.000000E+00),complex(+0.000000E+00,+1.532923E-01),complex(+1.532923E-01,+0.000000E+00),complex(+0.000000E+00,+2.742176E-01),complex(-6.249994E-02,+0.000000E+00),complex(+0.000000E+00,+2.320194E-01),complex(-2.672286E-01,+0.000000E+00),complex(+0.000000E+00,+7.450581E-09),complex(-2.981061E-01,+0.000000E+00),complex(+0.000000E+00,-3.688874E-01),complex(+2.172802E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.347515E-01),complex(+2.941382E-01,+0.000000E+00),complex(+0.000000E+00,+3.697549E-01),complex(-2.480392E-01,+0.000000E+00),complex(+0.000000E+00,+3.314560E-02),complex(-2.672286E-01,+0.000000E+00),complex(+0.000000E+00,-2.519456E-01),complex(+4.842877E-08,+0.000000E+00),complex(+0.000000E+00,-2.471764E-01),complex(+2.471765E-01,+0.000000E+00),complex(+0.000000E+00,-2.235174E-08),complex(+2.519456E-01,+0.000000E+00),complex(+0.000000E+00,+2.672286E-01),complex(-3.314563E-02,+0.000000E+00),complex(+0.000000E+00,+2.480392E-01),complex(-3.697549E-01,+0.000000E+00),complex(+0.000000E+00,-2.941383E-01),complex(+1.347515E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+7.202776E-02),complex(-1.921621E-01,+0.000000E+00),complex(+0.000000E+00,-3.211690E-01),complex(+3.646020E-01,+0.000000E+00),complex(+0.000000E+00,+2.480392E-01),complex(-4.284084E-08,+0.000000E+00),complex(+0.000000E+00,+2.356735E-01),complex(-2.954324E-01,+0.000000E+00),complex(+0.000000E+00,-1.321214E-01),complex(-1.321213E-01,+0.000000E+00),complex(+0.000000E+00,-2.954324E-01),complex(+2.356735E-01,+0.000000E+00),complex(+0.000000E+00,-9.499490E-08),complex(+2.480392E-01,+0.000000E+00),complex(+0.000000E+00,+3.646020E-01),complex(-3.211688E-01,+0.000000E+00),complex(+0.000000E+00,-1.921621E-01),complex(+7.202774E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-3.221174E-02),complex(+1.015625E-01,+0.000000E+00),complex(+0.000000E+00,+2.099223E-01),complex(-3.211688E-01,+0.000000E+00),complex(+0.000000E+00,-3.697550E-01),complex(+2.981060E-01,+0.000000E+00),complex(+0.000000E+00,+1.053964E-01),complex(+1.321214E-01,+0.000000E+00),complex(+0.000000E+00,+2.954324E-01),complex(-2.954324E-01,+0.000000E+00),complex(+0.000000E+00,-1.321214E-01),complex(-1.053964E-01,+0.000000E+00),complex(+0.000000E+00,-2.981060E-01),complex(+3.697550E-01,+0.000000E+00),complex(+0.000000E+00,+3.211688E-01),complex(-2.099224E-01,+0.000000E+00),complex(+0.000000E+00,-1.015624E-01),complex(+3.221175E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,+1.138859E-02),complex(-4.143208E-02,+0.000000E+00),complex(+0.000000E+00,-1.015626E-01),complex(+1.921621E-01,+0.000000E+00),complex(+0.000000E+00,+2.941385E-01),complex(-3.688875E-01,+0.000000E+00),complex(+0.000000E+00,-3.726324E-01),complex(+2.802718E-01,+0.000000E+00),complex(+0.000000E+00,+1.044512E-01),complex(+1.044511E-01,+0.000000E+00),complex(+0.000000E+00,+2.802718E-01),complex(-3.726325E-01,+0.000000E+00),complex(+0.000000E+00,-3.688873E-01),complex(+2.941382E-01,+0.000000E+00),complex(+0.000000E+00,+1.921621E-01),complex(-1.015625E-01,+0.000000E+00),complex(+0.000000E+00,-4.143200E-02),complex(+1.138858E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-2.762089E-03),complex(+1.138853E-02,+0.000000E+00),complex(+0.000000E+00,+3.221178E-02),complex(-7.202769E-02,+0.000000E+00),complex(+0.000000E+00,-1.347514E-01),complex(+2.172802E-01,+0.000000E+00),complex(+0.000000E+00,+3.072806E-01),complex(-3.851966E-01,+0.000000E+00),complex(+0.000000E+00,-4.306630E-01),complex(+4.306629E-01,+0.000000E+00),complex(+0.000000E+00,+3.851967E-01),complex(-3.072806E-01,+0.000000E+00),complex(+0.000000E+00,-2.172802E-01),complex(+1.347515E-01,+0.000000E+00),complex(+0.000000E+00,+7.202769E-02),complex(-3.221183E-02,+0.000000E+00),complex(+0.000000E+00,-1.138857E-02),complex(+2.762161E-03,-0.000000E+00)]$
            ]
          end

      16: begin
            matrix[136:152,136:152]=[$
            [complex(+3.906270E-03,+0.000000E+00),complex(+0.000000E+00,-1.562497E-02),complex(-4.279088E-02,+0.000000E+00),complex(+0.000000E+00,+9.243869E-02),complex(+1.666463E-01,+0.000000E+00),complex(+0.000000E+00,-2.581674E-01),complex(-3.495602E-01,+0.000000E+00),complex(+0.000000E+00,+4.178044E-01),complex(+4.431485E-01,+0.000000E+00),complex(+0.000000E+00,-4.178044E-01),complex(-3.495603E-01,+0.000000E+00),complex(+0.000000E+00,+2.581674E-01),complex(+1.666463E-01,+0.000000E+00),complex(+0.000000E+00,-9.243876E-02),complex(-4.279083E-02,+0.000000E+00),complex(+0.000000E+00,+1.562500E-02),complex(+3.906284E-03,-0.000000E+00)],$
            [complex(-1.562496E-02,+0.000000E+00),complex(+0.000000E+00,+5.468754E-02),complex(+1.283725E-01,+0.000000E+00),complex(+0.000000E+00,-2.310969E-01),complex(-3.332927E-01,+0.000000E+00),complex(+0.000000E+00,+3.872511E-01),complex(+3.495602E-01,+0.000000E+00),complex(+0.000000E+00,-2.089022E-01),complex(+4.062895E-08,+0.000000E+00),complex(+0.000000E+00,-2.089022E-01),complex(-3.495603E-01,+0.000000E+00),complex(+0.000000E+00,+3.872510E-01),complex(+3.332926E-01,+0.000000E+00),complex(+0.000000E+00,-2.310968E-01),complex(-1.283725E-01,+0.000000E+00),complex(+0.000000E+00,+5.468753E-02),complex(+1.562497E-02,-0.000000E+00)],$
            [complex(+4.279085E-02,+0.000000E+00),complex(+0.000000E+00,-1.283724E-01),complex(-2.500000E-01,+0.000000E+00),complex(+0.000000E+00,+3.544155E-01),complex(+3.651038E-01,+0.000000E+00),complex(+0.000000E+00,-2.356734E-01),complex(+1.210719E-08,+0.000000E+00),complex(+0.000000E+00,-2.288409E-01),complex(-3.236299E-01,+0.000000E+00),complex(+0.000000E+00,+2.288409E-01),complex(+2.561137E-08,+0.000000E+00),complex(+0.000000E+00,+2.356735E-01),complex(+3.651038E-01,+0.000000E+00),complex(+0.000000E+00,-3.544154E-01),complex(-2.500000E-01,+0.000000E+00),complex(+0.000000E+00,+1.283725E-01),complex(+4.279087E-02,-0.000000E+00)],$
            [complex(-9.243874E-02,+0.000000E+00),complex(+0.000000E+00,+2.310970E-01),complex(+3.544154E-01,+0.000000E+00),complex(+0.000000E+00,-3.515625E-01),complex(-1.690102E-01,+0.000000E+00),complex(+0.000000E+00,-1.090956E-01),complex(-2.954323E-01,+0.000000E+00),complex(+0.000000E+00,+2.471764E-01),complex(-4.379894E-10,+0.000000E+00),complex(+0.000000E+00,+2.471765E-01),complex(+2.954323E-01,+0.000000E+00),complex(+0.000000E+00,-1.090955E-01),complex(+1.690102E-01,+0.000000E+00),complex(+0.000000E+00,-3.515626E-01),complex(-3.544155E-01,+0.000000E+00),complex(+0.000000E+00,+2.310969E-01),complex(+9.243879E-02,-0.000000E+00)],$
            [complex(+1.666463E-01,+0.000000E+00),complex(+0.000000E+00,-3.332926E-01),complex(-3.651038E-01,+0.000000E+00),complex(+0.000000E+00,+1.690102E-01),complex(-1.406249E-01,+0.000000E+00),complex(+0.000000E+00,+3.025768E-01),complex(+1.638764E-01,+0.000000E+00),complex(+0.000000E+00,+1.371088E-01),complex(+2.908517E-01,+0.000000E+00),complex(+0.000000E+00,-1.371088E-01),complex(+1.638764E-01,+0.000000E+00),complex(+0.000000E+00,-3.025768E-01),complex(-1.406250E-01,+0.000000E+00),complex(+0.000000E+00,-1.690103E-01),complex(-3.651038E-01,+0.000000E+00),complex(+0.000000E+00,+3.332927E-01),complex(+1.666463E-01,+0.000000E+00)],$
            [complex(-2.581674E-01,+0.000000E+00),complex(+0.000000E+00,+3.872512E-01),complex(+2.356734E-01,+0.000000E+00),complex(+0.000000E+00,+1.090955E-01),complex(+3.025770E-01,+0.000000E+00),complex(+0.000000E+00,-1.328125E-01),complex(+1.904071E-01,+0.000000E+00),complex(+0.000000E+00,-2.655101E-01),complex(+2.500546E-08,+0.000000E+00),complex(+0.000000E+00,-2.655101E-01),complex(-1.904072E-01,+0.000000E+00),complex(+0.000000E+00,-1.328125E-01),complex(-3.025769E-01,+0.000000E+00),complex(+0.000000E+00,+1.090957E-01),complex(-2.356735E-01,+0.000000E+00),complex(+0.000000E+00,+3.872511E-01),complex(+2.581674E-01,-0.000000E+00)],$
            [complex(+3.495603E-01,+0.000000E+00),complex(+0.000000E+00,-3.495603E-01),complex(-2.980232E-08,+0.000000E+00),complex(+0.000000E+00,-2.954322E-01),complex(-1.638764E-01,+0.000000E+00),complex(+0.000000E+00,-1.904071E-01),complex(-2.500001E-01,+0.000000E+00),complex(+0.000000E+00,-6.536402E-02),complex(-2.773162E-01,+0.000000E+00),complex(+0.000000E+00,+6.536414E-02),complex(-2.499999E-01,+0.000000E+00),complex(+0.000000E+00,+1.904072E-01),complex(-1.638763E-01,+0.000000E+00),complex(+0.000000E+00,+2.954324E-01),complex(+1.490116E-08,+0.000000E+00),complex(+0.000000E+00,+3.495603E-01),complex(+3.495604E-01,-0.000000E+00)],$
            [complex(-4.178044E-01,+0.000000E+00),complex(+0.000000E+00,+2.089022E-01),complex(-2.288409E-01,+0.000000E+00),complex(+0.000000E+00,+2.471766E-01),complex(-1.371088E-01,+0.000000E+00),complex(+0.000000E+00,+2.655101E-01),complex(-6.536409E-02,+0.000000E+00),complex(+0.000000E+00,+2.734375E-01),complex(+1.556443E-08,+0.000000E+00),complex(+0.000000E+00,+2.734375E-01),complex(+6.536406E-02,+0.000000E+00),complex(+0.000000E+00,+2.655101E-01),complex(+1.371088E-01,+0.000000E+00),complex(+0.000000E+00,+2.471765E-01),complex(+2.288409E-01,+0.000000E+00),complex(+0.000000E+00,+2.089022E-01),complex(+4.178045E-01,-0.000000E+00)],$
            [complex(+4.431485E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+3.236299E-01,+0.000000E+00),complex(+0.000000E+00,+1.192093E-07),complex(+2.908517E-01,+0.000000E+00),complex(+0.000000E+00,+2.980232E-08),complex(+2.773162E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+2.734375E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+2.773163E-01,+0.000000E+00),complex(+0.000000E+00,+2.980232E-08),complex(+2.908518E-01,+0.000000E+00),complex(+0.000000E+00,+5.960464E-08),complex(+3.236299E-01,+0.000000E+00),complex(+0.000000E+00,-8.940697E-08),complex(+4.431486E-01,-0.000000E+00)],$
            [complex(-4.178044E-01,+0.000000E+00),complex(+0.000000E+00,-2.089022E-01),complex(-2.288409E-01,+0.000000E+00),complex(+0.000000E+00,-2.471764E-01),complex(-1.371089E-01,+0.000000E+00),complex(+0.000000E+00,-2.655101E-01),complex(-6.536411E-02,+0.000000E+00),complex(+0.000000E+00,-2.734375E-01),complex(-2.423982E-08,+0.000000E+00),complex(+0.000000E+00,-2.734375E-01),complex(+6.536403E-02,+0.000000E+00),complex(+0.000000E+00,-2.655101E-01),complex(+1.371089E-01,+0.000000E+00),complex(+0.000000E+00,-2.471764E-01),complex(+2.288410E-01,+0.000000E+00),complex(+0.000000E+00,-2.089022E-01),complex(+4.178044E-01,-0.000000E+00)],$
            [complex(+3.495604E-01,+0.000000E+00),complex(+0.000000E+00,+3.495603E-01),complex(-4.470348E-08,+0.000000E+00),complex(+0.000000E+00,+2.954325E-01),complex(-1.638764E-01,+0.000000E+00),complex(+0.000000E+00,+1.904072E-01),complex(-2.500001E-01,+0.000000E+00),complex(+0.000000E+00,+6.536405E-02),complex(-2.773163E-01,+0.000000E+00),complex(+0.000000E+00,-6.536397E-02),complex(-2.500000E-01,+0.000000E+00),complex(+0.000000E+00,-1.904072E-01),complex(-1.638763E-01,+0.000000E+00),complex(+0.000000E+00,-2.954324E-01),complex(+2.980232E-08,+0.000000E+00),complex(+0.000000E+00,-3.495602E-01),complex(+3.495603E-01,-0.000000E+00)],$
            [complex(-2.581674E-01,+0.000000E+00),complex(+0.000000E+00,-3.872510E-01),complex(+2.356734E-01,+0.000000E+00),complex(+0.000000E+00,-1.090956E-01),complex(+3.025768E-01,+0.000000E+00),complex(+0.000000E+00,+1.328125E-01),complex(+1.904071E-01,+0.000000E+00),complex(+0.000000E+00,+2.655101E-01),complex(-2.115846E-08,+0.000000E+00),complex(+0.000000E+00,+2.655101E-01),complex(-1.904071E-01,+0.000000E+00),complex(+0.000000E+00,+1.328125E-01),complex(-3.025768E-01,+0.000000E+00),complex(+0.000000E+00,-1.090957E-01),complex(-2.356734E-01,+0.000000E+00),complex(+0.000000E+00,-3.872510E-01),complex(+2.581674E-01,+0.000000E+00)],$
            [complex(+1.666463E-01,+0.000000E+00),complex(+0.000000E+00,+3.332928E-01),complex(-3.651039E-01,+0.000000E+00),complex(+0.000000E+00,-1.690103E-01),complex(-1.406251E-01,+0.000000E+00),complex(+0.000000E+00,-3.025768E-01),complex(+1.638764E-01,+0.000000E+00),complex(+0.000000E+00,-1.371088E-01),complex(+2.908517E-01,+0.000000E+00),complex(+0.000000E+00,+1.371089E-01),complex(+1.638764E-01,+0.000000E+00),complex(+0.000000E+00,+3.025767E-01),complex(-1.406250E-01,+0.000000E+00),complex(+0.000000E+00,+1.690102E-01),complex(-3.651037E-01,+0.000000E+00),complex(+0.000000E+00,-3.332925E-01),complex(+1.666463E-01,-0.000000E+00)],$
            [complex(-9.243876E-02,+0.000000E+00),complex(+0.000000E+00,-2.310968E-01),complex(+3.544154E-01,+0.000000E+00),complex(+0.000000E+00,+3.515625E-01),complex(-1.690102E-01,+0.000000E+00),complex(+0.000000E+00,+1.090957E-01),complex(-2.954323E-01,+0.000000E+00),complex(+0.000000E+00,-2.471764E-01),complex(+3.906834E-08,+0.000000E+00),complex(+0.000000E+00,-2.471764E-01),complex(+2.954323E-01,+0.000000E+00),complex(+0.000000E+00,+1.090956E-01),complex(+1.690102E-01,+0.000000E+00),complex(+0.000000E+00,+3.515624E-01),complex(-3.544155E-01,+0.000000E+00),complex(+0.000000E+00,-2.310968E-01),complex(+9.243876E-02,-0.000000E+00)],$
            [complex(+4.279077E-02,+0.000000E+00),complex(+0.000000E+00,+1.283724E-01),complex(-2.500000E-01,+0.000000E+00),complex(+0.000000E+00,-3.544155E-01),complex(+3.651038E-01,+0.000000E+00),complex(+0.000000E+00,+2.356736E-01),complex(-6.845221E-08,+0.000000E+00),complex(+0.000000E+00,+2.288409E-01),complex(-3.236299E-01,+0.000000E+00),complex(+0.000000E+00,-2.288411E-01),complex(-3.539026E-08,+0.000000E+00),complex(+0.000000E+00,-2.356735E-01),complex(+3.651038E-01,+0.000000E+00),complex(+0.000000E+00,+3.544155E-01),complex(-2.500000E-01,+0.000000E+00),complex(+0.000000E+00,-1.283725E-01),complex(+4.279094E-02,-0.000000E+00)],$
            [complex(-1.562506E-02,+0.000000E+00),complex(+0.000000E+00,-5.468757E-02),complex(+1.283724E-01,+0.000000E+00),complex(+0.000000E+00,+2.310969E-01),complex(-3.332927E-01,+0.000000E+00),complex(+0.000000E+00,-3.872510E-01),complex(+3.495603E-01,+0.000000E+00),complex(+0.000000E+00,+2.089022E-01),complex(+5.820793E-10,+0.000000E+00),complex(+0.000000E+00,+2.089022E-01),complex(-3.495603E-01,+0.000000E+00),complex(+0.000000E+00,-3.872511E-01),complex(+3.332927E-01,+0.000000E+00),complex(+0.000000E+00,+2.310969E-01),complex(-1.283724E-01,+0.000000E+00),complex(+0.000000E+00,-5.468746E-02),complex(+1.562506E-02,-0.000000E+00)],$
            [complex(+3.906293E-03,+0.000000E+00),complex(+0.000000E+00,+1.562496E-02),complex(-4.279085E-02,+0.000000E+00),complex(+0.000000E+00,-9.243858E-02),complex(+1.666463E-01,+0.000000E+00),complex(+0.000000E+00,+2.581674E-01),complex(-3.495603E-01,+0.000000E+00),complex(+0.000000E+00,-4.178044E-01),complex(+4.431485E-01,+0.000000E+00),complex(+0.000000E+00,+4.178045E-01),complex(-3.495603E-01,+0.000000E+00),complex(+0.000000E+00,-2.581673E-01),complex(+1.666463E-01,+0.000000E+00),complex(+0.000000E+00,+9.243878E-02),complex(-4.279085E-02,+0.000000E+00),complex(+0.000000E+00,-1.562493E-02),complex(+3.906303E-03,-0.000000E+00)]$
            ]
          end

      15: begin
            matrix[120:135,120:135]=[$
            [complex(+0.000000E+00,-5.524215E-03),complex(-2.139550E-02,+0.000000E+00),complex(+0.000000E+00,+5.660695E-02),complex(+1.178367E-01,+0.000000E+00),complex(+0.000000E+00,-2.040992E-01),complex(-3.027281E-01,+0.000000E+00),complex(+0.000000E+00,+3.908203E-01),complex(+4.431485E-01,+0.000000E+00),complex(+0.000000E+00,-4.431486E-01),complex(-3.908203E-01,+0.000000E+00),complex(+0.000000E+00,+3.027281E-01),complex(+2.040993E-01,+0.000000E+00),complex(+0.000000E+00,-1.178368E-01),complex(-5.660694E-02,+0.000000E+00),complex(+0.000000E+00,+2.139542E-02),complex(+5.524225E-03,-0.000000E+00)],$
            [complex(+0.000000E+00,+2.139542E-02),complex(+7.181553E-02,+0.000000E+00),complex(+0.000000E+00,-1.607744E-01),complex(-2.738278E-01,+0.000000E+00),complex(+0.000000E+00,+3.688874E-01),complex(+3.908203E-01,+0.000000E+00),complex(+0.000000E+00,-3.027281E-01),complex(-1.144205E-01,+0.000000E+00),complex(+0.000000E+00,-1.144205E-01),complex(-3.027281E-01,+0.000000E+00),complex(+0.000000E+00,+3.908203E-01),complex(+3.688874E-01,+0.000000E+00),complex(+0.000000E+00,-2.738278E-01),complex(-1.607744E-01,+0.000000E+00),complex(+0.000000E+00,+7.181549E-02),complex(+2.139541E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-5.660708E-02),complex(-1.607743E-01,+0.000000E+00),complex(+0.000000E+00,+2.927863E-01),complex(+3.794898E-01,+0.000000E+00),complex(+0.000000E+00,-3.386068E-01),complex(-1.477162E-01,+0.000000E+00),complex(+0.000000E+00,-1.144204E-01),complex(-3.027281E-01,+0.000000E+00),complex(+0.000000E+00,+3.027281E-01),complex(+1.144204E-01,+0.000000E+00),complex(+0.000000E+00,+1.477162E-01),complex(+3.386068E-01,+0.000000E+00),complex(+0.000000E+00,-3.794897E-01),complex(-2.927864E-01,+0.000000E+00),complex(+0.000000E+00,+1.607744E-01),complex(+5.660692E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,+1.178366E-01),complex(+2.738279E-01,+0.000000E+00),complex(+0.000000E+00,-3.794896E-01),complex(-3.148835E-01,+0.000000E+00),complex(+0.000000E+00,+6.697825E-02),complex(-2.128817E-01,+0.000000E+00),complex(+0.000000E+00,+3.114729E-01),complex(+1.454258E-01,+0.000000E+00),complex(+0.000000E+00,+1.454259E-01),complex(+3.114730E-01,+0.000000E+00),complex(+0.000000E+00,-2.128817E-01),complex(+6.697821E-02,+0.000000E+00),complex(+0.000000E+00,-3.148835E-01),complex(-3.794897E-01,+0.000000E+00),complex(+0.000000E+00,+2.738279E-01),complex(+1.178367E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-2.040992E-01),complex(-3.688874E-01,+0.000000E+00),complex(+0.000000E+00,+3.386068E-01),complex(+6.697830E-02,+0.000000E+00),complex(+0.000000E+00,+2.375436E-01),complex(+2.867838E-01,+0.000000E+00),complex(+0.000000E+00,-3.173452E-02),complex(+2.518850E-01,+0.000000E+00),complex(+0.000000E+00,-2.518850E-01),complex(+3.173450E-02,+0.000000E+00),complex(+0.000000E+00,-2.867837E-01),complex(-2.375437E-01,+0.000000E+00),complex(+0.000000E+00,-6.697824E-02),complex(-3.386067E-01,+0.000000E+00),complex(+0.000000E+00,+3.688874E-01),complex(+2.040992E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.027280E-01),complex(+3.908202E-01,-0.000000E+00),complex(+0.000000E+00,-1.477161E-01),complex(+2.128817E-01,+0.000000E+00),complex(-0.000000E+00,-2.867836E-01),complex(+5.524345E-03,+0.000000E+00),complex(-0.000000E+00,-2.781403E-01),complex(-1.698208E-01,+0.000000E+00),complex(-0.000000E+00,-1.698208E-01),complex(-2.781404E-01,+0.000000E+00),complex(+0.000000E+00,+5.524285E-03),complex(-2.867836E-01,+0.000000E+00),complex(+0.000000E+00,+2.128816E-01),complex(-1.477161E-01,+0.000000E+00),complex(+0.000000E+00,+3.908202E-01),complex(+3.027280E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-3.908203E-01),complex(-3.027281E-01,+0.000000E+00),complex(+0.000000E+00,-1.144204E-01),complex(-3.114729E-01,+0.000000E+00),complex(+0.000000E+00,+3.173459E-02),complex(-2.781404E-01,+0.000000E+00),complex(+0.000000E+00,+1.381068E-01),complex(-2.192377E-01,+0.000000E+00),complex(+0.000000E+00,+2.192378E-01),complex(-1.381068E-01,+0.000000E+00),complex(+0.000000E+00,+2.781404E-01),complex(-3.173442E-02,+0.000000E+00),complex(+0.000000E+00,+3.114730E-01),complex(+1.144205E-01,+0.000000E+00),complex(+0.000000E+00,+3.027280E-01),complex(+3.908203E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+4.431488E-01),complex(+1.144205E-01,+0.000000E+00),complex(+0.000000E+00,+3.027281E-01),complex(+1.454259E-01,+0.000000E+00),complex(+0.000000E+00,+2.518850E-01),complex(+1.698208E-01,+0.000000E+00),complex(+0.000000E+00,+2.192378E-01),complex(+1.933494E-01,+0.000000E+00),complex(+0.000000E+00,+1.933495E-01),complex(+2.192377E-01,+0.000000E+00),complex(+0.000000E+00,+1.698208E-01),complex(+2.518849E-01,+0.000000E+00),complex(+0.000000E+00,+1.454259E-01),complex(+3.027280E-01,+0.000000E+00),complex(+0.000000E+00,+1.144204E-01),complex(+4.431485E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.431485E-01),complex(+1.144205E-01,+0.000000E+00),complex(+0.000000E+00,-3.027281E-01),complex(+1.454259E-01,+0.000000E+00),complex(+0.000000E+00,-2.518850E-01),complex(+1.698210E-01,+0.000000E+00),complex(+0.000000E+00,-2.192377E-01),complex(+1.933496E-01,+0.000000E+00),complex(+0.000000E+00,-1.933496E-01),complex(+2.192378E-01,+0.000000E+00),complex(+0.000000E+00,-1.698208E-01),complex(+2.518850E-01,+0.000000E+00),complex(+0.000000E+00,-1.454259E-01),complex(+3.027280E-01,+0.000000E+00),complex(+0.000000E+00,-1.144205E-01),complex(+4.431486E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.908204E-01),complex(-3.027281E-01,+0.000000E+00),complex(+0.000000E+00,+1.144204E-01),complex(-3.114728E-01,+0.000000E+00),complex(+0.000000E+00,-3.173453E-02),complex(-2.781402E-01,+0.000000E+00),complex(+0.000000E+00,-1.381068E-01),complex(-2.192376E-01,+0.000000E+00),complex(+0.000000E+00,-2.192378E-01),complex(-1.381067E-01,+0.000000E+00),complex(+0.000000E+00,-2.781403E-01),complex(-3.173448E-02,+0.000000E+00),complex(+0.000000E+00,-3.114730E-01),complex(+1.144204E-01,-0.000000E+00),complex(+0.000000E+00,-3.027281E-01),complex(+3.908202E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-3.027281E-01),complex(+3.908202E-01,+0.000000E+00),complex(+0.000000E+00,+1.477162E-01),complex(+2.128817E-01,+0.000000E+00),complex(+0.000000E+00,+2.867837E-01),complex(+5.524248E-03,+0.000000E+00),complex(+0.000000E+00,+2.781404E-01),complex(-1.698208E-01,+0.000000E+00),complex(+0.000000E+00,+1.698208E-01),complex(-2.781405E-01,+0.000000E+00),complex(+0.000000E+00,-5.524322E-03),complex(-2.867838E-01,+0.000000E+00),complex(+0.000000E+00,-2.128816E-01),complex(-1.477161E-01,+0.000000E+00),complex(+0.000000E+00,-3.908203E-01),complex(+3.027281E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+2.040992E-01),complex(-3.688874E-01,+0.000000E+00),complex(+0.000000E+00,-3.386067E-01),complex(+6.697821E-02,+0.000000E+00),complex(+0.000000E+00,-2.375437E-01),complex(+2.867836E-01,+0.000000E+00),complex(+0.000000E+00,+3.173448E-02),complex(+2.518850E-01,+0.000000E+00),complex(+0.000000E+00,+2.518850E-01),complex(+3.173444E-02,+0.000000E+00),complex(+0.000000E+00,+2.867837E-01),complex(-2.375438E-01,+0.000000E+00),complex(+0.000000E+00,+6.697825E-02),complex(-3.386068E-01,+0.000000E+00),complex(+0.000000E+00,-3.688874E-01),complex(+2.040993E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-1.178367E-01),complex(+2.738278E-01,+0.000000E+00),complex(+0.000000E+00,+3.794898E-01),complex(-3.148836E-01,+0.000000E+00),complex(+0.000000E+00,-6.697829E-02),complex(-2.128818E-01,+0.000000E+00),complex(+0.000000E+00,-3.114730E-01),complex(+1.454259E-01,+0.000000E+00),complex(+0.000000E+00,-1.454258E-01),complex(+3.114728E-01,+0.000000E+00),complex(+0.000000E+00,+2.128817E-01),complex(+6.697825E-02,+0.000000E+00),complex(-0.000000E+00,+3.148834E-01),complex(-3.794898E-01,+0.000000E+00),complex(+0.000000E+00,-2.738279E-01),complex(+1.178367E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+5.660691E-02),complex(-1.607743E-01,+0.000000E+00),complex(+0.000000E+00,-2.927864E-01),complex(+3.794897E-01,+0.000000E+00),complex(+0.000000E+00,+3.386067E-01),complex(-1.477161E-01,+0.000000E+00),complex(+0.000000E+00,+1.144205E-01),complex(-3.027281E-01,+0.000000E+00),complex(+0.000000E+00,-3.027281E-01),complex(+1.144204E-01,+0.000000E+00),complex(+0.000000E+00,-1.477162E-01),complex(+3.386068E-01,+0.000000E+00),complex(-0.000000E+00,+3.794897E-01),complex(-2.927864E-01,+0.000000E+00),complex(+0.000000E+00,-1.607743E-01),complex(+5.660695E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-2.139549E-02),complex(+7.181545E-02,+0.000000E+00),complex(+0.000000E+00,+1.607743E-01),complex(-2.738279E-01,+0.000000E+00),complex(+0.000000E+00,-3.688875E-01),complex(+3.908203E-01,+0.000000E+00),complex(+0.000000E+00,+3.027281E-01),complex(-1.144205E-01,+0.000000E+00),complex(+0.000000E+00,+1.144205E-01),complex(-3.027280E-01,+0.000000E+00),complex(+0.000000E+00,-3.908203E-01),complex(+3.688874E-01,+0.000000E+00),complex(-0.000000E+00,+2.738279E-01),complex(-1.607744E-01,+0.000000E+00),complex(+0.000000E+00,-7.181549E-02),complex(+2.139546E-02,-0.000000E+00)],$
            [complex(-0.000000E+00,+5.524323E-03),complex(-2.139536E-02,+0.000000E+00),complex(-0.000000E+00,-5.660698E-02),complex(+1.178368E-01,+0.000000E+00),complex(+0.000000E+00,+2.040993E-01),complex(-3.027280E-01,+0.000000E+00),complex(+0.000000E+00,-3.908203E-01),complex(+4.431486E-01,+0.000000E+00),complex(+0.000000E+00,+4.431486E-01),complex(-3.908202E-01,+0.000000E+00),complex(+0.000000E+00,-3.027281E-01),complex(+2.040992E-01,+0.000000E+00),complex(-0.000000E+00,+1.178367E-01),complex(-5.660696E-02,+0.000000E+00),complex(+0.000000E+00,-2.139546E-02),complex(+5.524276E-03,-0.000000E+00)]$
            ]
          end

      14: begin
            matrix[105:119,105:119]=[$
            [complex(-7.812563E-03,+0.000000E+00),complex(+0.000000E+00,+2.923165E-02),complex(+7.452651E-02,+0.000000E+00),complex(+0.000000E+00,-1.490530E-01),complex(-2.471764E-01,+0.000000E+00),complex(+0.000000E+00,+3.495604E-01),complex(+4.281222E-01,+0.000000E+00),complex(+0.000000E+00,-4.576819E-01),complex(-4.281222E-01,+0.000000E+00),complex(+0.000000E+00,+3.495602E-01),complex(+2.471764E-01,+0.000000E+00),complex(+0.000000E+00,-1.490530E-01),complex(-7.452646E-02,+0.000000E+00),complex(+0.000000E+00,+2.923172E-02),complex(+7.812493E-03,-0.000000E+00)],$
            [complex(+2.923160E-02,+0.000000E+00),complex(+0.000000E+00,-9.375009E-02),complex(-1.991805E-01,+0.000000E+00),complex(+0.000000E+00,+3.186888E-01),complex(+3.963641E-01,+0.000000E+00),complex(+0.000000E+00,-3.736956E-01),complex(-2.288409E-01,+0.000000E+00),complex(+0.000000E+00,-9.313226E-09),complex(-2.288409E-01,+0.000000E+00),complex(+0.000000E+00,+3.736956E-01),complex(+3.963641E-01,+0.000000E+00),complex(+0.000000E+00,-3.186888E-01),complex(-1.991804E-01,+0.000000E+00),complex(+0.000000E+00,+9.375004E-02),complex(+2.923169E-02,-0.000000E+00)],$
            [complex(-7.452656E-02,+0.000000E+00),complex(+0.000000E+00,+1.991805E-01),complex(+3.359374E-01,+0.000000E+00),complex(+0.000000E+00,-3.906249E-01),complex(-2.850223E-01,+0.000000E+00),complex(+0.000000E+00,+3.664374E-02),complex(-2.243971E-01,+0.000000E+00),complex(+0.000000E+00,+3.358467E-01),complex(+2.243969E-01,+0.000000E+00),complex(+0.000000E+00,+3.664386E-02),complex(+2.850225E-01,+0.000000E+00),complex(+0.000000E+00,-3.906251E-01),complex(-3.359375E-01,+0.000000E+00),complex(+0.000000E+00,+1.991805E-01),complex(+7.452654E-02,-0.000000E+00)],$
            [complex(+1.490530E-01,+0.000000E+00),complex(+0.000000E+00,-3.186888E-01),complex(-3.906250E-01,+0.000000E+00),complex(+0.000000E+00,+2.500000E-01),complex(-5.182231E-02,+0.000000E+00),complex(+0.000000E+00,+2.931509E-01),complex(+2.692763E-01,+0.000000E+00),0,complex(+2.692764E-01,+0.000000E+00),complex(+0.000000E+00,-2.931510E-01),complex(-5.182227E-02,+0.000000E+00),complex(+0.000000E+00,-2.500001E-01),complex(-3.906250E-01,+0.000000E+00),complex(+0.000000E+00,+3.186888E-01),complex(+1.490531E-01,-0.000000E+00)],$
            [complex(-2.471765E-01,+0.000000E+00),complex(+0.000000E+00,+3.963639E-01),complex(+2.850223E-01,+0.000000E+00),complex(+0.000000E+00,+5.182222E-02),complex(+3.046875E-01,+0.000000E+00),complex(+0.000000E+00,-2.099224E-01),complex(+1.217849E-01,+0.000000E+00),complex(+0.000000E+00,-3.037848E-01),complex(-1.217848E-01,+0.000000E+00),complex(+0.000000E+00,-2.099223E-01),complex(-3.046875E-01,+0.000000E+00),complex(+0.000000E+00,+5.182226E-02),complex(-2.850224E-01,+0.000000E+00),complex(+0.000000E+00,+3.963641E-01),complex(+2.471764E-01,-0.000000E+00)],$
            [complex(+3.495603E-01,+0.000000E+00),complex(-0.000000E+00,-3.736957E-01),complex(-3.664391E-02,+0.000000E+00),complex(-0.000000E+00,-2.931510E-01),complex(-2.099223E-01,+0.000000E+00),complex(+0.000000E+00,-1.562500E-01),complex(-2.870496E-01,+0.000000E+00),0,complex(-2.870496E-01,+0.000000E+00),complex(+0.000000E+00,+1.562500E-01),complex(-2.099223E-01,+0.000000E+00),complex(+0.000000E+00,+2.931510E-01),complex(-3.664388E-02,+0.000000E+00),complex(+0.000000E+00,+3.736957E-01),complex(+3.495603E-01,-0.000000E+00)],$
            [complex(-4.281221E-01,+0.000000E+00),complex(+0.000000E+00,+2.288409E-01),complex(-2.243970E-01,+0.000000E+00),complex(+0.000000E+00,+2.692764E-01),complex(-1.217848E-01,+0.000000E+00),complex(+0.000000E+00,+2.870495E-01),complex(-3.906253E-02,+0.000000E+00),complex(+0.000000E+00,+2.923170E-01),complex(+3.906254E-02,+0.000000E+00),complex(+0.000000E+00,+2.870495E-01),complex(+1.217848E-01,+0.000000E+00),complex(+0.000000E+00,+2.692764E-01),complex(+2.243970E-01,+0.000000E+00),complex(+0.000000E+00,+2.288409E-01),complex(+4.281221E-01,-0.000000E+00)],$
            [complex(+4.576819E-01,+0.000000E+00),complex(+0.000000E+00,+2.980232E-08),complex(+3.358465E-01,+0.000000E+00),0,complex(+3.037847E-01,+0.000000E+00),0,complex(+2.923169E-01,+0.000000E+00),0,complex(+2.923170E-01,+0.000000E+00),complex(+0.000000E+00,+5.960464E-08),complex(+3.037847E-01,+0.000000E+00),0,complex(+3.358466E-01,+0.000000E+00),complex(+0.000000E+00,+4.470348E-08),complex(+4.576819E-01,-0.000000E+00)],$
            [complex(-4.281221E-01,+0.000000E+00),complex(+0.000000E+00,-2.288409E-01),complex(-2.243970E-01,+0.000000E+00),complex(+0.000000E+00,-2.692764E-01),complex(-1.217848E-01,+0.000000E+00),complex(+0.000000E+00,-2.870496E-01),complex(-3.906251E-02,+0.000000E+00),complex(+0.000000E+00,-2.923170E-01),complex(+3.906249E-02,+0.000000E+00),complex(+0.000000E+00,-2.870495E-01),complex(+1.217848E-01,+0.000000E+00),complex(+0.000000E+00,-2.692763E-01),complex(+2.243970E-01,+0.000000E+00),complex(+0.000000E+00,-2.288409E-01),complex(+4.281221E-01,-0.000000E+00)],$
            [complex(+3.495604E-01,+0.000000E+00),complex(+0.000000E+00,+3.736957E-01),complex(-3.664395E-02,+0.000000E+00),complex(+0.000000E+00,+2.931509E-01),complex(-2.099223E-01,+0.000000E+00),complex(+0.000000E+00,+1.562500E-01),complex(-2.870495E-01,+0.000000E+00),complex(+0.000000E+00,+1.490116E-08),complex(-2.870496E-01,+0.000000E+00),complex(+0.000000E+00,-1.562500E-01),complex(-2.099223E-01,+0.000000E+00),complex(+0.000000E+00,-2.931510E-01),complex(-3.664388E-02,-0.000000E+00),complex(+0.000000E+00,-3.736957E-01),complex(+3.495603E-01,-0.000000E+00)],$
            [complex(-2.471764E-01,+0.000000E+00),complex(+0.000000E+00,-3.963641E-01),complex(+2.850224E-01,+0.000000E+00),complex(+0.000000E+00,-5.182227E-02),complex(+3.046875E-01,+0.000000E+00),complex(+0.000000E+00,+2.099223E-01),complex(+1.217847E-01,+0.000000E+00),complex(-0.000000E+00,+3.037848E-01),complex(-1.217849E-01,+0.000000E+00),complex(+0.000000E+00,+2.099223E-01),complex(-3.046875E-01,+0.000000E+00),complex(+0.000000E+00,-5.182222E-02),complex(-2.850225E-01,+0.000000E+00),complex(+0.000000E+00,-3.963641E-01),complex(+2.471764E-01,-0.000000E+00)],$
            [complex(+1.490530E-01,+0.000000E+00),complex(+0.000000E+00,+3.186887E-01),complex(-3.906250E-01,-0.000000E+00),complex(+0.000000E+00,-2.499999E-01),complex(-5.182232E-02,-0.000000E+00),complex(+0.000000E+00,-2.931510E-01),complex(+2.692763E-01,+0.000000E+00),complex(+0.000000E+00,+2.235174E-08),complex(+2.692764E-01,+0.000000E+00),complex(+0.000000E+00,+2.931510E-01),complex(-5.182226E-02,+0.000000E+00),complex(+0.000000E+00,+2.500000E-01),complex(-3.906249E-01,+0.000000E+00),complex(+0.000000E+00,-3.186887E-01),complex(+1.490530E-01,+0.000000E+00)],$
            [complex(-7.452655E-02,+0.000000E+00),complex(+0.000000E+00,-1.991806E-01),complex(+3.359375E-01,+0.000000E+00),complex(+0.000000E+00,+3.906249E-01),complex(-2.850224E-01,+0.000000E+00),complex(+0.000000E+00,-3.664384E-02),complex(-2.243971E-01,+0.000000E+00),complex(+0.000000E+00,-3.358467E-01),complex(+2.243969E-01,+0.000000E+00),complex(+0.000000E+00,-3.664390E-02),complex(+2.850225E-01,+0.000000E+00),complex(-0.000000E+00,+3.906250E-01),complex(-3.359375E-01,+0.000000E+00),complex(+0.000000E+00,-1.991804E-01),complex(+7.452650E-02,-0.000000E+00)],$
            [complex(+2.923160E-02,+0.000000E+00),complex(+0.000000E+00,+9.374999E-02),complex(-1.991805E-01,+0.000000E+00),complex(+0.000000E+00,-3.186887E-01),complex(+3.963641E-01,+0.000000E+00),complex(+0.000000E+00,+3.736957E-01),complex(-2.288409E-01,+0.000000E+00),complex(+0.000000E+00,+7.450581E-09),complex(-2.288409E-01,+0.000000E+00),complex(+0.000000E+00,-3.736957E-01),complex(+3.963641E-01,+0.000000E+00),complex(-0.000000E+00,+3.186887E-01),complex(-1.991804E-01,+0.000000E+00),complex(+0.000000E+00,-9.375001E-02),complex(+2.923170E-02,-0.000000E+00)],$
            [complex(-7.812543E-03,+0.000000E+00),complex(+0.000000E+00,-2.923160E-02),complex(+7.452649E-02,+0.000000E+00),complex(+0.000000E+00,+1.490530E-01),complex(-2.471764E-01,+0.000000E+00),complex(+0.000000E+00,-3.495603E-01),complex(+4.281222E-01,+0.000000E+00),complex(-0.000000E+00,+4.576819E-01),complex(-4.281222E-01,+0.000000E+00),complex(+0.000000E+00,-3.495603E-01),complex(+2.471764E-01,+0.000000E+00),complex(-0.000000E+00,+1.490530E-01),complex(-7.452650E-02,+0.000000E+00),complex(+0.000000E+00,-2.923170E-02),complex(+7.812484E-03,-0.000000E+00)]$
            ]
          end

      13: begin
            matrix[91:104,91:104]=[$
            [complex(+0.000000E+00,+1.104856E-02),complex(+3.983606E-02,+0.000000E+00),complex(+0.000000E+00,-9.757807E-02),complex(-1.868479E-01,+0.000000E+00),complex(+0.000000E+00,+2.954322E-01),complex(+3.963641E-01,+0.000000E+00),complex(+0.000000E+00,-4.576819E-01),complex(-4.576818E-01,+0.000000E+00),complex(+0.000000E+00,+3.963642E-01),complex(+2.954323E-01,+0.000000E+00),complex(+0.000000E+00,-1.868478E-01),complex(-9.757809E-02,+0.000000E+00),complex(+0.000000E+00,+3.983602E-02),complex(+1.104853E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,-3.983605E-02),complex(-1.215340E-01,+0.000000E+00),complex(+0.000000E+00,+2.435697E-01),complex(+3.627558E-01,+0.000000E+00),complex(+0.000000E+00,-4.096911E-01),complex(-3.297948E-01,+0.000000E+00),complex(+0.000000E+00,+1.269380E-01),complex(-1.269382E-01,+0.000000E+00),complex(+0.000000E+00,+3.297948E-01),complex(+4.096909E-01,+0.000000E+00),complex(+0.000000E+00,-3.627558E-01),complex(-2.435696E-01,+0.000000E+00),complex(+0.000000E+00,+1.215340E-01),complex(+3.983614E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,+9.757815E-02),complex(+2.435696E-01,+0.000000E+00),complex(+0.000000E+00,-3.756506E-01),complex(-3.808143E-01,+0.000000E+00),complex(+0.000000E+00,+2.007068E-01),complex(-8.975875E-02,+0.000000E+00),complex(+0.000000E+00,+3.109335E-01),complex(+3.109336E-01,+0.000000E+00),complex(+0.000000E+00,-8.975876E-02),complex(+2.007068E-01,+0.000000E+00),complex(+0.000000E+00,-3.808143E-01),complex(-3.756505E-01,-0.000000E+00),complex(+0.000000E+00,+2.435697E-01),complex(+9.757811E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-1.868477E-01),complex(-3.627558E-01,+0.000000E+00),complex(+0.000000E+00,+3.808142E-01),complex(+1.546797E-01,+0.000000E+00),complex(+0.000000E+00,+1.746928E-01),complex(+3.281250E-01,+0.000000E+00),complex(+0.000000E+00,-1.623798E-01),complex(+1.623797E-01,+0.000000E+00),complex(+0.000000E+00,-3.281251E-01),complex(-1.746928E-01,+0.000000E+00),complex(+0.000000E+00,-1.546796E-01),complex(-3.808143E-01,+0.000000E+00),complex(+0.000000E+00,+3.627558E-01),complex(+1.868478E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+2.954324E-01),complex(+4.096910E-01,+0.000000E+00),complex(+0.000000E+00,-2.007068E-01),complex(+1.746929E-01,+0.000000E+00),complex(+0.000000E+00,-3.204077E-01),complex(-7.411587E-02,+0.000000E+00),complex(+0.000000E+00,-2.567449E-01),complex(-2.567450E-01,+0.000000E+00),complex(+0.000000E+00,-7.411592E-02),complex(-3.204077E-01,+0.000000E+00),complex(+0.000000E+00,+1.746928E-01),complex(-2.007067E-01,-0.000000E+00),complex(+0.000000E+00,+4.096909E-01),complex(+2.954323E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-3.963640E-01),complex(-3.297949E-01,+0.000000E+00),complex(+0.000000E+00,-8.975881E-02),complex(-3.281249E-01,+0.000000E+00),complex(+0.000000E+00,+7.411585E-02),complex(-2.762136E-01,+0.000000E+00),complex(+0.000000E+00,+1.913664E-01),complex(-1.913664E-01,+0.000000E+00),complex(+0.000000E+00,+2.762136E-01),complex(-7.411583E-02,+0.000000E+00),complex(+0.000000E+00,+3.281249E-01),complex(+8.975881E-02,+0.000000E+00),complex(+0.000000E+00,+3.297948E-01),complex(+3.963641E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+4.576818E-01),complex(+1.269380E-01,+0.000000E+00),complex(+0.000000E+00,+3.109335E-01),complex(+1.623799E-01,+0.000000E+00),complex(+0.000000E+00,+2.567450E-01),complex(+1.913663E-01,+0.000000E+00),complex(+0.000000E+00,+2.209710E-01),complex(+2.209709E-01,+0.000000E+00),complex(+0.000000E+00,+1.913664E-01),complex(+2.567450E-01,+0.000000E+00),complex(+0.000000E+00,+1.623798E-01),complex(+3.109336E-01,+0.000000E+00),complex(+0.000000E+00,+1.269381E-01),complex(+4.576819E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.576818E-01),complex(+1.269381E-01,+0.000000E+00),complex(+0.000000E+00,-3.109336E-01),complex(+1.623799E-01,+0.000000E+00),complex(+0.000000E+00,-2.567450E-01),complex(+1.913663E-01,+0.000000E+00),complex(+0.000000E+00,-2.209708E-01),complex(+2.209708E-01,+0.000000E+00),complex(+0.000000E+00,-1.913664E-01),complex(+2.567450E-01,+0.000000E+00),complex(+0.000000E+00,-1.623798E-01),complex(+3.109336E-01,+0.000000E+00),complex(+0.000000E+00,-1.269381E-01),complex(+4.576818E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.963640E-01),complex(-3.297948E-01,+0.000000E+00),complex(+0.000000E+00,+8.975881E-02),complex(-3.281249E-01,-0.000000E+00),complex(+0.000000E+00,-7.411592E-02),complex(-2.762135E-01,+0.000000E+00),complex(+0.000000E+00,-1.913664E-01),complex(-1.913664E-01,-0.000000E+00),complex(+0.000000E+00,-2.762136E-01),complex(-7.411583E-02,+0.000000E+00),complex(+0.000000E+00,-3.281250E-01),complex(+8.975883E-02,+0.000000E+00),complex(+0.000000E+00,-3.297949E-01),complex(+3.963640E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-2.954323E-01),complex(+4.096910E-01,+0.000000E+00),complex(+0.000000E+00,+2.007068E-01),complex(+1.746929E-01,+0.000000E+00),complex(+0.000000E+00,+3.204078E-01),complex(-7.411580E-02,+0.000000E+00),complex(+0.000000E+00,+2.567449E-01),complex(-2.567450E-01,+0.000000E+00),complex(+0.000000E+00,+7.411580E-02),complex(-3.204077E-01,+0.000000E+00),complex(+0.000000E+00,-1.746928E-01),complex(-2.007067E-01,+0.000000E+00),complex(+0.000000E+00,-4.096910E-01),complex(+2.954322E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+1.868479E-01),complex(-3.627557E-01,+0.000000E+00),complex(+0.000000E+00,-3.808143E-01),complex(+1.546797E-01,+0.000000E+00),complex(+0.000000E+00,-1.746928E-01),complex(+3.281251E-01,+0.000000E+00),complex(+0.000000E+00,+1.623797E-01),complex(+1.623798E-01,+0.000000E+00),complex(+0.000000E+00,+3.281249E-01),complex(-1.746928E-01,+0.000000E+00),complex(+0.000000E+00,+1.546795E-01),complex(-3.808143E-01,+0.000000E+00),complex(+0.000000E+00,-3.627559E-01),complex(+1.868477E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-9.757812E-02),complex(+2.435697E-01,+0.000000E+00),complex(+0.000000E+00,+3.756505E-01),complex(-3.808143E-01,+0.000000E+00),complex(+0.000000E+00,-2.007067E-01),complex(-8.975878E-02,+0.000000E+00),complex(+0.000000E+00,-3.109336E-01),complex(+3.109336E-01,+0.000000E+00),complex(+0.000000E+00,+8.975876E-02),complex(+2.007068E-01,+0.000000E+00),complex(+0.000000E+00,+3.808143E-01),complex(-3.756505E-01,+0.000000E+00),complex(+0.000000E+00,-2.435696E-01),complex(+9.757807E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.983614E-02),complex(-1.215339E-01,+0.000000E+00),complex(+0.000000E+00,-2.435697E-01),complex(+3.627559E-01,+0.000000E+00),complex(+0.000000E+00,+4.096909E-01),complex(-3.297949E-01,+0.000000E+00),complex(+0.000000E+00,-1.269381E-01),complex(-1.269381E-01,+0.000000E+00),complex(+0.000000E+00,-3.297949E-01),complex(+4.096909E-01,+0.000000E+00),complex(+0.000000E+00,+3.627558E-01),complex(-2.435697E-01,+0.000000E+00),complex(+0.000000E+00,-1.215340E-01),complex(+3.983609E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-1.104848E-02),complex(+3.983610E-02,+0.000000E+00),complex(+0.000000E+00,+9.757810E-02),complex(-1.868478E-01,+0.000000E+00),complex(+0.000000E+00,-2.954324E-01),complex(+3.963641E-01,+0.000000E+00),complex(+0.000000E+00,+4.576818E-01),complex(-4.576819E-01,+0.000000E+00),complex(+0.000000E+00,-3.963641E-01),complex(+2.954323E-01,+0.000000E+00),complex(+0.000000E+00,+1.868478E-01),complex(-9.757807E-02,+0.000000E+00),complex(+0.000000E+00,-3.983606E-02),complex(+1.104853E-02,-0.000000E+00)]$
            ]
          end

      12: begin
            matrix[78:90,78:90]=[$
            [complex(+1.562498E-02,+0.000000E+00),complex(+0.000000E+00,-5.412656E-02),complex(-1.269381E-01,+0.000000E+00),complex(+0.000000E+00,+2.317562E-01),complex(+3.476343E-01,+0.000000E+00),complex(+0.000000E+00,-4.397266E-01),complex(-4.749589E-01,+0.000000E+00),complex(+0.000000E+00,+4.397265E-01),complex(+3.476343E-01,+0.000000E+00),complex(+0.000000E+00,-2.317562E-01),complex(-1.269381E-01,+0.000000E+00),complex(+0.000000E+00,+5.412656E-02),complex(+1.562502E-02,+0.000000E+00)],$
            [complex(-5.412657E-02,+0.000000E+00),complex(+0.000000E+00,+1.562500E-01),complex(+2.931510E-01,+0.000000E+00),complex(+0.000000E+00,-4.014137E-01),complex(-4.014136E-01,+0.000000E+00),complex(+0.000000E+00,+2.538761E-01),complex(-2.980232E-08,+0.000000E+00),complex(+0.000000E+00,+2.538762E-01),complex(+4.014135E-01,+0.000000E+00),complex(+0.000000E+00,-4.014134E-01),complex(-2.931510E-01,+0.000000E+00),complex(+0.000000E+00,+1.562500E-01),complex(+5.412663E-02,+0.000000E+00)],$
            [complex(+1.269381E-01,+0.000000E+00),complex(+0.000000E+00,-2.931509E-01),complex(-4.062500E-01,+0.000000E+00),complex(+0.000000E+00,+3.423265E-01),complex(+8.558162E-02,+0.000000E+00),complex(+0.000000E+00,+2.165063E-01),complex(+3.507804E-01,+0.000000E+00),complex(+0.000000E+00,-2.165063E-01),complex(+8.558160E-02,+0.000000E+00),complex(+0.000000E+00,-3.423265E-01),complex(-4.062500E-01,+0.000000E+00),complex(+0.000000E+00,+2.931510E-01),complex(+1.269381E-01,+0.000000E+00)],$
            [complex(-2.317561E-01,+0.000000E+00),complex(+0.000000E+00,+4.014136E-01),complex(+3.423267E-01,+0.000000E+00),complex(+0.000000E+00,-3.125003E-02),complex(+2.812501E-01,+0.000000E+00),complex(+0.000000E+00,-2.964635E-01),complex(-2.980232E-08,+0.000000E+00),complex(+0.000000E+00,-2.964636E-01),complex(-2.812500E-01,+0.000000E+00),complex(+0.000000E+00,-3.124994E-02),complex(-3.423265E-01,+0.000000E+00),complex(+0.000000E+00,+4.014134E-01),complex(+2.317561E-01,+0.000000E+00)],$
            [complex(+3.476344E-01,+0.000000E+00),complex(-0.000000E+00,-4.014135E-01),complex(-8.558163E-02,+0.000000E+00),complex(-0.000000E+00,-2.812500E-01),complex(-2.656250E-01,+0.000000E+00),complex(+0.000000E+00,-9.882118E-02),complex(-3.202173E-01,+0.000000E+00),complex(+0.000000E+00,+9.882119E-02),complex(-2.656250E-01,+0.000000E+00),complex(+0.000000E+00,+2.812500E-01),complex(-8.558162E-02,+0.000000E+00),complex(+0.000000E+00,+4.014135E-01),complex(+3.476343E-01,+0.000000E+00)],$
            [complex(-4.397265E-01,+0.000000E+00),complex(+0.000000E+00,+2.538762E-01),complex(-2.165064E-01,+0.000000E+00),complex(+0.000000E+00,+2.964636E-01),complex(-9.882117E-02,+0.000000E+00),complex(+0.000000E+00,+3.125000E-01),complex(-5.960464E-08,+0.000000E+00),complex(+0.000000E+00,+3.125001E-01),complex(+9.882123E-02,+0.000000E+00),complex(+0.000000E+00,+2.964635E-01),complex(+2.165064E-01,+0.000000E+00),complex(+0.000000E+00,+2.538762E-01),complex(+4.397265E-01,+0.000000E+00)],$
            [complex(+4.749588E-01,+0.000000E+00),complex(+0.000000E+00,-1.788139E-07),complex(+3.507804E-01,+0.000000E+00),complex(+0.000000E+00,+2.980232E-08),complex(+3.202172E-01,+0.000000E+00),complex(+0.000000E+00,-2.980232E-08),complex(+3.125000E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+3.202173E-01,+0.000000E+00),complex(+0.000000E+00,-8.940697E-08),complex(+3.507804E-01,+0.000000E+00),complex(+0.000000E+00,+1.192093E-07),complex(+4.749588E-01,+0.000000E+00)],$
            [complex(-4.397265E-01,+0.000000E+00),complex(+0.000000E+00,-2.538762E-01),complex(-2.165063E-01,+0.000000E+00),complex(+0.000000E+00,-2.964635E-01),complex(-9.882119E-02,-0.000000E+00),complex(+0.000000E+00,-3.125000E-01),0,complex(+0.000000E+00,-3.125001E-01),complex(+9.882117E-02,+0.000000E+00),complex(+0.000000E+00,-2.964636E-01),complex(+2.165064E-01,+0.000000E+00),complex(+0.000000E+00,-2.538761E-01),complex(+4.397265E-01,+0.000000E+00)],$
            [complex(+3.476343E-01,+0.000000E+00),complex(+0.000000E+00,+4.014135E-01),complex(-8.558162E-02,+0.000000E+00),complex(+0.000000E+00,+2.812499E-01),complex(-2.656250E-01,+0.000000E+00),complex(+0.000000E+00,+9.882119E-02),complex(-3.202173E-01,+0.000000E+00),complex(+0.000000E+00,-9.882128E-02),complex(-2.656250E-01,+0.000000E+00),complex(+0.000000E+00,-2.812500E-01),complex(-8.558162E-02,-0.000000E+00),complex(+0.000000E+00,-4.014135E-01),complex(+3.476343E-01,+0.000000E+00)],$
            [complex(-2.317562E-01,+0.000000E+00),complex(+0.000000E+00,-4.014137E-01),complex(+3.423268E-01,+0.000000E+00),complex(+0.000000E+00,+3.125004E-02),complex(+2.812500E-01,+0.000000E+00),complex(+0.000000E+00,+2.964636E-01),complex(+4.470348E-08,+0.000000E+00),complex(+0.000000E+00,+2.964635E-01),complex(-2.812500E-01,+0.000000E+00),complex(+0.000000E+00,+3.124994E-02),complex(-3.423266E-01,+0.000000E+00),complex(+0.000000E+00,-4.014134E-01),complex(+2.317562E-01,+0.000000E+00)],$
            [complex(+1.269381E-01,+0.000000E+00),complex(-0.000000E+00,+2.931510E-01),complex(-4.062499E-01,+0.000000E+00),complex(+0.000000E+00,-3.423266E-01),complex(+8.558162E-02,+0.000000E+00),complex(+0.000000E+00,-2.165064E-01),complex(+3.507804E-01,+0.000000E+00),complex(+0.000000E+00,+2.165064E-01),complex(+8.558164E-02,+0.000000E+00),complex(-0.000000E+00,+3.423266E-01),complex(-4.062500E-01,+0.000000E+00),complex(+0.000000E+00,-2.931510E-01),complex(+1.269381E-01,+0.000000E+00)],$
            [complex(-5.412662E-02,+0.000000E+00),complex(+0.000000E+00,-1.562500E-01),complex(+2.931510E-01,+0.000000E+00),complex(+0.000000E+00,+4.014135E-01),complex(-4.014136E-01,+0.000000E+00),complex(+0.000000E+00,-2.538763E-01),complex(+1.490116E-08,+0.000000E+00),complex(+0.000000E+00,-2.538761E-01),complex(+4.014135E-01,+0.000000E+00),complex(-0.000000E+00,+4.014135E-01),complex(-2.931509E-01,+0.000000E+00),complex(+0.000000E+00,-1.562500E-01),complex(+5.412662E-02,+0.000000E+00)],$
            [complex(+1.562494E-02,+0.000000E+00),complex(+0.000000E+00,+5.412659E-02),complex(-1.269381E-01,+0.000000E+00),complex(+0.000000E+00,-2.317562E-01),complex(+3.476343E-01,+0.000000E+00),complex(+0.000000E+00,+4.397264E-01),complex(-4.749589E-01,+0.000000E+00),complex(+0.000000E+00,-4.397264E-01),complex(+3.476343E-01,-0.000000E+00),complex(-0.000000E+00,+2.317562E-01),complex(-1.269381E-01,+0.000000E+00),complex(+0.000000E+00,-5.412659E-02),complex(+1.562501E-02,+0.000000E+00)]$
            ]
          end

      11: begin
            matrix[66:77,66:77]=[$
            [complex(+0.000000E+00,-2.209715E-02),complex(-7.328776E-02,+0.000000E+00),complex(+0.000000E+00,+1.638764E-01),complex(+2.838423E-01,+0.000000E+00),complex(+0.000000E+00,-4.014135E-01),complex(-4.749588E-01,+0.000000E+00),complex(+0.000000E+00,+4.749590E-01),complex(+4.014136E-01,+0.000000E+00),complex(+0.000000E+00,-2.838422E-01),complex(-1.638764E-01,+0.000000E+00),complex(+0.000000E+00,+7.328776E-02),complex(+2.209708E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,+7.328772E-02),complex(+1.988738E-01,+0.000000E+00),complex(+0.000000E+00,-3.458741E-01),complex(-4.279082E-01,+0.000000E+00),complex(+0.000000E+00,+3.630922E-01),complex(+1.432055E-01,+0.000000E+00),complex(+0.000000E+00,+1.432056E-01),complex(+3.630922E-01,+0.000000E+00),complex(+0.000000E+00,-4.279082E-01),complex(-3.458741E-01,+0.000000E+00),complex(+0.000000E+00,+1.988738E-01),complex(+7.328776E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-1.638763E-01),complex(-3.458742E-01,+0.000000E+00),complex(+0.000000E+00,+4.198447E-01),complex(+2.679128E-01,+0.000000E+00),complex(+0.000000E+00,+5.412655E-02),complex(+3.202174E-01,+0.000000E+00),complex(+0.000000E+00,-3.202171E-01),complex(-5.412658E-02,+0.000000E+00),complex(+0.000000E+00,-2.679130E-01),complex(-4.198447E-01,+0.000000E+00),complex(+0.000000E+00,+3.458741E-01),complex(+1.638765E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+2.838423E-01),complex(+4.279081E-01,+0.000000E+00),complex(+0.000000E+00,-2.679130E-01),complex(+1.104853E-01,+0.000000E+00),complex(+0.000000E+00,-3.437501E-01),complex(-1.848777E-01,+0.000000E+00),complex(+0.000000E+00,-1.848774E-01),complex(-3.437500E-01,+0.000000E+00),complex(+0.000000E+00,+1.104855E-01),complex(-2.679129E-01,+0.000000E+00),complex(+0.000000E+00,+4.279082E-01),complex(+2.838423E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.014135E-01),complex(-3.630922E-01,+0.000000E+00),complex(+0.000000E+00,-5.412661E-02),complex(-3.437501E-01,+0.000000E+00),complex(+0.000000E+00,+1.325825E-01),complex(-2.614563E-01,+0.000000E+00),complex(+0.000000E+00,+2.614563E-01),complex(-1.325825E-01,+0.000000E+00),complex(+0.000000E+00,+3.437501E-01),complex(+5.412661E-02,+0.000000E+00),complex(+0.000000E+00,+3.630922E-01),complex(+4.014136E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+4.749590E-01),complex(+1.432056E-01,+0.000000E+00),complex(+0.000000E+00,+3.202171E-01),complex(+1.848774E-01,+0.000000E+00),complex(+0.000000E+00,+2.614563E-01),complex(+2.209709E-01,+0.000000E+00),complex(+0.000000E+00,+2.209709E-01),complex(+2.614563E-01,+0.000000E+00),complex(+0.000000E+00,+1.848775E-01),complex(+3.202172E-01,+0.000000E+00),complex(+0.000000E+00,+1.432055E-01),complex(+4.749588E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.749589E-01),complex(+1.432055E-01,+0.000000E+00),complex(+0.000000E+00,-3.202172E-01),complex(+1.848774E-01,+0.000000E+00),complex(+0.000000E+00,-2.614563E-01),complex(+2.209709E-01,+0.000000E+00),complex(+0.000000E+00,-2.209708E-01),complex(+2.614563E-01,+0.000000E+00),complex(+0.000000E+00,-1.848775E-01),complex(+3.202173E-01,+0.000000E+00),complex(+0.000000E+00,-1.432055E-01),complex(+4.749589E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+4.014135E-01),complex(-3.630922E-01,+0.000000E+00),complex(+0.000000E+00,+5.412665E-02),complex(-3.437501E-01,-0.000000E+00),complex(+0.000000E+00,-1.325825E-01),complex(-2.614562E-01,+0.000000E+00),complex(+0.000000E+00,-2.614563E-01),complex(-1.325825E-01,+0.000000E+00),complex(+0.000000E+00,-3.437500E-01),complex(+5.412659E-02,-0.000000E+00),complex(+0.000000E+00,-3.630922E-01),complex(+4.014136E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-2.838423E-01),complex(+4.279084E-01,+0.000000E+00),complex(+0.000000E+00,+2.679128E-01),complex(+1.104855E-01,+0.000000E+00),complex(+0.000000E+00,+3.437499E-01),complex(-1.848774E-01,+0.000000E+00),complex(+0.000000E+00,+1.848776E-01),complex(-3.437499E-01,+0.000000E+00),complex(+0.000000E+00,-1.104854E-01),complex(-2.679129E-01,+0.000000E+00),complex(+0.000000E+00,-4.279083E-01),complex(+2.838423E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+1.638764E-01),complex(-3.458741E-01,+0.000000E+00),complex(+0.000000E+00,-4.198447E-01),complex(+2.679130E-01,+0.000000E+00),complex(+0.000000E+00,-5.412655E-02),complex(+3.202172E-01,+0.000000E+00),complex(+0.000000E+00,+3.202173E-01),complex(-5.412658E-02,+0.000000E+00),complex(+0.000000E+00,+2.679129E-01),complex(-4.198447E-01,+0.000000E+00),complex(+0.000000E+00,-3.458742E-01),complex(+1.638764E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,-7.328776E-02),complex(+1.988738E-01,+0.000000E+00),complex(+0.000000E+00,+3.458740E-01),complex(-4.279084E-01,+0.000000E+00),complex(+0.000000E+00,-3.630921E-01),complex(+1.432055E-01,+0.000000E+00),complex(-0.000000E+00,-1.432054E-01),complex(+3.630922E-01,-0.000000E+00),complex(-0.000000E+00,+4.279082E-01),complex(-3.458742E-01,+0.000000E+00),complex(+0.000000E+00,-1.988738E-01),complex(+7.328777E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,+2.209707E-02),complex(-7.328773E-02,+0.000000E+00),complex(+0.000000E+00,-1.638764E-01),complex(+2.838422E-01,+0.000000E+00),complex(+0.000000E+00,+4.014136E-01),complex(-4.749589E-01,+0.000000E+00),complex(+0.000000E+00,-4.749589E-01),complex(+4.014135E-01,-0.000000E+00),complex(-0.000000E+00,+2.838422E-01),complex(-1.638763E-01,+0.000000E+00),complex(+0.000000E+00,-7.328772E-02),complex(+2.209713E-02,-0.000000E+00)]$
            ]
          end

      10: begin
            matrix[55:65,55:65]=[$
            [complex(-3.124998E-02,+0.000000E+00),complex(+0.000000E+00,+9.882118E-02),complex(+2.096314E-01,+0.000000E+00),complex(+0.000000E+00,-3.423266E-01),complex(-4.528556E-01,+0.000000E+00),complex(+0.000000E+00,+4.960784E-01),complex(+4.528555E-01,+0.000000E+00),complex(+0.000000E+00,-3.423266E-01),complex(-2.096314E-01,+0.000000E+00),complex(+0.000000E+00,+9.882114E-02),complex(+3.124998E-02,-0.000000E+00)],$
            [complex(+9.882127E-02,+0.000000E+00),complex(+0.000000E+00,-2.500000E-01),complex(-3.977476E-01,+0.000000E+00),complex(+0.000000E+00,+4.330126E-01),complex(+2.864110E-01,+0.000000E+00),complex(+0.000000E+00,-4.470348E-08),complex(+2.864109E-01,+0.000000E+00),complex(+0.000000E+00,-4.330127E-01),complex(-3.977476E-01,+0.000000E+00),complex(+0.000000E+00,+2.500000E-01),complex(+9.882116E-02,-0.000000E+00)],$
            [complex(-2.096313E-01,+0.000000E+00),complex(+0.000000E+00,+3.977475E-01),complex(+4.062500E-01,+0.000000E+00),complex(+0.000000E+00,-1.530932E-01),complex(+2.025232E-01,+0.000000E+00),complex(+0.000000E+00,-3.697550E-01),complex(-2.025232E-01,+0.000000E+00),complex(+0.000000E+00,-1.530930E-01),complex(-4.062501E-01,+0.000000E+00),complex(+0.000000E+00,+3.977475E-01),complex(+2.096313E-01,+0.000000E+00)],$
            [complex(+3.423266E-01,+0.000000E+00),complex(+0.000000E+00,-4.330129E-01),complex(-1.530930E-01,+0.000000E+00),complex(+0.000000E+00,-2.500000E-01),complex(-3.307188E-01,+0.000000E+00),complex(+0.000000E+00,-8.940697E-08),complex(-3.307189E-01,+0.000000E+00),complex(+0.000000E+00,+2.500001E-01),complex(-1.530930E-01,+0.000000E+00),complex(+0.000000E+00,+4.330128E-01),complex(+3.423266E-01,-0.000000E+00)],$
            [complex(-4.528555E-01,+0.000000E+00),complex(+0.000000E+00,+2.864109E-01),complex(-2.025231E-01,+0.000000E+00),complex(+0.000000E+00,+3.307189E-01),complex(-6.250003E-02,+0.000000E+00),complex(+0.000000E+00,+3.423266E-01),complex(+6.250004E-02,+0.000000E+00),complex(+0.000000E+00,+3.307189E-01),complex(+2.025232E-01,+0.000000E+00),complex(+0.000000E+00,+2.864110E-01),complex(+4.528555E-01,-0.000000E+00)],$
            [complex(+4.960785E-01,+0.000000E+00),complex(+0.000000E+00,+1.490116E-07),complex(+3.697550E-01,+0.000000E+00),complex(+0.000000E+00,+5.960464E-08),complex(+3.423267E-01,+0.000000E+00),complex(+0.000000E+00,+5.960464E-08),complex(+3.423266E-01,-0.000000E+00),complex(+0.000000E+00,+2.980232E-08),complex(+3.697549E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+4.960784E-01,-0.000000E+00)],$
            [complex(-4.528554E-01,+0.000000E+00),complex(+0.000000E+00,-2.864110E-01),complex(-2.025232E-01,+0.000000E+00),complex(+0.000000E+00,-3.307189E-01),complex(-6.249991E-02,+0.000000E+00),complex(+0.000000E+00,-3.423267E-01),complex(+6.250001E-02,+0.000000E+00),complex(+0.000000E+00,-3.307189E-01),complex(+2.025231E-01,+0.000000E+00),complex(+0.000000E+00,-2.864110E-01),complex(+4.528555E-01,-0.000000E+00)],$
            [complex(+3.423266E-01,+0.000000E+00),complex(+0.000000E+00,+4.330128E-01),complex(-1.530931E-01,-0.000000E+00),complex(+0.000000E+00,+2.500001E-01),complex(-3.307189E-01,-0.000000E+00),complex(+0.000000E+00,-2.980232E-08),complex(-3.307189E-01,+0.000000E+00),complex(+0.000000E+00,-2.499999E-01),complex(-1.530931E-01,+0.000000E+00),complex(+0.000000E+00,-4.330127E-01),complex(+3.423266E-01,+0.000000E+00)],$
            [complex(-2.096313E-01,+0.000000E+00),complex(+0.000000E+00,-3.977475E-01),complex(+4.062500E-01,+0.000000E+00),complex(+0.000000E+00,+1.530931E-01),complex(+2.025231E-01,+0.000000E+00),complex(+0.000000E+00,+3.697551E-01),complex(-2.025232E-01,+0.000000E+00),complex(+0.000000E+00,+1.530931E-01),complex(-4.062500E-01,+0.000000E+00),complex(+0.000000E+00,-3.977476E-01),complex(+2.096314E-01,-0.000000E+00)],$
            [complex(+9.882117E-02,+0.000000E+00),complex(+0.000000E+00,+2.500000E-01),complex(-3.977475E-01,+0.000000E+00),complex(+0.000000E+00,-4.330127E-01),complex(+2.864110E-01,+0.000000E+00),complex(+0.000000E+00,+8.195639E-08),complex(+2.864110E-01,+0.000000E+00),complex(+0.000000E+00,+4.330127E-01),complex(-3.977475E-01,+0.000000E+00),complex(+0.000000E+00,-2.500000E-01),complex(+9.882123E-02,-0.000000E+00)],$
            [complex(-3.125006E-02,+0.000000E+00),complex(+0.000000E+00,-9.882120E-02),complex(+2.096314E-01,+0.000000E+00),complex(+0.000000E+00,+3.423267E-01),complex(-4.528555E-01,+0.000000E+00),complex(+0.000000E+00,-4.960783E-01),complex(+4.528555E-01,+0.000000E+00),complex(+0.000000E+00,+3.423266E-01),complex(-2.096313E-01,+0.000000E+00),complex(+0.000000E+00,-9.882118E-02),complex(+3.125001E-02,-0.000000E+00)]$
            ]
          end

       9: begin
            matrix[45:54,45:54]=[$
            [complex(+0.000000E+00,+4.419409E-02),complex(+1.325825E-01,+0.000000E+00),complex(+0.000000E+00,-2.651650E-01),complex(-4.050463E-01,+0.000000E+00),complex(+0.000000E+00,+4.960783E-01),complex(+4.960784E-01,+0.000000E+00),complex(+0.000000E+00,-4.050463E-01),complex(-2.651650E-01,+0.000000E+00),complex(+0.000000E+00,+1.325825E-01),complex(+4.419418E-02,-0.000000E+00)],$
            [complex(+0.000000E+00,-1.325825E-01),complex(-3.093592E-01,+0.000000E+00),complex(+0.000000E+00,+4.419417E-01),complex(+4.050463E-01,+0.000000E+00),complex(+0.000000E+00,-1.653594E-01),complex(+1.653594E-01,+0.000000E+00),complex(+0.000000E+00,-4.050463E-01),complex(-4.419417E-01,+0.000000E+00),complex(+0.000000E+00,+3.093592E-01),complex(+1.325825E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+2.651650E-01),complex(+4.419417E-01,+0.000000E+00),complex(+0.000000E+00,-3.535534E-01),complex(-7.450581E-08,+0.000000E+00),complex(+0.000000E+00,-3.307189E-01),complex(-3.307189E-01,+0.000000E+00),complex(+0.000000E+00,+1.490116E-08),complex(-3.535534E-01,+0.000000E+00),complex(+0.000000E+00,+4.419418E-01),complex(+2.651651E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.050463E-01),complex(-4.050463E-01,+0.000000E+00),complex(+0.000000E+00,+4.828528E-08),complex(-3.535534E-01,+0.000000E+00),complex(+0.000000E+00,+2.165064E-01),complex(-2.165064E-01,+0.000000E+00),complex(+0.000000E+00,+3.535534E-01),complex(-2.414264E-08,+0.000000E+00),complex(+0.000000E+00,+4.050463E-01),complex(+4.050463E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+4.960784E-01),complex(+1.653594E-01,+0.000000E+00),complex(+0.000000E+00,+3.307189E-01),complex(+2.165065E-01,+0.000000E+00),complex(+0.000000E+00,+2.651650E-01),complex(+2.651650E-01,+0.000000E+00),complex(+0.000000E+00,+2.165064E-01),complex(+3.307190E-01,+0.000000E+00),complex(+0.000000E+00,+1.653594E-01),complex(+4.960784E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.960784E-01),complex(+1.653595E-01,+0.000000E+00),complex(+0.000000E+00,-3.307189E-01),complex(+2.165064E-01,+0.000000E+00),complex(+0.000000E+00,-2.651650E-01),complex(+2.651650E-01,+0.000000E+00),complex(+0.000000E+00,-2.165064E-01),complex(+3.307189E-01,+0.000000E+00),complex(+0.000000E+00,-1.653595E-01),complex(+4.960784E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+4.050463E-01),complex(-4.050461E-01,+0.000000E+00),complex(+0.000000E+00,+1.848296E-08),complex(-3.535534E-01,+0.000000E+00),complex(+0.000000E+00,-2.165063E-01),complex(-2.165064E-01,+0.000000E+00),complex(+0.000000E+00,-3.535534E-01),complex(-5.394496E-08,+0.000000E+00),complex(+0.000000E+00,-4.050464E-01),complex(+4.050463E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-2.651651E-01),complex(+4.419418E-01,+0.000000E+00),complex(+0.000000E+00,+3.535534E-01),complex(-1.490116E-08,+0.000000E+00),complex(+0.000000E+00,+3.307189E-01),complex(-3.307189E-01,+0.000000E+00),complex(+0.000000E+00,+1.490116E-08),complex(-3.535534E-01,+0.000000E+00),complex(+0.000000E+00,-4.419418E-01),complex(+2.651651E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+1.325824E-01),complex(-3.093593E-01,+0.000000E+00),complex(+0.000000E+00,-4.419418E-01),complex(+4.050463E-01,+0.000000E+00),complex(+0.000000E+00,+1.653595E-01),complex(+1.653594E-01,+0.000000E+00),complex(+0.000000E+00,+4.050463E-01),complex(-4.419417E-01,+0.000000E+00),complex(+0.000000E+00,-3.093593E-01),complex(+1.325825E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-4.419417E-02),complex(+1.325825E-01,+0.000000E+00),complex(+0.000000E+00,+2.651650E-01),complex(-4.050463E-01,+0.000000E+00),complex(+0.000000E+00,-4.960784E-01),complex(+4.960784E-01,+0.000000E+00),complex(+0.000000E+00,+4.050463E-01),complex(-2.651651E-01,+0.000000E+00),complex(+0.000000E+00,-1.325825E-01),complex(+4.419418E-02,-0.000000E+00)]$
            ]
          end

       8: begin
            matrix[36:44,36:44]=[$
            [complex(+6.249997E-02,+0.000000E+00),complex(+0.000000E+00,-1.767768E-01),complex(-3.307190E-01,+0.000000E+00),complex(+0.000000E+00,+4.677072E-01),complex(+5.229125E-01,+0.000000E+00),complex(+0.000000E+00,-4.677072E-01),complex(-3.307189E-01,+0.000000E+00),complex(+0.000000E+00,+1.767767E-01),complex(+6.250001E-02,-0.000000E+00)],$
            [complex(-1.767767E-01,+0.000000E+00),complex(+0.000000E+00,+3.750000E-01),complex(+4.677071E-01,+0.000000E+00),complex(+0.000000E+00,-3.307189E-01),complex(+4.759509E-08,+0.000000E+00),complex(+0.000000E+00,-3.307189E-01),complex(-4.677072E-01,+0.000000E+00),complex(+0.000000E+00,+3.750001E-01),complex(+1.767767E-01,+0.000000E+00)],$
            [complex(+3.307189E-01,+0.000000E+00),complex(+0.000000E+00,-4.677073E-01),complex(-2.499999E-01,+0.000000E+00),complex(+0.000000E+00,-1.767767E-01),complex(-3.952847E-01,+0.000000E+00),complex(+0.000000E+00,+1.767767E-01),complex(-2.500000E-01,+0.000000E+00),complex(+0.000000E+00,+4.677072E-01),complex(+3.307189E-01,-0.000000E+00)],$
            [complex(-4.677071E-01,+0.000000E+00),complex(+0.000000E+00,+3.307189E-01),complex(-1.767766E-01,+0.000000E+00),complex(+0.000000E+00,+3.750001E-01),complex(+2.206292E-09,+0.000000E+00),complex(+0.000000E+00,+3.750000E-01),complex(+1.767767E-01,+0.000000E+00),complex(+0.000000E+00,+3.307189E-01),complex(+4.677071E-01,-0.000000E+00)],$
            [complex(+5.229125E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+3.952848E-01,+0.000000E+00),0,complex(+3.750000E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(+3.952847E-01,+0.000000E+00),complex(+0.000000E+00,-2.980232E-08),complex(+5.229125E-01,-0.000000E+00)],$
            [complex(-4.677070E-01,+0.000000E+00),complex(+0.000000E+00,-3.307189E-01),complex(-1.767766E-01,-0.000000E+00),complex(+0.000000E+00,-3.750000E-01),complex(-3.721847E-08,+0.000000E+00),complex(+0.000000E+00,-3.750000E-01),complex(+1.767766E-01,+0.000000E+00),complex(+0.000000E+00,-3.307189E-01),complex(+4.677071E-01,+0.000000E+00)],$
            [complex(+3.307191E-01,+0.000000E+00),complex(+0.000000E+00,+4.677072E-01),complex(-2.499999E-01,+0.000000E+00),complex(+0.000000E+00,+1.767768E-01),complex(-3.952847E-01,+0.000000E+00),complex(+0.000000E+00,-1.767766E-01),complex(-2.500001E-01,+0.000000E+00),complex(+0.000000E+00,-4.677072E-01),complex(+3.307188E-01,-0.000000E+00)],$
            [complex(-1.767766E-01,+0.000000E+00),complex(+0.000000E+00,-3.750001E-01),complex(+4.677072E-01,+0.000000E+00),complex(+0.000000E+00,+3.307189E-01),complex(+2.891598E-09,+0.000000E+00),complex(+0.000000E+00,+3.307190E-01),complex(-4.677071E-01,+0.000000E+00),complex(+0.000000E+00,-3.750000E-01),complex(+1.767766E-01,+0.000000E+00)],$
            [complex(+6.250003E-02,+0.000000E+00),complex(+0.000000E+00,+1.767767E-01),complex(-3.307189E-01,+0.000000E+00),complex(+0.000000E+00,-4.677072E-01),complex(+5.229126E-01,+0.000000E+00),complex(+0.000000E+00,+4.677072E-01),complex(-3.307189E-01,+0.000000E+00),complex(+0.000000E+00,-1.767767E-01),complex(+6.249999E-02,-0.000000E+00)]$
            ]
          end

       7: begin
            matrix[28:35,28:35]=[$
            [complex(+0.000000E+00,-8.838837E-02),complex(-2.338536E-01,+0.000000E+00),complex(+0.000000E+00,+4.050463E-01),complex(+5.229126E-01,+0.000000E+00),complex(+0.000000E+00,-5.229126E-01),complex(-4.050463E-01,+0.000000E+00),complex(+0.000000E+00,+2.338536E-01),complex(+8.838837E-02,+0.000000E+00)],$
            [complex(+0.000000E+00,+2.338536E-01),complex(+4.419418E-01,+0.000000E+00),complex(+0.000000E+00,-4.592794E-01),complex(-1.976424E-01,+0.000000E+00),complex(+0.000000E+00,-1.976424E-01),complex(-4.592793E-01,+0.000000E+00),complex(+0.000000E+00,+4.419417E-01),complex(+2.338536E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,-4.050463E-01),complex(-4.592793E-01,+0.000000E+00),complex(+0.000000E+00,+8.838829E-02),complex(-3.423266E-01,+0.000000E+00),complex(+0.000000E+00,+3.423266E-01),complex(-8.838829E-02,+0.000000E+00),complex(+0.000000E+00,+4.592793E-01),complex(+4.050463E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+5.229124E-01),complex(+1.976423E-01,+0.000000E+00),complex(+0.000000E+00,+3.423266E-01),complex(+2.651651E-01,+0.000000E+00),complex(+0.000000E+00,+2.651650E-01),complex(+3.423266E-01,+0.000000E+00),complex(+0.000000E+00,+1.976423E-01),complex(+5.229126E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-5.229126E-01),complex(+1.976424E-01,+0.000000E+00),complex(+0.000000E+00,-3.423266E-01),complex(+2.651651E-01,+0.000000E+00),complex(+0.000000E+00,-2.651651E-01),complex(+3.423266E-01,+0.000000E+00),complex(+0.000000E+00,-1.976424E-01),complex(+5.229126E-01,-0.000000E+00)],$
            [complex(-0.000000E+00,+4.050463E-01),complex(-4.592793E-01,+0.000000E+00),complex(+0.000000E+00,-8.838831E-02),complex(-3.423266E-01,+0.000000E+00),complex(+0.000000E+00,-3.423266E-01),complex(-8.838832E-02,+0.000000E+00),complex(+0.000000E+00,-4.592793E-01),complex(+4.050463E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-2.338536E-01),complex(+4.419417E-01,+0.000000E+00),complex(+0.000000E+00,+4.592793E-01),complex(-1.976423E-01,+0.000000E+00),complex(+0.000000E+00,+1.976424E-01),complex(-4.592793E-01,+0.000000E+00),complex(+0.000000E+00,-4.419417E-01),complex(+2.338536E-01,+0.000000E+00)],$
            [complex(+0.000000E+00,+8.838834E-02),complex(-2.338536E-01,+0.000000E+00),complex(+0.000000E+00,-4.050463E-01),complex(+5.229126E-01,-0.000000E+00),complex(+0.000000E+00,+5.229125E-01),complex(-4.050463E-01,+0.000000E+00),complex(+0.000000E+00,-2.338536E-01),complex(+8.838834E-02,-0.000000E+00)]$
            ]
          end

       6: begin
            matrix[21:27,21:27]=[$
            [complex(-1.250000E-01,+0.000000E+00),complex(+0.000000E+00,+3.061862E-01),complex(+4.841229E-01,+0.000000E+00),complex(+0.000000E+00,-5.590170E-01),complex(-4.841229E-01,-0.000000E+00),complex(+0.000000E+00,+3.061862E-01),complex(+1.250000E-01,-0.000000E+00)],$
            [complex(+3.061861E-01,+0.000000E+00),complex(+0.000000E+00,-5.000000E-01),complex(-3.952847E-01,+0.000000E+00),complex(+0.000000E+00,-2.980232E-08),complex(-3.952847E-01,+0.000000E+00),complex(+0.000000E+00,+5.000000E-01),complex(+3.061863E-01,-0.000000E+00)],$
            [complex(-4.841229E-01,+0.000000E+00),complex(+0.000000E+00,+3.952847E-01),complex(-1.250000E-01,+0.000000E+00),complex(+0.000000E+00,+4.330127E-01),complex(+1.250000E-01,+0.000000E+00),complex(+0.000000E+00,+3.952847E-01),complex(+4.841229E-01,-0.000000E+00)],$
            [complex(+5.590170E-01,+0.000000E+00),complex(+0.000000E+00,-2.980232E-08),complex(+4.330127E-01,+0.000000E+00),0,complex(+4.330127E-01,+0.000000E+00),complex(+0.000000E+00,-2.980232E-08),complex(+5.590170E-01,-0.000000E+00)],$
            [complex(-4.841229E-01,+0.000000E+00),complex(+0.000000E+00,-3.952847E-01),complex(-1.250000E-01,+0.000000E+00),complex(+0.000000E+00,-4.330127E-01),complex(+1.250000E-01,+0.000000E+00),complex(+0.000000E+00,-3.952847E-01),complex(+4.841229E-01,-0.000000E+00)],$
            [complex(+3.061862E-01,+0.000000E+00),complex(-0.000000E+00,+5.000000E-01),complex(-3.952847E-01,+0.000000E+00),complex(+0.000000E+00,-5.960464E-08),complex(-3.952847E-01,+0.000000E+00),complex(+0.000000E+00,-5.000000E-01),complex(+3.061862E-01,-0.000000E+00)],$
            [complex(-1.250000E-01,-0.000000E+00),complex(+0.000000E+00,-3.061862E-01),complex(+4.841229E-01,+0.000000E+00),complex(+0.000000E+00,+5.590170E-01),complex(-4.841229E-01,+0.000000E+00),complex(+0.000000E+00,-3.061862E-01),complex(+1.250000E-01,+0.000000E+00)]$
            ]
          end

       5: begin
            matrix[15:20,15:20]=[$
            [complex(+0.000000E+00,+1.767767E-01),complex(+3.952847E-01,+0.000000E+00),complex(+0.000000E+00,-5.590170E-01),complex(-5.590169E-01,+0.000000E+00),complex(+0.000000E+00,+3.952847E-01),complex(+1.767767E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-3.952847E-01),complex(-5.303301E-01,+0.000000E+00),complex(+0.000000E+00,+2.500000E-01),complex(-2.500000E-01,+0.000000E+00),complex(+0.000000E+00,+5.303301E-01),complex(+3.952847E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+5.590169E-01),complex(+2.500000E-01,+0.000000E+00),complex(+0.000000E+00,+3.535534E-01),complex(+3.535534E-01,+0.000000E+00),complex(+0.000000E+00,+2.500000E-01),complex(+5.590169E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-5.590170E-01),complex(+2.500000E-01,+0.000000E+00),complex(+0.000000E+00,-3.535534E-01),complex(+3.535534E-01,+0.000000E+00),complex(+0.000000E+00,-2.500000E-01),complex(+5.590170E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.952847E-01),complex(-5.303301E-01,+0.000000E+00),complex(+0.000000E+00,-2.499999E-01),complex(-2.500001E-01,-0.000000E+00),complex(+0.000000E+00,-5.303301E-01),complex(+3.952847E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-1.767767E-01),complex(+3.952847E-01,+0.000000E+00),complex(+0.000000E+00,+5.590170E-01),complex(-5.590170E-01,+0.000000E+00),complex(+0.000000E+00,-3.952847E-01),complex(+1.767767E-01,+0.000000E+00)]$
            ]
          end

       4: begin
            matrix[10:14,10:14]=[$
            [complex(+2.500000E-01,+0.000000E+00),complex(-0.000000E+00,-5.000000E-01),complex(-6.123724E-01,+0.000000E+00),complex(+0.000000E+00,+5.000000E-01),complex(+2.500000E-01,+0.000000E+00)],$
            [complex(-5.000000E-01,+0.000000E+00),complex(+0.000000E+00,+5.000000E-01),0,complex(+0.000000E+00,+5.000000E-01),complex(+5.000000E-01,+0.000000E+00)],$
            [complex(+6.123724E-01,+0.000000E+00),complex(+0.000000E+00,+5.960464E-08),complex(+4.999999E-01,+0.000000E+00),0,complex(+6.123724E-01,+0.000000E+00)],$
            [complex(-5.000000E-01,+0.000000E+00),complex(+0.000000E+00,-5.000000E-01),0,complex(+0.000000E+00,-5.000000E-01),complex(+5.000000E-01,+0.000000E+00)],$
            [complex(+2.500000E-01,+0.000000E+00),complex(+0.000000E+00,+5.000000E-01),complex(-6.123724E-01,+0.000000E+00),complex(+0.000000E+00,-5.000000E-01),complex(+2.500000E-01,+0.000000E+00)]$
            ]
          end

       3: begin
            matrix[6:9,6:9]=[$
            [complex(+0.000000E+00,-3.535534E-01),complex(-6.123724E-01,+0.000000E+00),complex(+0.000000E+00,+6.123725E-01),complex(+3.535534E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+6.123725E-01),complex(+3.535534E-01,+0.000000E+00),complex(+0.000000E+00,+3.535534E-01),complex(+6.123725E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-6.123725E-01),complex(+3.535535E-01,+0.000000E+00),complex(+0.000000E+00,-3.535534E-01),complex(+6.123725E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,+3.535533E-01),complex(-6.123724E-01,+0.000000E+00),complex(+0.000000E+00,-6.123725E-01),complex(+3.535535E-01,+0.000000E+00)]$
            ]
          end

       2: begin
            matrix[3:5,3:5]=[$
            [complex(-4.999999E-01,+0.000000E+00),complex(+0.000000E+00,+7.071068E-01),complex(+5.000000E-01,+0.000000E+00)],$
            [complex(+7.071069E-01,+0.000000E+00),complex(+0.000000E+00,+5.960464E-08),complex(+7.071067E-01,+0.000000E+00)],$
            [complex(-5.000000E-01,+0.000000E+00),complex(+0.000000E+00,-7.071068E-01),complex(+5.000000E-01,+0.000000E+00)]$
            ]
          end

       1: begin
            matrix[1:2,1:2]=[$
            [complex(+0.000000E+00,+7.071068E-01),complex(+7.071068E-01,-0.000000E+00)],$
            [complex(+0.000000E+00,-7.071068E-01),complex(+7.071068E-01,-0.000000E+00)]$
            ]
          end

       0: matrix[0]=complex(1.,0.)
 
    endswitch
  endelse
endelse

; Code to write the matrix look-up tables used in this routine. Note that we
; can't print out the whole matrix because the read buffer isn't long enough
; to cope with entire matrices larger than n_max=4. Uncomment this to write
; out the necessary code to a temp file.
;openw,lun,"~/temp",/get_lun
;printf,lun,"pro temp, matrix=matrix"
;printf,lun,"n_max=",n_max,format='(A6,I0)'
;printf,lun,"n_coeffs=",n_coeffs,format='(A9,I0)'
;printf,lun
;printf,lun
;printf,lun,"matrix=complexarr(n_coeffs,n_coeffs)"
;printf,lun,"switch n_max of"
;printf,lun
;for n=n_max,0,-1 do begin
;  wn=where(n1+n2 eq n,n_n)
;  minn=min(wn) & maxn=max(wn) 
;  matrixn=matrix[minn:maxn,minn:maxn]
;  printf,lun,"  ",n,": begin",format='(A2,I2,A7)'
;  printf,lun,"        matrix[",minn,":",maxn,",",minn,":",maxn,"]","=[$",$
;        format='(A15,4(I0,A1),A3)'
;  for i=minn,maxn do begin
;    printf,lun,"        [",format='($,A9)'
;    for j=minn,maxn do begin
;      if matrix[j,i] eq 0 then begin
;        printf,lun,0,format='($,I1)'
;      endif else begin
;        printf,lun,"complex(",float(matrix[j,i]),",",imaginary(matrix[j,i]),")",$
;             format='($,A8,2(E+-13.6,A1))'
;             ;format='($,A8,2(F+-10.7,A1))'
;      endelse
;      if j lt maxn then printf,lun,",",format='($,A1)'
;    endfor
;    printf,lun,"]",format='($,A1)'
;  if i lt maxn then printf,lun,",",format='($,A1)'
;    printf,lun,"$",format='(A1)'
;  endfor
;  printf,lun,"        ]",format='(A9)'
;  printf,lun,"      end",format='(A9)'
;  printf,lun
;endfor
;printf,lun,"endswitch"
;printf,lun
;printf,lun
;printf,lun,"end"
;close,lun
;free_lun,lun
;print,"Now type:"
;print,".com ~/temp"
;print,"temp, matrix=matrix"

return,matrix

end
